/*  FatFs - FAT file system module  R0.09a                 (C)ChaN, 2012
 *
 *  FatFs module is a generic FAT file system module. This is a free software
 *  open for education, research and commercial developments under license
 *  policy of following terms.
 *
 *  Copyright (C) 2012, ChaN, all right reserved.
 *
 *  No restriction on use. You can use, modify and redistribute it for
 *  personal, non-profit or commercial use UNDER YOUR RESPONSIBILITY.
 *
 *  Redistributions of source code must retain the above copyright notice.
 *
/ Jun 01,'06 R0.02  Added FAT12 support.
/ Dec 11,'06 R0.03a Improved cluster scan algorithm to write files fast.
/ Apr 01,'07 R0.04a Supported multiple partitions on a physical drive.
/                   Added a capability of extending file size to f_lseek().
/                   Fixed an endian sensitive code in f_mkfs().
/ Apr 01,'08 R0.06  Added fputc(), fputs(), fprintf() and fgets().
/ Apr 01,'09 R0.07  Merged Tiny-FatFs as a configuration option. (_FS_TINY)
/                   Added long file name feature.
/                   Added multiple code page feature.
/                   Added rewind option to f_readdir().
/ Nov 03,'09 R0.07e Changed f_readdir() to return the SFN with
/                   always upper case on non-LFN cfg.
/ May 15,'10 R0.08  Added a memory configuration option. (_USE_LFN = 3)
/                   Added fast seek feature. (_USE_FASTSEEK)
/ Aug 16,'10 R0.08a Added f_getcwd(). (_FS_RPATH = 2)
/                   Added sector erase feature. (_USE_ERASE)
 * Oct 12,'12  406  modified for weAutSys weAut_01 (ATmega)
 *             424  added (made non static) and sub-divided some functions
 *                  needed to make shorter stints (between yields) for
 *                  non-preemptive runtime
 * Nov 14,'12  432  added file name length in FILINFO
 *
 * ff.c 2 2017-01-25 16:49:03Z albrecht
 */

/**
 * \file ff.c
 *
 *  FatFs - FAT file system module  (Grande FatFs module)
 *
 *  Copyright (C) 2012, ChaN, all rights reserved.
 *
 *  Modification for weAutSys by A. Weinert <br />
 *  Copyright (c) 2012 Albrecht Weinert <br />
 *  weinert - automation, Bochum <br />
 *  weinert-automation.de  weAut.de  a-weinert.de
 *
 *  See the notes on copyright and modifications in file
 *  \ref ff.h "fatFS/ff.h".
 *
 * 2
 * 25.01.2017
 */

#include "fatFS/ff.h"     // FatFs configurations and declarations

#include "fatFS/diskio.h" // Declarations of low level disk I/O functions
#include "we-aut_sys/utils.h"
#include "we-aut_sys/ll_system.h"

#if DEBUG_SMC
#  include "we-aut_sys/log_streams.h"
#  include "we-aut_sys/smc.h"
#endif

#include <string.h>

//---    Module's private definitions -------------------

#if _FATFS != 4004   // Revision ID
#error Wrong include file (ff.h).
#endif

#define ABORT(fs, res) { fp->flag |= FA__ERROR; return res; }


/* DBCS code ranges and SBCS extend char conversion table */

#if _CODE_PAGE == 932 // Japanese Shift-JIS
#define _DF1S   0x81   /* DBC 1st byte range 1 start */
#define _DF1E   0x9F   /* DBC 1st byte range 1 end */
#define _DF2S   0xE0   /* DBC 1st byte range 2 start */
#define _DF2E   0xFC   /* DBC 1st byte range 2 end */
#define _DS1S   0x40   /* DBC 2nd byte range 1 start */
#define _DS1E   0x7E   /* DBC 2nd byte range 1 end */
#define _DS2S   0x80   /* DBC 2nd byte range 2 start */
#define _DS2E   0xFC   /* DBC 2nd byte range 2 end */

#elif _CODE_PAGE == 936  // Simplified Chinese GBK
#define _DF1S   0x81
#define _DF1E   0xFE
#define _DS1S   0x40
#define _DS1E   0x7E
#define _DS2S   0x80
#define _DS2E   0xFE

#elif _CODE_PAGE == 949  // Korean
#define _DF1S   0x81
#define _DF1E   0xFE
#define _DS1S   0x41
#define _DS1E   0x5A
#define _DS2S   0x61
#define _DS2E   0x7A
#define _DS3S   0x81
#define _DS3E   0xFE

#elif _CODE_PAGE == 950  // Traditional Chinese Big5
#define _DF1S   0x81
#define _DF1E   0xFE
#define _DS1S   0x40
#define _DS1E   0x7E
#define _DS2S   0xA1
#define _DS2E   0xFE

#elif _CODE_PAGE == 437 // U.S. (OEM)
#define _DF1S   0
#define _EXCVT {0x80,0x9A,0x90,0x41,0x8E,0x41,0x8F,0x80,0x45,0x45,0x45,0x49,0x49,0x49,0x8E,0x8F,0x90,0x92,0x92,0x4F,0x99,0x4F,0x55,0x55,0x59,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F, \
            0x41,0x49,0x4F,0x55,0xA5,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0x21,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 720  // Arabic (OEM)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x45,0x41,0x84,0x41,0x86,0x43,0x45,0x45,0x45,0x49,0x49,0x8D,0x8E,0x8F,0x90,0x92,0x92,0x93,0x94,0x95,0x49,0x49,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F, \
            0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 737 // Greek (OEM)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x92,0x92,0x93,0x94,0x95,0x96,0x97,0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87, \
            0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0xAA,0x92,0x93,0x94,0x95,0x96,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0x97,0xEA,0xEB,0xEC,0xE4,0xED,0xEE,0xE7,0xE8,0xF1,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 775  // Baltic (OEM)
#define _DF1S   0
#define _EXCVT {0x80,0x9A,0x91,0xA0,0x8E,0x95,0x8F,0x80,0xAD,0xED,0x8A,0x8A,0xA1,0x8D,0x8E,0x8F,0x90,0x92,0x92,0xE2,0x99,0x95,0x96,0x97,0x97,0x99,0x9A,0x9D,0x9C,0x9D,0x9E,0x9F, \
            0xA0,0xA1,0xE0,0xA3,0xA3,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xB5,0xB6,0xB7,0xB8,0xBD,0xBE,0xC6,0xC7,0xA5,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xE0,0xE1,0xE2,0xE3,0xE5,0xE5,0xE6,0xE3,0xE8,0xE8,0xEA,0xEA,0xEE,0xED,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 850  //  Multilingual Latin 1 (OEM)
#define _DF1S   0
#define _EXCVT {0x80,0x9A,0x90,0xB6,0x8E,0xB7,0x8F,0x80,0xD2,0xD3,0xD4,0xD8,0xD7,0xDE,0x8E,0x8F,0x90,0x92,0x92,0xE2,0x99,0xE3,0xEA,0xEB,0x59,0x99,0x9A,0x9D,0x9C,0x9D,0x9E,0x9F, \
            0xB5,0xD6,0xE0,0xE9,0xA5,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0x21,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC7,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xE0,0xE1,0xE2,0xE3,0xE5,0xE5,0xE6,0xE7,0xE7,0xE9,0xEA,0xEB,0xED,0xED,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 852   // Latin 2 (OEM)
#define _DF1S   0
#define _EXCVT {0x80,0x9A,0x90,0xB6,0x8E,0xDE,0x8F,0x80,0x9D,0xD3,0x8A,0x8A,0xD7,0x8D,0x8E,0x8F,0x90,0x91,0x91,0xE2,0x99,0x95,0x95,0x97,0x97,0x99,0x9A,0x9B,0x9B,0x9D,0x9E,0x9F, \
            0xB5,0xD6,0xE0,0xE9,0xA4,0xA4,0xA6,0xA6,0xA8,0xA8,0xAA,0x8D,0xAC,0xB8,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBD,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC6,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD1,0xD1,0xD2,0xD3,0xD2,0xD5,0xD6,0xD7,0xB7,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xE0,0xE1,0xE2,0xE3,0xE3,0xD5,0xE6,0xE6,0xE8,0xE9,0xE8,0xEB,0xED,0xED,0xDD,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xEB,0xFC,0xFC,0xFE,0xFF}

#elif _CODE_PAGE == 855  //  Cyrillic (OEM)
#define _DF1S   0
#define _EXCVT {0x81,0x81,0x83,0x83,0x85,0x85,0x87,0x87,0x89,0x89,0x8B,0x8B,0x8D,0x8D,0x8F,0x8F,0x91,0x91,0x93,0x93,0x95,0x95,0x97,0x97,0x99,0x99,0x9B,0x9B,0x9D,0x9D,0x9F,0x9F, \
            0xA1,0xA1,0xA3,0xA3,0xA5,0xA5,0xA7,0xA7,0xA9,0xA9,0xAB,0xAB,0xAD,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB6,0xB6,0xB8,0xB8,0xB9,0xBA,0xBB,0xBC,0xBE,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC7,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD1,0xD1,0xD3,0xD3,0xD5,0xD5,0xD7,0xD7,0xDD,0xD9,0xDA,0xDB,0xDC,0xDD,0xE0,0xDF, \
            0xE0,0xE2,0xE2,0xE4,0xE4,0xE6,0xE6,0xE8,0xE8,0xEA,0xEA,0xEC,0xEC,0xEE,0xEE,0xEF,0xF0,0xF2,0xF2,0xF4,0xF4,0xF6,0xF6,0xF8,0xF8,0xFA,0xFA,0xFC,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 857  // Turkish (OEM)
#define _DF1S   0
#define _EXCVT {0x80,0x9A,0x90,0xB6,0x8E,0xB7,0x8F,0x80,0xD2,0xD3,0xD4,0xD8,0xD7,0x98,0x8E,0x8F,0x90,0x92,0x92,0xE2,0x99,0xE3,0xEA,0xEB,0x98,0x99,0x9A,0x9D,0x9C,0x9D,0x9E,0x9E, \
            0xB5,0xD6,0xE0,0xE9,0xA5,0xA5,0xA6,0xA6,0xA8,0xA9,0xAA,0xAB,0xAC,0x21,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC7,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xE0,0xE1,0xE2,0xE3,0xE5,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xEB,0xDE,0x59,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 858  // Multilingual Latin 1 + Euro (OEM)
#define _DF1S   0
#define _EXCVT {0x80,0x9A,0x90,0xB6,0x8E,0xB7,0x8F,0x80,0xD2,0xD3,0xD4,0xD8,0xD7,0xDE,0x8E,0x8F,0x90,0x92,0x92,0xE2,0x99,0xE3,0xEA,0xEB,0x59,0x99,0x9A,0x9D,0x9C,0x9D,0x9E,0x9F, \
            0xB5,0xD6,0xE0,0xE9,0xA5,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0x21,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC7,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD1,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xE0,0xE1,0xE2,0xE3,0xE5,0xE5,0xE6,0xE7,0xE7,0xE9,0xEA,0xEB,0xED,0xED,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 862  // Hebrew (OEM)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F, \
            0x41,0x49,0x4F,0x55,0xA5,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0x21,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 866  // Russian (OEM)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F, \
            0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0x90,0x91,0x92,0x93,0x9d,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F,0xF0,0xF0,0xF2,0xF2,0xF4,0xF4,0xF6,0xF6,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 874  // Thai (OEM, Windows)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F, \
            0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 1250 // Central Europe (Windows)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x8A,0x9B,0x8C,0x8D,0x8E,0x8F, \
            0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xA3,0xB4,0xB5,0xB6,0xB7,0xB8,0xA5,0xAA,0xBB,0xBC,0xBD,0xBC,0xAF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xF7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xFF}

#elif _CODE_PAGE == 1251 // Cyrillic (Windows)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x82,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x80,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x8A,0x9B,0x8C,0x8D,0x8E,0x8F, \
            0xA0,0xA2,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB2,0xA5,0xB5,0xB6,0xB7,0xA8,0xB9,0xAA,0xBB,0xA3,0xBD,0xBD,0xAF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF}

#elif _CODE_PAGE == 1252 // Latin 1 (Windows)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0xAd,0x9B,0x8C,0x9D,0xAE,0x9F, \
            0xA0,0x21,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xF7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0x9F}

#elif _CODE_PAGE == 1253 // Greek (Windows)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F, \
            0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xA2,0xB8,0xB9,0xBA, \
            0xE0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xF2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xFB,0xBC,0xFD,0xBF,0xFF}

#elif _CODE_PAGE == 1254 // Turkish (Windows)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x8A,0x9B,0x8C,0x9D,0x9E,0x9F, \
            0xA0,0x21,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xF7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0x9F}

#elif _CODE_PAGE == 1255 // Hebrew (Windows)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F, \
            0xA0,0x21,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 1256 // Arabic (Windows)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x8C,0x9D,0x9E,0x9F, \
            0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0x41,0xE1,0x41,0xE3,0xE4,0xE5,0xE6,0x43,0x45,0x45,0x45,0x45,0xEC,0xED,0x49,0x49,0xF0,0xF1,0xF2,0xF3,0x4F,0xF5,0xF6,0xF7,0xF8,0x55,0xFA,0x55,0x55,0xFD,0xFE,0xFF}

#elif _CODE_PAGE == 1257 // Baltic (Windows)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F, \
            0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xA8,0xB9,0xAA,0xBB,0xBC,0xBD,0xBE,0xAF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xF7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xFF}

#elif _CODE_PAGE == 1258 // Vietnam (OEM, Windows)
#define _DF1S   0
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0xAC,0x9D,0x9E,0x9F, \
            0xA0,0x21,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
            0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xEC,0xCD,0xCE,0xCF,0xD0,0xD1,0xF2,0xD3,0xD4,0xD5,0xD6,0xF7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xFE,0x9F}

#elif _CODE_PAGE == 1  // ASCII (for only non-LFN configuration)
#if _USE_LFN
#error Cannot use LFN feature without valid code page.
#endif
#define _DF1S   0

#else
#error Unknown code page
#endif


/* Character code support macros */
#define isUpper(c)   (((c)>='A')&&((c)<='Z'))
#define isLower(c)   (((c)>='a')&&((c)<='z'))
#define isDigit(c)   (((c)>='0')&&((c)<='9'))

#if _DF1S      /* Code page is DBCS */

#ifdef _DF2S   /* Two 1st byte areas */
#define IsDBCS1(c)   (((uint8_t)(c) >= _DF1S && (uint8_t)(c) <= _DF1E) || ((uint8_t)(c) >= _DF2S && (uint8_t)(c) <= _DF2E))
#else         /* One 1st byte area */
#define IsDBCS1(c)   ((uint8_t)(c) >= _DF1S && (uint8_t)(c) <= _DF1E)
#endif

#ifdef _DS3S   /* Three 2nd byte areas */
#define IsDBCS2(c)   (((uint8_t)(c) >= _DS1S && (uint8_t)(c) <= _DS1E) || ((uint8_t)(c) >= _DS2S && (uint8_t)(c) <= _DS2E) || ((uint8_t)(c) >= _DS3S && (uint8_t)(c) <= _DS3E))
#else         /* Two 2nd byte areas */
#define IsDBCS2(c)   (((uint8_t)(c) >= _DS1S && (uint8_t)(c) <= _DS1E) || ((uint8_t)(c) >= _DS2S && (uint8_t)(c) <= _DS2E))
#endif

#else         /* Code page is SBCS */

#define IsDBCS1(c)   0
#define IsDBCS2(c)   0

#endif /* _DF1S */


/* Name status flags */
#define NS      11   // Index of name status byte in fn[]
#define NS_LOSS 0x01 // Out of 8.3 format
#define NS_LFN  0x02 // Force to create LFN entry
#define NS_LAST 0x04 // Last segment
#define NS_BODY 0x08 // Lower case flag (body)
#define NS_EXT  0x10 // Lower case flag (ext)
#define NS_DOT  0x20 // Dot entry


/* FAT sub-type boundaries
   Note that the FAT specification by Microsoft says 4085
   but Windows works with 4087! */
#define MIN_FAT16  4086   // Minimum number of clusters for FAT16
#define MIN_FAT32  65526  // Minimum number of clusters for FAT32


/* FatFS refers the members in the FAT structures as byte array instead of
   structure member because the structure is not binary compatible between
   different platforms */

#define BS_jmpBoot       0  // Jump instruction (3)
#define BS_OEMName       3  // OEM name (8)
#define BPB_BytsPerSec  11  // Sector size [byte] (2)
#define BPB_SecPerClus  13  // Cluster size [sector] (1)
#define BPB_RsvdSecCnt  14  // Size of reserved area [sector] (2)
#define BPB_NumFATs     16  // Number of FAT copies (1)
#define BPB_RootEntCnt  17  // Number of root dir entries for FAT12/16 (2)
#define BPB_TotSec16    19  // Volume size [sector] (2)
#define BPB_Media       21  // Media descriptor (1)
#define BPB_FATSz16     22  // FAT size [sector] (2)
#define BPB_SecPerTrk   24  // Track size [sector] (2)
#define BPB_NumHeads    26  // Number of heads (2)
#define BPB_HiddSec     28  // Number of special hidden sectors (4)
#define BPB_TotSec32    32  // Volume size [sector] (4)
#define BS_DrvNum       36  // Physical drive number (2)
#define BS_BootSig      38  // Extended boot signature (1)
#define BS_VolID        39  // Volume serial number (4)
#define BS_VolLab       43  // Volume label (8)
#define BS_FilSysType   54  // File system type  (1)
#define BPB_FATSz32     36  // FAT size [sector] (4)
#define BPB_ExtFlags    40  // Extended flags (2)
#define BPB_FSVer       42  // File system version (2)
#define BPB_RootClus    44  // Root dir first cluster  (4)
#define BPB_FSInfo      48  // Offset of FSInfo sector (2)
#define BPB_BkBootSec   50  // Offset of backup boot sector (2)
#define BS_DrvNum32     64  // Physical drive number  (2)
#define BS_BootSig32    66  // Extended boot signature (1)
#define BS_VolID32      67  // Volume serial number (4)
#define BS_VolLab32     71  // Volume label (8)
#define BS_FilSysType32 82  // File system type (1)
#define  FSI_LeadSig    0   // FSI: Leading signature (4)
#define  FSI_StrucSig   484 // FSI: Structure signature (4)
#define  FSI_Free_Count 488 // FSI: Number of free clusters (4)
#define  FSI_Nxt_Free   492 // FSI: Last allocated cluster  (4)
#define MBR_Table       446 // MBR: Partition table offset  (2)
#define   SZ_PTE        16  // MBR: Size of a partition table entry
#define BS_55AA         510 // Boot sector signature (2)

#define   DIR_Name      0   // Short file name (11)
#define   DIR_Attr      11  // Attribute (1)
#define   DIR_NTres     12  // NT flag (1)
#define DIR_CrtTimeTenth   13 // Created time sub-second (1)
#define   DIR_CrtTime   14  // Created time (2)
#define   DIR_CrtDate   16  // Created date (2)
#define DIR_LstAccDate  18  // Last accessed date (2)
#define   DIR_FstClusHI 20  // Higher 16-bit of first cluster (2)
#define   DIR_WrtTime   22  // Modified time (2)
#define   DIR_WrtDate   24  // Modified date (2)
#define   DIR_FstClusLO 26  // Lower 16-bit of first cluster (2)
#define   DIR_FileSize  28  // File size (4)
#define   LDIR_Ord      0   // LFN entry order and LLE flag (1)
#define   LDIR_Attr     11  // LFN attribute (1)
#define   LDIR_Type     12  // LFN type (1)
#define   LDIR_Chksum   13  // Sum of corresponding SFN entry
#define   LDIR_FstClusLO  26 // Filled by zero (0)
#define   SZ_DIR        32   // Size of a directory entry
#define   LLE           0x40 // Last long entry flag in LDIR_Ord
#define   DDE           0xE5 // Deleted directory entry mark in DIR_Name[0]
#define   NDDE          0x05 // Replacement of the character collides with DDE



// ---- Module private work area  ------------------------------
// Note that uninitialized static variables are set 0 at start-up.

#if _VOLUMES
static FATFS *FatFs[_VOLUMES]; // Pointer to the file system objects (logical drives)
#else
#error Number of volumes must not be 0.
#endif

static uint16_t Fsid; // File system mount ID

#if _FS_RPATH
static uint8_t CurrVol; // Current drive
#endif

#if _USE_LFN == 0    // No LFN feature (i.e. short file names only)
#define  DEF_NAMEBUF    uint8_t sfn[12]
#define INIT_BUF(dobj) (dobj).fn = sfn
#define  FREE_BUF()

#elif _USE_LFN == 1   // LFN feature with static working buffer
static WCHAR LfnBuf[_MAX_LFN + 1];
#define   DEF_NAMEBUF uint8_t sfn[12]
#define INIT_BUF(dobj)      { (dobj).fn = sfn; (dobj).lfn = LfnBuf; }
#define   FREE_BUF()

#elif _USE_LFN == 2  // LFN feature with dynamic working buffer on the stack
#define   DEF_NAMEBUF         uint8_t sfn[12]; WCHAR lbuf[_MAX_LFN + 1]
#define INIT_BUF(dobj)      { (dobj).fn = sfn; (dobj).lfn = lbuf; }
#define   FREE_BUF()

#elif _USE_LFN == 3   // LFN feature with dynamic working buffer on the heap
#define   DEF_NAMEBUF         uint8_t sfn[12]; WCHAR *lfn
#define INIT_BUF(dobj)      { lfn = ff_memalloc((_MAX_LFN + 1) * 2); \
                       if (!lfn) return (FR_NOT_ENOUGH_CORE); \
                       (dobj).lfn = lfn;   (dobj).fn = sfn; }
#define   FREE_BUF()         ff_memfree(lfn)

#else
#error Wrong LFN configuration.
#endif


//-----   private functions     -------------------------------------------


// memory bulk functions  redefined back to C standard / AVR-GCC library

#define mem_cpy(dst, src, cnt) memcpy(dst, src, cnt)

#define mem_set(dst, val, cnt) memset(dst, val, cnt)

#define mem_cmp(dst, src, cnt) memcmp(dst, src, cnt)

/* Check if chr is contained in the string
 *
 * redefined to OS / weAutSys function
 * todo check to switch to the flush / programme memory variant
 * \ref isContainedIn_P
 */
#define chk_chr(str, chr) isContainedIn(str, chr)


/* Change window offset
 *
 * \param fs File system object
 * \param sector  Sector number to make appearance in the fs->win[]
 *                Move to zero only writes back dirty window
 */
static FRESULT move_window(FATFS *fs, uint32_t sector){
   uint32_t wsect = fs->winsect;
   if (wsect != sector) { // Changed current window ( = sector)
#if !_FS_READONLY
      if (fs->wflag) { // Write back dirty window if needed
         if (disk_write(fs->drv, fs->win, wsect)) return FR_DISK_ERR;
         fs->wflag = 0;
         if (wsect < (fs->fatbase + fs->fsize)) { // in FAT area
            uint8_t nf;
            for (nf = fs->n_fats; nf > 1; --nf) { // on  all FAT copies
               wsect += fs->fsize;
               disk_write(fs->drv, fs->win, wsect);
            } // for : Reflect the change to all FAT copies
         }
      }
#endif
      if (sector) {
         if (disk_read(fs->drv, fs->win, sector)) return FR_DISK_ERR;
         fs->winsect = sector;
      }
   }
   return FR_OK;
} // move_window(FATFS *, uint32_t)


#if !_FS_READONLY
/*  Clean-up cached data
 *
 * \param   fs  File system object (pointer to)
 * \return      FR_OK: successful, FR_DISK_ERR: failed
 */
static FRESULT sync(FATFS *fs){
   FRESULT res = move_window(fs, 0);
   if (res == FR_OK) {
      // Update FSInfo sector if needed
      if (fs->fs_type == FS_FAT32 && fs->fsi_flag) {
         fs->winsect = 0;
         // Create FSInfo structure
         mem_set(fs->win, 0, 512);
         ST_WORD(fs->win+BS_55AA, 0xAA55);
         ST_DWORD(fs->win+FSI_LeadSig, 0x41615252);
         ST_DWORD(fs->win+FSI_StrucSig, 0x61417272);
         ST_DWORD(fs->win+FSI_Free_Count, fs->free_clust);
         ST_DWORD(fs->win+FSI_Nxt_Free, fs->last_clust);
         // Write it into the FSInfo sector
         disk_write(fs->drv, fs->win, fs->fsi_sector);
         fs->fsi_flag = 0;
      }
   }
   return res;
} // sync
#endif



/* Get sector# from cluster#
 *
 * \param   fs  File system object (pointer to)
 * \param  clst Cluster# to be converted
 * \return     !=0: Sector number, 0: Failed - invalid cluster#
 */
uint32_t clust2sect(FATFS *fs, uint32_t clst){
   clst -= 2;
   if (clst >= (fs->n_fatent - 2)) return 0; // Invalid cluster#
   return clst * fs->csize + fs->database;
} // clust2sect




/* FAT access - Read value of a FAT entry
 *
 * \param   fs  File system object (pointer to)
 * \param  clst  Cluster# to get the link information
 * \return   0xFFFFFFFF: Disk error, 1: internal error,
 *           else: Cluster status
 */
uint32_t get_fat(FATFS *fs, uint32_t clst){
   if (clst < 2 || clst >= fs->n_fatent) return 1;  // Check range

   uint16_t wc, bc;
   uint8_t *p;

   switch (fs->fs_type) {
   case FS_FAT12 :
      bc = (uint16_t)clst; bc += bc / 2;
      if (move_window(fs, fs->fatbase + div16SSZ(bc))) break;
      wc = fs->win[modSSZ(bc)]; bc++;
      if (move_window(fs, fs->fatbase + div16SSZ(bc))) break;
      wc |= fs->win[modSSZ(bc)] << 8;
      return (clst & 1) ? (wc >> 4) : (wc & 0xFFF);

   case FS_FAT16 :
      if (move_window(fs, fs->fatbase + divSSZ2(clst))) break;
      p = &fs->win[modSSZ(clst * 2)];
      return LD_WORD(p);

   case FS_FAT32 :
//      if (move_window(fs, fs->fatbase + (clst / (SSZ / 4)))) break;
      if (move_window(fs, fs->fatbase + divSSZ4(clst))) break;
      p = &fs->win[modSSZ(clst * 4)];
      return LD_DWORD(p) & 0x0FFFFFFF;
   }

   return 0xFFFFFFFF;   /* An error occurred at the disk I/O layer */
} // get_fat(FATFS *, uint32_t)



#if !_FS_READONLY
/*  FAT access - Change value of a FAT entry
 *
 * \param   fs   File system object (pointer to)
 * \param  clst  Cluster# to be changed in range of 2 to fs->n_fatent - 1
 * \param   val  New value to mark the cluster
 * \return FR_INT_ERR or return of move_window
 */
FRESULT put_fat(FATFS *fs, uint32_t clst,  uint32_t val){
   FRESULT res;
   if (clst < 2 || clst >= fs->n_fatent) { // Check range
      res = FR_INT_ERR;
   } else {
      uint16_t bc;
      uint8_t *p;
      switch (fs->fs_type) {
      case FS_FAT12 :
         bc = (uint16_t)clst; bc += bc / 2;
         res = move_window(fs, fs->fatbase + div16SSZ(bc));
         if (res != FR_OK) break;
         p = &fs->win[bc % SSZ];
         *p = (clst & 1) ? ((*p & 0x0F) | ((uint8_t)val << 4)) : (uint8_t)val;
         bc++;
         fs->wflag = 1;
         res = move_window(fs, fs->fatbase + div16SSZ(bc));
         if (res != FR_OK) break;
         p = &fs->win[bc % SSZ];
         *p = (clst & 1) ? (uint8_t)(val >> 4)
                         : ((*p & 0xF0) | ((uint8_t)(val >> 8) & 0x0F));
         break;

      case FS_FAT16 :
         res = move_window(fs, fs->fatbase + divSSZ2(clst));
         if (res != FR_OK) break;
         p = &fs->win[clst * 2 % SSZ];
         ST_WORD(p, (uint16_t)val);
         break;

      case FS_FAT32 :
         res = move_window(fs, fs->fatbase + divSSZ4(clst));
         if (res != FR_OK) break;
         p = &fs->win[clst * 4 % SSZ];
         val |= LD_DWORD(p) & 0xF0000000;
         ST_DWORD(p, val);
         break;

      default :
         res = FR_INT_ERR;
         break;
      } // switch fs->fs_type
      fs->wflag = 1;
   }
   return res;
} // put_fat(FATFS *, 2*uint32_t)
#endif /* !_FS_READONLY */



#if ! _FS_READONLY
/* FAT handling - Remove a cluster chain
 *
 * \param   fs  File system object (pointer to)
 * \param  clst  Cluster# to remove a chain from
 * \return   FR_INT_ERR FR_OK 0xFFFFFFFF: Disk error, 1: internal error,
 *           else: Cluster status
 */
static FRESULT remove_chain(FATFS *fs, uint32_t clst){
   FRESULT res;
#if _USE_ERASE
   uint32_t scl = clst, ecl = clst, rt[2];
#endif

   if (clst < 2 || clst >= fs->n_fatent) {   /* Check range */
      res = FR_INT_ERR;

   } else {
      res = FR_OK;
      while (clst < fs->n_fatent) { // Not a last link?
         uint32_t nxt = get_fat(fs, clst);  // Get cluster status
         if (nxt == 0) break; // Empty cluster?
         if (nxt == 1) { res = FR_INT_ERR; break; } // Internal error?
         if (nxt == 0xFFFFFFFF) { res = FR_DISK_ERR; break; } // Disk error?
         res = put_fat(fs, clst, 0);  // Mark the cluster "empty"
         if (res != FR_OK) break;
         if (fs->free_clust != 0xFFFFFFFF) { // Update FSInfo
            fs->free_clust++;
            fs->fsi_flag = 1;
         }
#if _USE_ERASE
         if (ecl + 1 == nxt) {   /* Is next cluster contiguous? */
            ecl = nxt;
         } else {            /* End of contiguous clusters */ 
            rt[0] = clust2sect(fs, scl);               /* Start sector */
            rt[1] = clust2sect(fs, ecl) + fs->csize - 1;   /* End sector */
            disk_ioctl(fs->drv, CTRL_ERASE_SECTOR, rt);      /* Erase the block */
            scl = ecl = nxt;
         }
#endif
         clst = nxt;  //Next cluster
      }
   }
   return res;
} // remove_chain(FATFS *, uint32_t)


/*  FAT handling - Stretch or Create a cluster chain
 *
 * \param   fs  File system object (pointer to)
 * \param  clst  Cluster# to stretch. 0 means create a new chain.
 * \return   0: No free cluster, 1: Internal error, 0xFFFFFFFF: Disk error,
 *            >=2: New cluster#
 */
static uint32_t create_chain(FATFS *fs, uint32_t clst){
   uint32_t cs, scl;
   FRESULT res;
   if (clst == 0) { // Create a new chain
      scl = fs->last_clust; // Get suggested start point
      if (!scl || scl >= fs->n_fatent) scl = 1;
   } else { // Stretch the current chain
      cs = get_fat(fs, clst);  // Check the cluster status
      if (cs < 2) return 1;   // It is an invalid cluster
      if (cs < fs->n_fatent) return cs; // It is already followed by next cluster
      scl = clst;
   }

   uint32_t ncl = scl; // Start cluster
   for (;;) {
      ncl++;  // Next cluster
      if (ncl >= fs->n_fatent) {  // Wrap around
         ncl = 2;
         if (ncl > scl) return 0;  // No free cluster
      }
      cs = get_fat(fs, ncl); // Get the cluster status
      if (cs == 0) break;   // Found a free cluster
      if (cs == 0xFFFFFFFF || cs == 1) // An error occurred
         return cs;
      if (ncl == scl) return 0;  // No free cluster
   }

   res = put_fat(fs, ncl, 0x0FFFFFFF); // Mark the new cluster "last link"
   if (res == FR_OK && clst != 0) {
      res = put_fat(fs, clst, ncl); // Link it to the previous one if needed
   }
   if (res == FR_OK) {
      fs->last_clust = ncl; // Update FSINFO
      if (fs->free_clust != 0xFFFFFFFF) {
         fs->free_clust--;
         fs->fsi_flag = 1;
      }
   } else {
      ncl = (res == FR_DISK_ERR) ? 0xFFFFFFFF : 1;
   }

   return ncl; // Return new cluster number or error code
} // create_chain(FATFS *, uint32_t)
#endif /* !_FS_READONLY */



/*-----------------------------------------------------------------------*/
/* FAT handling - Convert offset into cluster with link map table        */
/*-----------------------------------------------------------------------*/

#if _USE_FASTSEEK
static uint32_t clmt_clust(   /* <2:Error, >=2:Cluster number */
   FIL* fp,      /* Pointer to the file object */
   uint32_t ofs){     /* File offset to be converted to cluster# */
   uint32_t cl, ncl, *tbl;

   tbl = fp->cltbl + 1;   /* Top of CLMT */
   cl = ofs / SSZ / fp->fs->csize;   /* Cluster order from top of the file */
   for (;;) {
      ncl = *tbl++;         /* Number of cluters in the fragment */
      if (!ncl) return 0;      /* End of table? (error) */
      if (cl < ncl) break;   /* In this fragment? */
      cl -= ncl; tbl++;      /* Next fragment */
   }
   return cl + *tbl;   /* Return the cluster number */
}
#endif   /* _USE_FASTSEEK */


/*  Directory handling - Set directory index
 *
 * \param   dj  directory object (pointer to)
 * \param  idx  index of directory table
 * \return   FR_INT_ERR FR_DISK_ERR FR_OK
 */
static FRESULT dir_sdi(DIR * dj, uint16_t idx){
   dj->index = idx;
   uint32_t  clst = dj->sclust;
   if (clst == 1 || clst >= dj->fs->n_fatent) // Check start cluster range
      return FR_INT_ERR;
   if (!clst && dj->fs->fs_type == FS_FAT32) // Replace cluster# 0 with root cluster# if in FAT32 */
      clst = dj->fs->dirbase;

   if (clst == 0) { // Static table (root-dir in FAT12/16)
      dj->clust = clst;
      if (idx >= dj->fs->n_rootdir) return FR_INT_ERR; // index is out of range
      dj->sect = dj->fs->dirbase + idx / (SSZ / SZ_DIR); // Sector#
   } else {   // Dynamic table (sub-dirs or root-dir in FAT32)
      uint16_t ic = SSZ / SZ_DIR * dj->fs->csize;   // Entries per cluster
      while (idx >= ic) { // Follow cluster chain
         clst = get_fat(dj->fs, clst); // Get next cluster
         if (clst == 0xFFFFFFFF) return FR_DISK_ERR; // Disk error
         if (clst < 2 || clst >= dj->fs->n_fatent) // Reached to end of table or int error
            return FR_INT_ERR;
         idx -= ic;
      }
      dj->clust = clst;
      dj->sect = clust2sect(dj->fs, clst) + idx / (SSZ / SZ_DIR); // Sector#
   }

   dj->dir = dj->fs->win + (idx % (SSZ / SZ_DIR)) * SZ_DIR; // Ptr to the entry in the sector
   return FR_OK;   // Seek succeeded
} // dir_sdi(DIR *, uint16_t)


/* Directory handling - Move directory table index next
 *
 * \param   dj  directory object (pointer to)
 * \param  stretch 0: Do not stretch table, !=0: Stretch table if needed
 * \return   FR_NO_FILE FR_DENIED FR_DISK_ERR FR_OK
 */
static FRESULT dir_next(DIR * dj, const uint8_t stretch){
   uint16_t  i = dj->index + 1;
   if (!i || !dj->sect)  // Report EOT when index has reached 65535
      return FR_NO_FILE;

   if (!(i % (SSZ / SZ_DIR))) { // Sector changed?
      dj->sect++; // Next sector
      if (dj->clust == 0) { // Static table
         if (i >= dj->fs->n_rootdir) // Report EOT when end of table
            return FR_NO_FILE;
      } else { // Dynamic table
         if (((i / (SSZ / SZ_DIR)) & (dj->fs->csize - 1)) == 0) { // Cluster changed?
            uint32_t  clst = get_fat(dj->fs, dj->clust); // Get next cluster
            if (clst <= 1) return FR_INT_ERR;
            if (clst == 0xFFFFFFFF) return FR_DISK_ERR;
            if (clst >= dj->fs->n_fatent) { // When it reached end of dynamic table
#if !_FS_READONLY
               uint8_t c;
               if (!stretch) return FR_NO_FILE; // When do not stretch, report EOT
               clst = create_chain(dj->fs, dj->clust); // Stretch cluster chain
               if (clst == 0) return FR_DENIED; // No free cluster
               if (clst == 1) return FR_INT_ERR;
               if (clst == 0xFFFFFFFF) return FR_DISK_ERR;
               /* Clean-up stretched table */
               if (move_window(dj->fs, 0)) return FR_DISK_ERR; // Flush active window
               mem_set(dj->fs->win, 0, SSZ); // Clear window buffer
               dj->fs->winsect = clust2sect(dj->fs, clst); // Cluster start sector
               for (c = 0; c < dj->fs->csize; c++) { // Fill the new cluster with 0
                  dj->fs->wflag = 1;
                  if (move_window(dj->fs, 0)) return FR_DISK_ERR;
                  dj->fs->winsect++;
               }
               dj->fs->winsect -= c; // Rewind window address
#else
               return FR_NO_FILE; // Report EOT
#endif
            }
            dj->clust = clst;            /* Initialize data for new cluster */
            dj->sect = clust2sect(dj->fs, clst);
         }
      }
   }
   dj->index = i;
   dj->dir = dj->fs->win + mod2pow(i , (SSZ / SZ_DIR)) * SZ_DIR;
   return FR_OK;
} // dir_next(DIR *, const uint8_t)


/*  Directory handling - Load start cluster number
 *
 * \param   fs  File system object (pointer to)
 * \param   dir  Pointer to the directory entry
 * \return   ? the cluster ? number ?
 */
static uint32_t ld_clust(FATFS * fs, uint8_t * dir){
   uint32_t cl = LD_WORD(dir + DIR_FstClusLO);
   if (fs->fs_type == FS_FAT32)
      cl |= (uint32_t)LD_WORD(dir + DIR_FstClusHI) << 16;
   return cl;
} // ld_clust(FATFS *fs, uint8_t *dir

#if !_FS_READONLY
/*  Directory handling - store start cluster number
 *
 * \param   dir  Pointer to the directory entry
 * \param   cl  Value to be set
 */
static void st_clust(uint8_t * dir,  uint32_t cl){
   ST_WORD(dir + DIR_FstClusLO, cl);
   ST_WORD(dir + DIR_FstClusHI, cl >> 16);
}
#endif

/*-----------------------------------------------------------------------*/
/* LFN handling - Test/Pick/Fit an LFN segment from/to directory entry   */
/*-----------------------------------------------------------------------*/
#if _USE_LFN
static const uint8_t LfnOfs[] = {1,3,5,7,9,14,16,18,20,22,24,28,30};   /* Offset of LFN chars in the directory entry */


static int cmp_lfn(         /* 1:Matched, 0:Not matched */
   WCHAR *lfnbuf,      /* Pointer to the LFN to be compared */
   uint8_t *dir){      /* Pointer to the directory entry containing a part of LFN */

   uint16_t i, s;
   WCHAR wc, uc;

   i = ((dir[LDIR_Ord] & ~LLE) - 1) * 13;   /* Get offset in the LFN buffer */
   s = 0; wc = 1;
   do {
      uc = LD_WORD(dir+LfnOfs[s]);   /* Pick an LFN character from the entry */
      if (wc) {   /* Last char has not been processed */
         wc = ff_wtoupper(uc);      /* Convert it to upper case */
         if (i >= _MAX_LFN || wc != ff_wtoupper(lfnbuf[i++]))   /* Compare it */
            return 0;            /* Not matched */
      } else {
         if (uc != 0xFFFF) return 0;   /* Check filler */
      }
   } while (++s < 13);            /* Repeat until all chars in the entry are checked */

   if ((dir[LDIR_Ord] & LLE) && wc && lfnbuf[i])   /* Last segment matched but different length */
      return 0;

   return 1;                  /* The part of LFN matched */
} // cmp_lfn(...)



static int pick_lfn(         /* 1:Succeeded, 0:Buffer overflow */
   WCHAR *lfnbuf,      /* Pointer to the Unicode-LFN buffer */
   uint8_t *dir){         /* Pointer to the directory entry */

   uint16_t i, s;
   WCHAR wc, uc;
   i = ((dir[LDIR_Ord] & 0x3F) - 1) * 13;   /* Offset in the LFN buffer */
   s = 0; wc = 1;
   do {
      uc = LD_WORD(dir+LfnOfs[s]);      /* Pick an LFN character from the entry */
      if (wc) {   /* Last char has not been processed */
         if (i >= _MAX_LFN) return 0;   /* Buffer overflow? */
         lfnbuf[i++] = wc = uc;         /* Store it */
      } else {
         if (uc != 0xFFFF) return 0;      /* Check filler */
      }
   } while (++s < 13);                  /* Read all character in the entry */

   if (dir[LDIR_Ord] & LLE) {            /* Put terminator if it is the last LFN part */
      if (i >= _MAX_LFN) return 0;      /* Buffer overflow? */
      lfnbuf[i] = 0;
   }

   return 1;
} // pick_lfn( .. )


#if !_FS_READONLY
static void fit_lfn(
   const WCHAR *lfnbuf,   /* Pointer to the LFN buffer */
   uint8_t *dir,            /* Pointer to the directory entry */
   uint8_t ord,            /* LFN order (1-20) */
   uint8_t sum){           /* SFN sum */

   uint16_t i, s;
   WCHAR wc;
   dir[LDIR_Chksum] = sum;         /* Set check sum */
   dir[LDIR_Attr] = AM_LFN;      /* Set attribute. LFN entry */
   dir[LDIR_Type] = 0;
   ST_WORD(dir+LDIR_FstClusLO, 0);

   i = (ord - 1) * 13;            /* Get offset in the LFN buffer */
   s = wc = 0;
   do {
      if (wc != 0xFFFF) wc = lfnbuf[i++];   /* Get an effective char */
      ST_WORD(dir+LfnOfs[s], wc);   /* Put it */
      if (!wc) wc = 0xFFFF;      /* Padding chars following last char */
   } while (++s < 13);
   if (wc == 0xFFFF || !lfnbuf[i]) ord |= LLE;   /* Bottom LFN part is the start of LFN sequence */
   dir[LDIR_Ord] = ord;         /* Set the LFN order */
}

#endif
#endif



/*-----------------------------------------------------------------------*/
/* Create numbered name                                                  */
/*-----------------------------------------------------------------------*/
#if _USE_LFN
void gen_numname(
   uint8_t *dst,       // Pointer to generated SFN
   const uint8_t *src, // Pointer to source SFN to be modified
   const WCHAR *lfn,   // Pointer to LFN
   uint16_t seq){      // Sequence number

   uint8_t ns[8], c;
   uint16_t i, j;
   mem_cpy(dst, src, 11);
   if (seq > 5) {   /* On many collisions, generate a hash number instead of sequential number */
      do seq = (seq >> 1) + (seq << 15) + (uint16_t)*lfn++; while (*lfn);
   }

   /* itoa (hexdecimal) */
   i = 7;
   do {
      c = (seq % 16) + '0';
      if (c > '9') c += 7;
      ns[i--] = c;
      seq /= 16;
   } while (seq);
   ns[i] = '~';

   /* Append the number */
   for (j = 0; j < i && dst[j] != ' '; j++) {
      if (IsDBCS1(dst[j])) {
         if (j == i - 1) break;
         j++;
      }
   }
   do {
      dst[j++] = (i < 8) ? ns[i++] : ' ';
   } while (j < 8);
} // gen_numname
#endif




/*-----------------------------------------------------------------------*/
/* Calculate sum of an SFN                                               */
/*-----------------------------------------------------------------------*/
#if _USE_LFN
static uint8_t sum_sfn( const uint8_t * dir){
   uint8_t sum = 0;
   uint16_t n = 11;
   do {sum = (sum >> 1) + (sum << 7) + *dir++;} while (--n);
   return sum;
}
#endif



/*  Directory handling - Find an object in the directory
 *
 *  \param dj     Pointer to the directory object linked to the file name
 *  \return FR_OK, return of move_window or dir_sdi, FR_NO_FILE
 */
static FRESULT dir_find(DIR * dj){
   FRESULT res = dir_sdi(dj, 0);         /* Rewind directory object */
   if (res != FR_OK) return res;

   uint8_t c, *dir;
#if _USE_LFN
   uint8_t a, ord, sum;
   ord = sum = 0xFF;
#endif
   do {
      res = move_window(dj->fs, dj->sect);
      if (res != FR_OK) break;
      dir = dj->dir; // points to the directory entry of current index
      c = dir[DIR_Name];
      if (c == 0) { res = FR_NO_FILE; break; } // Reached to end of table
#if _USE_LFN   /* LFN configuration */
      a = dir[DIR_Attr] & AM_MASK;
      if (c == DDE || ((a & AM_VOL) && a != AM_LFN)) {   /* An entry without valid data */
         ord = 0xFF;
      } else {
         if (a == AM_LFN) {         /* An LFN entry is found */
            if (dj->lfn) {
               if (c & LLE) {      /* Is it start of LFN sequence? */
                  sum = dir[LDIR_Chksum];
                  c &= ~LLE; ord = c;   /* LFN start order */
                  dj->lfn_idx = dj->index;
               }
               /* Check validity of the LFN entry and compare it with given name */
               ord = (c == ord && sum == dir[LDIR_Chksum] && cmp_lfn(dj->lfn, dir)) ? ord - 1 : 0xFF;
            }
         } else {               /* An SFN entry is found */
            if (!ord && sum == sum_sfn(dir)) break;   /* LFN matched? */
            ord = 0xFF; dj->lfn_idx = 0xFFFF;   /* Reset LFN sequence */
            if (!(dj->fn[NS] & NS_LOSS) && !mem_cmp(dir, dj->fn, 11)) break;   /* SFN matched? */
         }
      }
#else // non LFN configuration
      if (!(dir[DIR_Attr] & AM_VOL)  // is it a valid entry?
            && !mem_cmp(dir, dj->fn, 11))  break;
#endif
      res = dir_next(dj, 0); //  Next entry
   } while (res == FR_OK);
   return res;
} // dir_find(DIR *dj


#if _FS_MINIMIZE <= 1
/*  Read an object from the directory
 *
 *  \param dj     Pointer to the directory object linked to the file name
 *  \return FR_OK, return of move_window or dir_sdi, FR_NO_FILE
 */
static FRESULT dir_read(DIR * dj){
   uint8_t c, *dir;
#if _USE_LFN
   uint8_t a, ord = 0xFF, sum = 0xFF;
#endif

   FRESULT res = FR_NO_FILE;
   while (dj->sect) {
      res = move_window(dj->fs, dj->sect);
      if (res != FR_OK) break;
      dir = dj->dir;  // points to the directory entry of current index
      c = dir[DIR_Name];
      if (c == 0) { res = FR_NO_FILE; break; } // Reached to end of table
#if _USE_LFN  // LFN configuration
      a = dir[DIR_Attr] & AM_MASK;
      if (c == DDE || (!_FS_RPATH && c == '.') || ((a & AM_VOL) && a != AM_LFN)) {   /* An entry without valid data */
         ord = 0xFF;
      } else {
         if (a == AM_LFN) {         /* An LFN entry is found */
            if (c & LLE) {         /* Is it start of LFN sequence? */
               sum = dir[LDIR_Chksum];
               c &= ~LLE; ord = c;
               dj->lfn_idx = dj->index;
            }
            /* Check LFN validity and capture it */
            ord = (c == ord && sum == dir[LDIR_Chksum] && pick_lfn(dj->lfn, dir)) ? ord - 1 : 0xFF;
         } else {               /* An SFN entry is found */
            if (ord || sum != sum_sfn(dir))   /* Is there a valid LFN? */
               dj->lfn_idx = 0xFFFF;      /* It has no LFN. */
            break;
         }
      }
#else  // Non LFN configuration
      if (c != DDE && (_FS_RPATH || c != '.') // is it  a valid entry? *
            && !(dir[DIR_Attr] & AM_VOL))  break;
#endif
      res = dir_next(dj, 0);  // Next entry
      if (res != FR_OK) break;
   }
   if (res != FR_OK) dj->sect = 0;
   return res;
} //  dir_read(DIR *)
#endif


#if !_FS_READONLY
/*  Register an object to the directory
 *
 *  \param dj     Pointer to target directory with object name to be created
 *  \return FR_OK: Successful, FR_DENIED: No free entry or too many
 *                SFN collision, FR_DISK_ERR: Disk error
 */
static FRESULT dir_register(DIR * dj){
   FRESULT res;
   uint8_t c, *dir;
#if _USE_LFN // LFN configuration
   uint16_t n, ne, is;
   uint8_t sn[12], *fn, sum;
   WCHAR *lfn;


   fn = dj->fn; lfn = dj->lfn;
   mem_cpy(sn, fn, 12);

   if (_FS_RPATH && (sn[NS] & NS_DOT))      /* Cannot create dot entry */
      return FR_INVALID_NAME;

   if (sn[NS] & NS_LOSS) {         /* When LFN is out of 8.3 format, generate a numbered name */
      fn[NS] = 0; dj->lfn = 0;         /* Find only SFN */
      for (n = 1; n < 100; n++) {
         gen_numname(fn, sn, lfn, n);   /* Generate a numbered name */
         res = dir_find(dj);            /* Check if the name collides with existing SFN */
         if (res != FR_OK) break;
      }
      if (n == 100) return FR_DENIED;      /* Abort if too many collisions */
      if (res != FR_NO_FILE) return res;   /* Abort if the result is other than 'not collided' */
      fn[NS] = sn[NS]; dj->lfn = lfn;
   }

   if (sn[NS] & NS_LFN) {         /* When LFN is to be created, reserve an SFN + LFN entries. */
      for (ne = 0; lfn[ne]; ne++) ;
      ne = (ne + 25) / 13;
   } else {                  /* Otherwise reserve only an SFN entry. */
      ne = 1;
   }

   /* Reserve contiguous entries */
   res = dir_sdi(dj, 0);
   if (res != FR_OK) return res;
   n = is = 0;
   do {
      res = move_window(dj->fs, dj->sect);
      if (res != FR_OK) break;
      c = *dj->dir;            /* Check the entry status */
      if (c == DDE || c == 0) {   /* Is it a blank entry? */
         if (n == 0) is = dj->index;   /* First index of the contiguous entry */
         if (++n == ne) break;   /* A contiguous entry that required count is found */
      } else {
         n = 0;               /* Not a blank entry. Restart to search */
      }
      res = dir_next(dj, 1);      /* Next entry with table stretch */
   } while (res == FR_OK);

   if (res == FR_OK && ne > 1) {   /* Initialize LFN entry if needed */
      res = dir_sdi(dj, is);
      if (res == FR_OK) {
         sum = sum_sfn(dj->fn);   /* Sum of the SFN tied to the LFN */
         ne--;
         do {               /* Store LFN entries in bottom first */
            res = move_window(dj->fs, dj->sect);
            if (res != FR_OK) break;
            fit_lfn(dj->lfn, dj->dir, (uint8_t)ne, sum);
            dj->fs->wflag = 1;
            res = dir_next(dj, 0);   /* Next entry */
         } while (res == FR_OK && --ne);
      }
   }

#else // non LFN configuration
   res = dir_sdi(dj, 0);
   if (res == FR_OK) {
      do { // Find a blank entry for the SFN
         res = move_window(dj->fs, dj->sect);
         if (res != FR_OK) break;
         c = *dj->dir;
         if (c == DDE || c == 0) break; // Is it a blank entry?
         res = dir_next(dj, 1); // Next entry with table stretch
      } while (res == FR_OK);
   }
#endif
   if (res == FR_OK) { // Initialize the SFN entry
      res = move_window(dj->fs, dj->sect);
      if (res == FR_OK) {
         dir = dj->dir;
         mem_set(dir, 0, SZ_DIR);   // Clean the entry
         mem_cpy(dir, dj->fn, 11);  // Put SFN
#if _USE_LFN
         dir[DIR_NTres] = *(dj->fn+NS) & (NS_BODY | NS_EXT);   /* Put NT flag */
#endif
         dj->fs->wflag = 1;
      }
   }
   return res;
} // dir_register(DIR *dj)
#endif /* !_FS_READONLY */



#if ! _FS_READONLY && ! _FS_MINIMIZE
/* Remove an object from the directory
 *
 *  \param dj     Pointer to the directory (entry) to be removed
 *  \return FR_OK: Successful, FR_DISK_ERR: Disk error
 */
static FRESULT dir_remove(DIR *dj){
   FRESULT res;
#if _USE_LFN // LFN configuration
   uint16_t i;

   i = dj->index;   /* SFN index */
   res = dir_sdi(dj, (uint16_t)((dj->lfn_idx == 0xFFFF) ? i : dj->lfn_idx));   /* Goto the SFN or top of the LFN entries */
   if (res == FR_OK) {
      do {
         res = move_window(dj->fs, dj->sect);
         if (res != FR_OK) break;
         *dj->dir = DDE;         /* Mark the entry "deleted" */
         dj->fs->wflag = 1;
         if (dj->index >= i) break;   /* When reached SFN, all entries of the object has been deleted. */
         res = dir_next(dj, 0);      /* Next entry */
      } while (res == FR_OK);
      if (res == FR_NO_FILE) res = FR_INT_ERR;
   }

#else  // non LFN configuration
   res = dir_sdi(dj, dj->index);
   if (res == FR_OK) {
      res = move_window(dj->fs, dj->sect);
      if (res == FR_OK) {
         *dj->dir = DDE;  // Mark the entry "deleted"
         dj->fs->wflag = 1;
      }
   }
#endif
   return res;
} // dir_remove(DIR *dj)
#endif /* !_FS_READONLY */


/*  Pick a segment and create the object name in directory form
 *
 *  \param dj     Pointer to the directory object
 *  \param path   Pointer to pointer to the segment in the path string
 *  \return FR_OK: Successful, FR_INVALID_NAME,  FR_DISK_ERR: Disk error
 */
static FRESULT create_name(DIR * dj, const TCHAR * * path){
   SAMPLE_MIN_SP(0);
#ifdef _EXCVT
   static const uint8_t excvt[] = _EXCVT;   /* Upper conversion table for extended chars */
#endif

#if _USE_LFN // LFN configuration
   uint8_t b, cf;
   WCHAR w, *lfn;
   uint16_t i, ni, si, di;
   const TCHAR *p;

   /* Create LFN in Unicode */
   for (p = *path; *p == '/' || *p == '\\'; p++) ;   /* Strip duplicated separator */
   lfn = dj->lfn;
   si = di = 0;
   for (;;) {
      w = p[si++];               /* Get a character */
      if (w < ' ' || w == '/' || w == '\\') break;   /* Break on end of segment */
      if (di >= _MAX_LFN)            /* Reject too long name */
         return FR_INVALID_NAME;
#if !_LFN_UNICODE
      w &= 0xFF;
      if (IsDBCS1(w)) {            /* Check if it is a DBC 1st byte (always false on SBCS cfg) */
         b = (uint8_t)p[si++];         /* Get 2nd byte */
         if (!IsDBCS2(b))
            return FR_INVALID_NAME;   /* Reject invalid sequence */
         w = (w << 8) + b;         /* Create a DBC */
      }
      w = ff_convert(w, 1);         /* Convert ANSI/OEM to Unicode */
      if (!w) return FR_INVALID_NAME;   /* Reject invalid code */
#endif
      if (w < 0x80 && chk_chr("\"*:<>\?|\x7F", w)) /* Reject illegal chars for LFN */
         return FR_INVALID_NAME;
      lfn[di++] = w;               /* Store the Unicode char */
   }
   *path = &p[si];                  /* Return pointer to the next segment */
   cf = (w < ' ') ? NS_LAST : 0;      /* Set last segment flag if end of path */
#if _FS_RPATH
   if ((di == 1 && lfn[di-1] == '.') || /* Is this a dot entry? */
      (di == 2 && lfn[di-1] == '.' && lfn[di-2] == '.')) {
      lfn[di] = 0;
      for (i = 0; i < 11; i++)
         dj->fn[i] = (i < di) ? '.' : ' ';
      dj->fn[i] = cf | NS_DOT;      /* This is a dot entry */
      return FR_OK;
   }
#endif
   while (di) {                  /* Strip trailing spaces and dots */
      w = lfn[di-1];
      if (w != ' ' && w != '.') break;
      di--;
   }
   if (!di) return FR_INVALID_NAME;   /* Reject nul string */

   lfn[di] = 0;                  /* LFN is created */

   /* Create SFN in directory form */
   mem_set(dj->fn, ' ', 11);
   for (si = 0; lfn[si] == ' ' || lfn[si] == '.'; si++) ;   /* Strip leading spaces and dots */
   if (si) cf |= NS_LOSS | NS_LFN;
   while (di && lfn[di - 1] != '.') di--;   /* Find extension (di<=si: no extension) */

   b = i = 0; ni = 8;
   for (;;) {
      w = lfn[si++];               /* Get an LFN char */
      if (!w) break;               /* Break on end of the LFN */
      if (w == ' ' || (w == '.' && si != di)) {   /* Remove spaces and dots */
         cf |= NS_LOSS | NS_LFN; continue;
      }

      if (i >= ni || si == di) {      /* Extension or end of SFN */
         if (ni == 11) {            /* Long extension */
            cf |= NS_LOSS | NS_LFN; break;
         }
         if (si != di) cf |= NS_LOSS | NS_LFN;   /* Out of 8.3 format */
         if (si > di) break;         /* No extension */
         si = di; i = 8; ni = 11;   /* Enter extension section */
         b <<= 2; continue;
      }

      if (w >= 0x80) {            /* Non ASCII char */
#ifdef _EXCVT
         w = ff_convert(w, 0);      /* Unicode -> OEM code */
         if (w) w = excvt[w - 0x80];   /* Convert extended char to upper (SBCS) */
#else
         w = ff_convert(ff_wtoupper(w), 0);   /* Upper converted Unicode -> OEM code */
#endif
         cf |= NS_LFN;            /* Force create LFN entry */
      }

      if (_DF1S && w >= 0x100) {      /* Double byte char (always false on SBCS cfg) */
         if (i >= ni - 1) {
            cf |= NS_LOSS | NS_LFN; i = ni; continue;
         }
         dj->fn[i++] = (uint8_t)(w >> 8);
      } else {                  /* Single byte char */
         if (!w || chk_chr("+,;=[]", w)) {   /* Replace illegal chars for SFN */
            w = '_'; cf |= NS_LOSS | NS_LFN;/* Lossy conversion */
         } else {
            if (isUpper(w)) {      /* ASCII large capital */
               b |= 2;
            } else {
               if (isLower(w)) {   /* ASCII small capital */
                  b |= 1; w -= 0x20;
               }
            }
         }
      }
      dj->fn[i++] = (uint8_t)w;
   }

   if (dj->fn[0] == DDE) dj->fn[0] = NDDE;   /* If the first char collides with deleted mark, replace it with 0x05 */

   if (ni == 8) b <<= 2;
   if ((b & 0x0C) == 0x0C || (b & 0x03) == 0x03)   /* Create LFN entry when there are composite capitals */
      cf |= NS_LFN;
   if (!(cf & NS_LFN)) {                  /* When LFN is in 8.3 format without extended char, NT flags are created */
      if ((b & 0x03) == 0x01) cf |= NS_EXT;   /* NT flag (Extension has only small capital) */
      if ((b & 0x0C) == 0x04) cf |= NS_BODY;   /* NT flag (Filename has only small capital) */
   }

   dj->fn[NS] = cf;   /* SFN is created */

   return FR_OK;


#else // non-LFN configuration
   uint8_t b, c, d, *sfn;
   uint16_t ni, si, i;
   const char *p;

   // Create file name in directory form
   for (p = *path; *p == '/' || *p == '\\'; p++) ; // Strip duplicated separator
   sfn = dj->fn;
   mem_set(sfn, ' ', 11);
   si = i = b = 0; ni = 8;
#if _FS_RPATH
   if (p[si] == '.') { // Is this a dot entry?
      for (;;) {
         c = (uint8_t)p[si++];
         if (c != '.' || si >= 3) break;
         sfn[i++] = c;
      }
      if (c != '/' && c != '\\' && c > ' ') return FR_INVALID_NAME;
      *path = &p[si];                           /* Return pointer to the next segment */
      sfn[NS] = (c <= ' ') ? NS_LAST | NS_DOT : NS_DOT;   /* Set last segment flag if end of path */
      return FR_OK;
   }
#endif
   for (;;) {
      c = (uint8_t)p[si++];
      if (c <= ' ' || c == '/' || c == '\\') break; // Break on end of segment
      if (c == '.' || i >= ni) {
         if (ni != 8 || c != '.') return FR_INVALID_NAME;
         i = 8; ni = 11;
         b <<= 2; continue;
      }
      if (c >= 0x80) { // Extended char?
         b |= 3;  // Eliminate NT flag
#ifdef _EXCVT
         c = excvt[c - 0x80]; // Upper conversion (SBCS)
#else
#if !_DF1S   /* ASCII only cfg */
         return FR_INVALID_NAME;
#endif
#endif
      }
      if (IsDBCS1(c)) { // Check if it is a DBC 1st byte (always false on SBCS cfg)
         d = (uint8_t)p[si++];  // Get 2nd byte
         if (!IsDBCS2(d) || i >= ni - 1)   // Reject invalid DBC
            return FR_INVALID_NAME;
         sfn[i++] = c;
         sfn[i++] = d;
      } else { // Single byte code
         if (chk_chr("\"*+,:;<=>\?[]|\x7F", c)) // Reject illegal chrs for SFN
            return FR_INVALID_NAME;
         if (isUpper(c)) {         /* ASCII large capital? */
            b |= 2;
         } else {
            if (isLower(c)) {      /* ASCII small capital? */
               b |= 1; c -= 0x20;
            }
         }
         sfn[i++] = c;
      }
   }
   *path = &p[si];                  /* Return pointer to the next segment */
   c = (c <= ' ') ? NS_LAST : 0;      /* Set last segment flag if end of path */

   if (!i) return FR_INVALID_NAME;      /* Reject nul string */
   if (sfn[0] == DDE) sfn[0] = NDDE;   /* When first char collides with DDE, replace it with 0x05 */

   if (ni == 8) b <<= 2;
   if ((b & 0x03) == 0x01) c |= NS_EXT;   /* NT flag (Name extension has only small capital) */
   if ((b & 0x0C) == 0x04) c |= NS_BODY;   /* NT flag (Name body has only small capital) */

   sfn[NS] = c;      /* Store NT flag, File name is created */
   return FR_OK;
#endif
} // create_name(DIR *dj, const TCHAR **path)


#if _FS_MINIMIZE <= 1
/*  Get file information from directory entry
 *
 *  \param dj    Pointer to the directory object
 *  \param fno   Pointer to the file information to be filled
 */
static void get_fileinfo(DIR * dj, FILINFO * fno){
   uint8_t nt, *dir;
   TCHAR  c;
   TCHAR * p = fno->fname;
   if (dj->sect) {
      dir = dj->dir;
      nt = dir[DIR_NTres]; //  NT flag
      for (uint8_t i = 0; i < 8; i++) { // Copy name body
         c = dir[i];
         if (c == ' ') break;
         if (c == NDDE) c = (TCHAR)DDE;
#if _USE_LFN
         if ((nt & NS_BODY) && isUpper(c)) c += 0x20;
#endif
#if _LFN_UNICODE
         if (IsDBCS1(c) && i < 7 && IsDBCS2(dir[i+1]))
            c = (c << 8) | dir[++i];
         c = ff_convert(c, 1);
         if (!c) c = '?';
#endif
         *p++ = c;
      } // for: Copy name body
      if (dir[8] != ' ') {  // Copy name extension
         *p++ = '.';
         for (uint8_t i = 8; i < 11; i++) {
            c = dir[i];
            if (c == ' ') break;
#if _USE_LFN
            if ((nt & NS_EXT) && isUpper(c)) c += 0x20;
#endif
#if _LFN_UNICODE
            if (IsDBCS1(c) && i < 10 && IsDBCS2(dir[i+1]))
               c = (c << 8) | dir[++i];
            c = ff_convert(c, 1);
            if (!c) c = '?';
#endif
            *p++ = c;
         }
      }
      fno->fattrib = dir[DIR_Attr]; // Attribute
      fno->fsize = LD_DWORD(dir+DIR_FileSize);  // Size
      fno->fdate = LD_WORD(dir+DIR_WrtDate);    // Date
      fno->ftime = LD_WORD(dir+DIR_WrtTime);    // Time
   }
   *p = 0;  // Terminate SFN string by a \0
   fno->namLen = (uint8_t)(p - fno->fname);

#if _USE_LFN
   if (fno->lfname && fno->lfsize) {
      TCHAR *tp = fno->lfname;
      WCHAR w, *lfn;

      uint16_t i = 0;
      if (dj->sect && dj->lfn_idx != 0xFFFF) {/* Get LFN if available */
         lfn = dj->lfn;
         while ((w = *lfn++) != 0) {         /* Get an LFN char */
#if !_LFN_UNICODE
            w = ff_convert(w, 0);         /* Unicode -> OEM conversion */
            if (!w) { i = 0; break; }      /* Could not convert, no LFN */
            if (_DF1S && w >= 0x100)      /* Put 1st byte if it is a DBC (always false on SBCS cfg) */
               tp[i++] = (TCHAR)(w >> 8);
#endif
            if (i >= fno->lfsize - 1) { i = 0; break; }   /* Buffer overflow, no LFN */
            tp[i++] = (TCHAR)w;
         }
      }
      tp[i] = 0;   /* Terminate the LFN str by a \0 */
   }
#endif
} //  get_fileinfo(DIR *dj, FILINFO *fno)
#endif /* _FS_MINIMIZE <= 1 */


/* Follow a file path
 *
 *  \param dj    Pointer to the directory object to return last
 *               directory and found object
 *  \param path  Full-path string to find a file or directory
 *  \return      FR_OK(0): successful, !=0: error code
 */
static FRESULT follow_path(DIR * dj, const TCHAR * path){
   FRESULT res;
   uint8_t *dir, ns;

#if _FS_RPATH
   if (*path == '/' || *path == '\\') { /* There is a heading separator */
      path++;   dj->sclust = 0;      /* Strip it and start from the root dir */
   } else {                     /* No heading separator */
      dj->sclust = dj->fs->cdir;   /* Start from the current dir */
   }
#else
   if (*path == '/' || *path == '\\') // Strip heading separator if exist
      path++;
   dj->sclust = 0; // Start from the root dir
#endif

   if ((uint16_t)*path < ' ') { // Nul path means the start directory itself
      res = dir_sdi(dj, 0);
      dj->dir = 0;
   } else {  // Follow path
      for (;;) {
         res = create_name(dj, &path); // Get a segment
         if (res != FR_OK) break;
         res = dir_find(dj); // Find it
         ns = *(dj->fn+NS);
         if (res != FR_OK) {   // Failed to find the object
            if (res != FR_NO_FILE) break; // Abort if any hard error occurred
            // Object not found
            if (_FS_RPATH && (ns & NS_DOT)) { // If dot entry is not exit
               dj->sclust = 0; dj->dir = 0;  // It is the root dir
               res = FR_OK;
               if (!(ns & NS_LAST)) continue;
            } else {  //Could not find the object
               if (!(ns & NS_LAST)) res = FR_NO_PATH;
            }
            break;
         }
         if (ns & NS_LAST) break;   // Last segment match. Function completed.
         dir = dj->dir;  // There is next segment. Follow the sub directory
         if (!(dir[DIR_Attr] & AM_DIR)) { // Cannot follow cause it is a file
            res = FR_NO_PATH; break;
         }
         dj->sclust = ld_clust(dj->fs, dir);
      }
   }
   return res;
} //  follow_path(DIR *dj, const TCHAR *path)


/*  Load a sector and check if it is an FAT Volume Boot Record
 *
 *  \param fs    Pointer to the file system object
 *               directory and found object
 *  \param sect  Sector# (lba) to check if it is an FAT boot record
 *  \return   0: FAT-VBR, 1: Any BR but not FAT, 2: Not a BR, 3: Disk error
 */
static uint8_t check_fs(FATFS * fs, uint32_t sect){
#if DEBUG_SMC >= 1
   //           0 123456789.123456789v123456789t1 2
   char tS[] = "\n  SMC chckFS aaaaaaaa rr f.0  \n";
   eightHexs(&tS[14], sect);
#endif
   uint8_t ret = 0;
   if (disk_read(fs->drv, fs->win, sect)) { // Load boot record
      ret =  3;
      goto weGo;
   }
   if (LD_WORD(&fs->win[BS_55AA]) != 0xAA55) { //Check record signature (always placed at offset 510 even if the sector size is >512) */
      ret =  2;
      goto weGo;
   }

   if ((LD_DWORD(&fs->win[BS_FilSysType]) & 0xFFFFFF) == 0x544146) // Check "FAT" string
      goto weGo;
   if ((LD_DWORD(&fs->win[BS_FilSysType32]) & 0xFFFFFF) != 0x544146) ret = 1;

   weGo:
#if DEBUG_SMC >= 1
   twoHexs(&tS[23], smcState.lastCmdResp[0]);
   twoHexs(&tS[27], ret);
   bufLogPutSt(tS, &bufLogStreams);
#endif
   return ret;
} // check_fs(FATFS *fs, uint32_t sect)


/*  Check if there's a valid mounted file system on the drive
 *
 *  This function checks if drive number \c vol is ready and has a valid
 *  recognised file system. In that case \ref FR_OK is returned and file
 *  system operations can be used without further preparations or delays.
 *
 *  \ref FR_NO_FILESYSTEM means all else is OK but no file system has been
 *  detected yet. That means it has either not yet been looked for or the
 *  medium is not formatted with a file system usable by this fatFS
 *  implementation.
 *
 *  \ref FR_INVALID_DRIVE means no valid volume number. <br />
 *  \ref FR_NOT_ENABLED  means no FS structure assigned.  <br />
 *  \ref FR_NOT_READY  means drive down or medium removed
 *
 *  \param vol drive number 0.. (at present only 0 is valid)
 *  \return FR_OK (0); FR_INVALID_DRIVE FR_NOT_ENABLED  FR_NO_FILESYSTEM
 *          FR_NOT_READY
 *  \sa getValidFS
 */
FRESULT checkValidFS(uint8_t vol){
   if (vol >= _VOLUMES) return FR_INVALID_DRIVE;
   FATFS * fs = FatFs[vol];   // Get corresponding file system object
   if (!fs) return FR_NOT_ENABLED;  //  no file system object available
   fs->drv = vol; // bind the logical drive and a physical drive
   if (disk_initialize(vol) & (STA_NOINIT | STA_NODISK)) {
      fs->fs_type = 0;
      return FR_NOT_READY;
   }
   return fs->fs_type ? FR_OK : FR_NO_FILESYSTEM;
} // checkValidFS(uint8_t vol)


/*  Get the file system for the drive
 *
 *  Exactly like \ref checkValidFS() this function checks if drive number
 *  \c vol is ready and has a valid recognised file system. In that sense
 *  \ref getValidFS() is a substitute for \ref checkValidFS().
 *
 *  Additionally it returns the (pointer to the) file system structure
 *  assigned with drive \c vol in parameter \c fatfs if that is possible.
 *
 *  \param vol drive number 0.. (at present only 0 is valid)
 *  \param fatfs  Pointer to pointer to the file system object to return
 *  \return FR_OK (0); FR_INVALID_DRIVE FR_NOT_ENABLED  FR_NO_FILESYSTEM
 *          FR_NOT_READY
 *  \sa checkValidFS()
 */
FRESULT getValidFS(uint8_t vol, FATFS * * fatfs){
   if (vol >= _VOLUMES) return FR_INVALID_DRIVE;
   FATFS * fs = FatFs[vol];   // Get corresponding file system object
   if (fatfs) *fatfs = fs;
   if (!fs) return FR_NOT_ENABLED;  //  no file system object available
   fs->drv = vol; // bind the logical drive and a physical drive
   if (disk_initialize(vol) & (STA_NOINIT | STA_NODISK)) {
      fs->fs_type = 0;
      return FR_NOT_READY;
   }
   return fs->fs_type ? FR_OK : FR_NO_FILESYSTEM;
} // checkValidFS(uint8_t vol)


/*  Check if the drive is write protected
 *
 *  This function checks if drive number \c vol is writable. In that case
 *  \ref FR_OK is returned. This result only makes sense it / respectively
 *  requires that the drive is been checked / initialised as usable. Otherwise
 *  \ref   FR_NOT_READY may be returned.
 *
 *  For a usable but write protected drive and file \ref FR_WRITE_PROTECTED
 *  will be returned. At present this will not happen on
 *  \ref intro_sec "weAutSys" /  \ref intro_secH "weAut_01" \ref smcCom SMC
 *  implementations as those slots do not have write protect switches.
 *  Nevertheless later versions may add write protect as software feature.
 *
 *  \param vol drive number 0.. (at present only 0 is valid)
 *  \return FR_OK (0); FR_NOT_READY FR_WRITE_PROTECTED
 */
FRESULT checkWriteProtect(uint8_t vol){
   const uint8_t ds = disk_initialize(vol);
   if (ds & (STA_NOINIT | STA_NODISK))      return FR_NOT_READY;
   if (!_FS_READONLY && (ds & STA_PROTECT)) return FR_WRITE_PROTECTED;
   return FR_OK;  // The drive is writeable
} // checkWriteProtect(uint8_t vol)

/*  Look for a recognised file system on the drive
 *
 *  This function checks if drive number \c vol is ready and looks for a valid
 *  recognised file system on it. If that is or was discovered \ref FR_OK is
 *  returned and file system operations can be used without further
 *  preparations or delays.
 *
 *  \ref FR_NO_FILESYSTEM means all else is OK but no file system can be
 *  detected on that drive respectively medium. The only remedy (on present
 *  \ref intro_sec "weAutSys" /  \ref intro_secH "weAut_01" \ref smcCom SMC
 *  implementations) is to remove the medium and format it with a suitable
 *  file system (preferably FAT32).
 *
 *  \ref FR_INVALID_DRIVE means no valid volume number. <br />
 *  \ref FR_NOT_ENABLED  means no FS structure assigned.  <br />
 *  \ref FR_NOT_READY  means drive down or medium removed
 *
 *  \param vol drive number 0.. (at present only 0 is valid)
 *  \return FR_OK (0); FR_INVALID_DRIVE FR_NOT_ENABLED  FR_NO_FILESYSTEM
 *          FR_NOT_READY
 *  \internal todo make a yieldable protothread version
 */
FRESULT lookForFS(uint8_t vol){
   if (vol >= _VOLUMES) return FR_INVALID_DRIVE;
   FATFS * fs = FatFs[vol];   // Get corresponding file system object
   if (!fs) return FR_NOT_ENABLED;  // Is the file system object available?
   fs->drv = vol; // bind the logical drive and a physical drive
   if (disk_status(vol) & (STA_NOINIT | STA_NODISK)) {
      fs->fs_type = 0;
      return FR_NOT_READY;
   }
   if (fs->fs_type) return FR_OK;  // already discovered do not again
 // below is copied from ChaN's chk_mounted

#if _MAX_SS != 512      // Get disk sector size (variable sector size cfg only)
   if (disk_ioctl(fs->drv, GET_SECTOR_SIZE, &fs->ssize)) return FR_DISK_ERR;
#endif
   uint32_t bsect;
   // Search FAT partition on the drive. Supports only generic partitions, FDISK and SFD
   uint8_t fmt = check_fs(fs, bsect = 0);   // Load sector 0 and check if it is an FAT-VBR (in SFD)
   if (LD2PT(vol) && !fmt) fmt = 1; // Force non-SFD if the volume is forced partition
   if (fmt == 1) { // Not an FAT-VBR, the physical drive can be partitioned
      // Check the partition listed in the partition table
      uint8_t pi = LD2PT(vol);
      if (pi) pi--;
      uint8_t * tbl = &fs->win[MBR_Table + pi * SZ_PTE]; // Partition table
      if (tbl[4]) {  // Is the partition existing?
         bsect = LD_DWORD(&tbl[8]);   //  Partition offset in LBA
         fmt = check_fs(fs, bsect);   //  Check the partition
      }
   }
   if (fmt == 3) return FR_DISK_ERR;
   if (fmt) return FR_NO_FILESYSTEM;   // No FAT volume is found

   // An FAT volume is found. Following code initializes the file system object

   if (LD_WORD(fs->win+BPB_BytsPerSec) != SSZ) // (BPB_BytsPerSec must be equal to the physical sector size)
      return FR_NO_FILESYSTEM;

   uint32_t fasize = LD_WORD(fs->win+BPB_FATSz16);   // Number of sectors per FAT
   if (!fasize) fasize = LD_DWORD(fs->win+BPB_FATSz32);
   fs->fsize = fasize;

   uint8_t b;
   fs->n_fats = b = fs->win[BPB_NumFATs];   // Number of FAT copies
   if (b != 1 && b != 2) return FR_NO_FILESYSTEM;  // (must be 1 or 2)
   fasize *= b;      // Number of sectors for FAT area

   fs->csize = b = fs->win[BPB_SecPerClus]; // Number of sectors per cluster
   if (!b || (b & (b - 1))) return FR_NO_FILESYSTEM;   // (Must be power of 2)

   fs->n_rootdir = LD_WORD(fs->win+BPB_RootEntCnt); // Number of root directory entries
   if (mod2pow(fs->n_rootdir, SSZ / SZ_DIR)) return FR_NO_FILESYSTEM; // (BPB_RootEntCnt must be sector aligned) */

   uint32_t tsect = LD_WORD(fs->win+BPB_TotSec16);   // Number of sectors on the volume
   if (!tsect) tsect = LD_DWORD(fs->win+BPB_TotSec32);

   uint16_t nrsv = LD_WORD(fs->win+BPB_RsvdSecCnt);   // Number of reserved sectors
   if (!nrsv) return FR_NO_FILESYSTEM; // (BPB_RsvdSecCnt must not be 0)

   /* Determine the FAT sub type */
   uint32_t sysect = nrsv + fasize + fs->n_rootdir / (SSZ / SZ_DIR); // RSV+FAT+DIR
   if (tsect < sysect) return FR_NO_FILESYSTEM; // (Invalid volume size)
   uint32_t nclst = (tsect - sysect) / fs->csize;  // Number of clusters
   if (!nclst) return FR_NO_FILESYSTEM; // (Invalid volume size)
   fmt = FS_FAT12;
   if (nclst >= MIN_FAT16) fmt = FS_FAT16;
   if (nclst >= MIN_FAT32) fmt = FS_FAT32;

   // Boundaries and Limits
   fs->n_fatent = nclst + 2;         // Number of FAT entries
   fs->database = bsect + sysect;   // Data start sector
   fs->fatbase = bsect + nrsv; // FAT start sector
   uint32_t szbfat;
   if (fmt == FS_FAT32) {
      if (fs->n_rootdir) return FR_NO_FILESYSTEM;  // (BPB_RootEntCnt must be 0)
      fs->dirbase = LD_DWORD(fs->win+BPB_RootClus); // Root directory start cluster
      szbfat = fs->n_fatent * 4; // (Required FAT size)
   } else {
      if (!fs->n_rootdir) return FR_NO_FILESYSTEM; // (BPB_RootEntCnt must not be 0)
      fs->dirbase = fs->fatbase + fasize; // Root directory start sector
      szbfat = (fmt == FS_FAT16)   // (Required FAT size)
                ? fs->n_fatent * 2
                : fs->n_fatent * 3 / 2 + (fs->n_fatent & 1);
   }
   if (fs->fsize < div32SSZ(szbfat + (SSZ - 1))) // (BPB_FATSz must not be less than required)
      return FR_NO_FILESYSTEM;

#if !_FS_READONLY
   // Initialize cluster allocation information
   fs->free_clust = 0xFFFFFFFF;
   fs->last_clust = 0;

   // Get fsinfo if available
   if (fmt == FS_FAT32) {
       fs->fsi_flag = 0;
      fs->fsi_sector = bsect + LD_WORD(fs->win+BPB_FSInfo);
      if (!disk_read(fs->drv, fs->win, fs->fsi_sector) &&
         LD_WORD(fs->win+BS_55AA) == 0xAA55 &&
         LD_DWORD(fs->win+FSI_LeadSig) == 0x41615252 &&
         LD_DWORD(fs->win+FSI_StrucSig) == 0x61417272) {
            fs->last_clust = LD_DWORD(fs->win+FSI_Nxt_Free);
            fs->free_clust = LD_DWORD(fs->win+FSI_Free_Count);
      }
   }
#endif
   fs->fs_type = fmt; //  FAT sub-type
   fs->id = ++Fsid;  // File system mount ID
   fs->winsect = 0; // Invalidate sector cache
   fs->wflag = 0;
#if _FS_RPATH
   fs->cdir = 0;  // Current directory (root dir)
#endif
   return FR_OK;
} // lookForFS(uint8_t vol)

/*  Extract drive number from path
 *
 *  This function checks if \c **path points to character sequence starting
 *  with "x:" where x may be '0..9', 'a..j' or 'A..J'. If so, \c *path is
 *  incremented by 2 (to point past the :) and 0.. 9 is returned according to
 *  the first character. Otherwise 0 is returned (respectively the
 *  current drive if \ref _FS_RPATH "enabled").
 *
 * \param   path  Pointer to pointer to the path name ([d:]path)
 * \return  drive / volume number in the range 0..9 (default 0)
 */
uint8_t extractDrive(const TCHAR * * path){
   if (!path) goto weGo;
   const TCHAR *p = *path;
   if (!p) goto weGo;
   const uint8_t c = p[0];
   if (!c) goto weGo;
   if (p[1] != ':') goto weGo;
   *path = p + 2; // adjust pointer to path to behind colon
   if (c < '0') goto weGo;
   if (c <= '9') return c - '0';
   if (c < 'A') goto weGo;
   if (c <= 'J')  return c - 'A';
   if (c < 'a') goto weGo;
   if (c <= 'j')  return c - 'a';

   weGo:
#if _FS_RPATH
   return CurrVol;   // Use current drive
#else
   return 0;   // Use drive 0
#endif
} //  extractDrive

/* Check if the file system object is valid or not
 *
 * \param   path  Pointer to pointer to the path name (drive number)
 * \param   rfs   Pointer to pointer to the found file system object
 * \param   wmode !=0: Check write protection for write access
 * \return  FR_OK (0); FR_INVALID_DRIVE FR_NOT_ENABLED FR_WRITE_PROTECTED
 *                     FR_NO_FILESYSTEM
 *  \internal inline this function optimised
 */
static FRESULT chk_mounted(TCHAR const * * path, FATFS * * rfs, uint8_t wmode){
   const uint8_t vol = extractDrive(path);
   FRESULT res = checkValidFS(vol);
   if (res == FR_NO_FILESYSTEM) res = lookForFS(vol);
   if (!res) {
      *rfs =  FatFs[vol];
      if (!_FS_READONLY && wmode) res = checkWriteProtect(vol);
   } else {
      *rfs = 0;
   }
   return res;
} // chk_mounted


/* Check if the file/dir object is valid
 *
 * \param    obj  Pointer to the object FIL/DIR to check validity
 * \return       FR_OK(0): The object is valid, !=0: Invalid
 */
static FRESULT validate(void * obj){
   FIL * fil = (FIL*)obj;   // Assuming start of FIL/DIR is identical
   if (!fil->fs || !fil->fs->fs_type || fil->fs->id != fil->id)
                                                     return FR_INVALID_OBJECT;
   //  ENTER_FF(fil->fs); // Lock file system (does nothing)
   if (disk_status(fil->fs->drv) & STA_NOINIT)
      return FR_NOT_READY;
   SAMPLE_MIN_SP(1); // 1: disk_status above
   return FR_OK;
} // validate (DIR or FIL)


//--------  Public Functions --------------------------------------------

/*  Mount/Unmount a logical drive
 *
 *  This function registers/unregisters a work area to the FatFs module. The
 *  work area must be given to the each volume with this function prior to
 *  use any other file function. To unregister a work area use NULL as
 *  parameter \c fs.
 *
 *  This function always succeeds for a legal \c vol value regardless of the
 *  drive status. No media is accessed by this function. It only clears the
 *  given work area and registers its address to the internal table. The
 *  volume mount process is performed on first file access after calling
 *  \ref f_mount  or a media change.
 *
 *  \param vol Logical drive number to be mounted/unmounted
 *  \param fs  Pointer to new file system object (NULL for unmount only)
 *  \return  \ref FR_OK, \ref FR_INVALID_DRIVE
 *  \sa fileSystSMC
 */
FRESULT f_mount(uint8_t vol, FATFS * fs){
   if (vol >= _VOLUMES) return FR_INVALID_DRIVE;
   FATFS *rfs = FatFs[vol];    // Get current fs object
   if (rfs) rfs->fs_type = 0; // Clear old fs object
   if (fs)  fs->fs_type = 0; // Clear new fs object
   FatFs[vol] = fs;         // Register new fs object
   return FR_OK;
} // f_mount (uint8_t, FATFS *)


FRESULT f_open( FIL * fp,   // Pointer to the blank file object
   const TCHAR * path,  // Pointer to the file name
   uint8_t mode){  // Access mode and file open mode flags

   if (!fp) return FR_INVALID_OBJECT;
   fp->fs = 0;   // Clear file object
   SAMPLE_MIN_SP(1);

   DIR dj;
   uint8_t *dir;
   DEF_NAMEBUF;

#if !_FS_READONLY
   mode &= FA_READ | FA_WRITE | FA_CREATE_ALWAYS | FA_OPEN_ALWAYS | FA_CREATE_NEW;
   FRESULT res = chk_mounted(&path, &dj.fs, (uint8_t)(mode & ~FA_READ));
#else
   mode &= FA_READ;
   FRESULT res = chk_mounted(&path, &dj.fs, 0);
#endif
   if (res == FR_OK) {
      INIT_BUF(dj);
      res = follow_path(&dj, path); /* Follow the file path */
      dir = dj.dir;
#if !_FS_READONLY /* R/W configuration */
      if (res == FR_OK) {
         if (!dir)   /* Current dir itself */
            res = FR_INVALID_NAME;
#if _FS_LOCK
         else
            res = chk_lock(&dj, (mode & ~FA_READ) ? 1 : 0);
#endif
      }
      /* Create or Open a file */
      if (mode & (FA_CREATE_ALWAYS | FA_OPEN_ALWAYS | FA_CREATE_NEW)) {
         uint32_t dw, cl;

         if (res != FR_OK) {              /* No file, create new */
            if (res == FR_NO_FILE)        /* There is no file to open, create a new entry */
#if _FS_LOCK
               res = enq_lock() ? dir_register(&dj) : FR_TOO_MANY_OPEN_FILES;
#else
               res = dir_register(&dj);
#endif
            mode |= FA_CREATE_ALWAYS;     /* File is created */
            dir = dj.dir;              /* New entry */
         }
         else {                        /* Any object is already existing */
            if (dir[DIR_Attr] & (AM_RDO | AM_DIR)) {  /* Cannot overwrite it (R/O or DIR) */
               res = FR_DENIED;
            } else {
               if (mode & FA_CREATE_NEW)  /* Cannot create as new file */
                  res = FR_EXIST;
            }
         }
         if (res == FR_OK && (mode & FA_CREATE_ALWAYS)) {   /* Truncate it if overwrite mode */
            dw = get_fattime();              /* Created time */
            ST_DWORD(dir+DIR_CrtTime, dw);
            dir[DIR_Attr] = 0;               /* Reset attribute */
            ST_DWORD(dir+DIR_FileSize, 0);      /* size = 0 */
            cl = ld_clust(dj.fs, dir);       /* Get start cluster */
            st_clust(dir, 0);             /* cluster = 0 */
            dj.fs->wflag = 1;
            if (cl) {                     /* Remove the cluster chain if exist */
               dw = dj.fs->winsect;
               res = remove_chain(dj.fs, cl);
               if (res == FR_OK) {
                  dj.fs->last_clust = cl - 1;   /* Reuse the cluster hole */
                  res = move_window(dj.fs, dw);
               }
            }
         }
      }
      else {   /* Open an existing file */
         if (res == FR_OK) {                 /* Follow succeeded */
            if (dir[DIR_Attr] & AM_DIR) {    /* It is a directory */
               res = FR_NO_FILE;
            } else {
               if ((mode & FA_WRITE) && (dir[DIR_Attr] & AM_RDO)) /* R/O violation */
                  res = FR_DENIED;
            }
         }
      }
      if (res == FR_OK) {
         if (mode & FA_CREATE_ALWAYS)        /* Set file change flag if created or overwritten */
            mode |= FA__WRITTEN;
         fp->dir_sect = dj.fs->winsect;         /* Pointer to the directory entry */
         fp->dir_ptr = dir;
#if _FS_LOCK
         fp->lockid = inc_lock(&dj, (mode & ~FA_READ) ? 1 : 0);
         if (!fp->lockid) res = FR_INT_ERR;
#endif
      }

#else          /* R/O configuration */
      if (res == FR_OK) {              /* Follow succeeded */
         dir = dj.dir;
         if (!dir) {                /* Current dir itself */
            res = FR_INVALID_NAME;
         } else {
            if (dir[DIR_Attr] & AM_DIR)   /* It is a directory */
               res = FR_NO_FILE;
         }
      }
#endif
      FREE_BUF();

      if (res == FR_OK) {
         fp->flag = mode;              /* File access mode */
         fp->sclust = ld_clust(dj.fs, dir);  /* File start cluster */
         fp->fsize = LD_DWORD(dir+DIR_FileSize);   /* File size */
         fp->fptr = 0;                 /* File pointer */
         fp->dsect = 0;
#if _USE_FASTSEEK
         fp->cltbl = 0;                /* Normal seek mode */
#endif
         fp->fs = dj.fs; fp->id = dj.fs->id; /* Validate file object */
      }
   }
   return res;
} // f_open old


/*  Open or create a file
 *
 *  If this function \ref f_open function succeeded, the file object is valid.
 *  The file object is used for subsequent read/write functions to identify
 *  the file. To close an open file use \ref f_close(). If the modified file
 *  is not closed, the file data can be corrupted.
 *
 *  Before using any file function, a work area (file system object) must be
 *  registered to the logical drive with \ref f_mount(). All functions except
 *  \ref f_fdisk function can work after this procedure.
 *
 *  The parameter mode can be ORed using the predefined mode flags:
 *
 - \ref FA_READ Specifies read access to the object.
        Data can be read from the file. Combine with FA_WRITE for
        read-write access.
 - \ref FA_WRITE  Specifies write access to the object.
        Data can be written to the file. Combine with FA_READ for
        read-write access.
 - \ref FA_OPEN_EXISTING (default)  Opens the file if it is existing. Otherwise the
             function fails.
 - \ref FA_OPEN_ALWAYS   Opens the file if it is existing. If not, a new file
            is created. To append data to the file, use \ref f_lseek()
            after the file open in this method.
 - \ref FA_CREATE_NEW Creates a new file. The function fails with
        \ref FR_EXIST if the file is existing.
 - \ref FA_CREATE_ALWAYS Creates a new file. If the file is existing,
        it is truncated and overwritten.
 *
 *  \param fp  Pointer to the blank file object
 *  \param path Pointer to the file name
 *  \param mode Access mode and file open mode flags
 *  \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *           \ref FR_NO_FILE,  \ref FR_NO_PATH, \ref FR_INVALID_NAME,
 *           \ref FR_DENIED, \ref FR_EXIST, \ref FR_WRITE_PROTECTED,
 *           \ref FR_INVALID_DRIVE, \ref FR_NOT_ENABLED,
 *           \ref FR_NO_FILESYSTEM, \ref FR_TIMEOUT, \ref FR_LOCKED,
 *           \ref FR_NOT_ENOUGH_CORE, \ref FR_TOO_MANY_OPEN_FILES
 */
FRESULT f_open2(FIL * fp, const TCHAR * path, uint8_t mode){
   if (!fp) return FR_INVALID_OBJECT;
   fp->fs = NULL;  // Clear file system
   SAMPLE_MIN_SP(1);

   // inlined new chk_mounted (not yet tested)
   const uint8_t vol = extractDrive(&path);
   FRESULT res = checkValidFS(vol);
   if (res == FR_NO_FILESYSTEM) res = lookForFS(vol);
   if (!res) {
#if !_FS_READONLY
      mode &= FA_READ | FA_WRITE | FA_CREATE_ALWAYS
            | FA_OPEN_ALWAYS | FA_CREATE_NEW;
      if (mode & ~FA_READ) res = checkWriteProtect(vol);
#else
      mode &= FA_READ;
#endif
   }

   if (res == FR_OK) {
      DIR dj;
      DEF_NAMEBUF;
      INIT_BUF(dj);
      dj.fs =  FatFs[vol];

      res = follow_path(&dj, path);  // Follow the file path
      uint8_t * dir = dj.dir;
#if !_FS_READONLY // R/W configuration
      if (res == FR_OK) {
         if (!dir)  // Current dir itself
            res = FR_INVALID_NAME;
      }
      // Create or Open a file
      if (mode & (FA_CREATE_ALWAYS | FA_OPEN_ALWAYS | FA_CREATE_NEW)) {
         uint32_t dw, cl;

         if (res != FR_OK) { // No file, create new
            if (res == FR_NO_FILE) { // There's no file to open, create a new entry
               res = dir_register(&dj);
            }
            mode |= FA_CREATE_ALWAYS;  // File is created/
            dir = dj.dir;  // New entry
         } else { // Any object is already existing
            if (dir[DIR_Attr] & (AM_RDO | AM_DIR)) { // Cannot overwrite it (R/O or DIR)
               res = FR_DENIED;
            } else {
               if (mode & FA_CREATE_NEW)  // Cannot create as new file
                  res = FR_EXIST;
            }
         }
         if (res == FR_OK && (mode & FA_CREATE_ALWAYS)) { // Truncate it if overwrite mode
            dw = get_fattime();  // Created time/
            ST_DWORD(dir + DIR_CrtTime, dw);
            dir[DIR_Attr] = 0;  // Reset attribute
            ST_DWORD(dir + DIR_FileSize, 0); // size = 0
            cl = ld_clust(dj.fs, dir);  // Get start cluster
            st_clust(dir, 0);  // cluster = 0
            dj.fs->wflag = 1;
            if (cl) { // Remove the cluster chain if exist
               dw = dj.fs->winsect;
               res = remove_chain(dj.fs, cl);
               if (res == FR_OK) {
                  dj.fs->last_clust = cl - 1; // Reuse the cluster hole
                  res = move_window(dj.fs, dw);
               }
            }
         }
      } else {  // Open an existing file
         if (res == FR_OK) {  // Follow succeeded
            if (dir[DIR_Attr] & AM_DIR) { // It is a directory
               res = FR_NO_FILE;
            } else {
               if ((mode & FA_WRITE) && (dir[DIR_Attr] & AM_RDO)) // R/O violation
                  res = FR_DENIED;
            }
         }
      }
      if (res == FR_OK) {
         if (mode & FA_CREATE_ALWAYS) // Set file change flag if created or overwritten
            mode |= FA__WRITTEN;
         fp->dir_sect = dj.fs->winsect; // Pointer to the directory entry
         fp->dir_ptr = dir;
      }

#else  // R/O configuration

      if (res == FR_OK) {               /* Follow succeeded */
         dir = dj.dir;
         if (!dir) {                  /* Current dir itself */
            res = FR_INVALID_NAME;
         } else {
            if (dir[DIR_Attr] & AM_DIR)   /* It is a directory */
               res = FR_NO_FILE;
         }
      }
#endif
      FREE_BUF();

      if (res == FR_OK) {
         fp->flag = mode;  // File access mode
         fp->sclust = ld_clust(dj.fs, dir);   // File start cluster
         fp->fsize = LD_DWORD(dir+DIR_FileSize);   // File size
         fp->fptr = 0;   // File pointer
         fp->dsect = 0;
#if _USE_FASTSEEK
         fp->cltbl = 0; // Normal seek mode
#endif
         fp->fs = dj.fs; fp->id = dj.fs->id;   // Validate file object
      }
   }
   return res;
} //  f_open(FIL *, const TCHAR *, uint8_t)


/*  Read data from a file
 *
 *  This function reads file data to the supplied buffer, The file pointer
 *  of the file object increases by the number of bytes read. After the
 *  function succeeded, *br should be checked to detect the end of file. In
 *  case of \c *br < \c btr the read/write pointer reached end of the file
 *  during the read operation.
 *
 * \param fp  Pointer to file object
 * \param buff Pointer to data buffer
 * \param btr Number of bytes to read
 * \param br Pointer to result: number of bytes read <br />
 *           This number must not be greater than the sector size (512)
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR,
 *          \ref FR_NOT_READY, \ref FR_INVALID_OBJECT, \ref FR_TIMEOUT
 */
FRESULT f_read(FIL * fp, void * buff, uint16_t btr,  uint16_t * br){
   uint32_t clst, sect, remain;
   uint16_t rcnt, cc;
   uint8_t csect, *rbuff = buff;

   *br = 0;   // Clear read byte counter

   FRESULT res = validate(fp);    // Check validity
   if (res != FR_OK) return res;
   if (fp->flag & FA__ERROR)    return FR_INT_ERR;
   if (!(fp->flag & FA_READ))  return FR_DENIED;
   remain = fp->fsize - fp->fptr;
   if (btr > remain) btr = (uint16_t)remain; // Truncate btr by remaining bytes

   for ( ;  btr;       // Repeat until all data read
      rbuff += rcnt, fp->fptr += rcnt, *br += rcnt, btr -= rcnt) {
      if (mod2pow(fp->fptr, SSZ) == 0) {  // On the sector boundary?
         csect = (uint8_t)(div32SSZ(fp->fptr) & (fp->fs->csize - 1)); // Sector offset in the cluster
         if (!csect) {  // On the cluster boundary?
            if (fp->fptr == 0) { // On the top of the file?
               clst = fp->sclust; // Follow from the origin
            } else {  // Middle or end of the file
#if _USE_FASTSEEK
               if (fp->cltbl)
                  clst = clmt_clust(fp, fp->fptr); // Get cluster# from the CLMT
               else
#endif
                  clst = get_fat(fp->fs, fp->clust); // Follow cluster chain on the FAT
            }
            if (clst < 2) ABORT(fp->fs, FR_INT_ERR);
            if (clst == 0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
            fp->clust = clst; // Update current cluster
         }
         sect = clust2sect(fp->fs, fp->clust); // Get current sector
         if (!sect) ABORT(fp->fs, FR_INT_ERR);
         sect += csect;
         cc = div16SSZ(btr);  // When remaining bytes >= sector size,
         if (cc) {  // one could read maximum contiguous sectors directly
           // if (csect + cc > fp->fs->csize) // Clip at cluster boundary
             //  cc = fp->fs->csize - csect;
            // todo check / remove
            cc = 1;
 //            if (disk_read(fp->fs->drv, rbuff, sect, (uint8_t)cc))
            if (disk_read(fp->fs->drv, rbuff, sect))
               ABORT(fp->fs, FR_DISK_ERR);
#if !_FS_READONLY && _FS_MINIMIZE <= 2   // Replace one of the read sectors with cached data if it contains a dirty sector
#if _FS_TINY
            if (fp->fs->wflag && fp->fs->winsect - sect < cc)
               mem_cpy(rbuff + (mul32SSZ(fp->fs->winsect - sect)), fp->fs->win, SSZ);
#else
            if ((fp->flag & FA__DIRTY) && fp->dsect - sect < cc)
               mem_cpy(rbuff + mul32SSZ(fp->dsect - sect), fp->buf, SSZ);
#endif
#endif
            rcnt = mul16SSZ(cc);  // Number of bytes transferred
            continue;
         }
#if !_FS_TINY
         if (fp->dsect != sect) {   // Load data sector if not in cache
#if !_FS_READONLY
            if (fp->flag & FA__DIRTY) {  // Write-back dirty sector cache
               if (disk_write(fp->fs->drv, fp->buf, fp->dsect, 1))
                  ABORT(fp->fs, FR_DISK_ERR);
               fp->flag &= ~FA__DIRTY;
            }
#endif
            if (disk_read(fp->fs->drv, fp->buf, sect, 1)) // Fill sector cache
               ABORT(fp->fs, FR_DISK_ERR);
         }
#endif
         fp->dsect = sect;
      }
      rcnt = SSZ - ((uint16_t)modSSZ(fp->fptr)); // Get partial sector data from sector buffer
      if (rcnt > btr) rcnt = btr;
#if _FS_TINY
      if (move_window(fp->fs, fp->dsect))      /* Move sector window */
         ABORT(fp->fs, FR_DISK_ERR);
      mem_cpy(rbuff, &fp->fs->win[modSSZ(fp->fptr)], rcnt);   /* Pick partial sector */
#else
      mem_cpy(rbuff, &fp->buf[modSSZ(fp->fptr)], rcnt);   /* Pick partial sector */
#endif
   }
    return FR_OK;
} // f_read(FIL *fp, void *buff, uint16_t btr,  uint16_t *br)



#if !_FS_READONLY
/*  Write data to a file
 *
 *  The read/write pointer in the \ref FIL "file object"  is increased by
 *  the number of bytes written. After the function succeeded, \c *bw
 *  should be checked to detect the disk full condition. In case of
 *  \c *bw < \c btw there was not enough space for the (complete) write
 *  operation.
 *
 *  The function may require more time when the volume is full or nearly so.
 *
 * \param  fp  Pointer to the file object
 * \param buff Pointer to data buffer
 * \param btw Number of bytes to write <br />
 *           This number must not be greater than the sector size (512)
 * \param bw  Pointer to result: number of bytes written
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_INVALID_OBJECT, FR_TIMEOUT
 */
FRESULT f_write(FIL * fp, const void * buff, uint16_t btw, uint16_t * bw){
   uint32_t clst, sect;
   uint16_t wcnt, cc;
   const uint8_t *wbuff = buff;
   uint8_t csect;

   *bw = 0;   /* Clear write byte counter */

   FRESULT res = validate(fp);   // Check validity
   if (res != FR_OK)  return res;
   if (fp->flag & FA__ERROR)   return FR_INT_ERR; // Aborted file?
   if (!(fp->flag & FA_WRITE))   return FR_DENIED; // Check access mode
   if ((uint32_t)(fp->fsize + btw) < fp->fsize) btw = 0;   // File size cannot reach 4GB

   for ( ;  btw;   // Repeat until all data written
      wbuff += wcnt, fp->fptr += wcnt, *bw += wcnt, btw -= wcnt) {
      if ((fp->fptr % SSZ) == 0) {   /* On the sector boundary? */
         csect = (uint8_t)(fp->fptr / SSZ & (fp->fs->csize - 1));   /* Sector offset in the cluster */
         if (!csect) {               /* On the cluster boundary? */
            if (fp->fptr == 0) {      /* On the top of the file? */
               clst = fp->sclust;      /* Follow from the origin */
               if (clst == 0)         /* When no cluster is allocated, */
                  fp->sclust = clst = create_chain(fp->fs, 0);   /* Create a new cluster chain */
            } else {               /* Middle or end of the file */
#if _USE_FASTSEEK
               if (fp->cltbl)
                  clst = clmt_clust(fp, fp->fptr);   /* Get cluster# from the CLMT */
               else
#endif
                  clst = create_chain(fp->fs, fp->clust);   /* Follow or stretch cluster chain on the FAT */
            }
            if (clst == 0) break;      /* Could not allocate a new cluster (disk full) */
            if (clst == 1) ABORT(fp->fs, FR_INT_ERR);
            if (clst == 0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
            fp->clust = clst;         /* Update current cluster */
         }
#if _FS_TINY
         if (fp->fs->winsect == fp->dsect && move_window(fp->fs, 0))   /* Write-back sector cache */
            ABORT(fp->fs, FR_DISK_ERR);
#else
         if (fp->flag & FA__DIRTY) {      /* Write-back sector cache */
            if (disk_write(fp->fs->drv, fp->buf, fp->dsect, 1))
               ABORT(fp->fs, FR_DISK_ERR);
            fp->flag &= ~FA__DIRTY;
         }
#endif
         sect = clust2sect(fp->fs, fp->clust);   /* Get current sector */
         if (!sect) ABORT(fp->fs, FR_INT_ERR);
         sect += csect;
         cc = btw / SSZ;         /* When remaining bytes >= sector size, */
         if (cc) {                  /* Write maximum contiguous sectors directly */
            if (csect + cc > fp->fs->csize)   /* Clip at cluster boundary */
               cc = fp->fs->csize - csect;
            // todo check / remove
            cc = 1;
           //if (disk_write(fp->fs->drv, wbuff, sect, (uint8_t)cc))
            if (disk_write(fp->fs->drv, wbuff, sect))
               ABORT(fp->fs, FR_DISK_ERR);
#if _FS_TINY
            if (fp->fs->winsect - sect < cc) {   /* Refill sector cache if it gets invalidated by the direct write */
               mem_cpy(fp->fs->win, wbuff + ((fp->fs->winsect - sect) * SSZ), SSZ);
               fp->fs->wflag = 0;
            }
#else
            if (fp->dsect - sect < cc) { /* Refill sector cache if it gets invalidated by the direct write */
               mem_cpy(fp->buf, wbuff + ((fp->dsect - sect) * SSZ), SSZ);
               fp->flag &= ~FA__DIRTY;
            }
#endif
            wcnt = SSZ * cc;      /* Number of bytes transferred */
            continue;
         }
#if _FS_TINY
         if (fp->fptr >= fp->fsize) {   /* Avoid silly cache filling at growing edge */
            if (move_window(fp->fs, 0)) ABORT(fp->fs, FR_DISK_ERR);
            fp->fs->winsect = sect;
         }
#else
         if (fp->dsect != sect) {      /* Fill sector cache with file data */
            if (fp->fptr < fp->fsize &&
               disk_read(fp->fs->drv, fp->buf, sect, 1))
                  ABORT(fp->fs, FR_DISK_ERR);
         }
#endif
         fp->dsect = sect;
      }
      wcnt = SSZ - ((uint16_t)fp->fptr % SSZ);/* Put partial sector into file I/O buffer */
      if (wcnt > btw) wcnt = btw;
#if _FS_TINY
      if (move_window(fp->fs, fp->dsect))   /* Move sector window */
         ABORT(fp->fs, FR_DISK_ERR);
      mem_cpy(&fp->fs->win[fp->fptr % SSZ], wbuff, wcnt);   /* Fit partial sector */
      fp->fs->wflag = 1;
#else
      mem_cpy(&fp->buf[fp->fptr % SSZ], wbuff, wcnt);   /* Fit partial sector */
      fp->flag |= FA__DIRTY;
#endif
   }

   if (fp->fptr > fp->fsize) fp->fsize = fp->fptr;   /* Update file size if needed */
   fp->flag |= FA__WRITTEN;                  /* Set file change flag */

   return FR_OK;
}

/*  Flush written / cached data to a file
 *
 *  This function transfers write data caches in memory / sector buffers to
 *  the device. (The process and the related functions is normally called
 *  flush.) The function \ref f_close() does, of course, the same before
 *  discarding the file object.
 *
 *  Calling this function periodically or after every \ref f_write() can
 *  reduce the risk of data loss due to power failure or SMC removal at all
 *  costs of otherwise unnecessary device accesses.
 *
 * \param  fp  Pointer to the file object
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_INVALID_OBJECT, FR_TIMEOUT
 */
FRESULT f_sync(FIL * fp ){
   FRESULT res = validate(fp);               /* Check validity of the object */
   if (res == FR_OK) {
      if (fp->flag & FA__WRITTEN) {   /* Has the file been written? */
#if !_FS_TINY   /* Write-back dirty buffer */
         if (fp->flag & FA__DIRTY) {
            if (disk_write(fp->fs->drv, fp->buf, fp->dsect, 1))
               return FR_DISK_ERR;
            fp->flag &= ~FA__DIRTY;
         }
#endif
         /* Update the directory entry */
         res = move_window(fp->fs, fp->dir_sect);
         if (res == FR_OK) {
            uint8_t * dir = fp->dir_ptr;
            dir[DIR_Attr] |= AM_ARC;               /* Set archive bit */
            ST_DWORD(dir+DIR_FileSize, fp->fsize);      /* Update file size */
            st_clust(dir, fp->sclust);               /* Update start cluster */
            uint32_t tim = get_fattime();                  /* Update updated time */
            ST_DWORD(dir+DIR_WrtTime, tim);
            ST_WORD(dir+DIR_LstAccDate, 0);
            fp->flag &= ~FA__WRITTEN;
            fp->fs->wflag = 1;
            res = sync(fp->fs);
         }
      }
   }
   return res;
} //  f_sync(FIL *)

#endif /* !_FS_READONLY */


/*  Close an open file object
 *
 *  This function closes an open file. If any data have been written to
 *  the file, the cached information is written back to the medium (SMC).
 *  After the function succeeded, the file object is no longer valid.
 *
 * \param fp  Pointer to file object
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR,
 *          \ref FR_NOT_READY, \ref FR_INVALID_OBJECT, FR_TIMEOUT
 */
FRESULT f_close(FIL * fp){
#if _FS_READONLY
   FRESULT const res = validate(fp);
#else
   FRESULT const res = f_sync(fp);      // Flush cached data
#endif
   if (res == FR_OK) fp->fs = 0;   /* Discard file object */
   return res;
} // f_close


// ---- Current Drive/Directory handling

#if _FS_RPATH >= 1

FRESULT f_chdrive( uint8_t drv){      // Drive number
   if (drv >= _VOLUMES) return FR_INVALID_DRIVE;
   CurrVol = drv;
   return FR_OK;
}


FRESULT f_chdir (const TCHAR *path){   /* Pointer to the directory path */
   DIR dj;
   DEF_NAMEBUF;

   FRESULT res = chk_mounted(&path, &dj.fs, 0);
   if (res == FR_OK) {
      INIT_BUF(dj);
      res = follow_path(&dj, path);      /* Follow the path */
      FREE_BUF();
      if (res == FR_OK) {               /* Follow completed */
         if (!dj.dir) {
            dj.fs->cdir = dj.sclust;   /* Start directory itself */
         } else {
            if (dj.dir[DIR_Attr] & AM_DIR)   /* Reached to the directory */
               dj.fs->cdir = ld_clust(dj.fs, dj.dir);
            else
               res = FR_NO_PATH;      /* Reached but a file */
         }
      }
      if (res == FR_NO_FILE) res = FR_NO_PATH;
   }
   return res;
}


#if _FS_RPATH >= 2
FRESULT f_getcwd(
   TCHAR *path,   /* Pointer to the directory path */
   uint16_t sz_path){  /* Size of path */
   FRESULT res;
   DIR dj;
   uint16_t i, n;
   uint32_t ccl;
   TCHAR *tp;
   FILINFO fno;
   DEF_NAMEBUF;

   *path = 0;
   res = chk_mounted((const TCHAR**)&path, &dj.fs, 0);   /* Get current volume */
   if (res == FR_OK) {
      INIT_BUF(dj);
      i = sz_path;      /* Bottom of buffer (dir stack base) */
      dj.sclust = dj.fs->cdir;         /* Start to follow upper dir from current dir */
      while ((ccl = dj.sclust) != 0) {   /* Repeat while current dir is a sub-dir */
         res = dir_sdi(&dj, 1);         /* Get parent dir */
         if (res != FR_OK) break;
         res = dir_read(&dj);
         if (res != FR_OK) break;
         dj.sclust = ld_clust(dj.fs, dj.dir);   /* Goto parent dir */
         res = dir_sdi(&dj, 0);
         if (res != FR_OK) break;
         do {                     /* Find the entry links to the child dir */
            res = dir_read(&dj);
            if (res != FR_OK) break;
            if (ccl == ld_clust(dj.fs, dj.dir)) break;   /* Found the entry */
            res = dir_next(&dj, 0);   
         } while (res == FR_OK);
         if (res == FR_NO_FILE) res = FR_INT_ERR;/* It cannot be 'not found'. */
         if (res != FR_OK) break;
#if _USE_LFN
         fno.lfname = path;
         fno.lfsize = i;
#endif
         get_fileinfo(&dj, &fno);      /* Get the dir name and push it to the buffer */
         tp = fno.fname;
         if (_USE_LFN && *path) tp = path;
         for (n = 0; tp[n]; n++) ;
         if (i < n + 3) {
            res = FR_NOT_ENOUGH_CORE; break;
         }
         while (n) path[--i] = tp[--n];
         path[--i] = '/';
      }
      tp = path;
      if (res == FR_OK) {
         *tp++ = '0' + CurrVol;         /* Put drive number */
         *tp++ = ':';
         if (i == sz_path) {            /* Root-dir */
            *tp++ = '/';
         } else {                  /* Sub-dir */
            do      /* Add stacked path str */
               *tp++ = path[i++];
            while (i < sz_path);
         }
      }
      *tp = 0;
      FREE_BUF();
   }

   return (res);
}
#endif /* _FS_RPATH >= 2 */
#endif /* _FS_RPATH >= 1 */


#if _FS_MINIMIZE <= 2

#if ! _weAutSys
/*  Move file pointer of a file object, expand file size
 *
 *  This function moves the file read/write pointer of an open file. The
 *  offset is absolute, i.e. from file origin resp. top of file. When an
 *  offset above the file size is specified in write mode, the file size is
 *  increased and the data in the expanded area are undefined. Thereby is is
 *  possible to create a large file quickly. After the \ref f_lseek function
 *  succeeded, current read/write pointer should be checked to make sure the
 *  read/write pointer has been moved correctly. In case of the current
 *  read/write pointer is not the expected value, either of followings has
 *  occurred:
  - End of file. The specified Offset was clipped at end of the file
         because the file has been opened in read-only mode.
  - Disk full. There is insufficient free space on the volume to expand
         the file size.
 *
 *  Fast seek feature is enabled when \ref _USE_FASTSEEK is set to 1 and the
 *  member .cltbl in the file object is not NULL. This feature enables fast
 *  backward/long seek operations without FAT access by cluster link map
 *  table (CLMT) stored in the user defined table. It is also applied to
 *  the functions \ref f_read() and \ref f_write(). In this mode, the file
 *  size cannot be increased by \ref f_write() or \ref f_lseek().
 *  The CLMT must be created in the user defined DWORD array prior to use
 *  fast seek feature. To create the CLMT, let the member .cltbl in the
 *  file object pointer to the DWORD array. Set the array size in unit of
 *  items into the first item and call the \ref f_lseek function with
 *  Offset = CREATE_LINKMAP. After the function succeeded and CLMT is
 *  created, no FAT access will occur in subsequent \ref f_read(),
 *  \ref f_write() or \ref f_lseek() calls on that file. If the function
 *  failed with FR_NOT_ENOUGH_CORE, the given array size is insufficient
 *  for the file and the required items is returned into the first item of
 *  the array. The required array size is (number of fragments + 1) * 2 items.
 *  For example, when the file is fragmented in 5, 12 items will be required
 *  for the CLMT.
 *
 * \param  fp  Pointer to the file object
 * \param  ofs  File pointer from top of file
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR,
 *          \ref FR_NOT_READY, \ref FR_INVALID_OBJECT, FR_TIMEOUT
 */
FRESULT f_lseek(FIL* fp, uint32_t ofs){
   FRESULT res = validate(fp);  // Check validity of the object
   if (res != FR_OK) return (res);
   if (fp->flag & FA__ERROR)  // Check abort flag
      return (FR_INT_ERR);

#if _USE_FASTSEEK
   if (fp->cltbl) {   /* Fast seek */
      uint32_t cl, pcl, ncl, tcl, dsc, tlen, ulen, *tbl;

      if (ofs == CREATE_LINKMAP) {   /* Create CLMT */
         tbl = fp->cltbl;
         tlen = *tbl++; ulen = 2;   /* Given table size and required table size */
         cl = fp->sclust;         /* Top of the chain */
         if (cl) {
            do {
               /* Get a fragment */
               tcl = cl; ncl = 0; ulen += 2;   /* Top, length and used items */
               do {
                  pcl = cl; ncl++;
                  cl = get_fat(fp->fs, cl);
                  if (cl <= 1) ABORT(fp->fs, FR_INT_ERR);
                  if (cl == 0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
               } while (cl == pcl + 1);
               if (ulen <= tlen) {      /* Store the length and top of the fragment */
                  *tbl++ = ncl; *tbl++ = tcl;
               }
            } while (cl < fp->fs->n_fatent);   /* Repeat until end of chain */
         }
         *fp->cltbl = ulen;   /* Number of items used */
         if (ulen <= tlen)
            *tbl = 0;      /* Terminate table */
         else
            res = FR_NOT_ENOUGH_CORE;   /* Given table size is smaller than required */

      } else {   // Fast seek
         if (ofs > fp->fsize)  // Clip offset at the file size
            ofs = fp->fsize;
         fp->fptr = ofs; // Set file pointer
         if (ofs) {
            fp->clust = clmt_clust(fp, ofs - 1);
            dsc = clust2sect(fp->fs, fp->clust);
            if (!dsc) ABORT(fp->fs, FR_INT_ERR);
            dsc += (ofs - 1) / SSZ & (fp->fs->csize - 1);
            if (fp->fptr % SSZ && dsc != fp->dsect) { // Refill sector cache if needed
#if !_FS_TINY
#if !_FS_READONLY
               if (fp->flag & FA__DIRTY) { // Write-back dirty sector cache
                  if (disk_write(fp->fs->drv, fp->buf, fp->dsect, 1))
                     ABORT(fp->fs, FR_DISK_ERR);
                  fp->flag &= ~FA__DIRTY;
               }
#endif
               if (disk_read(fp->fs->drv, fp->buf, dsc, 1)) // Load current sector
                  ABORT(fp->fs, FR_DISK_ERR);
#endif
               fp->dsect = dsc;
            }
         }
      }
   } else
#endif

   //  Normal Seek
   {
      uint32_t clst, bcs, nsect, ifptr;

      if (ofs > fp->fsize // In read-only mode, clip offset with the file size
#if !_FS_READONLY
          && !(fp->flag & FA_WRITE)
#endif
         ) ofs = fp->fsize;

      ifptr = fp->fptr;
      fp->fptr = nsect = 0;
      if (ofs) {
         bcs = (uint32_t)fp->fs->csize * SSZ; // Cluster size (byte)
         if (ifptr > 0 &&
            (ofs - 1) / bcs >= (ifptr - 1) / bcs) { //  When seek to same or following cluster,
            fp->fptr = (ifptr - 1) & ~(bcs - 1);  // start from the current cluster
            ofs -= fp->fptr;
            clst = fp->clust;
         } else {   // When seek to back cluster,
            clst = fp->sclust; // start from the first cluster
#if !_FS_READONLY
            if (clst == 0) {  // If no cluster chain, create a new chain
               clst = create_chain(fp->fs, 0);
               if (clst == 1) ABORT(fp->fs, FR_INT_ERR);
               if (clst == 0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
               fp->sclust = clst;
            }
#endif
            fp->clust = clst;
         }
         if (clst != 0) {
            while (ofs > bcs) { //  Cluster following loop
#if !_FS_READONLY
               if (fp->flag & FA_WRITE) {         /* Check if in write mode or not */
                  clst = create_chain(fp->fs, clst);   /* Force stretch if in write mode */
                  if (clst == 0) {            /* When disk gets full, clip file size */
                     ofs = bcs; break;
                  }
               } else
#endif
                  clst = get_fat(fp->fs, clst); // Follow cluster chain if not in write mode
               if (clst == 0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
               if (clst <= 1 || clst >= fp->fs->n_fatent) ABORT(fp->fs, FR_INT_ERR);
               fp->clust = clst;
               fp->fptr += bcs;
               ofs -= bcs;
            }
            fp->fptr += ofs;
            if (ofs % SSZ) {
               nsect = clust2sect(fp->fs, clst); // Current sector
               if (!nsect) ABORT(fp->fs, FR_INT_ERR);
               nsect += ofs / SSZ;
            }
         }
      }
      if (fp->fptr % SSZ && nsect != fp->dsect) { // Fill sector cache if needed
#if !_FS_TINY
#if !_FS_READONLY
         if (fp->flag & FA__DIRTY) { // Write-back dirty sector cache
            if (disk_write(fp->fs->drv, fp->buf, fp->dsect, 1))
               ABORT(fp->fs, FR_DISK_ERR);
            fp->flag &= ~FA__DIRTY;
         }
#endif
         if (disk_read(fp->fs->drv, fp->buf, nsect, 1)) // Fill sector cache
            ABORT(fp->fs, FR_DISK_ERR);
#endif
         fp->dsect = nsect;
      }
#if !_FS_READONLY
      if (fp->fptr > fp->fsize) { // Set file change flag if the file size is extended
         fp->fsize = fp->fptr;
         fp->flag |= FA__WRITTEN;
      }
#endif
   }

   return (res);
} //  f_lseek(FIL* ...)
#endif // ! _weAutSys

#if _FS_MINIMIZE <= 1

/* Open an existing directory
 *
 *  This function opens an existing directory and creates the directory
 *  object for subsequent calls of directory related functions. The
 *  \ref DIR "directory object" can be discarded at any time.
 *
 * \param path Pointer to the directory's file name (0-terminated)
 * \param dj Pointer to directory object to create
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_PATH, \ref FR_INVALID_NAME, \ref FR_INVALID_DRIVE,
 *          \ref FR_NOT_ENABLED, \ref FR_NO_FILESYSTEM, FR_TIMEOUT,
 *          \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_opendir(DIR * dj, TCHAR const * path){
   if (!dj) return FR_INVALID_OBJECT;

   const uint8_t vol = extractDrive(&path);
   FRESULT res = checkValidFS(vol);
   if (res == FR_NO_FILESYSTEM) res = lookForFS(vol);

   if (res == FR_OK) {
      FATFS * fs = dj->fs = FatFs[vol];
      DEF_NAMEBUF;
      INIT_BUF(*dj);
      res = follow_path(dj, path); // Follow the path to the directory/
      FREE_BUF();
      if (res == FR_OK) {   // Follow completed
         if (dj->dir) {      // It is not the root dir
            if (dj->dir[DIR_Attr] & AM_DIR) { // The object is a directory
               dj->sclust = ld_clust(fs, dj->dir);
            } else {   // The object is not a directory
               res = FR_NO_PATH;
            }
         }
         if (res == FR_OK) {
            dj->id = fs->id;
            res = dir_sdi(dj, 0); // Rewind dir
         }
      }
      if (res == FR_NO_FILE) res = FR_NO_PATH;
      if (res != FR_OK) dj->fs = 0;   // Invalidate dir object if function failed
   } else {
      dj->fs = 0;
   }
   return res;
} // f_opendir(DIR*, const TCHAR *)


/*  Read a directory item
 *
 *  This function reads directory entries in sequence. All items in the
 *  directory can be read by repeated calls. When all directory entries have
 *  been thus visited the function put a null string into fno.f_name[] without
 *  any error. When a null pointer is given to the FileInfo, the read index
 *  of the directory object will be rewound.
 *
 *  When \ref _USE_LFN "LFN" feature is enabled, lfname and lfsize in the file
 *  information structure must be initialized with valid value prior to use
 *  the f_readdir function. The lfname is a pointer to the string buffer to
 *  return the long file name. The lfsize is the size of the string buffer
 *  in unit of TCHAR. If either the size of read buffer or LFN working buffer
 *  is insufficient for the LFN or the object has no LFN, a null string will
 *  be returned to the LFN read buffer. If the LFN contains any character
 *  that cannot be converted to OEM code, a null string will be returned but
 *  this is not the case on Unicode API configuration. When lfname is NULL,
 *  nothing of the LFN is returned. When the object has no LFN, some small
 *  capitals can be contained in the SFN.
 *
 *  When relative path feature is enabled (\ref _FS_RPATH == 1), "." and
 *  ".." entries are not filtered out and it will appear in the read entries.
 *
 *  This function is available if \ref _FS_MINIMIZE <= 1.
 *
 * \param dj  Pointer to the open directory object
 * \param fno Pointer to the file information
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_INVALID_OBJECT, FR_TIMEOUT, \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_readdir(DIR * dj, FILINFO * fno){
   DEF_NAMEBUF;
   INIT_BUF(*dj);

   FRESULT res = validate(dj); // check validity of the object
   if (res == FR_OK) {
      if (!fno) {
         res = dir_sdi(dj, 0); //  Rewind the directory object
      } else {
         res = dir_read(dj);  // Read an directory item
         if (res == FR_NO_FILE) { // Reached end of dir
            dj->sect = 0;
            res = FR_OK;
         }
         if (res == FR_OK) {   // A valid entry is found
            get_fileinfo(dj, fno);   // Get the object information
            res = dir_next(dj, 0);   //  Increment index for next
            if (res == FR_NO_FILE) {
               dj->sect = 0;
               res = FR_OK;
            }
         }
         FREE_BUF();
      }
   }
   return res;
} // f_readdir(DIR*, FILINFO*)


#if _FS_MINIMIZE == 0
/*  Get file status
 *
 *  This function gets the size, timestamp and attribute of a file or
 *  directory. For details of the information, refer to the \ref FILINFO
 *  structure and \ref f_readdir().
 *
 *  This function is not available in
 *  \ref _FS_MINIMIZE "minimization level" of >= 1.
 *
 * \param path Pointer to the directory's file name (0-terminated)
 * \param fno Pointer to the file information
 * \return  FR_OK, FR_DISK_ERR, FR_INT_ERR, FR_NOT_READY, FR_NO_FILE,
 *          FR_NO_PATH, FR_INVALID_NAME, FR_INVALID_DRIVE, FR_NOT_ENABLED,
 *          FR_NO_FILESYSTEM, FR_TIMEOUT, FR_NOT_ENOUGH_CORE
 */
FRESULT f_stat(const TCHAR * path, FILINFO * fno){
   DIR dj;
   DEF_NAMEBUF;

   FRESULT res = chk_mounted(&path, &dj.fs, 0);
   if (res == FR_OK) {
      INIT_BUF(dj);
      res = follow_path(&dj, path); // Follow the file path
      if (res == FR_OK) {           // Follow completed
         if (dj.dir) // Found an object
            get_fileinfo(&dj, fno);
         else  // it is root dir
            res = FR_INVALID_NAME;
      }
      FREE_BUF();
   }
   return res;
} // f_stat(const TCHAR*, FILINFO*)


#if !_FS_READONLY

#if ! _weAutSys
/*  Get the number of free clusters on the drive
 *
 *  This function gets number of free clusters on the drive. The member
 *  \c csize in the file system object is reflecting the number of sectors
 *  per cluster, so that the free space in unit of sector can be calculated.
 *  When FSInfo structure on FAT32 volume is not in sync, this function can
 *  return an incorrect free cluster count.
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 *  \param vol    the drive's volume number (0..9); to extract form a
 *                path name use \ref extractDrive "extractDrive(&path)"
 *  \param nclst  Pointer to result: number of free clusters
 *  \param fatfs  Pointer to pointer to corresponding file system object to return
 *  \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *           \ref FR_INVALID_DRIVE, \ref FR_NOT_ENABLED,
 *           \ref FR_NO_FILESYSTEM, FR_TIMEOUT
 *  \internal fail on 2 gig card
 */
FRESULT f_getfree(const uint8_t vol, uint32_t * nclst){
   FATFS * fs;
   FRESULT res = checkValidFS(vol);
   if (res == FR_NO_FILESYSTEM) res = lookForFS(vol);
   if (res) {
      *nclst = 0;
      return res;
   }
   fs = FatFs[vol];


   // Get drive number */
   //FRESULT res = chk_mounted(&path, fatfs, 0);
   //* fatfs;
   //if (res == FR_OK) {
   /* If free_clust is valid, return it without full cluster scan */
   if (fs->free_clust <= fs->n_fatent - 2) {
      *nclst = fs->free_clust;
   } else {
      uint32_t n, clst, sect, stat;
      uint16_t i;
      uint8_t fat, *p;

      /* Get number of free clusters */
      fat = fs->fs_type;
      n = 0;
      if (fat == FS_FAT12) {
         clst = 2;
         do {
            stat = get_fat(fs, clst);
            if (stat == 0xFFFFFFFF) { res = FR_DISK_ERR; break; }
            if (stat == 1) { res = FR_INT_ERR; break; }
            if (stat == 0) n++;
         } while (++clst < fs->n_fatent);
      } else {
         clst = fs->n_fatent;
         sect = fs->fatbase;
         i = 0; p = 0;
         do {
            if (!i) {
               res = move_window(fs, sect++);
               if (res != FR_OK) break;
               p = fs->win;
               i = SSZ;
            }
            if (fat == FS_FAT16) {
               if (LD_WORD(p) == 0) n++;
               p += 2; i -= 2;
            } else {
               if ((LD_DWORD(p) & 0x0FFFFFFF) == 0) n++;
               p += 4; i -= 4;
            }
         } while (--clst);
      }
      fs->free_clust = n;
      if (fat == FS_FAT32) fs->fsi_flag = 1;
      *nclst = n;
   }
   //  }
   return res;
} // f_getfree(const TCHAR*, uint32_t*, FATFS**)
#endif // ! _weAutSys

/*  Truncate file
 *
 *  This function truncates the file size to the current file read/write
 *  point. This function has no effect if the file read/write pointer is
 *  already pointing end of the file.
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 *  \param  fp  Pointer to the file object
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_INVALID_OBJECT, FR_TIMEOUT
 */
FRESULT f_truncate(FIL * fp){
   if (!fp) return FR_INVALID_OBJECT;
   FRESULT res = validate(fp); // Check validity of the object
   if (res == FR_OK) {
      if (fp->flag & FA__ERROR) {  // Check abort flag
         res = FR_INT_ERR;
      } else {
         if (!(fp->flag & FA_WRITE)) // Check access mode
            res = FR_DENIED;
      }
   }
   if (res == FR_OK) {
      if (fp->fsize > fp->fptr) {
         fp->fsize = fp->fptr;  //  Set file size to current R/W point
         fp->flag |= FA__WRITTEN;
         if (fp->fptr == 0) { // When set file size to zero, remove entire cluster chain
            res = remove_chain(fp->fs, fp->sclust);
            fp->sclust = 0;
         } else { // When truncate a part of the file, remove remaining clusters
            uint32_t ncl = get_fat(fp->fs, fp->clust);
            res = FR_OK;
            if (ncl == 0xFFFFFFFF) res = FR_DISK_ERR;
            if (ncl == 1) res = FR_INT_ERR;
            if (res == FR_OK && ncl < fp->fs->n_fatent) {
               res = put_fat(fp->fs, fp->clust, 0x0FFFFFFF);
               if (res == FR_OK) res = remove_chain(fp->fs, ncl);
            }
         }
      }
      if (res != FR_OK) fp->flag |= FA__ERROR;
   }
   return res;
} // f_truncate(FIL*)


/*  Delete an existing file or directory
 *
 *  If the object to be erased / removed has a read-only attribute
 *  (\ref AM_RDO) the operation will be rejected with \ref FR_DENIED.
 *
 *  To remove a directory it must be empty and must not be current directory
 *  or the operation will be rejected with \ref FR_DENIED.
 *
 *  A file to be removed must not be open or the FAT volume can collapse.
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 *  \param path Pointer to the directory's file name (0-terminated)
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_FILE, \ref FR_NO_PATH, \ref FR_INVALID_NAME,
 *          \ref FR_DENIED, \ref FR_EXIST, \ref FR_WRITE_PROTECTED,
 *          \ref FR_INVALID_DRIVE, \ref FR_NOT_ENABLED, \ref FR_NO_FILESYSTEM,
 *          FR_TIMEOUT, \ref FR_LOCKED, \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_unlink(const TCHAR * path){
   DIR dj, sdj;
   uint8_t *dir;
   uint32_t dclst;
   DEF_NAMEBUF;

   FRESULT res = chk_mounted(&path, &dj.fs, 1);
   if (res == FR_OK) {
      INIT_BUF(dj);
      res = follow_path(&dj, path); // Follow the file path
      if (_FS_RPATH && res == FR_OK && (dj.fn[NS] & NS_DOT))
         res = FR_INVALID_NAME;     // Cannot remove dot entry
      if (res == FR_OK) {           // The object is accessible
         dir = dj.dir;
         if (!dir) {
            res = FR_INVALID_NAME;  // Cannot remove the start directory
         } else {
            if (dir[DIR_Attr] & AM_RDO)
               res = FR_DENIED;     // Cannot remove R/O object
         }
         dclst = ld_clust(dj.fs, dir);
         if (res == FR_OK && (dir[DIR_Attr] & AM_DIR)) { // Is it a sub-dir?
            if (dclst < 2) {
               res = FR_INT_ERR;
            } else {
               mem_cpy(&sdj, &dj, sizeof (DIR)); // Check if the sub-dir is empty
               sdj.sclust = dclst;
               res = dir_sdi(&sdj, 2); // Exclude dot entries
               if (res == FR_OK) {
                  res = dir_read(&sdj);
                  if (res == FR_OK     // Not empty dir
#if _FS_RPATH
                  || dclst == dj.fs->cdir  // Current dir
#endif
                  ) res = FR_DENIED;
                  if (res == FR_NO_FILE) res = FR_OK;  // Empty
               }
            }
         }
         if (res == FR_OK) {
            res = dir_remove(&dj); // Remove the directory entry
            if (res == FR_OK) {
               if (dclst)          // Remove the cluster chain if exist
                  res = remove_chain(dj.fs, dclst);
               if (res == FR_OK) res = sync(dj.fs);
            }
         }
      }
      FREE_BUF();
   }
   return (res);
} // f_unlink(const TCHAR)



/*  Create a new directory
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 *  \param path Pointer to the directory's file name (0-terminated)
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_PATH, \ref FR_INVALID_NAME, \ref FR_DENIED,
 *          \ref FR_EXIST, \ref FR_WRITE_PROTECTED, \ref FR_INVALID_DRIVE,
 *          \ref FR_NOT_ENABLED, \ref FR_NO_FILESYSTEM,
 *          FR_TIMEOUT, \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_mkdir(const TCHAR * path){
   DIR dj;
   // inlined new chk_mounted (not yet tested)
   const uint8_t vol = extractDrive(&path);
   FRESULT res = checkValidFS(vol);
   if (res == FR_NO_FILESYSTEM) res = lookForFS(vol);
   if (!res) {
      dj.fs=  FatFs[vol];
      if (!_FS_READONLY) res = checkWriteProtect(vol);
   } else {
      dj.fs = 0;
   }

   uint8_t *dir, n;
   uint32_t dsc, dcl, pcl, tim = get_fattime();
   DEF_NAMEBUF;
   INIT_BUF(dj);


   //  res = chk_mounted(&path, &dj.fs, 1);
   if (res == FR_OK) {
     res = follow_path(&dj, path);         /* Follow the file path */
      if (res == FR_OK) res = FR_EXIST;      /* Any object with same name is already existing */
      if (_FS_RPATH && res == FR_NO_FILE && (dj.fn[NS] & NS_DOT))
         res = FR_INVALID_NAME;
      if (res == FR_NO_FILE) {            /* Can create a new directory */
         dcl = create_chain(dj.fs, 0);      /* Allocate a cluster for the new directory table */
         res = FR_OK;
         if (dcl == 0) res = FR_DENIED;      /* No space to allocate a new cluster */
         if (dcl == 1) res = FR_INT_ERR;
         if (dcl == 0xFFFFFFFF) res = FR_DISK_ERR;
         if (res == FR_OK)               /* Flush FAT */
            res = move_window(dj.fs, 0);
         if (res == FR_OK) {               /* Initialize the new directory table */
            dsc = clust2sect(dj.fs, dcl);
            dir = dj.fs->win;
            mem_set(dir, 0, SSZ);
            mem_set(dir+DIR_Name, ' ', 8+3);   /* Create "." entry */
            dir[DIR_Name] = '.';
            dir[DIR_Attr] = AM_DIR;
            ST_DWORD(dir+DIR_WrtTime, tim);
            st_clust(dir, dcl);
            mem_cpy(dir+SZ_DIR, dir, SZ_DIR);    /* Create ".." entry */
            dir[33] = '.'; pcl = dj.sclust;
            if (dj.fs->fs_type == FS_FAT32 && pcl == dj.fs->dirbase)
               pcl = 0;
            st_clust(dir+SZ_DIR, pcl);
            for (n = dj.fs->csize; n; n--) { // Write dot entries and clear following sectors
               dj.fs->winsect = dsc++;
               dj.fs->wflag = 1;
               res = move_window(dj.fs, 0);
               if (res != FR_OK) break;
               mem_set(dir, 0, SSZ);
            }
         }
         if (res == FR_OK) res = dir_register(&dj); // Register the object to the directory
         if (res != FR_OK) {
            remove_chain(dj.fs, dcl);  // Could not register, remove cluster chain
         } else {
            dir = dj.dir;
            dir[DIR_Attr] = AM_DIR;   // Attribute
            ST_DWORD(dir+DIR_WrtTime, tim);  // Created time
            st_clust(dir, dcl);             //Table start cluster
            dj.fs->wflag = 1;
            res = sync(dj.fs);
         }
      }
      FREE_BUF();
   }
   SAMPLE_MIN_SP(1);
   return (res);
}


/*  Change attribute of the file or directory
 *
 *  This function will set and/or clear the attributes \ref AM_RDO (read only),
 *  \ref AM_ARC (archive), \ref AM_SYS (system) and \ref AM_HID (hidden) of
 *  the file or directory specified. The parameters \c value and \c mask are
 *  OR expressions of the attribute masks named. Setting (\c value ) has
 *  preference over clearing (\c mask ).
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 *  \param path Pointer to the file name (0-terminated)
 *  \param value Attribute flags to be set
 *  \param mask  Attribute flag  to be cleared
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_FILE, \ref FR_NO_PATH, \ref FR_INVALID_NAME,
 *          \ref FR_WRITE_PROTECTED, \ref FR_INVALID_DRIVE,
 *          \ref FR_NOT_ENABLED, \ref FR_NO_FILESYSTEM,
 *          FR_TIMEOUT, \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_chmod(const TCHAR* path, uint8_t value, uint8_t mask){
   DIR dj;
   DEF_NAMEBUF;
   INIT_BUF(dj);

   FRESULT res = chk_mounted(&path, &dj.fs, 1);
   if (res != FR_OK) goto weGo;
   res = follow_path(&dj, path);  // Follow the file path
   FREE_BUF();
   if (_FS_RPATH && res == FR_OK && (dj.fn[NS] & NS_DOT))
      res = FR_INVALID_NAME;
   if (res == FR_OK) {
      uint8_t * dir = dj.dir;
      if (!dir) {  //  Is it a root directory?
         res = FR_INVALID_NAME;
      } else {   // File or sub directory
         mask &= AM_RDO|AM_HID|AM_SYS|AM_ARC;   // Valid attribute clear mask
         value &= AM_RDO|AM_HID|AM_SYS|AM_ARC;   // Valid attribute set mask
         if (!(mask | value)) goto weGo; // nothing to do
         dir[DIR_Attr] = value | (dir[DIR_Attr] & (uint8_t)~mask);
     // (value & mask) | (dir[DIR_Attr] & (uint8_t)~mask);   /* Apply attribute change */
                  dj.fs->wflag = 1;
         res = sync(dj.fs);
      }
   }
   weGo:
   return res;
} // f_chmod



/*  Change time-stamps of a file or directory
 *
 * \param path Pointer to the file/directory name
 * \param fno  Pointer to a  FILINFO structure with .fdate and .ftime values
 *             to be set to the file respectively directory
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_FILE, \ref FR_NO_PATH, \ref FR_INVALID_NAME,
 *          \ref FR_WRITE_PROTECTED, \ref FR_INVALID_DRIVE,
 *          \ref FR_NOT_ENABLED, \ref FR_NO_FILESYSTEM,
 *          FR_TIMEOUT, \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_utime(TCHAR const * path, FILINFO const * fno){
   DIR dj;
   DEF_NAMEBUF;
   FRESULT  res = chk_mounted(&path, &dj.fs, 1);
   if (res == FR_OK) {
      INIT_BUF(dj);
      res = follow_path(&dj, path);   /* Follow the file path */
      FREE_BUF();
      if (_FS_RPATH && res == FR_OK && (dj.fn[NS] & NS_DOT))
         res = FR_INVALID_NAME;
      if (res == FR_OK) {
         uint8_t * dir = dj.dir;
         if (!dir) {               /* Root directory */
            res = FR_INVALID_NAME;
         } else {               /* File or sub-directory */
            ST_WORD(dir+DIR_WrtTime, fno->ftime);
            ST_WORD(dir+DIR_WrtDate, fno->fdate);
            dj.fs->wflag = 1;
            res = sync(dj.fs);
         }
      }
   }
   return res;
}


/*  Rename or move a file or directory
 *
 *  This function renames an object (file or directory) and can also move it
 *  to other directory. The logical drive number is determined by old name,
 *  new name must not contain a logical drive number.
 *
 *  Do not rename open objects.
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 * \param path_old Pointer to the current name / path
 * \param path_new Pointer to the new name / path
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_FILE, \ref FR_NO_PATH, \ref FR_INVALID_NAME,
 *          \ref FR_DENIED,  \ref FR_EXIST, \ref FR_WRITE_PROTECTED,
 *          \ref FR_INVALID_DRIVE, \ref FR_NOT_ENABLED,
 *          \ref FR_NO_FILESYSTEM, FR_TIMEOUT,
 *          \ref FR_LOCKED, \ref FR_NOT_ENOUGH_CORE */
FRESULT f_rename(TCHAR const * path_old, TCHAR const * path_new){
   DIR djo, djn;
   uint8_t buf[21], *dir;
   uint32_t dw;
   DEF_NAMEBUF;

   FRESULT res = chk_mounted(&path_old, &djo.fs, 1);
   if (res == FR_OK) {
      djn.fs = djo.fs;
      INIT_BUF(djo);
      res = follow_path(&djo, path_old);      /* Check old object */
      if (_FS_RPATH && res == FR_OK && (djo.fn[NS] & NS_DOT))
         res = FR_INVALID_NAME;
      if (res == FR_OK) {                  /* Old object is found */
         if (!djo.dir) {                  /* Is root dir? */
            res = FR_NO_FILE;
         } else {
            mem_cpy(buf, djo.dir+DIR_Attr, 21);      /* Save the object information except for name */
            mem_cpy(&djn, &djo, sizeof (DIR));      /* Check new object */
            res = follow_path(&djn, path_new);
            if (res == FR_OK) res = FR_EXIST;      /* The new object name is already existing */
            if (res == FR_NO_FILE) {             /* Is it a valid path and no name collision? */
/* Start critical section that an interruption or error can cause cross-link */
               res = dir_register(&djn);         /* Register the new entry */
               if (res == FR_OK) {
                  dir = djn.dir;               /* Copy object information except for name */
                  mem_cpy(dir+13, buf+2, 19);
                  dir[DIR_Attr] = buf[0] | AM_ARC;
                  djo.fs->wflag = 1;
                  if (djo.sclust != djn.sclust && (dir[DIR_Attr] & AM_DIR)) {      /* Update .. entry in the directory if needed */
                     dw = clust2sect(djo.fs, ld_clust(djo.fs, dir));
                     if (!dw) {
                        res = FR_INT_ERR;
                     } else {
                        res = move_window(djo.fs, dw);
                        dir = djo.fs->win+SZ_DIR;   /* .. entry */
                        if (res == FR_OK && dir[1] == '.') {
                           dw = (djo.fs->fs_type == FS_FAT32 && djn.sclust == djo.fs->dirbase) ? 0 : djn.sclust;
                           st_clust(dir, dw);
                           djo.fs->wflag = 1;
                        }
                     }
                  }
                  if (res == FR_OK) {
                     res = dir_remove(&djo);      /* Remove old entry */
                     if (res == FR_OK)
                        res = sync(djo.fs);
                  }
               }
/* End critical section */
            }
         }
      }
      FREE_BUF();
   }
   return res;
}

#endif /* !_FS_READONLY */
#endif /* _FS_MINIMIZE == 0 */
#endif /* _FS_MINIMIZE <= 1 */
#endif /* _FS_MINIMIZE <= 2 */



/*-----------------------------------------------------------------------*/
/* Forward data to the stream directly (available on only tiny cfg)      */
/*-----------------------------------------------------------------------*/
#if _USE_FORWARD && _FS_TINY

FRESULT f_forward(
   FIL *fp,       /* Pointer to the file object */
   uint16_t (*func)(const uint8_t*,uint16_t),   /* Pointer to the streaming function */
   uint16_t btr,      /* Number of bytes to forward */
   uint16_t *bf){        /* Pointer to number of bytes forwarded */

   FRESULT res;
   uint32_t remain, clst, sect;
   uint16_t rcnt;
   uint8_t csect;


   *bf = 0;   /* Clear transfer byte counter */

   if (!fp) return FR_INVALID_OBJECT;

   res = validate(fp);            /* Check validity of the object */
   if (res != FR_OK) return (res);
   if (fp->flag & FA__ERROR)          /* Check error flag */
      return (FR_INT_ERR);
   if (!(fp->flag & FA_READ))                  /* Check access mode */
      return (FR_DENIED);

   remain = fp->fsize - fp->fptr;
   if (btr > remain) btr = (uint16_t)remain;         /* Truncate btr by remaining bytes */

   for ( ;  btr && (*func)(0, 0);               /* Repeat until all data transferred or stream becomes busy */
      fp->fptr += rcnt, *bf += rcnt, btr -= rcnt) {
      csect = (uint8_t)(fp->fptr / SSZ & (fp->fs->csize - 1));   /* Sector offset in the cluster */
      if ((fp->fptr % SSZ) == 0) {         /* On the sector boundary? */
         if (!csect) {                     /* On the cluster boundary? */
            clst = (fp->fptr == 0) ?         /* On the top of the file? */
               fp->sclust : get_fat(fp->fs, fp->clust);
            if (clst <= 1) ABORT(fp->fs, FR_INT_ERR);
            if (clst == 0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
            fp->clust = clst;               /* Update current cluster */
         }
      }
      sect = clust2sect(fp->fs, fp->clust);      /* Get current data sector */
      if (!sect) ABORT(fp->fs, FR_INT_ERR);
      sect += csect;
      if (move_window(fp->fs, sect))            /* Move sector window */
         ABORT(fp->fs, FR_DISK_ERR);
      fp->dsect = sect;
      rcnt = SSZ - (uint16_t)(fp->fptr % SSZ);   /* Forward data from sector window */
      if (rcnt > btr) rcnt = btr;
      rcnt = (*func)(&fp->fs->win[(uint16_t)fp->fptr % SSZ], rcnt);
      if (!rcnt) ABORT(fp->fs, FR_INT_ERR);
   }

   return (FR_OK);
}
#endif /* _USE_FORWARD */


#if _USE_STRFUNC
/*-----------------------------------------------------------------------*/
/* Get a string from the file                                            */
/*-----------------------------------------------------------------------*/
TCHAR* f_gets (
   TCHAR* buff,   /* Pointer to the string buffer to read */
   int len,      /* Size of string buffer (characters) */
   FIL* fil      /* Pointer to the file object */
)
{
   int n = 0;
   TCHAR c, *p = buff;
   uint8_t s[2];
   uint16_t rc;


   while (n < len - 1) {         /* Read bytes until buffer gets filled */
      f_read(fil, s, 1, &rc);
      if (rc != 1) break;         /* Break on EOF or error */
      c = s[0];
#if _LFN_UNICODE               /* Read a character in UTF-8 encoding */
      if (c >= 0x80) {
         if (c < 0xC0) continue;   /* Skip stray trailer */
         if (c < 0xE0) {         /* Two-byte sequence */
            f_read(fil, s, 1, &rc);
            if (rc != 1) break;
            c = ((c & 0x1F) << 6) | (s[0] & 0x3F);
            if (c < 0x80) c = '?';
         } else {
            if (c < 0xF0) {      /* Three-byte sequence */
               f_read(fil, s, 2, &rc);
               if (rc != 2) break;
               c = (c << 12) | ((s[0] & 0x3F) << 6) | (s[1] & 0x3F);
               if (c < 0x800) c = '?';
            } else {         /* Reject four-byte sequence */
               c = '?';
            }
         }
      }
#endif
#if _USE_STRFUNC >= 2
      if (c == '\r') continue;   /* Strip '\r' */
#endif
      *p++ = c;
      n++;
      if (c == '\n') break;      /* Break on EOL */
   }
   *p = 0;
   return n ? buff : 0;         /* When no data read (eof or error), return with error. */
}



#if !_FS_READONLY
#include <stdarg.h>
/*-----------------------------------------------------------------------*/
/* Put a character to the file                                           */
/*-----------------------------------------------------------------------*/
int f_putc (
   TCHAR c,   /* A character to be output */
   FIL* fil   /* Pointer to the file object */
)
{
   uint16_t bw, btw;
   uint8_t s[3];


#if _USE_STRFUNC >= 2
   if (c == '\n') f_putc ('\r', fil);   /* LF -> CRLF conversion */
#endif

#if _LFN_UNICODE   /* Write the character in UTF-8 encoding */
   if (c < 0x80) {         /* 7-bit */
      s[0] = (uint8_t)c;
      btw = 1;
   } else {
      if (c < 0x800) {   /* 11-bit */
         s[0] = (uint8_t)(0xC0 | (c >> 6));
         s[1] = (uint8_t)(0x80 | (c & 0x3F));
         btw = 2;
      } else {         /* 16-bit */
         s[0] = (uint8_t)(0xE0 | (c >> 12));
         s[1] = (uint8_t)(0x80 | ((c >> 6) & 0x3F));
         s[2] = (uint8_t)(0x80 | (c & 0x3F));
         btw = 3;
      }
   }
#else            /* Write the character without conversion */
   s[0] = (uint8_t)c;
   btw = 1;
#endif
   f_write(fil, s, btw, &bw);      /* Write the char to the file */
   return (bw == btw) ? 1 : EOF;   /* Return the result */
}




/*-----------------------------------------------------------------------*/
/* Put a string to the file                                              */
/*-----------------------------------------------------------------------*/
int f_puts (
   const TCHAR* str,   /* Pointer to the string to be output */
   FIL* fil         /* Pointer to the file object */
)
{
   int n;


   for (n = 0; *str; str++, n++) {
      if (f_putc(*str, fil) == EOF) return EOF;
   }
   return n;
}




/*-----------------------------------------------------------------------*/
/* Put a formatted string to the file                                    */
/*-----------------------------------------------------------------------*/
int f_printf (
   FIL* fil,         /* Pointer to the file object */
   const TCHAR* str,   /* Pointer to the format string */
   ...               /* Optional arguments... */
)
{
   va_list arp;
   uint8_t f, r;
   uint16_t i, j, w;
   ULONG v;
   TCHAR c, d, s[16], *p;
   int res, chc, cc;


   va_start(arp, str);

   for (cc = res = 0; cc != EOF; res += cc) {
      c = *str++;
      if (c == 0) break;         /* End of string */
      if (c != '%') {            /* Non escape character */
         cc = f_putc(c, fil);
         if (cc != EOF) cc = 1;
         continue;
      }
      w = f = 0;
      c = *str++;
      if (c == '0') {            /* Flag: '0' padding */
         f = 1; c = *str++;
      } else {
         if (c == '-') {         /* Flag: left justified */
            f = 2; c = *str++;
         }
      }
      while (isDigit(c)) {      /* Precision */
         w = w * 10 + c - '0';
         c = *str++;
      }
      if (c == 'l' || c == 'L') {   /* Prefix: Size is long int */
         f |= 4; c = *str++;
      }
      if (!c) break;
      d = c;
      if (isLower(d)) d -= 0x20;
      switch (d) {            /* Type is... */
      case 'S' :               /* String */
         p = va_arg(arp, TCHAR*);
         for (j = 0; p[j]; j++) ;
         chc = 0;
         if (!(f & 2)) {
            while (j++ < w) chc += (cc = f_putc(' ', fil));
         }
         chc += (cc = f_puts(p, fil));
         while (j++ < w) chc += (cc = f_putc(' ', fil));
         if (cc != EOF) cc = chc;
         continue;
      case 'C' :               /* Character */
         cc = f_putc((TCHAR)va_arg(arp, int), fil); continue;
      case 'B' :               /* Binary */
         r = 2; break;
      case 'O' :               /* Octal */
         r = 8; break;
      case 'D' :               /* Signed decimal */
      case 'U' :               /* Unsigned decimal */
         r = 10; break;
      case 'X' :               /* Hexdecimal */
         r = 16; break;
      default:               /* Unknown type (pass-through) */
         cc = f_putc(c, fil); continue;
      }

      /* Get an argument and put it in numeral */
      v = (f & 4) ? (ULONG)va_arg(arp, long) : ((d == 'D') ? (ULONG)(long)va_arg(arp, int) : (ULONG)va_arg(arp, unsigned int));
      if (d == 'D' && (v & 0x80000000)) {
         v = 0 - v;
         f |= 8;
      }
      i = 0;
      do {
         d = (TCHAR)(v % r); v /= r;
         if (d > 9) d += (c == 'x') ? 0x27 : 0x07;
         s[i++] = d + '0';
      } while (v && i < sizeof s / sizeof s[0]);
      if (f & 8) s[i++] = '-';
      j = i; d = (f & 1) ? '0' : ' ';
      res = 0;
      while (!(f & 2) && j++ < w) res += (cc = f_putc(d, fil));
      do res += (cc = f_putc(s[--i], fil)); while(i);
      while (j++ < w) res += (cc = f_putc(' ', fil));
      if (cc != EOF) cc = res;
   }

   va_end(arp);
   return (cc == EOF) ? cc : res;
}

#endif /* !_FS_READONLY */
#endif /* _USE_STRFUNC */

