/**
 * \file mmc.c
 *
 *  Memory card FAT file system functions
 *
 *  This file is the (Grande) FatFs low level disk control module
 *  implementation for ATmel AVR. In this implementation it is just clue
 *  code to \ref intro_sec "weAutSys"'s SMC handling functions (often just
 *  transforming return codes).
 *
 *  Copyright (C) 2010, ChaN, all rights reserved.
 *
 *  Modification for weAutSys by A. Weinert <br />
 *  Copyright (c) 2012 Albrecht Weinert <br />
 *  weinert - automation, Bochum <br />
 *  weinert-automation.de  weAut.de  a-weinert.de
 *
 *  See the notes on copyright and modifications in file
 *  \ref ff.h "fatFS/ff.h". See also  \ref diskio.h "fatFS/diskio.h"
 *
 * 2
 * 25.01.2017
 */

#include "fatFS/diskio.h"
#include "we-aut_sys/smc.h"
#include "we-aut_sys/utils.h"

#include <string.h>

/*  Initialise disk drive
 *
 * \return 0: success, \ref STA_NOINIT | \ref STA_NODISK
 */
uint8_t disk_initialize(uint8_t drv){
   if (drv) return STA_NODISK; // we support only single drive (0)
   if (!smcInsPow()){
      smcState.flag |= FL_CAT | CT_INS_POWD;
      return STA_NODISK;
   }
   return getSMCtype() ? 0 : STA_NOINIT;
}  // disk_initialize(uint8_t)

/*  Get drive status
 *
 *  This function just informs about the SMC drive status.
 *
 *  \param drv the physical drive number to initialise
 *             (must be 0 in this implementation)
 *  \return 0: success, \ref STA_NOINIT, \ref STA_NODISK <br />
 *  The \ref STA_PROTECT would never be set in this SMC implementation
 *  without WP-switch.
 */
uint8_t disk_status(uint8_t drv){
   if (drv) return STA_NODISK;  // at present supports only single drive
   if (!smcInsPow()) return STA_NODISK;
   return smcState.cardType & CT_TYPE_MASK ? 0 :  STA_NOINIT;
} // disk_status(uint8_t)


/*  Read a sector (/ read sectors)
 *
 *  This function reads a sector into the buffer \c buff supplied. As the
 *  process takes 2 ms on SMCs with 5MHz SPI clock frequency multiple sectors
 *  are disallowed. This function basically delegates to
 *  \ref readDataBlock() "readDataBlock(sector, buff)". Hence the sector
 *  number \c sector is just that cause \ref intro_sec "weAutSys"'s
 *  \ref readDataBlock() takes care of address transformations required for
 *  some low capacity SMCs.
 *
 *  Instead of doing block operations by this functions it would be better
 *  to let \ref intro_sec "weAutSys" handle that in the respective
 *  background thread. That would then be done more compatibly with real time
 *  (process automation) tasks.
 *
 *  The same restrictions and recommendations apply to \ref disk_write().
 *
 * \param drv the physical drive number to initialise
 *             (must be 0 in this implementation)
 * \param buff Pointer to the read buffer (NULL not allowed)
 * \param sector  Sector number (LBA)
 */
DRESULT disk_read(uint8_t drv, uint8_t* buff, uint32_t sector){
   if (drv) return RES_PARERR;
   return readDataBlock(sector, buff); //  ?  RES_ERROR : RES_OK;
} // disk_read (uint8_t, uint8_t* buff, uint32_t)


/** Write a sector (/ write sectors)
 *
 * \internal the thing takes 2 ms on 5MHz for one sector
 *
 * \param drv the physical drive number to initialise
 *             (must be 0 in this implementation)
 * \param buff pointer to the write buffer (NULL: not allowed)
 * \param sector  Sector number (LBA)
 */
DRESULT disk_write(uint8_t drv, const uint8_t* buff, uint32_t sector){
   if (drv) return RES_PARERR;
   return writeDataBlock(sector, buff); //  ? RES_ERROR : RES_OK;
} // disk_write(uint8_t, const uint8_t*, uint32_t)


#if _USE_IOCTL
/*  Special control functions
 *
 *  Depending on the device and implementation this function allows to get
 *  (/put) certain information from the device: <br />
 *  \ref CTRL_POWER: no effect on weAut_01 module <br />
 *  \ref GET_SECTOR_COUNT: get number of sectors on the disk (uint32_t) <br />
 *  \ref GET_SECTOR_SIZE: get R/W sector size (uint16_t; this implementation
 *                         will always return 512) <br />
 *  \ref GET_BLOCK_SIZE: get the erase block size in unit of
 *                        sectors (uint16_t) <br />
 *  \ref MMC_GET_TYPE: get card type flags (1 byte) <br />
 *  \ref MMC_GET_CSD: receive CSD as a data block (16 bytes)  <br />
 *  \ref MMC_GET_CID: receive CID as a data block (16 bytes)  <br />
 *  \ref MMC_GET_OCR: receive OCR as an R3 (4 bytes) <br />
 *  \ref MMC_GET_SDSTAT: receive SD status as a data block (64 bytes) <br />

 * \param drv the physical drive number to initialise
 *             (must be 0 in this implementation)
 * \param ctrl  control code
 * \param buff  buffer to send/receive control data
 */
DRESULT disk_ioctl(uint8_t drv, uint8_t ctrl, uint8_t* buff){
   if (drv) return RES_PARERR; // yet only one drive

   DRESULT res = RES_ERROR;
   if (ctrl == CTRL_POWER) {
      switch (buff[0]) {
         case 0:      // Sub control code (POWER_OFF)
            //  not implemented power_off();
            res = 0; // no error
            break;
         case 1:      // Sub control code (POWER_GET)
            buff[1] = 1; // not implemented (BYTE)power_status();
            res = 0; // no error
            break;
         default :
            res = RES_PARERR;
            break;
      }
      return res;
   }
   if (! smcTypeD()) return RES_NOTRDY;

   // uint16_t csize;
   // uint8_t n;

   switch (ctrl) {

      case GET_SECTOR_COUNT :   // Get number of sectors on the disk (uint32_t)
         *(uint32_t*)buff = smcGetSectCount();
         /* out
            if ((smcState.csd[0] >> 6) == 1) {   // SDC version 2.00
               csize = smcState.csd[9] + ((uint16_t)smcState.csd[8] << 8) + 1;
               *(uint32_t*)buff = (uint32_t)csize << 10;
            } else {               // SDC version 1.XX or MMC
               n = (smcState.csd[5] & 15) + ((smcState.csd[10] & 128) >> 7)
                     + ((smcState.csd[9] & 3) << 1) + 2;
               csize = (smcState.csd[8] >> 6)
                     + ((uint16_t)smcState.csd[7] << 2)
                     + ((uint16_t)(smcState.csd[6] & 3) << 10) + 1;
               *(uint32_t*)buff = (uint32_t)csize << (n - 9);
            }
            */
            res = 0; // no error
         break;

      case GET_SECTOR_SIZE: // Get R/W sector size (WORD)
         *(uint16_t*)buff = 512;
         res = 0; // no error
         break;

      case GET_BLOCK_SIZE:   // Get erase block size in unit of sector (uint32_t)
         if (smcState.cardType & CT_SD2) {   // SDv2?
            *(uint32_t*)buff = 16UL << (smcState.csd[10] >> 4);
            res = 0; // no error
         } else {   // SDv1 or MMCv3
            if (smcState.cardType & CT_SD1) {   // SDv1
               *(uint32_t*)buff = (((smcState.csd[10] & 63) << 1)
                     + ((uint16_t)(smcState.csd[11] & 128) >> 7) + 1) << ((smcState.csd[13] >> 6) - 1);
            } else { // MMCv3
               *(uint32_t*)buff = ((uint16_t)((smcState.csd[10] & 124) >> 2) + 1)
                              * (((smcState.csd[11] & 3) << 3)
                                    + ((smcState.csd[11] & 224) >> 5) + 1);
            }
            res = 0; // no error
         }
         break;

      case MMC_GET_TYPE :   // Get card type flags (1 byte)
         *buff = smcState.cardType &  CT_TYPE_MASK;
         res = 0; // no error
         break;

      case MMC_GET_CSD :   // get CSD as a data block (16 bytes)
         memcpy(buff, smcState.csd, 16);
         res = 0; // no error
         break;

      case MMC_GET_CID :      /* Receive CID as a data block (16 bytes) */
         if (!smcReadCID(buff))  res = 0; // no error
         //if (!sendCmd0arg(SEND_CID)) {/* READ_CID */
           // spi2ReceiveSN(buff, 0, 16);
           // res = 0; // no error
         //}
         break;

      case MMC_GET_OCR :      /* Receive OCR as an R3 resp (4 bytes) */
         memcpy(buff, smcState.ocr, 4);
         res = 0; // no error
         break;

      case MMC_GET_SDSTAT :   /* Receive SD status as a data block (64 bytes) */
         if (sendAppCmd(13, 0) == 0) {   /* SD_STATUS */
            spi2ReceiveSN(buff, 1, 64);
            res = 0; // no error
         }
         break;

      default:
         res = RES_PARERR;
         break;
   }

   deSelectSMC();
   return res;
}
#endif

