/**
 * \addtogroup uipopt
 * @{
 *
 * \name Project-specific configuration options
 * @{
 *
 * uIP has a number of configuration options that can be overridden
 * for each project. These are kept in a project-specific uip-conf.h
 * file and all configuration names have the prefix UIP_CONF.
 */

/* Copyright (c) 2006, Swedish Institute of Computer Science.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**
 * \file  uip-conf.h
 * IP configuration file
 *
 * Configuration definitions given in this file are (indirectly) used to
 * define uIP configuration values.
 *
 *  Adam Dunkels' original copyright notice <br /> &nbsp;
 *  Copyright (c) 2006, Swedish Institute of Computer Science.
 *  All rights reserved. <br />
 *  (contained in this file) still holds. For modifications:
 *
 *  <br /> &nbsp;
 * \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author Adam Dunkels <adam@dunkels.com>
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

#ifndef __UIP_CONF_H__
#define __UIP_CONF_H__

#include "we-aut_sys/common.h"

// note by A.W. on A.D.'s httpd_state
//      httpd_state is used in different uip files in different meaning and
//      structure. But using more than application that causes trouble
//      (see httph.h). A union of all state structures by all potentially
//      used applications is what was probably meant in the end.
//      See \ref uip_tcp_appstate_t \ref uipConn_t
//      You'll find httpd_state in these sources only at applications not
//      (yet) used within weAutSys. those source files, kept for uip
//      reference only in past releases, are removed now.


/** Maximum number of UDP connections
 *
 *  This value determines the length of the  UDP connections list (array
 *  uip_udp_conns). As the uip_udp_conn structure is quite small
 *  (~ 11 byte) this value is not critical.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref UIP_UDP_CONNS
 *  \showinitializer
 *  \internal originally 10
 */
#define UIP_CONF_UDP_CONNS 10

/** Maximum number of TCP connections
 *
 *  This value determines the length of the TCP connections list (array
 *  \ref uip_conns). As the \ref uip_conn structure is very large
 *  (> kByte if http is used) this value is critical.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref UIP_CONNS
 *  \showinitializer
 *  \internal originally 40
 */
#define UIP_CONF_MAX_CONNECTIONS 10

/** Maximum number of listening TCP ports
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref UIP_LISTENPORTS
 *  \internal  originally 40
 *  \showinitializer
 */
#define UIP_CONF_MAX_LISTENPORTS 20

/** The size of the ARP table
 *
 *  This option should be set to a larger value if this uIP node will
 *  have many connections from the local network. (uIP default would be 8)
 *  \showinitializer
 */
#define UIP_CONF_ARPTAB_SIZE 12


/** uIP buffer size
 *
 *  This is the size of \ref uip "uIP"'s global transfer (RAM) buffer uIP
 *  uses for sending and receiving, also for prepended ARP. <br />
 *  A.D. used 420 byte. A value > 590 allows to work with some DHCP servers
 *  (like Fritz!boxes) without restrictions.
 *  \showinitializer
 */
#define UIP_CONF_BUFFER_SIZE  620

/** CPU byte order
 *
 *  ATmega and AVR GCC used in \ref intro_sec "weAutSys" are little endian.
 *
 *  With this setting for \ref uip "uIP" we are on the save side. uIP does
 *  have this macro to be architecture independent, but there is uIP code
 *  just assuming little endian.
 *  \showinitializer
 */
#define UIP_CONF_BYTE_ORDER UIP_LITTLE_ENDIAN

/** Logging on or off
 *
 *  Used to set \ref UIP_LOGGING
 * \hideinitializer
 */
#define UIP_CONF_LOGGING  0

/** UDP support on or off
 *
 * Used to set \ref UIP_UDP
 * \hideinitializer
 */
#define UIP_CONF_UDP  1

/** UDP checksums on or off
 *
 * Used to set \ref UIP_UDP_CHECKSUMS
 * \showinitializer
 */
#define UIP_CONF_UDP_CHECKSUMS  1

/** Check incoming checksums on or off
 *
 *  Consideration and experiment show this can safely be turned off. (ENC
 *  checks incoming aggregate checksum and drops spoiled packets.
 *
 * Used to set \ref UIP_INC_CHECKSUMS
 * \showinitializer
 */
#define UIP_CONF_INC_CHECKSUMS  0

/** uIP statistics on or off
 *
 * Used to set \ref UIP_STATISTICS
 * \showinitializer
 */
#define UIP_CONF_STATISTICS  0


#define UIP_CONF_BROADCAST 1

/** The uIP event function
 *
 *  The reason for being called will be in \c uip_flags.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref UIP_APPCALL
 */
void uip_appcall(void);

/** Macro to name the uIP event function
 *
 *  An uIP application is implemented using a single application function
 *  that is called by uIP whenever a TCP/IP event occurs. The name of this
 *  function must be registered with uIP at compile time using this
 *  UIP_APPCALL definition.
 *
 *  The function named must return void and take no arguments. It will be
 *  called on uIP events.
 *
 *  The applications provided by uIP (like Telnet etc.) will set this macro
 *  to their specific event function, if not yet set. If there are multiple
 *  applications the function set here must handle them all, e.g. by acting
 *  as forwarder to the application specific functions.
 *
 *  uIP applications can store the application state within the uip_conn
 *  structure by specifying the type of the application structure by
 *  typedef'ing the type uip_tcp_appstate_t and uip_udp_appstate_t.
 *
 *  Remark by A.W.: A.D.'s typedef'ing approach (recommended here some ten
 *  years ago) does not carry further than to single TCP/IP application.
 *  But using more than one application and hence contradictory definitions
 *  (distributed all over original uIP's sources) causes trouble. Obviously
 *  (by now) a union of all state structures by all applications used is what
 *  A.D. must have meant in the end and what is done by
 *  \ref intro_sec "weAutSys".
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref  uip_appcall
 */
#define UIP_APPCALL uip_appcall

/** The uIP udp event function
 *
 *  The reason for being called will be in the \c uip_udp_conn structure,
 *  especially in the ports.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref UIP_UDP_APPCALL
 */
void udp_appcall(void);

/** Macro to name the uIP udp event function
 *
 * \sa handle_dhcp()
 * \sa udp_appcall
 * \sa UIP_APPCALL
 */
#define UIP_UDP_APPCALL udp_appcall


/** The type of appstate in \ref uip_conn
 *
 *  uip_con.appstate will have to be cast to application specific types,
 *  usually struct \ref thr_data_t "struct thr_data_t".
 */
typedef uint8_t uip_tcp_appstate_t[SIZE_OF_BIGGEST_APPSTATE];

#endif /* __UIP_CONF_H__ */

/** @} */ // Project-specific configuration options
/** @} */
