/**
 * \addtogroup osSWinc
 * @{
 * \addtogroup fatfs
 * @{
 */

/**
 * \file diskio.h
 *
 *  Small memory card (SMC / MMC) driver function definitions
 *
 *  This is the basic include file for ChaN's (grande) fatFS. It defines the
 *  low level disk / media I/O driver functions.
 *
 *  Copyright (c) 2012, ChaN, all right reserved. <br />
 *
 *
 *  Modification for \ref intro_sec "weAutSys" by A. Weinert <br />
 *  \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert <br />
 *  weinert - automation, Bochum <br />
 *  weinert-automation.de  weAut.de  a-weinert.de
 *
 *  A.W.'s modifications for \ref intro_sec "weAutSys" are
  -  adaptions to ATmel ATmega1284P,
     i.e. \ref intro_secH "weAut_01" and similar small (petit) systems,
  -  performance improvements and removal of compiler warnings
  -  make usable on non pre-emptive (Protothreads based) run-time, i.e
     \ref intro_sec "weAutSys"
  -  documentation enhancements, mainly by adding (Doxygen) documentation
     to the code
 *
 *  See the notes on copyright and modifications in file
 *  \ref ff.h "fatFS/ff.h".
 *
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
/** @} */
/** @} */ // osSWinc

#ifndef _DISKIO_DEFINED
#define _DISKIO_DEFINED

#include <stdint.h>

/**
 * \addtogroup osSWinc
 * @{
 * \addtogroup fatfs
 * @{
 * \defgroup smcdvr Driver adaption to SMC
 * @{
 * \details
 * The functions in this architecture specific module mainly implement the
 * glue code between ChaN's file system adaption and
 * \ref intro_sec "weAutSys"' driver and handling functions for
 * small memory cards (\ref smcCom "SMCs").
 */

#define _USE_WRITE 1  //!< enable disk_write function \showinitializer
#define _USE_IOCTL 1  //!< enable disk_ioctl function \showinitializer


/** Results of disk (media driver) functions
 *
 *  Hint: The values assigned here for (Grande) FAT FS are not the same
 *        as with Petit FatFs.
 *  Remark 2: Not any usage of the (three) media access functions within
 *        this fatFS version really uses those values: Some just compare to
 *        0 (RES_OK) and some do not care at all. Hence we (this \ref intro_sec "weAutSys"
 *        port, A.W.) change it to uint8_t and "0 = false = no error = ".
 */
#define DRESULT uint8_t
#define RES_ERROR  1 //!< Any (hard) error occurred  \showinitializer
#define RES_NOTRDY  3 //!< The drive has not been initialised \showinitializer
#define RES_PARERR  3 //!< Invalid function argument \showinitializer

/** Initialise the SMC drive
 *
 *  This function is used in fatFS to initialise the drive.
 *  Legal return values are \ref STA_NOINIT, \ref STA_NODISK, \ref STA_PROTECT
 *  and 0 for OK. \ref STA_PROTECT cannot occur with SMCs as there is no
 *  hardware write protect mechanism.
 *
 *  With \ref intro_sec "weAutSys" this function just checks the
 *  initialisation, hence doing (in effect) same as \ref disk_status(). The
 *  reason is  \ref intro_sec "weAutSys" doing the SMC initialisation and
 *  other time consuming SMC actions in a background (Protothread) thread .
 *
 *  \param drv the number of the physical drive to initialise
 *             (must be 0 in this implementation)
 *  \return 0: success, \ref STA_NOINIT | \ref STA_NODISK
 */
uint8_t disk_initialize(uint8_t drv);

/** Get drive status
 *
 *  This function just informs about the drive status without actually
 *  accessing the device.
 *
 *  \param drv the physical drive number to check (must be 0 in
 *             this implementation)
 *  \return 0: success, \ref STA_NOINIT | \ref STA_NODISK
 *  \sa disk_initialize()
 */
uint8_t disk_status(uint8_t drv);


/** Read a sector
 *
 *  This function reads a sector into the buffer \c buff supplied. As the
 *  process takes 2 ms on SMCs with 5MHz SPI clock frequency multiple sectors
 *  are disallowed. This function basically delegates to
 *  \ref readDataBlock() "readDataBlock(sector, buff)". Hence the sector
 *  number \c sector is just that cause \ref intro_sec "weAutSys"'s
 *  \ref readDataBlock() takes care of address transformations required for
 *  some low capacity SMCs.
 *
 *  Instead of doing block operations by this functions it is preferable to
 *  let \ref intro_sec "weAutSys" handle that in the respective background
 *  thread. That would then be done more compatibly with real time
 *  (process automation) tasks.
 *
 *  The same restrictions and recommendations apply to \ref disk_write().
 *
 * \param drv the physical drive number to initialise
 *             (must be 0 in this implementation)
 * \param buff Pointer to the read buffer (NULL not allowed)
 * \param sector  Sector number (LBA)
 */
DRESULT disk_read(uint8_t drv, uint8_t* buff, uint32_t sector);

/** Write a sector
 *
 *  In the sense described with \ref disk_read() this function basically
 *  delegates to \ref writeDataBlock "writeDataBlock(sector, buff)".
 *
 * \param drv the physical drive number to initialise
 *             (must be 0 in this implementation)
 * \param buff pointer to the write buffer (NULL: not allowed)
 * \param sector  Sector number (LBA)
 */
DRESULT disk_write(uint8_t drv, const uint8_t* buff, uint32_t sector);


/** Special control functions
 *
 *  Depending on the device and implementation this function allows to get
 *  (/put) certain information from the device: <br />
 *  \ref CTRL_POWER : no effect on weAut_01 module <br />
 *  \ref GET_SECTOR_COUNT : get number of sectors on the disk (uint32_t) <br />
 *  \ref GET_SECTOR_SIZE : get R/W sector size (uint16_t; this implementation
 *                         will always return 512) <br />
 *  \ref GET_BLOCK_SIZE : get the erase block size in unit of
 *                        sectors (uint16_t) <br />
 *  \ref MMC_GET_TYPE : get card type flags (1 byte) <br />
 *  \ref MMC_GET_CSD : receive CSD as a data block (16 bytes)  <br />
 *  \ref MMC_GET_CID : &nbsp; receive CID as a data block (16 bytes)  <br />
 *  \ref MMC_GET_OCR : receive OCR as an R3 (4 bytes) <br />
 *  \ref MMC_GET_SDSTAT : receive SD status as a data block (64 bytes) <br />
 *
 *  Hint: This function is not implemented if \ref _USE_IOCTL is undefined or
 *        0.
 *
 * \param drv the physical drive number to initialise
 *             (must be 0 in this implementation)
 * \param ctrl  control code
 * \param buff  buffer to send/receive control data
 */
DRESULT disk_ioctl (uint8_t drv, uint8_t ctrl, uint8_t* buff);


//---- Disk Status Bits (DSTATUS) ----------------

/** Drive not initialized
 *
 *  This status (mask) bit indicates that the disk drive has not been
 *  initialised. The type of the (inserted) SMC is not determined yet .
 *
 * \showinitializer
 * \sa STA_NODISK
 * \sa disk_status()
 */
#define STA_NOINIT 0x01

/** No medium in the drive / slot
 *
 *  This status (mask) bit indicates no medium being inserted in the
 *  drive respectively in the small memory card (SMC / MMC) connector.
 *
 * \showinitializer
 * \sa STA_NOINIT
 * \sa disk_status()
 * \sa smcInsertSwitch()
 * \sa smcInsPow()
 */
#define STA_NODISK 0x02

/** Medium is write protected by hardware switch
 *
 *  Small memory cards (SMCs) do not have a write protect slider and
 *  the MMC-slot built in \ref intro_secH "weAut_01" accordingly has no
 *  WP-switch. Hence this feature is not implemented and this status won't
 *  occur.
 *
 * \showinitializer
 * \sa STA_NOINIT
 * \sa STA_NODISK
 */
#define STA_PROTECT 0x04


// Command codes for disk_ioctrl() function

/* Generic command (used by FatFs) */
#define GET_SECTOR_COUNT 1  //!< Generic command: get media size (for only f_mkfs()) \showinitializer
#define GET_SECTOR_SIZE 2  //!< Generic command: get sector size \showinitializer
#define GET_BLOCK_SIZE 3  //!< Generic command: get erase block size (for only f_mkfs()) \showinitializer
#define CTRL_ERASE_SECTOR 4  //!< Generic command: force erased a block of sectors (for only _USE_ERASE) \showinitializer

//--- Generic commands (not used by FatFs)
#define CTRL_POWER 5  //!< Generic command: get/set power status \showinitializer
//#define CTRL_LOCK 6  //!< Generic command: lock/unlock media removal \showinitializer
//#define CTRL_EJECT 7  //!< Generic command: eject media \showinitializer
//#define CTRL_FORMAT 8 //!< Generic command: create physical format on the media \showinitializer

//---   MMC/SDC specific ioctl commands
#define MMC_GET_TYPE  10  //!< Specific ioctl command: get card type \showinitializer
#define MMC_GET_CSD 11  //!< Specific ioctl command: get CSD \showinitializer
#define MMC_GET_CID 12  //!< Specific ioctl command: get CID \showinitializer
#define MMC_GET_OCR 13  //!< Specific ioctl command: get OCR \showinitializer
#define MMC_GET_SDSTAT 14  //!< Specific ioctl command: get SMC status \showinitializer

//----- ATA/CF specific ioctl commands
//#define ATA_GET_REV      20  //Get F/W revision
//#define ATA_GET_MODEL    21  // Get model name
//#define ATA_GET_SN      22  // Get serial number/


//----  Card type flags
#define CT_MMC 0x01 //!< Card type flag: MMC version 3 \showinitializer
#define CT_SD1 0x02 //!< Card type flag: SD version 1 \showinitializer
#define CT_SD2 0x04 //!< Card type flag: SD version 2 \showinitializer
#define CT_SDC (CT_SD1|CT_SD2) //!< Card type mask: SD \hideinitializer
#define CT_BLOCK 0x08 //!< Card type flag: block addressing \showinitializer

#endif
/** @} */
/** @} */
/** @} */ // osSWinc
