/* FatFs - FAT file system module include file  R0.09a    (C) ChaN, 2012
 *
 * FatFs module is a generic FAT file system module for small embedded
 * systems. This is a free software that opened for education, research
 * and commercial developments under license policy of following terms.
 *
 * Copyright (C) 2012, ChaN, all right reserved.
 *
 * The FatFs module is a free software and there is NO WARRANTY.
 * No restriction on use. You can use, modify and redistribute it for
 * personal, non-profit or commercial product UNDER YOUR RESPONSIBILITY.
 * Redistributions of source code must retain the above copyright notice.
*/
/**
 * \addtogroup osSWinc
 * @{
 * \addtogroup fatfs
 * @{
 */
/** \file ff.h
 *
 *  (Grande) fatFS configurations and declarations
 *
 *  This is the main include file for ChaN's (grande) fatFS for small memory
 *  cards (SMC, MMC) and  FAT file system functions. They are build upon the
 *  low level driver functions defined in file
 *  \ref diskio.h "fatFS/diskio.h".
 *
 *  Copyright (c) 2012, ChaN, all right reserved. <br />
 *  The original copyright notice is at the begin of the (this) source file
 *  fatFS/ff.h. It is still valid.
 *
 *  Modification for \ref intro_sec "weAutSys" by A. Weinert <br />
 \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert <br />
 *  weinert - automation, Bochum <br />
 *  weinert-automation.de  weAut.de  a-weinert.de
 *
 *  A.W.'s modifications for \ref intro_sec "weAutSys" are
  -  adaptions to ATmel AATmega1284P,
     i.e. \ref intro_secH "weAut_01" and similar small (petit) systems,
     -  restriction to SPI with small memory cards (SMCs)
     -  disallow multi-sector access for SMCs on SPI
     -  restriction to small file names at present for assumed license reasons
     -  removal of file lock feature
  -  performance improvements, mending of compiler warnings (bugs) etc.
  -  make usable on non pre-emptive (Protothreads based) run-time, i.e
     \ref intro_sec "weAutSys"
  -  documentation enhancements, mainly by adding (Doxygen) documentation
     to the code
 */
/** @} */
/** @} */ // osSWinc

#ifndef _FATFS
#define _FATFS  4004  // Revision ID

#include "fatFS/ffconf.h"      // FatFs configuration options
#include "we-aut_sys/timing.h" // for getFATtime

#if _FATFS != _FFCONF
#error Wrong configuration file (ffconf.h).
#endif


/**
 * \addtogroup osSWinc
 * @{
 * \addtogroup fatfs
 * @{
 * \defgroup fsop File system operations
 * @{
 * \details
 * The file operations provided by ChaN's fatFS were adapted to
 * \ref intro_secH "weAut_01" like hardware, the use with one slot for small
 * memory cards (\ref smcCom "SMCs") and the \ref intro_sec "weAutSys"
 * runtime. The necessary "cut to size" was quite far-reaching. But it was
 * not so radical, that the restriction to one device and one file system
 * could not be lifted up easily for future upgrading.
 */

//---- Definitions of volume management

#if _MULTI_PARTITION    /* Multiple partition configuration */
typedef struct {
  uint8_t pd;  /* Physical drive number */
  uint8_t pt;  /* Partition: 0:Auto detect, 1-4:Forced partition) */
} PARTITION;
extern PARTITION VolToPart[];  /* Volume - Partition resolution table */
#define LD2PD(vol) (VolToPart[vol].pd)  /* Get physical drive number */
#define LD2PT(vol) (VolToPart[vol].pt)  /* Get partition index */

#else              // Single partition configuration
#define LD2PD(vol) (uint8_t)(vol) //!< Each logical drive is bound to the same physical drive number */
#define LD2PT(vol) 0  //!< Always mounts the 1st partition or in SFD
#endif

//----------------------- sector size and related macros  -- start ----------

/** Maximum sector size to be handled
 *
 *  Always set 512 for memory card and hard disk but a larger value may be
 *  required for on-board flash memory, floppy disk and optical disk. When
 *  _MAX_SS is larger than 512, it configures FatFs to variable sector
 *  size and \ref GET_SECTOR_SIZE command must be implemented by the
 *  \ref disk_ioctl() function.
 *
 *  Hint: For small memory card (SMC, MMC) implementation \ref GET_SECTOR_SIZE
 *        is implemented in \ref disk_ioctl() blindly returning 512.
 *
 *  legal values for original fatFS: 512, 1024, 2048 or 4096 (configurable)<br />
 *  fixed value for this implementation:  512
 *
 * \showinitializer
 */
#define _MAX_SS 512
#define SSZ 512 //!< Fixed sector size  \sa _MAX_SS \showinitializer
#define SSM 511 //!< Mask for \ref SSZ "fixed sector size" \showinitializer

/** Multiply (32 bit, unsigned) with fixed sector size
 *
 * \sa SSZ
 * \param fac the 32 bit (unsigned) factor <br />
 *             fac's upper 9 bits are ignored and should be 0 to get a
 *             arithmetically correct result
 *  \return the product fac * \ref SSZ
 *  \sa div32SSZ(div)
 */
#define mul32SSZ(fac)  mul32withVal512(fac)

/** Multiply (16 bit, unsigned) with fixed sector size
 *
 * \sa SSZ
 * \param fac the 16 bit (unsigned) factor <br />
 *             fac's upper 9 bits are ignored and should be 0 to get a
 *             arithmetically correct result
 *  \return the product fac * \ref SSZ
 *  \sa mul32SSZ(div)
 */
#define mul16SSZ(fac)  mul16withVal512(fac)

/** Divide (32 bit, unsigned) by fixed sector size
 *
 *  \sa SSZ
 *  \param div the 32 bit (unsigned) dividend
 *  \return the quotient div /  \ref SSZ
 *  \sa mul32SSZ(fac)
 *  \sa div2SSZ(div)
 *  \sa div4SSZ(div)
 */
#define div32SSZ(div)  div32byVal512(div)

/** Divide (16 bit, unsigned) by fixed sector size
 *
 *  \sa SSZ
 *  \param div the 16 bit (unsigned) dividend
 *  \return the quotient div /  \ref SSZ
 *  \sa mul32SSZ(fac)
 *  \sa div32SSZ(div)
 */
#define div16SSZ(div)  div16byVal512(div)

/** Modulo (unsigned) by fixed sector size
 *
 *  \sa SSZ
 *  \param div the (unsigned) dividend
 *  \return the remainder div % \ref SSZ
 *  \sa mul32SSZ(fac)
 */
#define modSSZ(div)  (div & SSM)

/** Divide (32 bit, unsigned) by 2 times fixed sector size
 *
 *  \sa SSZ
 *  \param div the 32 bit (unsigned) dividend
 *  \return the quotient div / (\ref SSZ * 2)
 *  \sa mul32SSZ(fac)
 */
#define div2SSZ(div)  div32byVal1024(div)

/** Divide (32 bit, unsigned) by 4 times fixed sector size
 *
 *  \sa SSZ
 *  \param div the 32 bit (unsigned) dividend
 *  \return the quotient div / (\ref SSZ * 4)
 *  \sa mul32SSZ(fac)
 */
#define div4SSZ(div)  div32byVal2048(div)

/** Divide (32 bit, unsigned) by fixed sector size by 4
 *
 *  This divides a 32 bit number by a quarter of the
 *  \ref SSZ "fixed sector size".
 *
 *  \sa SSZ
 *  \param div the 32 bit (unsigned) dividend
 *  \return the quotient div / (\ref SSZ / 4)
 *  \sa mul32SSZ(fac)
 */
#define divSSZ4(div)  div32byVal128(div)

/** Divide (32 bit, unsigned) by fixed sector size by 2
 *
 *  This divides a 32 bit number by a quarter of the
 *  \ref SSZ "fixed sector size".
 *
 *  \sa SSZ
 *  \param div the 32 bit (unsigned) dividend
 *  \return the quotient div / (\ref SSZ / 2)
 *  \sa mul32SSZ(fac)
 */
#define divSSZ2(div)  div32byVal256(div)

//----------------------- sector size and related macros  -- end ------------

/** Tiny FS
 *
 *  When _FS_TINY is set to 1, fatFS uses the sector buffer in the
 *  \ref FATFS "file system" object instead of the sector buffer in the
 *  individual \ref FIL "file object" for file data transfer. This reduces
 *  the memory consumption by 512 bytes for each file object.
 *
 *  Legal values on original fatFS: 0 = normal or 1 = tiny &mdash;
 *  configurable there but fixed for this implementation.
 *  \internal todo add variant 2 to use underlined runtime's (weAutSys')
 *  own sector buffer directly.
 *  \showinitializer
 */
#define _FS_TINY 1

//  Type of path name strings on FatFS API

#if _LFN_UNICODE // Unicode string
#if !_USE_LFN
#error _LFN_UNICODE must be 0 in non-LFN cfg.
#endif
#ifndef _INC_TCHAR
typedef WCHAR TCHAR;
#define _T(x) L ## x
#define _TEXT(x) L ## x
#endif
#else  // Unicode string > ANSI/OEM string
#ifndef _INC_TCHAR
typedef char TCHAR;
#define _T(x) x
#define _TEXT(x) x
#endif
#endif // ANSI/OEM string



/** File system object structure (FATFS) */
typedef struct {
  uint8_t  fs_type; //!< FAT sub-type (0:Not mounted)
  uint8_t  drv;     //!< Physical drive number
  uint8_t  csize;   //!< Sectors per cluster (1,2,4...128)
  uint8_t  n_fats;  //!< Number of FAT copies (1,2)
  uint8_t  wflag;   //!< win[] dirty flag (1:must be written back)
  uint8_t  fsi_flag; //!< fsinfo dirty flag (1:must be written back)
  uint16_t id;       //!< File system mount ID
  uint16_t n_rootdir; //!< Number of root directory entries (FAT12/16)
#if _MAX_SS != 512
  uint16_t  ssize; //!< Bytes per sector (512, 1024, 2048 or 4096)
#endif
#if !_FS_READONLY
  uint32_t last_clust; //!< Last allocated cluster
  uint32_t free_clust; //!< Number of free clusters
  uint32_t fsi_sector; //!< fsinfo sector (FAT32)
#endif
#if _FS_RPATH
  uint32_t  cdir; //!<  Current directory start cluster (0:root)
#endif
  uint32_t n_fatent; //!< Number of FAT entries (= number of clusters + 2)
  uint32_t fsize;    //!<  Sectors per FAT
  uint32_t fatbase;  //!<  FAT start sector
  uint32_t dirbase;  //!<  Root directory start sector (FAT32: Cluster#)
  uint32_t database; //!<  Data start sector
  uint32_t winsect;  //!< Current sector appearing in the win[]

/** Disk access window for directory, FAT and data
 *
 *  This window respectively buffer is bound to the file system structure
 *  (FATFS) on \ref _FS_TINY "(fiexd) tiny" configuration.<br />
 *  Note (internal, future): On weAutSys' SMC this should be omitted as the
 *  (driver's) sector buffer is of same size and purpose. It also handles
 *  "dirty state" correctly / automatically.
 */
  uint8_t win[_MAX_SS];
} FATFS;


/** File object structure (FIL)
 *
 *  A structure of this type holds the state of an open file initialised by
 *  \ref f_open() and subsequently used by file functions.
 *
 *  Internal hint: some parts of ChaN's fatFS software assume the beginning of
 *  \ref DIR and \ref FIL being identical.
 *  \sa DIR
 */
typedef struct {
  FATFS *  fs;   //!< Pointer to the related file system object
  uint16_t id;   //!< File system mount ID of the related file system object
  uint8_t  flag; //!< File status flags
  uint8_t  pad1; //!< Fill byte
  uint32_t fptr;   //!< File read/write pointer (0ed on file open)
  uint32_t fsize;  //!< File size
  uint32_t sclust; //!< File data start cluster (0:no data cluster, always 0 when fsize is 0)
  uint32_t clust;  //!< Current cluster of fpter
  uint32_t dsect;  //!< Current data sector of fpter
#if !_FS_READONLY
  uint32_t dir_sect; //!< Sector containing the directory entry
  uint8_t * dir_ptr; //!< Pointer to the directory entry in the window
#endif
#if _USE_FASTSEEK
  uint32_t*  cltbl; //!< Pointer to the cluster link map table (null on file open)
#endif
#if !_FS_TINY
  uint8_t  buf[_MAX_SS]; //!< File data read/write buffer
#endif
}  FIL;

/** Check EOF
 *
 *  \param fp Pointer to the file's object structure (type \ref FIL)
 *  \return 1 if EOF reached
 */
#define f_eof(fp) (((fp)->fptr == (fp)->fsize) ? 1 : 0)

/** Check error state
 *
 *  \param fp Pointer to the file's object structure (type \ref FIL)
 *  \return 1 if in an erroneous state
 */
#define f_error(fp) (((fp)->flag & FA__ERROR) ? 1 : 0)

/** Get the current read/write pointer
 *
 *  \param fp Pointer to the open file's object structure (type \ref FIL)
 *  \return the current R/W pointer, i.e. byte index (type uint32_t)
 */
#define f_tell(fp) ((fp)->fptr)

/** Get the size
 *
 *  \param fp Pointer to the open file's object structure (type \ref FIL)
 *  \return the size (type uint32_t)
 */
#define f_size(fp) ((fp)->fsize)


/** Directory object structure (DIR)
 *
 *  A structure of this type holds the state of an open respectively selected
 *  directory initialised by \ref f_opendir and subsequently used by directory
 *  functions.
 *
 *  \sa FIL
 */
typedef struct {
  FATFS*   fs;   //!< Pointer to the owner file system object
  uint16_t id;   //!< Owner file system mount ID
  uint16_t index;  //!< Current read/write index number
  uint32_t sclust; //!< Table start cluster (0: Root dir)
  uint32_t clust;  //!< Current cluster
  uint32_t sect; //!< Current sector
  uint8_t* dir;  //!< Pointer to the current SFN entry in the win[]
  uint8_t* fn;   //!< Pointer to the SFN (in/out) {file[8],ext[3],status[1]}
#if _USE_LFN
  WCHAR*  lfn;       //!< Pointer to the LFN working buffer
  uint16_t  lfn_idx; //!< Last matched LFN index number (0xFFFF:No LFN)
#endif
} DIR;


/** File status structure (FILINFO)
 *
 *  A structure of this type holds some extra informations on a file or
 *  directory. It is determined by \ref f_stat() for later reference.
 * */
typedef struct {
  uint32_t fsize;   //!< File size
  uint16_t fdate;   //!< Last modified date
  uint16_t ftime;   //!< Last modified time
  uint8_t  fattrib; //!< Attribute See \ref AM_RDO, \ref AM_DIR
  TCHAR  fname[13]; //!< Short file name (8.3 format, 0-terminated)
  uint8_t namLen;   //!< Length of file name; 0..12 for short names
#if _USE_LFN
  TCHAR*   lfname;  //!< Pointer to the LFN buffer
  uint16_t lfsize;  //!< Size of LFN buffer in TCHAR
#endif
} FILINFO;


/** File functions return code  */
typedef enum {
  FR_OK = 0,   //!< (0) Succeeded
  FR_DISK_ERR, //!< (1) A hard error occurred in the low level disk I/O layer
  FR_INT_ERR,  //!< (2) Assertion failed
  FR_NOT_READY,//!< (3) The physical drive cannot work
  FR_NO_FILE,  //!< (4) Could not find the file
  FR_NO_PATH,  //!< (5) Could not find the path
  FR_INVALID_NAME, //!< (6) The path name format is invalid
  FR_DENIED, //!< (7) Access denied due to prohibited access or directory full
  FR_EXIST,  //!< (8) Access denied due to prohibited access as already existing
  FR_INVALID_OBJECT,  //!< (9) The file/directory object is invalid
  FR_WRITE_PROTECTED, //!< (10) The physical drive is write protected
  FR_INVALID_DRIVE, //!< (11) The logical drive number is invalid
  FR_NOT_ENABLED,   //!< (12) The volume has no work area
  FR_NO_FILESYSTEM, //!< (13) There is no valid FAT volume
  FR_MKFS_ABORTED,  //!< (14) The f_mkfs() aborted due to any parameter error
  FR_TIMEOUT, //!< (15) Could not get a grant to access the volume within defined period
  FR_LOCKED,  //!< (16) The operation is rejected according to the file sharing policy
  FR_NOT_ENOUGH_CORE,     //!< (17) LFN working buffer could not be allocated
  FR_TOO_MANY_OPEN_FILES, //!< (18) Number of open files > _FS_SHARE
  FR_INVALID_PARAMETER    //!< (19) Given parameter is invalid
} FRESULT;



//----- FatFs module application interface --------------------------

/** Mount / unmount a logical drive
 *
 *  This function registers / unregisters a work area, i.e. \ref FATFS
 *  structure, assigning it to a drive number (0...) respectively volume. A
 *  work area must be assigned to the a volume before the use of any other
 *  file function. To unregister a work area use NULL as parameter \c fs.
 *
 *  This function always succeeds for a \ref _VOLUMES "legal" \c vol value
 *  regardless of the drive's status. No media is accessed by this function.
 *  It only clears the given work area and registers it.
 *
 *  The "real" volume mount process is performed <ol>
 *  <li> once on first file access after calling \ref f_mount or </li>
 *  <li> once on first file access after after a media change. </li></ol>
 *
 *  \param vol Logical drive number to be mounted / unmounted
 *  \param fs  Pointer to new file system object (NULL for unmount only)
 *  \return  \ref FR_OK, \ref FR_INVALID_DRIVE
 *  \sa fileSystSMC
 *  \sa fMountSMC()
 */
FRESULT f_mount(uint8_t vol, FATFS *fs);


/** Extract drive number from path
 *
 *  This function checks if \c **path points to character sequence starting
 *  with "x:" where x may be '0..9', 'a..j' or 'A..J'. If so, \c *path is
 *  incremented by 2 (to point past the :) and 0.. 9 is returned according to
 *  the first character. Otherwise 0 is returned (respectively the
 *  current drive if \ref _FS_RPATH "enabled").
 *
 * \param   path  Pointer to pointer to the path name ([d:]path)
 * \return  drive / volume number in the range 0..9 (default 0)
 */
uint8_t extractDrive(const TCHAR **path);

/** Check if there's a valid mounted file system on the drive
 *
 *  This function checks if drive number \c vol is ready and has a valid
 *  recognised file system. In that case \ref FR_OK is returned and file
 *  system operations can be used without further preparations or delays.
 *
 *  \ref FR_NO_FILESYSTEM means, of course, no file system has been
 *  detected yet. That means it has either not yet been looked for (and all
 *  else is OK) or the medium is not formatted with a file system usable by
 *  this fatFS implementation.
 *
 *  \ref FR_INVALID_DRIVE means no valid volume number. <br />
 *  \ref FR_NOT_ENABLED  means no FS structure \ref f_mount "assigned". <br />
 *  \ref FR_NOT_READY  means drive down or medium removed
 *
 *  \param vol drive number 0.. (at present only 0 is valid)
 *  \return FR_OK (0); FR_INVALID_DRIVE FR_NOT_ENABLED  FR_NO_FILESYSTEM
 *          FR_NOT_READY
 *  \sa getValidFS
 */
FRESULT checkValidFS(uint8_t vol);

/** Get the file system for the drive
 *
 *  Exactly like \ref checkValidFS() this function checks if drive number
 *  \c vol is ready and has a valid recognised file system. In that sense
 *  \ref getValidFS() is a substitute for \ref checkValidFS().
 *
 *  Additionally it returns the (pointer to the) file system structure
 *  assigned with drive \c vol in parameter \c fatfs if that is possible.
 *
 *  \param vol drive number 0.. (at present only 0 is valid)
 *  \param fatfs  Pointer to pointer to the file system object to return
 *  \return FR_OK (0); FR_INVALID_DRIVE FR_NOT_ENABLED  FR_NO_FILESYSTEM
 *          FR_NOT_READY
 *  \sa checkValidFS()
 */
FRESULT getValidFS(uint8_t vol, FATFS** fatfs);

/** Check if the drive is write protected
 *
 *  This function checks if drive number \c vol is writable. In that case
 *  \ref FR_OK is returned. This result only makes sense it / respectively
 *  requires that the drive is been checked / initialised as usable. Otherwise
 *  \ref   FR_NOT_READY may be returned.
 *
 *  For a usable but write protected drive and file \ref FR_WRITE_PROTECTED
 *  will be returned. At present this will not happen on
 *  \ref intro_sec "weAutSys" /  \ref intro_secH "weAut_01" \ref smcCom SMC
 *  implementations as those slots do not have write protect switches.
 *  Nevertheless later versions may add write protect as software feature.
 *
 *  \param vol drive number 0.. (at present only 0 is valid)
 *  \return FR_OK (0); FR_NOT_READY FR_WRITE_PROTECTED
 */
FRESULT checkWriteProtect(uint8_t vol);

/** Look for a recognised file system on the drive
 *
 *  This function checks if drive number \c vol is ready and looks for a valid
 *  recognised file system on it. If that is or was discovered \ref FR_OK is
 *  returned and file system operations can be used without further
 *  preparations or delays.
 *
 *  \ref FR_NO_FILESYSTEM means all else is OK but no file system can be
 *  detected on that drive respectively medium. The only remedy (on present
 *  \ref intro_sec "weAutSys" /  \ref intro_secH "weAut_01" \ref smcCom SMC
 *  implementations) is to remove the medium and format it with a suitable
 *  file system (preferably FAT32).
 *
 *  \ref FR_INVALID_DRIVE means no valid volume number. <br />
 *  \ref FR_NOT_ENABLED  means no FS structure assigned. <br />
 *  \ref FR_NOT_READY  means drive down or medium removed
 *
 *  \param vol drive number 0.. (at present only 0 is valid)
 *  \return FR_OK (0); FR_INVALID_DRIVE FR_NOT_ENABLED  FR_NO_FILESYSTEM
 *          FR_NOT_READY
 */
FRESULT lookForFS(uint8_t vol);

/** Open or create a file
 *
 *  If this function \ref f_open function succeeded, the file object is valid.
 *  The file object is used for subsequent read/write functions to identify
 *  the file. To close an open file use \ref f_close(). If the modified file
 *  is not closed, the file data can be corrupted.
 *
 *  Before using any file function, a work area (file system object) must be
 *  registered to the logical drive with \ref f_mount(). All functions except
 *  \ref f_fdisk function can work after this procedure.
 *
 *  The parameter mode can be ORed using the predefined mode flags:
 *
 - \ref FA_READ Specifies read access to the object.
        Data can be read from the file. Combine with FA_WRITE for
        read-write access.
 - \ref FA_WRITE  Specifies write access to the object.
        Data can be written to the file. Combine with FA_READ for
        read-write access.
 - \ref FA_OPEN_EXISTING (default)  Opens the file if it is existing. Otherwise the
             function fails.
 - \ref FA_OPEN_ALWAYS   Opens the file if it is existing. If not, a new file
            is created. To append data to the file, use \ref f_lseek()
            after the file open in this method.
 - \ref FA_CREATE_NEW Creates a new file. The function fails with
        \ref FR_EXIST if the file is existing.
 - \ref FA_CREATE_ALWAYS Creates a new file. If the file is existing,
        it is truncated and overwritten.
 *
 *  \param fp  Pointer to the blank file object
 *  \param path Pointer to the file name
 *  \param mode Access mode and file open mode flags
 *  \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *           \ref FR_NO_FILE,  \ref FR_NO_PATH, \ref FR_INVALID_NAME,
 *           \ref FR_DENIED, \ref FR_EXIST, \ref FR_WRITE_PROTECTED,
 *           \ref FR_INVALID_DRIVE, \ref FR_NOT_ENABLED,
 *           \ref FR_NO_FILESYSTEM, \ref FR_TIMEOUT, \ref FR_LOCKED,
 *           \ref FR_NOT_ENOUGH_CORE, \ref FR_TOO_MANY_OPEN_FILES
 */
FRESULT f_open(FIL *fp, const TCHAR *path, uint8_t mode);

/** Read data from a file
 *
 *  This function reads file data to the buffer supplied. The file object's
 *  file pointer increases by the number of bytes read. After the function
 *  succeeded, \c *br should be checked to detect the end of file. In the
 *  case of \c *br < \c btr the read/write pointer reached end of the file
 *  during the read operation.
 *
 * \param fp  Pointer to file object
 * \param buff Pointer to data buffer
 * \param btr Number of bytes to read
 * \param br Pointer to result: number of bytes read <br />
 *           This number must not be greater than the sector size (512)
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR,
 *          \ref FR_NOT_READY, \ref FR_INVALID_OBJECT, \ref FR_TIMEOUT
 */
FRESULT f_read(FIL *fp, void *buff, uint16_t btr,  uint16_t *br);

/** Close an open file object
 *
 *  This function closes an open file. If any data have been written to
 *  the file, the cached information is written back to the medium (SMC).
 *  After the function succeeded, the file object is no longer valid.
 *
 * \param fp  Pointer to file object
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR,
 *          \ref FR_NOT_READY, \ref FR_INVALID_OBJECT, FR_TIMEOUT
 */
FRESULT f_close(FIL* fp);

#if ! _weAutSys
/** Move file pointer of a file object, expand file size
 *
 *  This function moves the file read/write pointer of an open file. The
 *  offset is absolute, i.e. from file origin resp. top of file. When an
 *  offset above the file size is specified in write mode, the file size is
 *  increased and the data in the expanded area are undefined. Thereby is is
 *  possible to create a large file quickly. After the \ref f_lseek function
 *  succeeded, current read/write pointer should be checked to make sure the
 *  read/write pointer has been moved correctly. In case of the current
 *  read/write pointer is not the expected value, either of followings has
 *  occured:
  - End of file. The specified Offset was clipped at end of the file
         because the file has been opened in read-only mode.
  - Disk full. There is insufficient free space on the volume to expand
         the file size.
 *
 *  Fast seek feature is enabled when \ref _USE_FASTSEEK is set to 1 and the
 *  member .cltbl in the file object is not NULL. This feature enables fast
 *  backward/long seek operations without FAT access by cluster link map
 *  table (CLMT) stored in the user defined table. It is also applied to
 *  the functions \ref f_read() and \ref f_write(). In this mode, the file
 *  size cannot be increased by \ref f_write() or \ref f_lseek().
 *  The CLMT must be created in the user defined DWORD array prior to use
 *  fast seek feature. To create the CLMT, let the member .cltbl in the
 *  file object pointer to the DWORD array. Set the array size in unit of
 *  items into the first item and call the \ref f_lseek function with
 *  Offset = CREATE_LINKMAP. After the function succeeded and CLMT is
 *  created, no FAT access will occur in subsequent \ref f_read(),
 *  \ref f_write() or \ref f_lseek() calls on that file. If the function
 *  failed with FR_NOT_ENOUGH_CORE, the given array size is insufficient
 *  for the file and the required items is returned into the first item of
 *  the array. The required array size is (number of fragments + 1) * 2 items.
 *  For example, when the file is fragmented in 5, 12 items will be required
 *  for the CLMT.
 *
 * \param  fp  Pointer to the file object
 * \param  ofs  File pointer from top of file
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR,
 *          \ref FR_NOT_READY, \ref FR_INVALID_OBJECT, FR_TIMEOUT
 */
FRESULT f_lseek(FIL* fp, uint32_t ofs);
#endif // !  _weAutSys

/** Open an existing directory
 *
 *  This function opens an existing directory and creates the directory
 *  object for subsequent calls of directory related functions. The
 *  \ref DIR "directory object" can be discarded at any time.
 *
 * \param path Pointer to the directory's file name (0-terminated)
 * \param dj Pointer to directory object to create
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_PATH, \ref FR_INVALID_NAME, \ref FR_INVALID_DRIVE,
 *          \ref FR_NOT_ENABLED, \ref FR_NO_FILESYSTEM, FR_TIMEOUT,
 *          \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_opendir(DIR* dj, const TCHAR* path);

/** Read a directory item
 *
 *  This function reads directory entries in sequence. All items in the
 *  directory can be read by repeated calls. When all directory entries have
 *  been thus visited the function just puts a null string into fno.f_name[].
 *  When a null pointer is given to the FileInfo, the read index
 *  of the directory object will be rewound.
 *
 *  When \ref _USE_LFN "LFN" feature is enabled, lfname and lfsize in the file
 *  information structure must be initialized with valid value prior to use
 *  the f_readdir function. The lfname is a pointer to the string buffer to
 *  return the long file name. The lfsize is the size of the string buffer
 *  in unit of TCHAR. If either the size of read buffer or LFN working buffer
 *  is insufficient for the LFN or the object has no LFN, a null string will
 *  be returned to the LFN read buffer. If the LFN contains any character
 *  that cannot be converted to OEM code, a null string will be returned but
 *  this is not the case on Unicode API configuration. When lfname is NULL,
 *  nothing of the LFN is returned. When the object has no LFN, some small
 *  capitals can be contained in the SFN.
 *
 *  When relative path feature is enabled (\ref _FS_RPATH == 1), "." and
 *  ".." entries are not filtered out and it will appear in the read entries.
 *
 *  This function is available if \ref _FS_MINIMIZE <= 1.
 *
 * \param dj  Pointer to the open directory object
 * \param fno Pointer to the file information to be set with next entry's
 *            information; fno == NUL means reset to first directory entry
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_INVALID_OBJECT, FR_TIMEOUT, \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_readdir(DIR* dj, FILINFO* fno);

/** Get file status
 *
 *  This function gets the size, timestamp and attribute of a file or
 *  directory. For details of the information, refer to the \ref FILINFO
 *  structure and \ref f_readdir().
 *
 *  This function is not available in
 *  \ref _FS_MINIMIZE "minimization level" of >= 1.
 *
 * \param path Pointer to the directory's file name (0-terminated)
 * \param fno Pointer to the file information
 * \return  FR_OK, FR_DISK_ERR, FR_INT_ERR, FR_NOT_READY, FR_NO_FILE,
 *          FR_NO_PATH, FR_INVALID_NAME, FR_INVALID_DRIVE, FR_NOT_ENABLED,
 *          FR_NO_FILESYSTEM, FR_TIMEOUT, FR_NOT_ENOUGH_CORE
 */
FRESULT f_stat(const TCHAR* path, FILINFO* fno);

/** Write data to a file
 *
 *  The read/write pointer in the \ref FIL "file object"  is increased by
 *  the number of bytes written. After the function succeeded, \c *bw
 *  should be checked to detect the disk full condition. In case of
 *  \c *bw < \c btw there was not enough space for the (complete) write
 *  operation.
 *
 *  The function may require more time when the volume is full or nearly so.
 *
 * \param  fp  Pointer to the file object
 * \param buff Pointer to data buffer
 * \param btw Number of bytes to write <br />
 *           This number must not be greater than the sector size (512)
 * \param bw  Pointer to result: number of bytes written
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_INVALID_OBJECT, FR_TIMEOUT
 */
FRESULT f_write(FIL* fp, const void * buff, uint16_t btw, uint16_t * bw);

#if ! _weAutSys
/** Get the number of free clusters on the drive
 *
 *  This function gets number of free clusters on the drive. The member
 *  \c csize in the \ref getValidFS "file system object" is the number of
 *  sectors per cluster, so that the free space in unit of sector can be
 *  calculated. When FSInfo structure on FAT32 volume is not in
 *  \ref f_sync "sync", this function may return an incorrect free cluster
 *  count.
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 *  \param vol    the drive's volume number (0..9); to extract form a
 *                path name use \ref extractDrive "extractDrive(&path)"
 *  \param nclst  Pointer to result: number of free clusters (not NULL!)
 *  \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *           \ref FR_INVALID_DRIVE, \ref FR_NOT_ENABLED,
 *           \ref FR_NO_FILESYSTEM, FR_TIMEOUT
 */
FRESULT f_getfree(const uint8_t vol, uint32_t* nclst);
#endif // ! _weAutSys

/** Truncate file
 *
 *  This function truncates the file size to the current file read/write
 *  point. This function has no effect if the file read/write pointer is
 *  already pointing end of the file.
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 *  \param  fp  Pointer to the file object
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_INVALID_OBJECT, FR_TIMEOUT
 */
FRESULT f_truncate(FIL* fp);

/** Flush written / cached data to a file
 *
 *  This function transfers write data caches in memory / sector buffers to
 *  the device. (The process and the related functions is normally called
 *  flush.) The function \ref f_close() does, of course, the same before
 *  discarding the file object.
 *
 *  Calling this function periodically or after every \ref f_write() can
 *  reduce the risk of data loss due to power failure or SMC removal at all
 *  costs of otherwise unnecessary device accesses.
 *
 * \param  fp  Pointer to the file object
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_INVALID_OBJECT, FR_TIMEOUT
 */
FRESULT f_sync (FIL* fp );

/** Delete an existing file or directory
 *
 *  If the object to be erased / removed has a read-only attribute
 *  (\ref AM_RDO) the operation will be rejected with \ref FR_DENIED.
 *
 *  To remove a directory it must be empty and must not be current directory
 *  or the operation will be rejected with \ref FR_DENIED.
 *
 *  A file to be removed must not be open or the FAT volume can collapse.
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 *  \param path Pointer to the directory's file name (0-terminated)
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_FILE, \ref FR_NO_PATH, \ref FR_INVALID_NAME,
 *          \ref FR_DENIED, \ref FR_EXIST, \ref FR_WRITE_PROTECTED,
 *          \ref FR_INVALID_DRIVE, \ref FR_NOT_ENABLED, \ref FR_NO_FILESYSTEM,
 *          FR_TIMEOUT, \ref FR_LOCKED, \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_unlink(const TCHAR* path);

/** Create a new directory
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 *  \param path Pointer to the directory's file name (0-terminated)
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_PATH, \ref FR_INVALID_NAME, \ref FR_DENIED,
 *          \ref FR_EXIST, \ref FR_WRITE_PROTECTED, \ref FR_INVALID_DRIVE,
 *          \ref FR_NOT_ENABLED, \ref FR_NO_FILESYSTEM,
 *          FR_TIMEOUT, \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_mkdir(const TCHAR* path);

/** Change attribute of the file or directory
 *
 *  This function will set and/or clear the attributes \ref AM_RDO (read only),
 *  \ref AM_ARC (archive), \ref AM_SYS (system) and \ref AM_HID (hidden) of
 *  the file or directory specified. The parameters \c value and \c mask are
 *  OR expressions of the attribute masks named. Setting (\c value ) has
 *  preference over clearing (\c mask ).
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 *  \param path Pointer to the file name (0-terminated)
 *  \param value Attribute flags to be set
 *  \param mask  Attribute flag  to be cleared
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_FILE, \ref FR_NO_PATH, \ref FR_INVALID_NAME,
 *          \ref FR_WRITE_PROTECTED, \ref FR_INVALID_DRIVE,
 *          \ref FR_NOT_ENABLED, \ref FR_NO_FILESYSTEM,
 *          FR_TIMEOUT, \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_chmod(const TCHAR* path, uint8_t value, uint8_t mask);

/** Change time-stamps of a file or directory
 *
 * \param path Pointer to the file/directory name
 * \param fno  Pointer to a  FILINFO structure with .fdate and .ftime values
 *             to be set to the file respectively directory
 *  \return \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_FILE, \ref FR_NO_PATH, \ref FR_INVALID_NAME,
 *          \ref FR_WRITE_PROTECTED, \ref FR_INVALID_DRIVE,
 *          \ref FR_NOT_ENABLED, \ref FR_NO_FILESYSTEM,
 *          FR_TIMEOUT, \ref FR_NOT_ENOUGH_CORE
 */
FRESULT f_utime(const TCHAR* path, const FILINFO* fno);

/** Rename or move a file or directory
 *
 *  This function renames an object (file or directory) and can also move it
 *  to other directory. The logical drive number is determined by old name,
 *  new name must not contain a logical drive number.
 *
 *  Do not rename open objects.
 *
 *  This function is available if \ref _FS_READONLY == 0 and
 *  \ref _FS_MINIMIZE == 0.
 *
 * \param path_old Pointer to the current name / path
 * \param path_new Pointer to the new name / path
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_INT_ERR, \ref FR_NOT_READY,
 *          \ref FR_NO_FILE, \ref FR_NO_PATH, \ref FR_INVALID_NAME,
 *          \ref FR_DENIED,  \ref FR_EXIST, \ref FR_WRITE_PROTECTED,
 *          \ref FR_INVALID_DRIVE, \ref FR_NOT_ENABLED,
 *          \ref FR_NO_FILESYSTEM, FR_TIMEOUT,
 *          \ref FR_LOCKED, \ref FR_NOT_ENOUGH_CORE */
FRESULT f_rename(const TCHAR* path_old, const TCHAR* path_new);

FRESULT f_chdrive(uint8_t); //!< Change current drive (not implemented)
FRESULT f_chdir(const TCHAR*); //!< Change current directory (not implemented)
FRESULT f_getcwd(TCHAR*, uint16_t); //!< Get current directory (not implemented)


/** Divide a physical drive into some partitions
 *
 *  This function creates a partition table into the MBR of the physical
 *  drive. The partitioning rule is in generic FDISK format so that it can
 *  create up to four primary partitions. Extended partition is not supported.
 *  The parameter \c szt specifies how to divide the physical drive. The first
 *  item specifies the size of first primary partition and fourth item
 *  specifies the fourth primary partition. If the value is less than or
 *  equal to 100, it means percentage of the partition in the entire disk
 *  space. If it is larger than 100, it means partition size in unit of
 *  sector.
 *
 *  This function is available if \ref _MULTI_PARTITION == 2
 *
 * \param pdrv Physical drive number
 * \param szt  Pointer to the size table for each partition
 * \param work Pointer to the working buffer. The size must be at least
 *             \ref _MAX_SS bytes.
 * \return  \ref FR_OK, \ref FR_DISK_ERR, \ref FR_NOT_READY,
 *          \ref FR_WRITE_PROTECTED, \ref FR_INVALID_PARAMETER
 */
FRESULT f_fdisk(uint8_t  pdrv, const uint32_t szt[], void* work);

int f_putc (TCHAR, FIL*); //!< Put a character to the file (not implemented)
int f_puts (const TCHAR*, FIL*); //!< Put a string to the file  (not implemented)
int f_printf (FIL*, const TCHAR*, ...); //!< Put a formatted string to the file  (not implemented)
TCHAR* f_gets (TCHAR*, int, FIL*); //!< Get a string from the file  (not implemented)


#ifndef EOF
#define EOF (-1)
#endif




//---------   Additional user defined functions   ----------------

#define  get_fattime() getFATtime() //!< Get current local FAT time \sa getFATtime

/* Unicode support functions */
#if _USE_LFN            /* Unicode - OEM code conversion */
WCHAR ff_convert (WCHAR, uint16_t);    /* OEM-Unicode bidirectional conversion */
WCHAR ff_wtoupper (WCHAR);      /* Unicode upper-case conversion */
#if _USE_LFN == 3          /* Memory functions */
void* ff_memalloc (uint16_t);      /* Allocate memory block */
void ff_memfree (void*);      /* Free memory block */
#endif
#endif


//-----  Flags and offset address
//----   File access control and file status flags (FIL.flag)

/** Mode flag: read access to the file object
 *
 *  Combine with \ref FA_WRITE for read-write access.
 *  \showinitializer
 */
#define FA_READ 0x01

/** Mode flag: open the existing file object
 *
 *  The function fails if the file is not existing (default).
 *  \showinitializer
 */
#define FA_OPEN_EXISTING 0

#define FA__ERROR 0x80 // mode flag bit set if operation aborted \showinitializer

#if !_FS_READONLY

/** Mode flag: write access to the file object
 *
 *  Combine with \ref FA_READ for read-write access.
 *  \showinitializer
 */
#define FA_WRITE 0x02

/** Mode flag: create a new file
 *
 *  The function fails with  \ref FR_EXIST if the file is existing.
 *  \showinitializer
 */
#define FA_CREATE_NEW 0x04

/** Mode flag: create anyway
 *
 *  Opens the file if it is existing. If not, a new file is created.
 *  \showinitializer
 */
#define FA_CREATE_ALWAYS 0x08

/** Mode flag: open anyway
 *
 *  Opens the file if it is existing. If not, a new file is created.
 *  To append data to the file, use \ref f_lseek after file opening with
 *  this mode.
 *
 *  \showinitializer
 */
#define FA_OPEN_ALWAYS 0x10

#define FA__WRITTEN 0x20
#define FA__DIRTY 0x40
#endif


/* FAT sub type (FATFS.fs_type) */

#define FS_FAT12 1 //!< FATFS.fs_type value  \showinitializer
#define FS_FAT16 2 //!< FATFS.fs_type value  \showinitializer
#define FS_FAT32 3 //!< FATFS.fs_type value  \showinitializer


//--- File attribute bits for directory entry

#define AM_RDO 0x01  //!< directory entry attribute bit: read only \showinitializer
#define AM_HID 0x02  //!< directory entry attribute bit: hidden \showinitializer
#define AM_SYS 0x04  //!< directory entry attribute bit: system \showinitializer
#define AM_VOL 0x08  //!< directory entry attribute bit: Volume label \showinitializer
#define AM_LFN 0x0F  //!< directory entry attribute bit: LFN entry \showinitializer
#define AM_DIR 0x10  //!< directory entry attribute bit: directory \showinitializer
#define AM_ARC 0x20  //!< directory entry attribute bit: archive \showinitializer
#define AM_MASK 0x3F //!< mask of defined directory entry attribute bits \showinitializer

/* Fast seek feature */
#define CREATE_LINKMAP  0xFFFFFFFF


//--- Multi-byte word access macros
// \internal catastrophic code generated for _WORD_ACCESS == 1

#if _WORD_ACCESS == 1  /* Enable word access to the FAT structure */
#define  LD_WORD(ptr)    (uint16_t)(*(uint16_t*)(uint8_t*)(ptr))
#define  LD_DWORD(ptr)    (uint32_t)(*(uint32_t*)(uint8_t*)(ptr))
#define  ST_WORD(ptr,val)  *(uint16_t*)(uint8_t*)(ptr)=(uint16_t)(val)
#define  ST_DWORD(ptr,val)  *(uint32_t*)(uint8_t*)(ptr)=(uint32_t)(val)
#else          /* Use byte-by-byte access to the FAT structure */
#define  LD_WORD(ptr)    (uint16_t)(((uint16_t)*((uint8_t*)(ptr)+1)<<8)|(uint16_t)*(uint8_t*)(ptr))
#define  LD_DWORD(ptr)    (uint32_t)(((uint32_t)*((uint8_t*)(ptr)+3)<<24)|((uint32_t)*((uint8_t*)(ptr)+2)<<16)|((uint16_t)*((uint8_t*)(ptr)+1)<<8)|*(uint8_t*)(ptr))
#define  ST_WORD(ptr,val)  *(uint8_t*)(ptr)=(uint8_t)(val); *((uint8_t*)(ptr)+1)=(uint8_t)((uint16_t)(val)>>8)
#define  ST_DWORD(ptr,val)  *(uint8_t*)(ptr)=(uint8_t)(val); *((uint8_t*)(ptr)+1)=(uint8_t)((uint16_t)(val)>>8); *((uint8_t*)(ptr)+2)=(uint8_t)((uint32_t)(val)>>16); *((uint8_t*)(ptr)+3)=(uint8_t)((uint32_t)(val)>>24)
#endif

/** @} */
/** @} */
/** @} */ // osSWinc
#endif /* _FATFS */
