// Copyright (c) 2012 Albrecht Weinert, Bochum
// log_streams.h 2 2017-01-25 16:49:03Z albrecht

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup appLaySup
 * @{
 * \addtogroup stdStreams
 * @{
 */

/**
 * \file log_streams.h
 *
 * Buffered log stream (definitions)
 *
 * This file contains the implementations of weAutSys' non blocking buffered
 * stream, mainly meant for logging. Any (standard) output stream may be
 * decorated with this buffered stream.
 *
 * According to embedded or real time requirements non-blocking has the
 * highest priority. Hence older buffered and not yet forwarded output is
 * allowed to be overwritten. Hence this is for logging and debugging purposes
 * only, when un-spoiled un-blocked operation is favored over non loosing
 * output and newer logs are more valued than older ones.
 *
 * With regard the main logging or debugging purpose of this buffered stream
 * there are some extra convenience functions for combined constant flash
 * string output and formatting. The rationale is to reduce the temptation
 * to use the fprint sort of output functions &mdash; usually a deadly sin in
 * embedded real time systems.
 *
 *  This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert
/** @} */ // Doxygen addtogroup stdStreams
/** @} */ // Doxygen addtogroup appLaySup

#ifndef __LOG_STREAMS_H__
#define __LOG_STREAMS_H__

#include <avr/pgmspace.h>
#include "arch/config.h"

/**
 * \addtogroup appLaySup
 * @{
 * \addtogroup stdStreams
 * @{
 */
/// \name The buffered log streams
/// @{

/** The buffered log streams capacity
 *
 *  The value must be (a power of 2 - 1) and at least 255.
 *  \showinitializer
 */
#define BUF_STREAMS_CAP 2047

/** The stream "device" being a buffer for logging
 *
 *  This stream is a buffer (or pipe or decorator) for output. Its intended
 *  use is as buffered log output to humans.
 *
 *  The buffered content is available by \ref bufLogGetC "input" or by
 *  \ref bufLogGetChars "bulk copying" out of the buffered text to be consumed
 *  and send forth by an appropriate device respectively
 *  \ref bufLogCons "application", be it the UART or one Telnet connection.
 *
 *  Every usage of the log buffered stream device via standard streams shall
 *  be embraced by \ref bufLogStreamsOpen() and \ref streamsClose().
 */
extern FILE bufLogStreams;

/** The consumer of buffered log streams output (text) data
 *
 *  This (16 bit) value points to the state data \ref thr_data_t "structure"
 *  of the application to regularly \ref bufLogGetChars "consume" (and send
 *  forward) the \ref bufLogStreams text buffered so far.
 *
 *  Special values are: <br />
 *  NULL or 0 : no consuming device or application <br \>
 *  1 : the \ref serStreams "UART" shall output all buffered text.
 *  \sa bufLogSetConsumer
 */
extern struct thr_data_t * bufLogCons;

/** The UART is the consumer of buffered log streams output (text) data
 *
 *  This evaluates to true if the serial output is to consume the buffered
 *  logs.
 */
#define bufLogConsIsUART (bufLogCons == (struct thr_data_t *)1)

/** Set the consumer of buffered log streams output (text) data
 *
 *  If the consumer is set to \c 0 or \c NULL there is no consumer and all
 *  output will go to the log streams (except the last \ref BUF_STREAMS_CAP
 *  bytes will be forgotten). <br />
 *  A value of \c 1 means the serial interface is the consumer. This
 *  presupposes the UART being connected to a (human readable) text mode
 *  device. <br />
 *  Any other value must be a pointer to a thread data structure. A suitable
 *  thread function has to recognise this and handle the buffered log output.
 *  \ref intro_sec "weAutSys"' \ref telnetServer "Telnet server" e.g. does
 *  this.
 *
 *  At any given moment there can be just one or none \c (0) consumer for the
 *  buffered log streams output.
 *
 *  \hideinitializer
 *  \param tD pointer to the consumer's thread data structure or 0 or 1
 *  \sa bufLogCons
 *  \sa bufLogConsIsUART
 *  \sa bufLogSetUART
 */
#define bufLogSetConsumer(tD) bufLogCons = ((struct thr_data_t *)(tD))

/** Set the UART as consumer of buffered log streams output
 *
 *  This sets the UART as consumer of log output. This presupposes the UART
 *  being connected to a (human readable) text mode device. <br />
 *
 *  If this pre-condition is met this may be called by user software's
 *  \ref appInitThreadF "initialisation" (phase 2) to prevent the first active
 *  \ref telnetServer "Telnet" client from automatically get that role. This
 *  may be mandatory if \ref DEBUG_TEL "Telnet debugging" is on. <br />
 *
 *  \hideinitializer
 *  \sa bufLogCons
 *  \sa bufLogConsIsUART
 *  \sa bufLogSetConsumer
 */
#define bufLogSetUART() bufLogCons = ((struct thr_data_t *)(1))


/** Initialise the buffered log streams
 *
 *  This function initialises and clears the \ref bufLogStreams.
 */
void bufLogStreamsInit(void);

/** Open the buffered log streams
 *
 *  This function sets the standard streams stdin/out/err to the
 *  \ref bufLogStreams.<br />
 *  This "switching" will be rescinded by \ref streamsClose() .
 *
 *  This function just (re-) opens leaving all else state un-changed.
 */
void bufLogStreamsOpen(void);


/** The number of characters buffered in buffered log streams for input
 *
 *  This function returns the number of bytes that can be fetched from
 *  buffered log streams input by \ref bufLogGetChar(FILE*) respectively
 *  \ref bufLogGetC() without getting EOF.
 *
 *  The returned value is in the range 0 .. \ref BUF_STREAMS_CAP.
 *  If 0 is returned nothing should be read.
 */
uint16_t bufLogInBufferd(FILE * streams);

/** The number of characters buffered in buffered log streams for input
 *
 *  \sa bufLogInBufferd
 */
#define BufLogInBufferd ((bufLogBufWi - bufLogBufRi) & BUF_STREAMS_CAP)

/** Get one byte from buffered stream input
 *
 *  This function never blocks. It returns a byte value read or EOF. <br />
 *
 *  \return a character 0..255 or EOF
 */
int bufLogGetC(void);

/** Get a number of bytes from buffered stream input
 *
 *  This function never blocks. It reads up to \c n bytes from the buffered
 *  stream to the buffer \c dst.
 *
 *  This is mainly intended as helper function for \ref bufLogCons "consuming"
 *  and forward the log output buffered so far.
 *
 *  \return the number of characters transferred 0 ..\c n
 */
uint16_t bufLogGetChars(char * dst, uint16_t n);

/** Put characters from buffered log streams to serial output
 *
 *  This function will put as much as possible characters from the buffered
 *  log streams to the UART. Returned is the number of characters transferred;
 *  its type is byte as long as we stick to UART buffers capacity below 256.
 *
 *  \return the number of characters transferred
 */
uint8_t uartPutLogBuf(void);

//----------------------  in / out ----------------------------------------

/** The buffer space available for buffered log streams output
 *
 *  \return 0 .. \ref BUF_STREAMS_CAP
 */
uint16_t bufLogStreamsOutSpace(FILE * streams) __attribute__ ((pure));

/** Check the space available buffered log streams output
 *
 *  This function might give a positive answer to repeated request for more
 *  characters than space \ref bufLogStreamsOutSpace "available" even if that
 *  might lead to loosing the oldest output. The rationale for this "lie" is
 *  that a standard programmed \ref PT_YIELD_OUT_SPACE "yielding" for output
 *  space wont't hold a thread on logging (debug) output more than twice.
 *
 *  \return The required space <br />
 *       4: will never be available cause asked for more than
 *         \ref BUF_STREAMS_CAP. I.e. give up or request less space;
 *       2: is not available, but may become (automatically) in the future,
 *          i.e. yielding / waiting is OK .<br />
 *       1: may be available, but can't be determined. I.e. blindly go ahead
 *          or give up. This answer is (wrongly) given on second request
 *          instead of 2; the third request will get 0. <br />
 *       0: is available or instead of 2 at third request or with
 *          no \ref bufLogCons "consumer".
 *          I.e. go ahead with planned output. <br />
 *  \sa str2CheckOutSpace
 */
uint8_t bufLogCheckOutSpace(uint16_t const reqSpace, FILE * const streams);



/** Put one byte to buffered stream output
 *
 *  This function never blocks and does the output. If there was no
 *  \ref bufLogStreamsOutSpace "space" in the buffer the "oldest" character
 *  there will be overwritten.
 *  Hence this operation should be (indirectly) guarded by
 *  \ref bufLogStreamsOutSpace().
 *
 *  \param c the character to be output
 */
void bufLogPutC(char const c);

/** Put two bytes to buffered stream output
 *
 *  This function acts like two times \ref bufLogPutC but is a bit more
 *  efficient (and readable).
 *
 *  \param c1 the first character to be output
 *  \param c2 the second character to be output
 */
void bufLogPut2C(char const c1, char const c2);


/** Get one byte from buffered stream input
 *
 *  This function never blocks. It returns a byte value read or EOF. <br />
 *
 *  This function may be used directly for input but then
 *  \ref bufLogGetC() is a better choice. Usually this function is used
 *  indirectly via stdin of stdio.h as this functions signature is what
 *  stdio expects of underlying input streams.
 *
 *  \param stream (passed by stdio input) totally ignored by this function
 *  \return a character 0..255 or EOF
 */
int bufLogGetChar(FILE * const stream);

/** Put one byte to buffered stream output
 *
 *  This function does exactly the the same as the preferable
 *  \ref  bufLogPutC() except for the extra parameter of type FILE * that is
 *  for sake of the standard I/O usage.
 *
 *  \param c the character to be output
 *  \param stream (passed by stdio output) totally ignored by this function
 *  \return c in case of success, EOF if output buffers full.
 */
int bufLogPutChar(char c, FILE * const stream); //  __attribute__ ((pure));


/** Put a RAM string (some characters) to buffered stream output
 *
 *  This function never blocks. It returns the number of characters
 *  transferred. If not enough buffer space is available the returned value
 *  will be smaller than the length of the RAM string \c src.
 *  This space restriction applies to the \ref BUF_STREAMS_CAP "size" of the
 *  buffer and not to the \ref bufLogStreamsOutSpace "free space". If the
 *  latter is insufficient this function will just overwrite the oldest
 *  output.
 *
 *  This operation should be (indirectly) guarded by
 *  \ref bufLogStreamsOutSpace() or \ref bufLogCheckOutSpace().
 *
 *  The output stops at the first 0 in \c src. By that this is a string
 *  output function. To send a (binary) 0 use \ref bufLogPutC "bufLogPutC(0)".
 *
 *  \param src the characters to be output
 *  \param stream the current stream (not used)
 *  \return the number of characters output
 */
int bufLogPutSt(char * src, FILE const * const stream);

/** Put some characters from program space to buffered stream output
 *
 *  This function never blocks. It returns the number of characters
 *  transferred. If not enough buffer space is available the returned value
 *  will be smaller than the length of the flash memory string \c src.
 *  This space restriction applies to the \ref BUF_STREAMS_CAP "size" of the
 *  buffer and not to the \ref bufLogStreamsOutSpace "free space". If the
 *  latter is insufficient this function will just overwrite the oldest
 *  output.
 *
 *  This operation should be (indirectly) guarded by
 *  \ref bufLogStreamsOutSpace() or \ref bufLogCheckOutSpace().
 *
 *  The output stops at the first 0 in \c src. By that this is a string
 *  output function. To send a (binary) 0 use \ref bufLogPutC "bufLogPutC(0)".
 *
 *  \param src the characters to be output in flash memory
 *  \param stream the current stream (not used)
 *  \return the number of characters output
 */
int bufLogPutSt_P(char * src, FILE const * const stream);

/// @}  // end member group bufLog device


/** @} */ // Doxygen addtogroup stdStreams
/** @} */ // Doxygen addtogroup appLaySup



/**
 * \addtogroup llSySer
 * @{
 * \addtogroup swDebaid
 * @{
 */

/** Log some characters from program space to buffered log output
 *
 *  This debugging function does essentially the same as \ref bufLogPutSt_P
 *  but it will not drop anything if not enough output buffer space is
 *  available. Instead it will just overwrite the oldest output.
 *
 *  Anyway the number of characters output will be restricted to
 *  \ref BUF_STREAMS_CAP. And, of course, the output stops at the
 *  first 0 in \c src.
 *
 *  \param src the characters to be output in flash memory (0-terminated)
 *  \sa bufLogDec_P
 *  \sa bufLogHex_P
 */
void bufLogTxt_P(char const * src);

/** Log some characters from RAM to buffered log output
 *
 *  This debugging function does the same as \ref bufLogTxt_P except that
 *  the string to be output is in RAM and the extra length limit \c n .
 *  Anyway the output will stop at the first 0 (zero) character.
 *
 *  All the restrictions and warnings are the same as for \ref bufLogTxt_P .
 *
 *  \param src the characters to be output in RAM
 *  \param n   the maximum number of characters to be output
 *  \sa bufLogTxt_P
 *  \sa bufLogHex_P
 */
void bufLogTxt(char * src, uint8_t n);

/** Log a string from program space + a 4 digit decimal number to buffered log output
 *
 *  This debugging function outputs the flash memory text \c src and appends
 *  directly a four digit decimal number /c info right aligned.
 *
 *  The same restrictions and warnings apply as for \ref bufLogTxt_P.
 *
 *  \param src the characters to be output in Flash memory (not NULL !)
 *  \param info the four digit number as extra info (right aligned)
 *  \sa bufLogTxt_P
 *  \sa bufLogHex_P
 *  \sa bufLog2Dec_P
 */
void bufLogDec_P(char const * src, uint16_t info);

/** Log a program space string + two 3 digit decimal numbers to buffered log output
 *
 *  This debugging function is similar to \ref  bufLogDec_P. It outputs the
 *  flash memory text (src, n) appends directly two three digit decimal
 *  numbers \c info1 and \c info2 with leading zeroes + a line feed.
 *
 *  The same restrictions and warnings apply as for \ref bufLogTxt_P.
 *
 *  \param src the characters to be output in Flash memory
 *  \param info1 the three digit number as extra info (right aligned, 0..999)
 *  \param info2 the second three digit number, 0..999
 *  \sa bufLogTxt_P
 *  \sa bufLogHex_P
 *  \sa bufLogDec_P
 */
void bufLog2Dec_P(char const * src, uint16_t info1, uint16_t info2);

/** Log a program space string + a 3 digit decimal + an 8 digit hex number to buffered log output
 *
 *  This debugging function is similar to \ref  bufLogDec_P. It outputs the
 *  flash memory text \c src) appends directly a three digit decimal
 *  number \c info1 with leading zeroes and an 8 digit hex number \c info2
 *  + a line feed.
 *
 *  The same restrictions and warnings apply as for \ref bufLogTxt_P.
 *
 *  \param src the characters to be output in Flash memory
 *  \param info1 the three digit number as extra info (right aligned 0..999)
 *  \param info2 the eight digit hex number
 *  \sa bufLogTxt_P
 *  \sa bufLogHex_P
 *  \sa bufLogDec_P
 */
void bufLogDecHex_P(char const * src, uint16_t info1, uint32_t info2);

/** Log a program space string + a 2 digit and an 8 digit hex number to buffered log output
 *
 *  This debugging function is similar to \ref  bufLogDec_P. It outputs the
 *  flash memory text \c src appends directly a two digit haxadecimal
 *  number \c info1, a blank and an 8 digit hex number \c info2
 *  + a line feed.
 *
 *  The same restrictions and warnings apply as for \ref bufLogTxt_P.
 *
 *  \param src the characters to be output in Flash memory
 *  \param info1 the two digit hex number as extra info
 *  \param info2 the eight digit hex number
 *  \sa bufLogTxt_P
 *  \sa bufLogHex_P
 *  \sa bufLogDec_P
 */
void bufLogHexHex_P(char const * src, uint8_t info1, uint32_t info2);

/** Log a string from program space + a two digit hex number to buffered log output
 *
 *  This debugging function outputs the flash memory text \c src and
 *  appends directly a two digit hexadecimal number \c info.
 *
 *  The same restrictions and warnings apply as for \ref bufLogTxt_P.
 *
 *  \param src the characters to be output in Flash memory
 *  \param info the hex number (two digits)
 *  \sa bufLogTxt_P
 *  \sa bufLogDec_P
 */
void bufLogHex_P(char const * src, uint8_t info);

/** Log two flash strings + a two digit hex number to buffered log output
 *
 *  This debugging function outputs the flash memory texts \c tx1 and \c tx2
 *  directly concatenated. After a space the two digit hexadecimal number
 *  \c info + a line feed is output.
 *
 *  The same restrictions and warnings apply as for \ref bufLogTxt_P.
 *
 *  \param tx1 the characters to be output from flash memory (not NULL !)
 *  \param tx2 more characters to be output from flash memory (not NULL !)
 *  \param info the hex number (two digits)
 *  \sa bufLogTxt_P
 *  \sa bufLogDec_P
 */
void bufLogTThex_P(char const * tx1, char const * tx2, uint8_t info);

/** Log a 4 digit decimal number to buffered log output
 *
 *  This debugging function outputs a four digit decimal number /c info
 *  right aligned + a space.
 *
 *  \param info the four digit number
 *  \sa bufLogTxt_P
 *  \sa bufLogHex_P
 *  \sa bufLog2Dec_P
 *  \sa bufLogDecB
 *  \sa bufLogDec3
 */
void bufLogDec(uint16_t const info);

/** Log a three decimal digit number with leading zeroes to buffered log output
 *
 *  This debugging function outputs a three digit decimal number /c info
 *  with leading zeros ("000" .. "255").
 *
 *  \param info the three digit number
 *  \sa bufLogDec
 *  \sa bufLogDecB
 */
void bufLogDec3(uint8_t const info);

/** Log a two decimal digit number with leading zeroes to buffered log output
 *
 *  This debugging function outputs a two digit decimal number /c info
 *  with leading zero.  A value 0..99 will be output as two characters
 *  " 0" .. "99". Values of 100 and above will get " *".
 *
 *  \param info the two digit number
 *  \sa bufLogDec
 *  \sa bufLogDec3
 */
void bufLogDecB(uint8_t const info);

/** Log a two digit hex number to buffered log output
 *
 *  This debugging function logs a two digit hexadecimal number.
 *
 *  \param info the hex number (two digits)
 *  \sa bufLogTxt_P
 *  \sa bufLogDec_P
 */
void bufLogHex(uint8_t const info);

/** Log a four digit big endian hex number to buffered log output
 *
 *  This debugging function logs a 4 digits two big endian bytes hexadecimal
 *  number.
 *
 *  \param info the hex number (4 digits, big endian)
 */
void bufLog4HexBE(uint16_t const info);

/** Log an eight digit big endian hex number + one space to buffered log output
 *
 *  The same restrictions and warnings apply as for \ref bufLogTxt_P.
 *
 *  \param info the hex number (8 digits, big endian)
 */
void bufLog8HexBE(uint32_t const info);

/* Log the date and time
 *
 * This function logs the date and time in one line (like does the CLI date
 * command). Involving some (32 bit) divisions, this is an expensive
 * operation.
 */
void bufLogDate(void);

/** Output a (debug) marker text
 *
 *  The text output is " # # @ <ms-stamp>/<12,8 s> : <info>" <br />
 *  Both time stamps are hexadecimal. The first one is the system run time
 *  in ms since start / reset. The second is a 12,8s tick counter that wraps
 *  independently about every 3,27 ms.
 *
 *  \param info marker text   (short !, might be NULL)
 *  \param len info's length  (0..31)
 *  \sa cnt12u8_8
 */
void bufLogMark(char *  info, uint8_t const len);


/** Log space and linefeed to buffered log output
 *
 *  This debugging function outputs  a space + a linefeed at least once
 *  to max. 16 times if n is in the range 2..16.
 *
 *  The same restrictions and warnings apply as for \ref bufLogTxt_P.
 *
 *  \param n the number of lines
 *  \sa bufLogTxt_P
 *  \sa bufLogDec_P
 */
void bufLogLF(uint8_t n);


/** @} */ // Doxygen addtogroup swDebaid
/** @} */ // Doxygen addtogroup llSySer

#endif /* __STREAMS_H__ */

/** @} */ // Doxygen addtogroup weAutSys
