// Copyright (c) 2011 Albrecht Weinert, Bochum
// $Id: network.h 2 2017-01-25 16:49:03Z albrecht $

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup sysSer
 * @{
 * \addtogroup lanCom
 * @{
 *
 * \file network.h
 *
 * \ref intro_sec "weAutSys"' (low level) system calls, services and types
 *           for LAN / Ethernet communication
 *
 * This file contains the definitions for  \ref intro_sec "weAutSys"' (low level) system
 * calls, services and types for LAN / Ethernet communication. They imply an
 * ENC28J60 used for Ethernet driver and\ref uip "uIP" used as TCP/IP stack.
 * To change that this file and many other sources would have to be touched
 * (and not just configuration macros).
 *
 * This is system software and must not be modified for user or
 * application programs.
 *
 * Please find detailed description of the variables, functions etc.
 * in the modules \ref ethCom "Ethernet communications" and
 * \ref lanCom "LAN communications".
 *
 * This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

/** @} */ // Doxygen addtogroup lanCom
/** @} */ // Doxygen addtogroup sysSer

#ifndef NETWORK_H
#define NETWORK_H

#include "uip/uip.h"
#include "uip/uip_arp.h"
#include "we-aut_sys/syst_threads.h"


/**
 * \addtogroup sysSer
 * @{
 * \addtogroup lanCom
 * @{
 */

/** The well known Telnet port
 *  \showinitializer
 */
#define TELNET_PORT 23

/** The well known Echo port
 *  \showinitializer
 */
#define ECHO_PORT 7

/** The well known Modbus port
 *  \showinitializer
 */
#define MODBUS_PORT 502


/** Representation of a 48-bit Ethernet address / MAC address
 *
 *  This is \ref uip "uIP's" type \ref uip_eth_addr defined in uip.h
 *  holding an array uint8_t addr[6] in the end.
 *
 *  addr[0] would be octet [1] of the MAC address defining<br />
 *  Bit 0 =  0: unicast / = 1: multi/broadcast address <br />
 *  Bit 1 =  0: manufacturer  / =1: locally assigned address
 */
typedef  struct uip_eth_addr   eth_addr_t;


/** Set the IP stack's MAC address
 *
 *  This sets the the Ethernet stack's i.e. \ref uip "uIP"'s
 *  MAC address but not the driver's ( ENC28J60).
 *
 *  Of course, the driver's MAC setting has to be same. This is done by
 *  \ref encSetMacAdd(),  \ref encInit() or \ref networkInit().
 *
 *  \sa  uip_setethaddr
 *  \sa actMACadd
 *  \sa encSetMacAdd
 *  \param mac pointer to a structure (in RAM) holding the new MAC address
 */
void setMACadd(eth_addr_t * mac);

/** Set the stack's MAC address
 *
 *  This is the same as \ref  setMACadd except for the new MAC address
 *  structure held in program (flash) memory
 *
 *  \param mac pointer to a structure (in flash) holding the new MAC address
 */
void setMACaddP(const eth_addr_t * mac);


/** Initializer of an IP (V4) address from four byte values
 *
 *  This function constructs an {embraced} initialiser for an
 *  \ref uip_ipaddr_t (array) type.
 *
 * Example:
 \code
 uip_ipaddr_t defaultIPaddr = IPADD(192, 168, 89, 31);
 \endcode
 * will result in the correct array initialisation in network byte order
 * (big endian).
 *
 * \param addr0 The first octet of the IP address
 * \param addr1 The second octet
 * \param addr2 The third octet
 * \param addr3 The forth octet
 *
 * \hideinitializer
 */
#define IP_ADD(addr0, addr1, addr2, addr3)\
   {(((addr1) << 8) | (addr0)), (((addr3) << 8) | (addr2))}



//--------------------------------------------------   IP configuration ---

#define DHCP_MSK 0x01  //!< Mask for DHCP (use / set by) bit \sa ipConf_t

#define DNS1_MSK 0x10  //!< Mask for DNS server 1 set bit \sa ipConf_t
#define DNS2_MSK 0x20  //!< Mask for DNS server 2 set bit \sa ipConf_t
#define DNS_MSK 0x30   //!< Mask for DNS servers (set / use if set) bits \sa ipConf_t
#define DNS_CLIENT_MSK 0x10  //!< Mask for be DNS client bit \sa ipConf_t


#define NTP1_MSK 0x40  //!< Mask for NTP server 1 set  bit \sa ipConf_t
#define NTP2_MSK 0x80  //!< Mask for NTP server 2 set  bit \sa ipConf_t
#define NTP_MSK 0xC0   //!< Mask for NTP server bits \sa ipConf_t
#define NTP_CLIENT_MSK 0x40  //!< Mask for be NTP client bit \sa ipConf_t
#define NTP_SERVERT_MSK 0x80 //!< Mask for be NTP server bit \sa ipConf_t

/** The IP configuration */
typedef struct {

   uip_ipaddr_t ipAddr;  //!< The IP address
   uip_ipaddr_t netMask; //!< The netmask
   uip_ipaddr_t defaultRouter; //!< The default router

/** The name server
 *
 *  This is the IP address of the name server 1.
 *
 *  \ref intro_sec "weAutSys" and ref uip "uIP" dispense with having
 *  multiple DNS servers.
 */
   uip_ipaddr_t dnsAddr;

/** The name server 2
 *
 *  This is the IP address of the name server 2. <br />
 *  If just one DNS server is available / set this should be \ref dnsAddr.
 *
 *  \ref intro_sec "weAutSys" and ref uip "uIP" dispense with having
 *  multiple DNS servers.
 */
   uip_ipaddr_t dns2Addr;

/** The time server
 *
 *  This is the IP address of the NTP server 1 (if set).
 *
 */
   uip_ipaddr_t ntpAddr;

/** The time server 2
 *
 *  This is the IP address of the NTP server 2. <br />
 *  If just one NTP server is available / set this should be \ref ntpAddr.
 */
   uip_ipaddr_t ntp2Addr;

/** Lease time in seconds
 *
 *  If bit 1 (\ref DHCP_MSK) is set in setFlags (part of) this configuration
 *  was set by a DHCP server and this is the real lease time in seconds
 *  (normal byte order).
 *
 *  Otherwise this element might be used as the requested lease time.
 */
  uint32_t leaseTime;

/** Set time
 *
 *  This is the \ref secClock() "run time" (with seconds resolution) when this
 *  configuration was set. If bit 1 (\ref DHCP_MSK)
 *  is set this is the time of the DHCP server's positive answer.
 *
 *  Rationale to use the \ref secClock() "run time" instead of an
 *  \ref secTime308Loc() "absolute local" time is a) the first one having no
 *  settings and gaps giving allways a correct "age" of the setting and b)
 *  the second one usually being set after the just set DHCP named one or two
 *  NTP servers and one of them gave the first correct
 *  \ref secTime308UTC() "absolute time".
 */
  uint32_t setTime;

/** Flags for the field is valid / is set
 *
 *  Bit 0 : This configuration (or part of it) was set by responses
 *          from a DHCP server (\ref DHCP_MSK).  <br />
 *  Bit 4 : name server 1 is set (\ref DNS1_MSK) <br />
 *  Bit 5 : name server 2 is set (\ref DNS2_MSK) <br />
 *  Bit 6 : time server 1 is set (\ref NTP1_MSK) <br />
 *  Bit 7 : time server 2 is set (\ref NTP2_MSK) <br />
 */
   uint8_t setFlags;

/* The network usage flags
 *
 *  Bit 0 : try to use DHCP before using the defaultConfData.ipConf
 *            default configuration.  (\ref DHCP_MSK). <br />
 *  Bit 4 : may use a name server if one is set / be DNS client
 *            (\ref DNS1_MSK \ref DNS2_MSK \ref DNS_CLIENT_MSK)  <br />
 *  Bit 6 : may use a time server  if one is set / be NTP client
 *            (\ref NTP1_MSK \ref NTP2_MSK \ref NTP_CLIENT_MSK) <br />
 *  Bit 7 : act as NPT time server if date time is set
 *            (\ref NTP_SERVER_MSK) <br />
 */
   uint8_t useFlags;

} ipConf_t;


/** Set the addresses in the uIP stack
 *
 *  This function takes the fields ipAddr, netMask, defaultRouter and
 *  dnsAddr and makes the appropriate settings in the \ref uip "uIP" stack.
 *  The bit 4 ("stack's IP is not yet set") is cleared in
 *  \ref networkNotReady.
 *
 *  \c ipConf's cFlags is irrelevant: the settings are made even if the
 *  flag says "use DHCP". Hence \c ipConf must not be null and it must have
 *  correct entries for \c ipAddr, \c netMask and \c defaultRouter.
 *
 *  \sa uip_sethostaddr
 *  \sa uip_setnetmask
 *  \sa uip_setdraddr
 *  \sa resolv_conf
 *
 *  \param ipConf  pointer to the source of the addresses and masks
 */
void uipSetAddresses(ipConf_t * ipConf);


/** Get the addresses from the uIP stack
 *
 *  This function fills the fields ipAddr, netMask, defaultRouter and
 *  dnsAddr from the respective settings in the \ref uip "uIP" stack.
 *
 *  ipConf's cFlags is left unchanged.
 *
 *  \sa uipSetAddresses
 *
 *  \param ipConf  pointer to the destination of the addresses and masks
 */
void uipGetAddresses(ipConf_t * ipConf);

/** The actual IP configuration
 *
 *  \sa  defaultConfData defaultConfData.ipConf
 */
extern ipConf_t curIpConf;

/** The period of the ARP timeout tick in seconds
 *
 *  The standard / default value is 10. User / application software
 *  might set a higher value to reduce ARP Controller load.
 *
 *  \sa UIP_ARP_MAXAGE
 *  \sa arpTimeOut
 */
extern uint8_t arpTickPeriod; // count down in about 10 s

/** The ARP timeout counter in seconds
 *
 *  This counter is decremented every second. At 0 it is reset to
 *  \ref arpTickPeriod and \ref uip_arp_timer() is called. The latter
 *  \ref uip "uIP" function handles the aging of ARP entries.
 *
 *  \sa UIP_ARP_MAXAGE
 *  \sa arparpTickPeriod
 */
extern uint8_t arpTimeOut; // count down in about 10 s

/** Network not ready
 *
 *  If this state flag 0x00 means the network is ready.<br />
 *  One of the lower four bits stands for a network state or problem that
 *  impedes communication. <br />
 *  The upper four bits do the same for the hardware driver, which in the case
 *  of \ref intro_secH "weAut_01" it is an ENC28J60.
 *
 *  Bit 7 : ENC is not initialised yet / was reset (mask \ref ENC_NRD) <br />
 *  Bit 6 : ENC has hardware (\ref CLKRDY) trouble (mask \ref ENC_HWP) <br />
 *  Bit 5 : ENC is in power save (\ref PWRSV) mode (mask \ref ENC_PWS) <br />
 *
 *  Bit 3 : stack's IP is not yet set / determined
 *          (by DHCP e.g., mask \ref STC_NST) <br />
 *  Bit 2 : link is down (status of last last request by e.g.
 *          \ref linkState(), mask \ref LNC_IDN) <br />
 *  Bit 1 : link was down (status before last last request by e.g.
 *          \ref linkState(), mask \ref LNC_WDN) <br />
 */
extern uint8_t networkNotReady;

#define ENC_MSK 0xF0  //!< Mask for the ENC bits in \ref networkNotReady
#define ENC_NRD 0x80  //!< Mask for the ENC not ready bit in \ref networkNotReady
#define ENC_PWS 0x20  //!< Mask for the ENC powersave bit in \ref networkNotReady
#define ENC_HWP 0x40  //!< Mask for the ENC HW problem bit in \ref networkNotReady

#define STC_MSK 0x0F  //!< Mask for the stack bits in \ref networkNotReady
#define STC_NST 0x08  //!< Mask for the stack' IP not set bit in \ref networkNotReady
#define LNC_IDN 0x04  //!< Mask for the link is down (at last status request) in \ref networkNotReady
#define LNC_WDN 0x02  //!< Mask for the link was down (before last status request) in \ref networkNotReady


/** The uip buffer cast to Ethernet header
 *
 *  This macro provides the buffer \ref uip_buf as type \ref uip_eth_hdr .
 */
#define BUF2EH ((struct uip_eth_hdr *)uip_buf_alias)

/** The uip buffer cast to Ethernet IP header
 *
 *  This macro provides the buffer \ref uip_buf as type \ref ethip_hdr .
 */
#define IPBUF ((struct ethip_hdr *)uip_buf_alias)

/*  Internal hint: macros like BUF2EH above and cognates will get
 *        "dereferencing type-punned pointer will break strict-aliasing rules"
 *  by newer optimising (GCC) compilers. This may be safely avoided (and is
 *  in \ref intro_sec "weAutSys") by having an extra char * alias  pointing to uip_buf which
 *  is of type uint8_t[]
 */


/** Send the actual uIP buffer
 *
 *  See function devicedriver_send respectively ethernetdevicedriver_send
 *  in \ref uip "uIP" documentation.
 *
 *  This function does nothing if the NIC's transmit logic is not (yet) idle.
 *  In case of  \ref intro_secH "weAut_01" the NIC is an ENC28J60.
 */
void deviceDriverSend(void);

/** Receive and then perhaps send the actual uIP buffer
 *
 *  This is standard control structure around calling \ref uip_input
 *  as described in the \ref uip "uIP" documentation in one function.
 *
 *  To have all this in one block by calling this function may be
 *  questionable. It will be put in (pieces in) appropriate threads. And as
 *  of this writing this function was not used yet.
 */
void networkPolling(void);

/** Hand over received packages
 *
 *  This function checks if the network driver IC (NIC), an ENC28J60 in case
 *  of  \ref intro_secH "weAut_01", has received a package. If so it is copied
 *  to \ref uip "uIP"'s \ref uip_buf and the number of bytes is returned.
 *
 *  See function devicedriver_poll respectively ethernetdevicedrver_poll
 *  in \ref uip "uIP" documentation.
 *
 *  \return the length of the (new) package or 0 if there was none
 */
uint16_t deviceDriverPoll(void);


/** Output Ethernet and driver (ENC28J60) state info, the thread function
 *
 *  This thread will output some info on Ethernet and driver configuration
 *  and state to \c toStream (as standard output).
 *
 *  This thread function acts on \ref hierThr_data_t "threadD"'s secondary
 *  thread and secondary flag.  \c flag2's bits  signal which informations
 *  are to be output: <br />
 *  Bit 0 : MAC <br />
 *  Bit 1 : current IP setting <br />
 *  Bit 2 : default IP setting <br />
 *  Bit 3 : ARP <br />
 *  Bit 4 : Link state <br />
 *  Bit 5 : DHCP state
 *
 *  As secondary thread it  will usually be started as the primary threads
 *  child.
 *
 *  \param threadD the thread state data (pointer to; not NULL)
 *  \param toStream streams to switch the standard streams to (if not MULL)
 *  \return \ref PT_YIELDED if waiting for conditions that may become
 *           automatically true; \ref PT_WAITING if blocked by conditions
 *           that other's action may set true;  \ref PT_EXITED if stopped
 *           by conditions that never will become true (sorry, not all work
 *           done and never will; \ref PT_ENDED if ready with all work
 *
 */
ptfnct_t outEncLanInfoThreadF(hierThr_data_t * threadD, FILE * toStream);


/** Check the link state
 *
 *  This function checks the actual link state and sets the relevant bits
 *  (\ref LNC_IDN and \ref LNC_WDN) in \ref networkNotReady.
 *
 *  \return 0: link is down; 1: link is and was up since last call;
 *          2: link is up, but was down since last call
 */
uint8_t linkState(void);


/** Get the MAC address for an IP address
 *
 *  This function checks the ARP table for an entry about ipAddr and returns
 *  the (pointer to its) MAC address.
 *
 *  The caller must not manipulate the returned MAC.
 *
 *
 *  \param  ipAddr  IP address the MAC is to be determined for
 *  \return pointer to MAC if determinable or NULL else
 */
const struct uip_eth_addr * getMACforIPaddr(uip_ipaddr_t ipAddr);

/** The uIP udp event function for the application software
 *
 *  The reason for being called will be in the \c uip_udp_conn structure,
 *  especially in the ports.
 *
 *  This function will be called for events, ports etc. not handled by
 *  \ref udp_appcall "system" software (like NTP or DHCP e.g.). <br />
 *  If no extra protocols are implemented by user / application software the
 *  implementation should do nothing (but \c return; ).
 *
 *  Otherwise it is best to use (at least) the remote port to distribute
 *  to different protocol handling functions (best being protothreads).
 *  Client example: \code
void udpAppcall(void){

   if(uip_udp_conn->rport == HTONS(MTP_PORT)) {  // MTP protocol
     mtpAppcall(); // protothread function / state machine handling MTP
     return;
   }  // MTP protocol

   if(uip_udp_conn->rport == HTONS(THCPC_SERVER_PORT)) {  // THCP protocol
      thcpc_appcall(); // protothread function / state machine handling THCP
      return;
   } // THCP protocol (client)

   // add further protocols here

} //  udpAppcall()
 *  \endcode
 *  \sa udp_appcall
 */
void udpAppcall(void);

/** The uIP event function for the application software
 *
 *  The reason for being called will be in \c uip_flags.
 *
 *  This function will be called for events, ports etc. not handled by
 *  \ref uip_appcall "system" software. <br />
 *  If no extra protocols are implemented by user / application software the
 *  implementation should do nothing (but \c return; ).
 *
 *  Otherwise it is best to use (at least) the local port to distribute
 *  to different protocol handlers. The handlers would best be implemented as
 *  protothread functions using appropriate \c appstate fields in the
 *  \c uip_conn as state.<br />
 *  The server example is simplified (i.e. stateless): \code
void uipAppcall(void){

   // for state use something
   // like : struct cliThr_data_t *s = (cliThr_data_t *)&(uip_conn->appstate);

   if(uip_conn->lport == HTONS(7)   // Echo protocol .. and
         && (uip_newdata() || uip_acked())) { // new (or ack'ed) data
      uip_send(uip_appdata, uip_datalen());
      return;
   } // echo server (new data only)

   // add further handling and protocols here

} // uipAppcall()
 *  \endcode
 *  \sa uip_appcall
 */
void uipAppcall(void);


/** @} */ // Doxygen addtogroup  lanCom
/** @} */ // Doxygen addtogroup  sysSer

//-- Initialisation functions ------------------------

/**
 * \addtogroup llSySer
 * @{
 *
 * \addtogroup  sysInit
 * @{
 */

/** Initialise the network stack
 *
 *  This function initialises the Ethernet driver ENC28J60 and the
 *  \ref uip "uIP" stack.
 *
 *  The driver IC ENC28J60 is not reset (\ref encReset) here.
 *
 *  \sa encInit()
 *  \sa lanComStdInit()
 */
void networkInit(void);

/** Initialise the LAN communication (standard way)
 *
 *  This function initialises the LAN communication in a standard way.
 *  If  \ref DHCP_MSK "use DHCP" is set in \ref curIpConf  (.useFlags) DHCP is
 *  initialiesd and nothing else is done.
 *
 *  Otherwise the \ref curIpConf "IP configuration" is set by the
 *  \ref  defaultConfData "default configuration" and NTP is
 *  \ref ntpInit "initialised" if \ref NTP_CLIENT_MSK "be NTP client" is set.
 */
void lanComStdInit(void);

/** @} */ // Doxygen addtogroup sysInit
/** @} */ // Doxygen addtogroup llSySer

//-- Formatter  ------------------------

/**
 * \addtogroup  utilHelp
 * @{
 * \addtogroup  utilForm
 * @{
 */

/** Format a MAC address
 *
 *  The MAC address from the structure \c mac will be formatted as
 *  11:22:33:44:55:66 delivering 18 characters in \c s including a
 *  trailing zero.
 *
 *  \param s start position within string respectively char array
 *  \param mac the (6 byte) address structure
 *  \return pointer to the (0-) character following the formatted MAC-address
 *  \sa formIpAdd
 *  \sa formIpConf
 */
char * formMacAdd(char * s, eth_addr_t * mac);


/** Format an ipV4 Ethernet address
 *
 *  The IP address from the structure \c ipAddr supplied will be formatted as
 *  112.12.0.1 delivering 8..16 characters including a trailing zero.
 *
 *  \param s start position within string respectively char array (length 16)
 *  \param ipAddr the (uint16-t[2] big endian) address structure
 *  \return pointer to the (0-) character following the formatted address
 *  \sa formMacAdd
 *  \sa formIpConf
 */
char * formIpAdd(char * s, uip_ipaddr_t ipAddr);

/** Format an IP configuration (w/o DHCP)
 *
 *  This function formats the IP configuration \c ipConf to a multi-line
 *  text.<br />
 *  The length of the text delivered (to \c *s) is up to 165 characters
 *  including a trailing zero.
 *
 *  The output does not contain the DHCP information.
 *
 *  \param s start position within string respectively char array
 *  \param ipConf  pointer to the IP configuration (addresses, masks, flags)
 *  \return pointer to the (0-) character following the formatted configuration
 *  \sa formMacAdd
 *  \sa formIPAdd
 *  \sa formIpConfDHCP
 */
char * formIpConf(char * s, ipConf_t * ipConf);

/** Format an IP configuration, the DHCP part
 *
 *  This function formats the IP configuration to a multi-line text.<br />
 *  The length of the text delivered (to \c *s) is up to 65 characters
 *  including a trailing zero.
 *
 *  The output does only contain the DHCP information.
 *
 *  \param s start position within string respectively char array
 *  \param ipConf  pointer to the IP configuration (addresses, masks, flags)
 *  \return pointer to the (0-) character following the formatted configuration
 *  \sa formMacAdd
 *  \sa formIPAdd
 *  \sa formIpConf
 */
char * formIpConfDHCP(char * s, ipConf_t * ipConf);


/** @} */ // Doxygen addtogroup utilForm
/**
 * \addtogroup utilPars
 * @{
 */

/** Parse an ipV4 Ethernet address
 *
 *  This function tries to interpret a part of the string s starting at
 *  \c si as an IP address expected in the form "112.12.0.1" i.e. four dot
 *  separated numbers of up to three decimal digits each.
 *
 *  Leading blanks will be skipped.
 *
 *  \param ipAddr pointer to the result (ipaddr_t is an array type); the
 *             result is changed only if an IP address could be parsed
 *             successfully
 *  \param s   the RAM string containing the digit(s)
 *  \param si  the first digit's index in s
 *  \return    the index of the first character after the parsed IP address
 *             in case of success or 0 else
 */
uint8_t parseIpAdd(uip_ipaddr_t ipAddr, char s[], uint8_t si);


/** Parse a MAC address
 *
 *  This function tries to interpret a part of the string s starting at
 *  \c si as a MAC address expected in the form "11:2B:3C:4D:5E:6F" i.e. six
 *  colon separated hexadecimal numbers of (up to) two digits each.
 *
 *  Leading blanks will be skipped.
 *
 *  \param macAddr pointer to the result (eth_addr_t is a structure
 *             containing an array type); the result is changed only
 *             if a MAC address could be parsed successfully
 *  \param s   the RAM string containing the digit(s)
 *  \param si  the first digit's index in s
 *  \return    the index of the first character after the parsed IP address
 *             in case of success or 0 else
 */
uint8_t parseMACadd(eth_addr_t * macAddr, char s[], uint8_t si);


/** @} */ // Doxygen addtogroup utilPars
/** @} */ // Doxygen defgroup utilHelp
#endif /* NETWORK_H */
/** @} */ // Doxygen addtogroup weAutSys
