// Copyright (c) 2012 Albrecht Weinert, Bochum
// ntp.h 2 2017-01-25 16:49:03Z albrecht

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup appLaySup
 * @{
 * \addtogroup ntpClient
 * @{
 *
 * \file ntp.h
 *
 * \ref intro_sec "weAutSys"' system calls and types for network time services
 *
 * This file contains the definitions for \ref intro_sec "weAutSys"'
 * system calls, services and types for using network time. They imply an
 * \ref uip "uIP" used as TCP/IP stack and \ref intro_sec "weAutSys"' timing
 * / time keeping modules.
 *
 * Please see also the modules \ref ethCom "Ethernet communications" and
 * \ref timeKeep "Time keeping (date and time)".
 *
 * This is system software and must not be modified for user or
 * application programs.
 *
 * This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012  Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert

/** @} */ // Doxygen addtogroup ntpClient
/** @} */ // Doxygen addtogroup appLaySup

#ifndef NTP_H
#define NTP_H

#include "we-aut_sys/network.h"
#include "we-aut_sys/uart0.h"
#include "we-aut_sys/timing.h"

/**
 * \addtogroup appLaySup
 * @{
 * \addtogroup ntpClient
 * @{
 */

/** The well known NTP port
 *
 *  NTP normally uses UDP port 123.
 */
#define NTP_PORT 123


/** NTP time stamp type
 *
 *  The NTP time stamp type as whole is a 64 bit unsigned number the upper
 *  32 bits being the seconds since era start and the lower 32 bits being
 *  their fraction in 2**-32 = ~0.23ns resolution.
 *
 *  Even being unsigned by declaration differences may be negative appearing
 *  as very large numbers. Modulo arithmetic will tread them right in
 *  addition and subtraction &mdash; only careless comparisons might go
 *  wrong.
 *
 *  /note Both parts (and the whole thing) is in wrong (big endian / network)
 *  byte order.
 */
struct ntpTimestamp_t {

/** The seconds
 *
 *  The zero date for NTP (era 0) is January 1st 1900 (1900-1-1; 1.1.1900)
 *  UTC. The wrap around at February 7th 2036 (next "era") is happily defined
 *  using  2**32 modulo arithmetic in a way one practically hasn't to worry
 *  about.
 *
 *  /note This field is in wrong (big endian / network) byte order.
 */
   ucnt32_t seconds;

/** The fractions of the second
 *
 *  The Bits of fraction.v8[3] are to be interpreted as [7|...|0]: <br />
 *  1/2 s | 1/4 s | 1/8 s | 1/16 s | ... <br />
 *  500 ms | 250 ms | 125 ms | ~62 ms ... The fraction digits 1/16 s and below
 *  cannot be expressed in ms. This is a complication for (most) systems using
 *  ms (or s) clocks respectively counters.
 */
   ucnt32_t  fraction;
}; // ntpTimestamp_t


/** Fraction byte significance
 *
 *  As said NTP \ref ntpTimestamp_t "time stamp"'s fraction has four bytes
 *  allowing 2**-32 = ~0.23ns resolution. That is far beyond any practical
 *  use for e.g. ATmel ATmega1284P based automation modules, like weAut_01
 *  with 20MHz processor clock (50ns period) at best.
 *
 *  Hence this value determines the number of bytes being ignored for all
 *  64 bit NTP time stamp operations.
 *
 *  Range: 0 (full 64 bit arithmetic) .. 3 (fraction is 8 bit only) <br />
 *  Recommended / default: 2 (16 bit fraction, sub-ms resolution)
 *
 *  The recommended value 2 is a good choice also in the light of most NTP
 *  servers. Observations of Windows 2008 R2 as NTP server showed the least
 *  16 fraction bits being always the same (rubbish ?) value no matter the
 *  circumstances. <br />
 *  \c NTP_FRACT_DROP 2 reduces \ref ntpTimestampSum, \ref ntpTimestampDif
 *  and \ref ntpTimestampHalf() from 64 bit to 48 bit arithmetic
 *  (saving ten / six cycles). \c NTP_FRACT_DROP 2  also
 *  fits well to \ref ntpTimestampToMillies() and \ref setNTPstampAct().
 */
#define NTP_FRACT_DROP 2


/** Sum of two NTP time stamps
 *
 *  This function calculates *result = *a + *b; <br />
 *  No pointer must be null, but two or three of them might point to the
 *  same \ref ntpTimestamp_t structure.
 *
 *  \ref ntpTimestampSum, \ref ntpTimestampDif and \ref ntpTimestampHalf are
 *  optimised implementations for 64 bit wrong endian (i.e. network byte
 *  order). <br />
 *  The former two take 74 cycles and the latter 43. This values are for
 *  \ref NTP_FRACT_DROP 0, the recommended value 2 makes all a bit faster.
 *
 *  \param a points to first operand (not NULL)
 *  \param b   points to second operand (not NULL)
 *  \param result points to result structure (not NULL)
 *  \sa  NTP_FRACT_DROP
 */
void ntpTimestampSum(struct ntpTimestamp_t * result,
                       struct ntpTimestamp_t * a, struct ntpTimestamp_t * b);

/** Difference of two NTP time stamps
 *
 *  This function calculates *result = *a - *b; <br />
 *  No pointer must be null, but two of them might point to the
 *  same \ref ntpTimestamp_t structure.
 *
 *  \ref ntpTimestampSum, \ref ntpTimestampDif and \ref ntpTimestampHalf are
 *  optimised implementations for 64 bit big endian.
 *
 *  \param a points to first operand (not NULL)
 *  \param b   points to second operand (not NULL)
 *  \param result points to result structure (not NULL)
 *  \sa  NTP_FRACT_DROP
 */
void ntpTimestampDif(struct ntpTimestamp_t * result,
                        struct ntpTimestamp_t * a, struct ntpTimestamp_t * b);

/** Half a NTP time stamp value
 *
 *  This function calculates *result = *result / 2 <br />
 *
 *  The division by 2 is performed for a signed (two complement) 64 bit
 *  number. This also gives correct results for large differences in
 *  modulo 2**64 bit arithmetic, but will of course fail on very large
 *  unsigned numbers >= 2**63.
 *
 *  \ref ntpTimestampSum, \ref ntpTimestampDif and \ref ntpTimestampHalf are
 *  optimised implementations for 64 bit big endian.
 *
 *  \param result points to operand and result structure (not NULL)
 *  \sa  NTP_FRACT_DROP
 */
void ntpTimestampHalf(struct ntpTimestamp_t * result) __attribute__ ((pure));

/** Milliseconds from a NTP time stamp's fraction
 *
 *  This function converts a NTP time stamp's fraction value (32 bit) into
 *  milliseconds in the range 0 .. 999 (16 bit).<br /> The fraction is
 *  expected in (wrong) network endianess while the result will be set in the
 *  correct byte order.
 *
 *  The implementation is (inline) ASM optimized. It sacrifices a tiny bit
 *  of exactness in favor of a slow RISC processor with only byte operations.
 *  The effect is the 0 .. 999ms being mapped to 0..992ms. <br />
 *  The assumption behind is the worst case 1% of a second error a being far
 *  within the NTP synchronisation abilities of an ATmega-uIP-ENC trio. <br />
 *  Round trip times of 50ms .. 80ms with a Windows 2008 R2 time server in a
 *  local network with some switches in between suggest the un-symmetry under
 *  these near ideal conditions being at least some milliseconds. Hence the
 *  above assumption seems quite adequate.
 *
 *  \param ntpStamp structure pointed to has fraction in network byte order;
 *          never null!<br />
 *          Using a pointer for this 16 bit result instead of a return
 *          value is just a work around a GCC inline ASM bug.
 *  \param ms points to the result variable where to put <br />
 *    value.fraction converted to ms (0 .. 999) to; never null
 *  \sa ntpTimestampSum
 *  \sa ntpTimestampHalf
 *  \sa setNTPstampAct
 */
void  ntpTimestampToMillies(uint16_t * ms, struct ntpTimestamp_t * ntpStamp);


/** Set a NTP time stamp to actual time
 *
 *  This function sets the \c .seconds and \c .fraction of \c ntpTimestamp to
 *  the actual UTC system time (\ref secTime308UTC(), \ref  msAbsClockCount).
 *
 *  On \c .fraction all but the upper 10 bits are set to zero.
 *
 *  The ms to fraction algorithm implementation is (ASM) optimised for the
 *  8bit AVR RISC. It sacrifices a tiny bit of exactness in favor of great
 *  CPU usage savings. The effect is the 0 .. 999ms of \ref msAbsClockCount
 *  being mapped to effectively 0..992ms in the NTP stamp fraction.
 *
 *  \param ntpTimestamp points to the stamp to be set (never null)
 *  \sa ntpTimestampToMillies
 */
void setNTPstampAct(struct ntpTimestamp_t * ntpTimestamp);


/** Adjust the actual (system) time by a NTP time stamp difference
 *
 *  This function adjusts the internal date and time (\ref secTime308UTC())
 *  by an amount given as a a NTP time stamp \ref ntpTimestampDif "difference"
 *  to the\ref secTime308Loc "actual time in s", \ref msAbsClockCount "ms".
 *
 *  Big (>=2**31 in the seconds part) differences will quite naturally act
 *  as negative in the sense of \ref secTime308Loc "second based" timing's
 *  (and NTP's) modulo  2**32 arithmetic.
 *
 *  This function should be used for any absolute date time adjustments
 *  caused by NTP client functions (or may be also other time sources).
 *
 *  The action taken and the flags set in \ref ntpState \c .lastAdjust
 *  depend on the (absolute) difference amount<ul>
 *  <li> below 8 ms : no action (considered as jitter) </li>
 *  <li> -100 .. +50ms : no setting; <br /> &nbsp; &nbsp;
 *        only oscillator speed up / slow down </li>
 *  <li> -999 .. + 999ms : setting of time and <br /> &nbsp; &nbsp;
 *        oscillator speed up / slow down </li>
 *  <li> above -1 .. +1s : setting of time only </li></ul>
 *
 *  \param ntpTimeDif  the adjustment difference in seconds,fraction;
 *                big values act as negative in the sense of seconds
 *                modulo 2**32 arithmetic
 *  \sa secTime308Loc(void)
 *  \sa secTime308UTC(void)
 */
 void adjustNTPtime308UTC(struct ntpTimestamp_t * ntpTimeDif);



/* the NTP message
 *
       0 1 2 3 4 5 6 7 8 9 X 1 2 3 4 5 6 7 8 9 V 1 2 3 4 5 6 7 8 9  T  1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |LI | VN  |Mode |    Stratum    |     Poll      |   Precision    |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                          Root  Delay                           |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                       Root  Dispersion                         |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                     Reference Identifier                       |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                    Reference Timestamp (64)                    |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                    Originate Timestamp (64)                    |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                     Receive Timestamp (64)                     |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                     Transmit Timestamp (64)                    |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                 Key Identifier (optional) (32)                 |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                 Message Digest (optional) (128)                |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

      LI       Meaning
      ---------------------------------------------
      0        no warning
      1        last minute has 61 seconds
      2        last minute has 59 seconds
      3        alarm condition (clock not synchronized)

       Stratum  Meaning
      ----------------------------------------------
      0        kiss-o'-death message (see below)
      1        primary reference (e.g., synchronized by radio clock)
      2-15     secondary reference (synchronized by NTP or SNTP)
      16-255   reserved
 */

/** NTP message  type
 *
 *  See  <br /> &nbsp;
 *   Network Time Protocol Version 4 Protocol and Algorithms Specification
 *   [draft]<br />
 *  or see [RFC 4330] (p. 8) <br />  &nbsp;
 *  Simple Network Time Protocol (SNTP) Version 4 for IPv4, IPv6 and OSI <br />
 *  and see also  [RFC 2030, cites below] <br />
 *  <br />  &nbsp;
 *  NTP client operations [RFC 2030 quoted]:
 *  \code
      Timestamp name         ID   is set when
      -----------------------------------------------------------------------
      originateTimestamp     T1   request sent by client as transmitTimestamp
      receiveTimestamp       T2   request received by server
      transmitTimestamp      T3   reply sent by server
      destinationTimestamp   T4   reply received by client

   The roundtrip delay d and system clock offset t are defined as:

      d = (T4 - T1) - (T3 - T2)     t = ((T2 - T1) + (T3 - T4)) / 2  \endcode
 *
 * For the arithmetic operations on the \ref ntpTimestamp_t "timestamps" use
 * the functions \ref ntpTimestampSum, \ref ntpTimestampDif and
 * \ref ntpTimestampHalf.
 */
struct ntpMess_t {

/** The flags
 *
 *  This byte consists of the leap indicator (2 Bits), a version number
 *  (3 bits) and  mode (3 bits).
 */
   uint8_t flags;
   uint8_t stratum; //!< A kind of time server hierarchy level

/** Maximum interval between successive messages
 *
 *  The value is 2**poll s. <br />
 *  Range: 4 .. 17 (16s .. 36h)
 */
   uint8_t poll;

/** Servers time precision
 *
 *  The value is 2**precision s. <br />
 *  Range: -6 .. -20  (16ms .. 1s)
 */
   int8_t precision;

   ucnt32_t rootDelay; //!< see RFC 2030
   ucnt32_t rootDispersion;  //!< see RFC 2030

/** Identifies the particular reference clock
 *
 *  Not used or evaluated here. See RFC 2030.
 */
   ucnt32_t referenceIdentifier;

/** The timestamp of the last setting of the local clock
 *
 *  If the local clock has never been synchronized, the value is zero;
 *  see  RFC 2030. <br />
 *  Not used or evaluated here.
 */
   struct ntpTimestamp_t referenceTimestamp;
   struct ntpTimestamp_t originateTimestamp; //!< The T1, see RFC 2030
   struct ntpTimestamp_t receiveTimestamp;   //!< The T2, see RFC 2030
   struct ntpTimestamp_t transmitTimestamp;  //!< The T3, see RFC 2030

} ; // struct ntpMess_t



#define FLAGS_LI_MASK 0xC0 //!< leap indicator bits in \ref ntpMess_t.flags
#define LI_NOWARN  0    //!< leap indicator 0 : no warning
#define LI_61SEC  0x40  //!< leap indicator 1 : minute = 61s
#define LI_59SEC  0x80  //!< leap indicator 2 : minute = 59s
#define LI_ALARM  0xC0  //!< leap indicator 3 : alarm (clock not synchronized)

#define FLAGS_VERSION_MASK 0x38 //!< version bits in \ref ntpMess_t.flags
#define NTP_VERSION  0x20 //!< current version (4) within \ref FLAGS_VERSION_MASK

#define FLAGS_MODE_MASK 0x07 //!< mode bits in \ref ntpMess_t.flags
#define NTP_CLIENT  3    //!< NTP mode client (3)
#define NTP_SERVER  4    //!< NTP mode server (4)
#define NTP_BROADCAST  5 //!< NTP mode broadcast (5)


#define NTP_STATE_RESET  0 //!< No known NTP server, no NTP client operation \showinitializer

#define NTP_STATE_CONN1  1 //!< Server known, port connected to server 1 \showinitializer
#define NTP_STATE_CONN2  2 //!< Server known, port connected to server 2 \showinitializer

#define NTP_STATE_SRVMSK  3   //!< Server (bits) mask
#define NTP_STATE_W4REPL  8   //!< flag: waiting for server's reply \showinitializer
#define NTP_STATE_OPMSK 0xF0  //!< Operation (bits) mask

#define NTP_STATE_TRYREQ 0x10 //!< try (first) request   \showinitializer
#define NTP_STATE_CNPROB 0xF0 //!< connection problematic \showinitializer
#define NTP_STATE_SEPROB 0xB0 //!< server answer problematic (bogus, alarm) \showinitializer
#define NTP_STATE_CNETAB 0xE0 //!< connection established \showinitializer
#define NTP_STATE_CONTSY 0xC0 //!< continuous synchronisation \showinitializer

/** NTP asks for priority
 *
 *  This expression is true if the NTP client sent a request and waits for
 *  the response. Postponing the reaction to this response by using needed
 *  resources (NIC, SPI) or having long running thread steps (by SMC block
 *  operations e.g.) will detriment the synchronisation accuracy.
 *
 *  As NTP synchronisations are quite seldom (about every minute in good
 *  sync state) and turn-around for (recommended) local NTP servers is quite
 *  fast it should be feasible for all others to respect this request for
 *  priority.
 */
#define ntpAsksPrio() (ntpState.state & NTP_STATE_W4REPL)

/** Structure for NTP (client) application state */
struct ntpState_t {
   struct pt pt;  //!< the protothreads (raw) structure
   struct timer_t timer; //!< the protocol timer, seconds resolution, duration type
   struct uip_udp_conn *conn;  //!< pointer to the connection (structure)

/** NTP state machine state
 *
 *  The lower \ref  NTP_STATE_SRVMSK two bits are 0 = no connection or the
 *  number (1, 2) of a configured NTP server connected to. <br />
 *  \ref NTP_STATE_W4REPL "Bit 3" (8) means waiting for server's reply.<br />
 *  At the \ref NTP_STATE_OPMSK "upper 4" bits
 *  \ref  NTP_STATE_TRYREQ "0x10" means first try respectively first
 *  request,<br />
 *  \ref NTP_STATE_CNPROB "0xF0" says connection problematic or faulty, <br />
 *  \ref NTP_STATE_SEPROB "0xB0" is server answer problematic (bogus,
 *  alarm), <br />
 *  \ref NTP_STATE_CNETAB "0xE0" means connection established and <br />
 *  \ref NTP_STATE_CONTSY "0xC0" signals continuous synchronisation.
 */
   uint8_t state;
   uint8_t count; //!< NTP state machine counter (used in some states)

 /** NTP repoprt level
  *
  *  The higher the value the more events will be reported to the buffered
  *  log device.<br />
  *  0: uninitialised; will be set according to macro \ref DEBUG_NTP + 3 when
  *  \ref ntpInit "initialising" NTP. <br />
  *  0..3: silent <br />
  *  4 : quiet (almost silent) <br />
  *  5 : info <br />
  *  6 : verbose <br />
  *  7 : debug
  */
   uint8_t logLev;

   struct ntpTimestamp_t lastSendTimestamp;  //!< our send / server' originate (T1)
   struct ntpTimestamp_t lastReceiveTimestamp; //!< our receive (T4)

   struct ntpTimestamp_t lastT34;  //!< the last T3-T4
   struct ntpTimestamp_t lastT21;  //!< the last T2-T1

/** State / actions of actual and previous adjustment
 *
 *  Bits 7,6 : 1 = last positive, 2 = last negative, 0 = last 0 <br />
 *  Bits 5,4 : 0 = in the range -100 .. + 50ms <br /> &nbsp; &nbsp;
 *       1 = in the range -999 .. +999 ms <br /> &nbsp; &nbsp;
 *       2 = in the range +1s .. +1999 ms; 3 = outside -999 .. +1999 ms <br />
 *
 *  Bits 4,5 : same as 6,7 for previous action <br />
 *  Bits 0,1 : same as 2,3 for previous action
 */
   uint8_t  adjuState;

/** Actual deviation in seconds
 *
 *  This is the current seconds deviation (mod 2**32 in normal byte order).
 *  The effective sign of the correction is Bit 7 of \ref adjuState.
 *  If bit \ref NTP_ADJUSTAT_RNG_1S "5" of \ref adjuState is not set
 *  this value is old / irrelevant.
 */
   uint32_t secDif;

/** Actual ms deviation
 *
 *  This is the current fraction deviation converted to milliseconds (0..999).
 *  The sign is Bit 7 of \ref adjuState.
 */
   uint16_t msDif;

/** Previous ms deviation
 *
 *  This is the previous fraction deviation converted to milliseconds (0..999).
 *  The sign is Bit 3 of \ref adjuState.
 */
   uint16_t msDifPrev;
}; // struct ntpState_t

#define NTP_ADJUSTAT_MINUS   0x80 //!< last adjustment negative
#define NTP_ADJUSTAT_PLUS    0x40 //!< last adjustment positive
#define NTP_ADJUSTAT_DIRMSK  0xC0 //!< last adjustment direction mask

#define NTP_ADJUSTAT_RNGMSK  0x30 //!< last adjustment range mask
#define NTP_ADJUSTAT_RNG_LM  0x00 //!< last adjustment range low milliseconds
#define NTP_ADJUSTAT_RNG_HM  0x10 //!< last adjustment range milliseconds
#define NTP_ADJUSTAT_RNG_1S  0x20 //!< last adjustment range +1 second
#define NTP_ADJUSTAT_RNG_HS  0x30 //!< last adjustment range seconds

#define NTP_ADJUSTAT_PRV_MSK  0x0F //!< previous (not last) adjustment mask
#define NTP_ADJUSTAT_PRV_MINUS 0x08 //!< previous (not last)  adjustment negative
#define NTP_ADJUSTAT_PRV_PLUS  0x04 //!< lprevious (not last)  adjustment positive

extern struct ntpState_t ntpState; //!< the NTP state


/** Initialise the NTP (client)
 *
 *  This function initialises the NTP structures and (protothread) state
 *  machine. The protocol is not started by this function; the work is done
 *  by later calls of \ref ntpAppcall(). There is no need to call
 *  \ref ntpReset() before.
 *
 *  If "be NTP client" is not set in \ref curIpConf nothing will be done
 *  except setting the state to \ref NTP_STATE_RESET.
 *
 *  If the parameter \c pref2 is true (not 0) the second NTP server in the
 *  \ref curIpConf "current" IP configuration will be used, if set there.
 *  Successful bind to a set NTP server will set the (NTP client) state to
 *  \ref NTP_STATE_CONN1 respectively \ref NTP_STATE_CONN2.
 *
 *  \param pref2 if true use second NTP server (if available)
 */
void ntpInit(uint8_t pref2);

/** Reset the NTP (client) to initial state
 *
 *  This function initialises the NTP structures and (protothread) state
 *  machine.
 *  This function resets all NTPP state to initial. In contrast to
 *  \ref ntpInit no trial to connect or bind to NTP ports or servers.
 *  The protocol is not (not even indirectly) started.
 *
 *  \sa ntpInit()
 */
void ntpReset(void);

/** Handle NTP (server) events
 *
 *  This function handles NTP events. And it is a (as a protothread) the
 *  NTP client state machine. It has to be called in the \ref udp_appcall
 *  in a manner like \code
   const uint16_t remPort = convert16endian(uip_udp_conn->rport);

   if(remPort == NTP_PORT) {  // NTP protocol
      ntpAppcall();
      return;
   } // NTP protocol
 *  \endcode
 */
ptfnct_t ntpAppcall(void);


/** @} */ // Doxygen addtogroup ntpClient
/** @} */ // Doxygen addtogroup appLaySup

#endif /* NTP_H */
/** @} */ // Doxygen addtogroup weAutSys
