// Copyright (c) 2012 Albrecht Weinert, Bochum
// smc.h 2 2017-01-25 16:49:03Z albrecht

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup sysSer
 * @{
 * \addtogroup smcCom
 * @{
 *
 * \file smc.h
 *
 *  \ref intro_sec "weAutSys"' (low level) system calls, services and types
 *           for communication with a small memory card
 *
 *  This file contains the definitions for \ref intro_sec "weAutSys"' (low level) system
 *  calls, services and types to communicate via SPI with a
 *  \ref memCard_sec "small memory card".
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert

/** @} */ // Doxygen addtogroup smcCom
/** @} */ // Doxygen addtogroup sysSer

#ifndef SMC_H
#define SMC_H

#include "we-aut_sys/spi.h"
#include "we-aut_sys/ntp.h" // for othersAskPrio()
#include <avr/io.h>

/**
 * \addtogroup sysSer
 * @{
 * \addtogroup smcCom
 * @{
 */

/** Other devices asks for priority
 *
 *  This expression is true if other software respectively devices need
 *  resources used by SMC operations in critical situations.
 *
 *  As some SMC operations need SPI communication or thread time uninterrupted
 *  for quite a long time they should refrain from doing so as long as this
 *  request is pending.
 *
 *  For SMC processing controlled by this module's functions this time may be
 *  up to 2ms as documented in each case. File system implementations
 *  implemented upon these (driver) functions would increase this value quite
 *  considerably if programmed unaware of the (Protothread) threading model.
 *
 *  As of Revision 422++ this request comes from (seldom)
 *  \ref ntpAsksPrio() "NTP synchronisation" actions only.
 */
#define othersAskPrio() ntpAsksPrio()

/** Hardware card detect
 *
 *  On the  \ref intro_secH "weAut_01" module the card switch must be
 *  "jumpered" to port D6 for this function  to work properly.
 *
 *
 *  \note To see a card that was correctly powered up and put out of
 *        \ref getSMCtype "idle" state has vanished is easily done by every
 *        \ref sendCmd "command". So in this respect the card switch is
 *        hardly needed. On the other hand the detection of a potentially
 *        inserted SMC by pure (SPI) access requires the complete power
 *        up and \ref getSMCtype "type recognition" process.
 *
 *  Remark: It might well be considered a misconception in
 *  \ref intro_secH "weAut_01" not to tie the respective port (D5) directly to
 *  the card insert switch and have a lot of jumper options instead. This can
 *  be mended by an (1K) resistor soldered in 2 via holes near the jumper.
 *
 *  \return true as 'i': switch closed = card inserted, true as 'm': may be
 *               inserted, switch not configured; false (0): no card
 *               inserted switch open.
 */
uint8_t smcInsertSwitch(void);

/** Have 80 dummy SPI clocks
 *
 *  Some small memory cards need at least 74 (dummy) SPI  clocks with
 *  their chip select inactive (high) and data in (DI, MoSi) all ones after
 *  power up.
 *
 *  This function sends 80 clocks (sClk) on the SPI (2) interface used for
 *  small memory cards with all (SPI2) devices de-selected. The stream of
 *  80 clocks delivered by this function has no gaps.
 *
 *  At SPI clock frequency of 2 MHz the whole thing would take 40,4 s. At
 *  the \ref SPI400kHz "400 kHz" usually required for power up it will take
 *  200 s.
 *
 *  \sa smcSetIdle()
 */
void clk80(void);

/** Set the SPI clock frequency for the small memory card
 *
 *  Some (older) small memory cards require SPI clock frequencies of
 *  \ref SPI400kHz "400 kHz" (or lower) during \ref clk80 "initialisation".
 *  Afterwards higher SPI baud-rates are to be used. This function sets the
 *  latter / normal mode frequency.
 *
 *  A 20 MHz clocked micro-controller (like ATmega1284P) has a maximum
 *  \ref SPI10MHz "10 MHz" SPI baud rate. There are few SMC types with a
 *  maximum SPI clock frequency of 6.5 MHz around. Most other types can handle
 *  much higher clock rates. So the principally possible
 *  \ref SPI10MHz "10 MHz" would hardly give them a wet shirt.
 *
 *  The next possible lower frequency is \ref SPI5MHz "5 MHz". It would
 *  fit all card types, anyway. Additionally in case of resistor divide level
 *  shifters &mdash; from ATmega's 5V to SMC's 3.3V &mdash; the waveforms
 *  provided for some card types dictate \ref SPI5MHz "5 MHz" as maximum SPI
 *  clock frequency.
 *
 *  \param frqUse control value to set SPI 2 clock for the memory card
 *            (use the \ref SPI5MHz "predefined values" or the UBRR1 formula)
 *  \sa spi2EtherChipBaud
 *  \sa spi2MemCardBaud
 */
void smcSetFrq(uint8_t frqUse) __attribute__((always_inline));

/** Calculate CRC7 (one step)
 *
 *  This function calculates the CRC7 (used in some small memory cards for
 *  command transmission). It is a 7 bit CRC with polynomial x7 + x3 + 1.
 *  This function implements just the step for next 8 bits given by
 *  \c datByte. The result returned would have to be fed as \c crcIn for the
 *  next step. Use 0 for \c crcIn in the first step.
 *
 *  To make the final 7 bit CRC sendable to the small memory card via SPI
 *  it must be put in the upper seven bit of the CRC byte padded with an
 *  one in bit 0 by:
 *
 *  &nbsp; &nbsp;
 *    crc = (crc<<1) | 1;  // make so sendable after the last step
 *
 *  The function uses 10 CPU cycles if inlined as intended or 18 including
 *  call and return. Hence the algorithm is "faster than one byte" for all
 *  SPI clocks not faster than 1/2 CPU clock. Sending one byte at
 *  \ref SPI5MHz "5 MHz" takes 48 CPU clocks with a 20 MHz clocked ATmega.
 *
 *  \param crcIn the crc before; use 0 in first step
 *  \param datByte the next 8 bits for the CRC
 *  \return the new CRC
 */
uint8_t crc7stp(uint8_t crcIn, uint8_t datByte) __attribute__((always_inline));


/** De-select the small memory card
 *
 *  This function (chip-) de-selects the SMC. It has to be used when all
 *  \ref sendCmd "commands" and write / reads are done. An explicit select
 *  is not needed as all \ref sendCmd0arg "command" functions do so
 *  automatically.
 *
 *  Hint: Most operations in most SMCs (probably all) fail if de-selected
 *  and re-selected. Keep in mind that small memory cards are not very
 *  friendly to sharing a SPI bus with other devices.
 */
void deSelectSMC(void) __attribute__((always_inline));

/** Maximum extra waits for command response
 *
 *  The maximum number of (8 SPI clock) cycles to receive a valid command
 *  response is this value + 1. The specification says the number (NCR) of
 *  those waiting cycles is in the range of 1 .. 8 (depending on card type
 *  and command). Hence 7 would be the minimum for \ref NCR_EXTR_WAIT;
 *  anything larger than 10 would just waste time in case of a (command)
 *  failure.
 *
 *  \showinitializer
 */
#define NCR_EXTR_WAIT 8

/** Send a command to the small memory card
 *
 *  This function sends the command, the arguments and the \ref crc7stp "CRC7"
 *  to the memory card. Returned is the card's answer if any. Several
 *  (i.e. \ref NCR_EXTR_WAIT + 1) attempts are made to get the cards command
 *  response recognised by bit 7 zero. If the response returned is 0xFF the
 *  card may be gone.
 *
 *  In this (worst) case this function takes 74 s at a SPI clock frequency
 *  of 2 MHz compared to 32.5 s for a card answering fast.
 *
 *  Hint: To send an application specific command use the function
 *        \ref sendAppCmd().
 *
 *  Hint2: If the response is 0 (all OK) and if the command's specified
 *         answer is not of type R1 (or if it is a block read command), its up
 *         to the caller of \ref sendCmd "this" or
 *         \ref sendCmd0arg "similar" \ref sendAppCmd "functions" to
 *         receive all coming response bytes, using e.g. \ref smcReceive()
 *         or \ref smcReceiveN().
 *
 *  Hint3: The returned response is also stored in
 *         \ref smcState "smcState.lastCmdResp[0]".
 *
 * \param cmdNum   command number, only the lower 6 bits are relevant
 * \param cmdArg   the 4 byte (32 bits) command arguments
 * \return the first (and in case of R1 only) response byte
 * \sa GO_IDLE_STATE
 * \sa deselectSCM()
 * \sa sendCmd0arg
 * \sa sendAppCmd
 */
uint8_t sendCmd(uint8_t cmdNum, uint32_t cmdArg);

/** Send a command with 0 argument to the small memory card
 *
 *  This function is equivalent to <br /> &nbsp;  &nbsp;  &nbsp;
 *     \ref sendCmd "sendCmd(cmdNr, 0)";   <br />
 *  with some improvements for this frequent case.
 *
 *  <b> &nbsp; See also: &nbsp; </b> The \ref sendCmd "hints" at
 *      \ref sendCmd().
 *
 * \param cmdNum   command number, only the lower 6 bits are relevant
 * \return the first (and in case of R1 only) response byte
 */
uint8_t sendCmd0arg(uint8_t cmdNum);

/** Send an application specific command to the small memory card
 *
 *  This function is equivalent to <br /> &nbsp; &nbsp; &nbsp;
 *     \ref sendCmd0arg "sendCmd0arg(APP_CMD)";  <br /> &nbsp; &nbsp; &nbsp;
 *     \ref sendCmd "sendCmd(cmdNr, cmdArg)";    <br />
 *  besides some slight improvements.
 *
 *  <b> &nbsp; See also: &nbsp; </b> The \ref sendCmd "hints" at
 *      \ref sendCmd() and \ref SEND_OP_COND_APP.
 *
 * \param appCmdNum application specific command number (lower 6 bits)
 * \param cmdArg   the 4 byte (32 bits) command arguments
 * \return the first (and in case of R1 only) response byte
 */
uint8_t sendAppCmd(uint8_t appCmdNum, uint32_t cmdArg);

/** Put card to idle state
 *
 *  This function should run once a small memory card is inserted (or is
 *  believed to have been). Returned is the "R1" answer to the
 *  \ref GO_IDLE_STATE "soft reset" command. Only the value 1 is OK meaning
 *  card is present and (now) in idle state. All other values are faults.
 *  0xFF (no valid R1 response) usually means no card inserted or card not
 *  responding at all.
 *
 *  Most specifications require  \ref SPI400kHz "400 kHz" for the initial
 *  ref clk80 "dummy clocks" and the  \ref GO_IDLE_STATE "command". Though
 *  this seems out-dated for most modern cards it can for compatibility
 *  reasons be done by setting bit 0 in \c mode. The whole thing takes 370s
 *  then.
 *
 *  In \ref smcState "smcState.cardType" this function sets bits
 *  \ref CT_POWERD_UP and \ref CT_INSERTED on success; all other bits
 *   are cleared.
 *
 *  \param mode <br />
 *           Bit 0 set (1): force  \ref SPI400kHz "400 kHz" and restore
 *           \ref  spi2MemCardBaud to previous value afterwards <br />
 *           Bit 1 set (2): omit the 80 dummy clocks before the
 *            \ref GO_IDLE_STATE "go idle" command
 *           Bit 2 set (4): omit the \ref GO_IDLE_STATE "go idle" command
 *                          (0 is returned in that case)
 *  \return  answer (R1) for the \ref GO_IDLE_STATE "command" sent: 0x01 is
 *           OK (card in slot and in idle state)
 */
uint8_t smcSetIdle(uint8_t mode);

/**  Check if card is busy
 *
 *  This function returns 0 (false not busy) if the SMC is inserted and
 *  responds 0xFF at least twice to two dummy reads. The two consecutive
 *  not busy answers are looked for 3 .. (3 + \c tries) times.
 *
 *  If the outcome is not busy the card is not de-selected.
 *  \param tries 2..255 maximum number of dummy check reads
 *         (byte clock cycles) used
 *  \return 0: not busy; 0xFF: no card inserted and powered up; 1: busy
 */
uint8_t checkBusy(uint8_t tries);

#define GO_IDLE_STATE 0 //!< SMC command: soft reset \showinitializer
#define SEND_OP_COND 1  //!< SMC command: init. process \showinitializer
#define SEND_CID 10 //!< SMC command: get CID \showinitializer
#define SEND_STATUS  13 //!< SMC command: get status (R2) \showinitializer
#define SEND_IF_COND 8  //!< SMC command: send interface condition \showinitializer
#define SET_BLOCKLEN 16 //!< SMC command: set block length for  \ref LOCK_UNLOCK \showinitializer
#define LOCK_UNLOCK 42  //!< SMC command: lock / unlock the card \sa SET_BLOCKLEN \showinitializer

#define SEND_CSD 9  //!< SMC command: send card specific data \showinitializer
#define SEND_CID 10 //!< SMC command: send card idendification \showinitializer
#define READ_SINGLE_BLOCK 17 //!< SMC command: read one block \showinitializer
#define WRITE_BLOCK 24 //!< SMC command: write one block \showinitializer
#define APP_CMD 55 //!< SMC command: next command is application specific \showinitializer
#define READ_OCR 58 //!< SMC command: read the card's OCR register (R3) \showinitializer

/** SMC application specific command: initialisation process
 *
 *  Like \ref SEND_OP_COND this command activates the card's initialisation
 *  process. But as an application specific command it has to be
 *  prepended by \ref APP_CMD() or simply used in \ref sendAppCmd instead of
 *  \ref sendCmd().
 *
 *  Note: In SPI mode \ref SEND_OP_COND_APP and \ref SEND_OP_COND should have
 *  the same behaviour, though most implementors prefer the application
 *  specific variant. This tradition is respected \ref getSMCtype "here".
 */
#define SEND_OP_COND_APP 41

//---------------------- end driver functions / start thread and state ------

/** Transmit a single byte to the small memory card
 *
 *  \param datByte the byte to be sent to the SMC
 *  \return the byte received (on MiSo); can be disregarded for SMCs (0 may
 *          indicate an \ref spi2Errors "error")
 */
#define smcXmit(datByte) spi2Tranceive(datByte)

/** Send two bytes and receive one byte to/from the small memory card
 *
 *  \param sendB1 the byte to be sent first
 *  \param sendB2 the second byte to be sent
 *  \return the byte read from SPI while transmitting; can be disregarded
 *          for SMCs  \c sendB2 (0 may indicate an \ref spi2Errors "error")
 */
#define smcXmit2(sendB1, sendB2) spi2Tranceive2(sendB1, sendB2)


/** Receive a single byte from the small memory card
 *
 *  \return the byte received
 *  \sa smcReceiveN
 */
#define smcReceive() spi2Tranceive(0xFF)

/** Receive n bytes from the SMC to a buffer with optional skip
 *
 *  This function does \c skip + \c n receptions. The (last) \c n bytes
 *  received are put into the buffer \c receiveB. It returns nothing (void).
 *
 *  Compared to doing this by \c skip + \c n times \ref smcReceive() this
 *  function is at least 20 % faster.
 *
 *  \param receiveB pointer to the buffer (type uint8_t *) to receive \c n
 *                  bytes to (must not be NULL if \c n != 0)
 *  \param skip     if > 0 \c skip bytes are received and forgotten
 *                  before \c receiveB will be filled
 *  \param n        number of bytes to be received and filled into \c receiveB
 */
#define smcReceiveN(receiveB, skip, n) void spi2ReceiveSN(receiveB, skip, n)


struct smcThr_data_t; // the SMC (thread) structure type ------------

/** The small memory card (smc) handling thread
 *
 *  If some basic handling, like initialising, has to be done with a SMC
 *  this protothread thread will do it reasonable steps, dividing long
 *  running procedures by yielding.
 *
 *  Hint: As this is for small (petit) systems with just one SMC slot this
 *        thread uses the one \ref smcState (without any choice by parameter).
 */
ptfnct_t smcThreadF(void);

/** Initialise the  small memory card (smc) handling thread
 *
 *  This function resets all SMC state. It should be called on reset /
 *  restart and on SMC removal, failures and insertion.
 *
 * \param actionFlag start value for the SMC thread's flag
 * \sa smcState
 */
void initSMCthreadState(uint8_t actionFlag);

/** The (one) small memory card (SMC) state */
extern struct smcThr_data_t smcState;


/** The organisational data for a small memory card (SMC) handling thread
 *
 *  A structure of this type hold the state of one small memory card (SMC)
 *  as well as of a \ref smcThreadF "handling thread".
 *
 *  \sa
 *  \sa appCliThreadF
 *  \sa appCliThread
 */
struct smcThr_data_t {

/** The (raw) protothread data structure for the SMC handling thread */
   pt_t pt;


/** The action flag
 *
 *  0 means nothing to do.
 *
 *  Bit 4 : power up (80 dummy clocks), \ref CT_POWERD_UP <br />
 *  Bit 5 : set idle \ref CT_INSERTED <br />
 *  Bit 6 : determine card type \ref FL_CAT <br />
 *  Bit 7 : if no card inserted and powered up, check for it.
 *
 *  Bit 1 : read sector \ref FL_RDS <br />
 *  Bit 2 : write sector \ref FL_WRS <br />

 *
 *  Bits 0..6 will be cleared by the respective action performed.
 *
 *  Bit 7 will not be cleared by the "check for card" action. If a card is
 *  (newly) detected bits 4 to 6 will be set
 */
 uint8_t flag;

#define FL_CAT 0x40 //!< Action/state flag: card type \showinitializer
#define FL_INI 0xF0 //!< Action mask: initialisation actions \showinitializer
#define FL_IMP 0x76 //!< Action mask: implemented actions (except Bit 7) \showinitializer

#define FL_RDS 0x02 //!< Action/state flag: sector read \showinitializer
#define FL_WRS 0x04 //!< Action/state flag: sector write \showinitializer
#define FL_R_W 0x06 //!< Action mask: sector access (read, write) \showinitializer

/** The return code of the last schedule
 *
 *  This is intended to hold the return code,  i.e. the SMC thread function's
 *  last  return value.
 *
 *  \internal reserve and structure compatibility only at the moment.
 */
  uint8_t retC;

/** The SMC thread function (pointer to)
 *
 *  The function pointer will be initialised pointing to \ref smcThreadF, the
 *  system provided SMC handling thread. It is provided for future use to
 *  put an application specific handler in.
 *
 *  \sa smcThreadF
 *  \sa initSMCthread
 */
  p2ptFunV theSMCfun;

//-- End common structure ---------------------


/** The buffer state
 *
 *  Bit 1 : sector in sync with SMC by just so read or written \ref FL_RDS <br />
 *  Bit 2 : sector to be written after being modified \ref FL_WRS <br />
 *
 *  If neither \ref FL_RDS "bit 1" nor \ref FL_WRS "bit 2" is set the buffer
 *  \c .sectorBuff has no relation to the SMC's sector \c rwSector. If either
 *  is set \c .sectorBuff is the actual or intended content implying a
 *  \ref readDataBlock "read operation" on  \c rwSector redundant.
 */
  uint8_t bufferState;

/** Last data response
 *
 *  This byte is set by some (block) write / read functions as received
 *  from the device. The bit pattern is xxx0rrr1.
 *
 *  rrr = 010 : data accepted <br />
 *  rrr = 101 : data rejected due to a CRC error <br />
 *  rrr = 110 : data rejected due to a write error <br />
 */
  uint8_t lastDatResp;

/** Last command response - first or single (R1) byte
  *
  * \sa lastCmdResp
  */
#define SMC_LAST_R1 smcState.lastCmdResp[0]

/** Last command response
 *
 *  In this field the functions \ref sendCmd(), \ref sendCmd0arg() and alike
 *  store their last return value for later reference. A R1 response would be
 *  in the first byte (\ref SMC_LAST_R1 or lastCmdResp[0]).
 */
  uint8_t lastCmdResp[5];

/** The cards OCR
 *
 *  This is the informational part of the R3 response.
 *  It is valid if (one bit of) \ref CT_TYPE_MASK bit is set in \c .cardType
 */
  uint8_t ocr[4];

/** The cards CSD
 *
 *  This is the informational part of the R3 response.
 *  It is valid if (one bit of) \ref CT_TYPE_MASK bit is set in \c .cardType
 */
 uint8_t csd[16];


/** The recognised small memory card (SMC)
 *
 *  The value 0 means no usable card recognised. The lower 4 bits contain
 *  the \ref  CT_TYPE_MASK "card type" information; the upper 4 bits show
 *  recognised \ref CT_OPCN_MASK "operational" conditions
 *
 *  Bit 0 (\ref CT_V_3) : MMC version 3 <br />
 *  Bit 1 (\ref CT_V_1) : SD version 1 <br />
 *  Bit 2 (\ref CT_V_2) : SD version 2 <br />
 *  Bit 3 (\ref CT_HC ) : high capacity (\ref CT_V_2 "V.2")
 *         respectively block addressing mode <br />
 *  <br />
 *  Bit 4 (\ref CT_POWERD_UP) : card has been successfully powered up (to idle state) <br />
 *  Bit 5 (\ref CT_INSERTED) : a card is (seems to be) inserted <br />
 *
 *
 *  Bit 4 (\ref CT_POWERD_UP) is set by getting a card successfully to
 *  \ref smcSetIdle "idle" state. Bit 5 (\ref CT_INSERTED) is set by any
 *  valid command response (or card switch). It is reset by no valid card
 *  response (i.e. MoSi always tri-stated).
 *
 *  Hint1: Bits 0..4 are compatible to petit FATFS for the purpose of
 *        using its file system implementation.
 *
 *  Hint2: For high capacity a 32-bit argument of memory access commands means
 *         the block address and the block length is fixed to 512 bytes. (For
 *         standard capacity cards this would be a byte address and the
 *         block length is determined by CMD16,
 */
  uint8_t cardType;

#define CT_TYPE_MASK 0x0F //!< Card type mask: type bits \showinitializer
#define CT_HC  0x08 //!< Card type flag: high capacity \showinitializer
#define CT_V_3 0x01 //!< Card type: MMC version 3 \showinitializer
#define CT_V_2 0x04 //!< Card type: SD version 2 \showinitializer
#define CT_V_1 0x02 //!< Card type: SD version 2 \showinitializer

#define CT_OPCN_MASK 0xF0 //!< Card type mask: operation bits \showinitializer
#define CT_POWERD_UP 0x10 //!< Card type flag: was powered up \showinitializer
#define CT_INSERTED 0x20 //!< Card type flag: inserted \showinitializer
#define CT_INS_POWD 0x30 //!< Card type mask: power & inserted \showinitializer
//  #define CT_OCR_RD   0x40 //!< Card type flag: the OCR has been read to \c .ocr \showinitializer


/** Buffer address
 *
 *  This is the address in the form of a 512 byte sector number
 *  (independent of card type) of the buffer \c .sectorBuff.
 */
  uint32_t rwSector;


#define SMC_SEC_BUF_SZ 516 //!< Sector buffer size \showinitializer
#define SMC_DAT_TOK 0xFE //!< data token \showinitializer

/** The sector buffer
 *
 *  This is a buffer for a 512 byte sector. The buffer's size is larger
 *  (\ref SMC_SEC_BUF_SZ bytes) to handle some extra bytes like CRC. If the
 *  SMC is not used the space can serve other purposes.
 *
 *  If the bit \ref FL_RDS is set in \c bufferState the content is the
 *  (cached) SMC's content at \c readSector. If the bit \ref FL_WRS is set in
 *  \c bufferState the content is the (cached) SMC's content at \c writeSector.
 *
 *  Any software that changes  \c writeSector, \c readSector or the first 512
 *  bytes of \c sectorBuff must clear the respective flag bits in
 *  \c bufferState.
 *
 * <b> &nbsp; See also: &nbsp; </b> \ref FL_CAT \ref FL_RDS  \ref FL_WRS
 * \internal keep at the end
 */
  uint8_t sectorBuff[SMC_SEC_BUF_SZ];

}; // smcThr_data_t

/** Inserted card is powered up
 *
 *  This expression is true if a SMC is inserted and powered up according to
 *  \ref smcState.
 */
#define smcInsPow() ((smcState.cardType & CT_INS_POWD) == CT_INS_POWD)

/** The SMC's type is determined and OCR is known
 *
 *  This expression is true if a (inserted and powered up) SMC's type has been
 *  determined. In the process also the OCR has been read to
 *  \ref smcState "smcState.ocr".
 */
#define smcTypeD() ((smcState.cardType & CT_TYPE_MASK) && smcInsPow())

/** Read the SMC's ORC
 *
 *  This function reads the operation condition register (OCR, 4 byte) of an
 *  inserted and powered up small memory card. In case of success 0 is
 *  returned and the result is stored in \ref smcState "smcState.ocr".
 *
 *  It is usually not necessary to call this function if \ref smcTypeD() is
 *  true as the OCR and the CSD have been read during the card type
 *  determination process.
 *
 *  \return 0: OK  or negative response of command \ref READ_OCR or
 *                or 0x80 if SCM isn't powered up.
 */
uint8_t smcReadOCR(void);

/** Read the SMC's CSD
 *
 *  This function reads the card specific data register (CSD, 16 byte) of an
 *  inserted and powered up small memory card. In case of success 0 is
 *  returned and the result is stored in \ref smcState "smcState.csd".
 *
 *  It is usually not necessary to call this function if \ref smcTypeD() is
 *  true as the OCR and CSD have been read during the card type
 *  determination process.
 *
 *  \return 0: OK  or negative response of command \ref SEND_CSD or
 *                or 0x80 if SCM isn't powered up.
 */
uint8_t smcReadCSD(void);

/** Get the sector count
 *
 *  If the card type is \ref smcTypeD() "determined" this function returns the
 *  number of (512 byte) sectors on the card and 0 otherwise.
 */
uint32_t smcGetSectCount(void);

/** Read the SMC's CID
 *
 *  This function reads the card identification register (CID, 16 byte) of an
 *  inserted and powered up small memory card. In case of success 0 is
 *  returned and the result is stored in \c buff.
 *
 *  0: Manufacturer ID  <br />
 *  1,2: OEM/Application ID <br />
 *  3..7: Product name <br />
 *  8: Product revision <br />
 *  9..12: Serial number <br />
 *  13 15: Manufaturing date <br />
 *
 *  \param buff 16 byte to write the CID to
 *  \return 0: OK  or negative response of command \ref SEND_CSD or
 *                or 0x80 if SCM isn't powered up.
 */
uint8_t smcReadCID(uint8_t * buff);


/** Read single data block (512 byte)
 *
 *  This function does one single block respectively sector read &mdash;
 *  with block size being fixed to 512 byte here. In case of success 0 is
 *  returned. If the given \c sector number has already been cached the SMC
 *  is not accessed.
 *
 *  In case of failure non 0 is returned. More information can be found in
 *  \ref smcState "smcState".
 *
 *  In the normal case of SCC access the time needed by this function heavily
 *  depends on the SMC make. The (astonishing) differences are mainly due
 *  to different waits for the so called \ref SMC_DAT_TOK "data token". The
 *  time needed for this With ~31 waits for data token (512MByte KData card)
 *  this takes about 1 ms. With 181 checks for the
 *  \ref SMC_DAT_TOK "data token" (8GByte Verbatim) we need 1,35 ms but this
 *  varies up to 250 waits and about 2 ms.
 *
 *  \ref intro_sec "weAutSys" is a non pre-emptive system &mdash; and
 *  besides that SMC operations (the /CS) could not be interrupted anyway.
 *  Hence it is clear and was said at other places this block read operation
 *  is the most critical for an automation modul / system with an 1ms
 *  \ref appThr_sec "cycle". No two such operations shall follow each
 *  other without \ref PT_YIELD "yielding" &mdash; and multi-block SMC
 *  operations are, of course, out of the question.
 *
 *  Hint: This function as well as \ref writeDataBlock() handle the sector to
 *  byte address transformation for commands \ref READ_SINGLE_BLOCK resp.
 *  \ref WRITE_BLOCK needed with low capacity SMC types internally. Parameters
 *  \c sector (and \ref smcState "smcState.rwSector") always reflect the
 *  sector number (0, 1 ...).
 *
 * \param  sector the sector number
 * \param  buff   pointer to where to write the received data
 *                (NULL not allowed)
 * \return 0: OK  or negative response of command \ref READ_SINGLE_BLOCK or
 *                or 0x80 if no data token came (time out; may try later) or
 *                no no type info available
 * \sa writeDataBlock
 */
uint8_t readDataBlock(uint32_t sector, uint8_t * buff);

/**  Write single data block (512 byte)
 *
 * \param  sector the sector number
 * \param  buff   pointer to where to get the send data from (NULL not allowed)
 * \return 0: OK  or negative response of command \ref WRITE_BLOCK or
 *                or 0x80 if no type info available
 * \sa readDataBlock
 */
uint8_t writeDataBlock(uint32_t sector, const uint8_t * buff);

/** Determine the card type
 *
 *  If the an SMC is inserted and its type has been \ref smcTypeD "determined"
 *  this function returns the \ref CT_TYPE_MASK "type bits" set in
 *  \ref smcState "smcState.cardType".
 *
 *  Otherwise 0 is returned and the card type determination is triggered in
 *  the SMC handling thread.
 *
 *  \return 0: no type determined yet,  \ref CT_V_1,
 *             \ref CT_V_2 [| \ref CT_HC], \ref CT_V_3,
 */
uint8_t getSMCtype(void);

/** Order sector read (as background task)
 *
 *  This function sets \ref smcState "smcState.rwSector" by \c sector and
 *  initialises \ref smcState "smcState.sectorBuff" to be read from that
 *  sector.
 *
 * \param  sector the (new) sector number
 *                for \ref smcState "smcState.rwSector/.sectorBuff"
 * \return
 *   0: OK already done (\ref smcState "smcState.sectorBuff" in sync.), <br />
 *   1: will be done ASAP (in background task), call again later, <br />
 *   0x80: action not possible (no card, no type or other error)
 * \sa setSectorModified
 */
uint8_t doSectorRead(uint32_t sector);

/** Mark sector buffer data as modified (start)
 *
 *  This function sets \ref smcState "smcState.rwSector" by \c sector and
 *  marks as \ref smcState "smcState.sectorBuff" as modified (new) data for
 *  the respective sector. This neither performs any read or write operations
 *  nor orders such as background task.
 *
 *  Hint: If this functions changes the current buffer's
 *  (\ref smcState "smcState.sectorBuff") sector number and the buffer
 *  contained modified data for that previous sector those modifications
 *  will be lost. Call \ref doSectorSync() before if that would not be
 *  intended.
 *
 * \param  sector the (new) sector number
 *                for \ref smcState "smcState.rwSector/.sectorBuff"
 * \sa doSectorRead
 */
void setSectorModified(uint32_t sector);

/** Order sector synchronisation (as background task)
 *
 *  This function checks if \ref smcState "smcState.sectorBuff" contains
 *  modified data (for sector \ref smcState "smcState.rwSector") and if so
 *  initialises \ref smcState "smcState.sectorBuff" to be written to that
 *  sector.
 *
 *  Example sector copy: \code
 *  PT_WAIT_ASYIELD_WHILE(&myThread, doSectorRead(sourceSect) == 1);
 *  if (doSectorSync()) {  optional error handling / exit  }
 *  setSectorModified(destinationSect);
 *  PT_WAIT_ASYIELD_WHILE(&myThread, doSectorSync() == 1);
 *  if (doSectorSync()) {  optional error handling / exit  }
 *  \endcode
 *
 * \return
 *   0: OK already done (\ref smcState "smcState.sectorBuff" in sync.), <br />
 *   1: will be done ASAP (in background task), call again later, <br />
 *   0x80: action not possible (no card, no type or other error)
 */
uint8_t doSectorSync(void);

/** @} */ // Doxygen addtogroup smcCom
/** @} */ // Doxygen addtogroup sysSer


#endif /* SMC_H */
/** @} */ // Doxygen addtogroup weAutSys
