// Copyright (c) 2011 Albrecht Weinert, Bochum
// spi.h 2 2017-01-25 16:49:03Z albrecht


/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup llSySer
 * @{
 * \addtogroup embDriv
 * @{
 *
 * \file spi.h
 *
 * \ref intro_sec "weAutSys"'/weAut_01' system calls and services for the SPI interfaces
 *
 *  This file contains the definitions for \ref intro_sec "weAutSys" /
 *  weAut_01  (low level) system calls and services for the serial peripheral
 *  interfaces (SPI).
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert
/** @} */ // Doxygen addtogroup   embDriv

#ifndef SPI_H
#define SPI_H

#include <avr/io.h>
#include "we-aut_sys/ll_common.h"
// #include "we-aut_sys/common.h"

/**
 * \addtogroup embDriv
 * @{
 */


/** De-select all SPI1 devices
 *
 *  This (low level helper) function de-activates the chip selects
 *  (sometimes aka slave selects) of all SPI 1 devices. For configurations
 *  without SPI1 devices nothing is done.
 *
 *  By polling SPI1's status it is checked before de-selecting if the
 *  interface is still busy. If so, it will be waited until ready. With
 *  higher transfer rates (CPU clock /2) that won't be long.
 *
 *  Returned is the first reading of the SPI1 status.
 *
 *  \note If the SPI is set for the highest transfer rate this function
 *  should never wait (long) for ready if a tiny bit of useful work is
 *  done between start send/receive and this call. At high transfer rates
 *  this approach is far cheaper &mdash; and quicker &mdash; than utilising
 *  the SPI ready interrupt.
 *
 *  \return the first status reading; can be checked for "was waiting"
 *           and for past collisions
 *  \sa  deSelectSPI1_impl
 */
uint8_t deSelectSPI1(void)  __attribute__((always_inline));

/** De-select all SPI1 devices (raw)
 *
 *  This function does so without any checks or waits.
 *  \sa   deSelectSPI1()
 */
void deSelectSPI1_impl(void)  __attribute__((always_inline));

/** Chip-select a SPI 1 device: the digital input display LEDS (HMI)
 *
 *  This (low level helper) function activates the chip select of a SPI 1
 *  device and de-selects all others at the same SPI bus. For configurations
 *  without that device nothing is done.
 *
 *  This select / de-select is done without any checks or waits.
 *  \sa   deSelectSPI1()
 */
void selectDIdisplayLEDs(void)  __attribute__((always_inline));


/** Chip-select two SPI 1 devices: the DO driver and the DI LEDs (HMI)
 *
 *  After this "double select" the same value can be output to both
 *  devices in one SPI (shift) operation.
 *
 *  So far this is only used for initialising both with 0 by system
 *  (start) software.
 *
 *  \sa   selectDIdisplayLEDs()
 */
void  selectDOdrivDIleds(void)  __attribute__((always_inline));





/** Chip-select a SPI 1 device: the digital output driver (DO)
 *
 *  \sa   selectDIdisplayLEDs()
 */
void  selectDOdriver(void)  __attribute__((always_inline));


//---------------------------------------------------------------


/** De-select all SPI2 devices
 *
 *  This (low level helper) function de-activates the chip selects
 *  (sometimes aka slave selects) of all SPI 2 devices. For configurations
 *  without such devices nothing is done.
 *
 *  By polling ISP2's status, it is checked before de-selecting, if the
 *  interface's transmitter is still busy. If so, it will be waited until
 *  ready, i.e. for transmission complete (TXC). If this waiting
 *  \ref SPI2COND_ERR_RET "times out" the appropriate
 *  \ref spi2Errors "error bits" are set.
 *
 *  \note If the SPI is set for a high transfer rate this function should
 *  never wait long for ready if a tiny bit of useful work is done between
 *  the start of send/receive and this call. At high transfer rates this
 *  "do something and than wait a bit" approach is far cheaper &mdash; and
 *  quicker &mdash; than utilising a SPI ready interrupt.
 *
 *  \sa   deSelectSPI2_impl()
 */
void deSelectSPI2(void)  __attribute__((always_inline));


/** Chip-select a SPI 2 device: the Ethernet driver chip
 *
 *  This (low level helper) function activates the chip select of a SPI 2
 *  device and de-selects all others at the same SPI bus. For configurations
 *  without that device nothing is done.
 *
 *  This select / de-select is done without any checks or waits.
 *
 *  \sa  deSelectSPI2()
 */
void selectEtherChip(void)  __attribute__((always_inline));

/** The Ethernet driver is chip-selected
 *
 *  This evaluates to true if the ENB28J60's / NIC's chip select is active.
 *  \sa selectEtherChip
 *  \hideinitializer
 */
#if defined(weAut_00)
// Port C  C6 /CS memory card  Port D  D5 Ethernet CS
#    define nicIsSelected() (!(PORTD & (1<<5)))
#elif defined(weAut_01)
// Port C  C6 /CS memory card  C7 Ethernet CS
#    define nicIsSelected() (!(PORTC & (1<<7)))
#else
// this should never be used on platforms without NIC
#    define nicIsSelected() (0)
#endif


#ifndef __DOXYGEN__
#  undef nicIsSelected
#  if defined(weAut_00)
// Port C  C6 /CS memory card  Port D  D5 Ethernet CS
#    define nicIsSelected() (!(PORTD & (1<<5)))
#  endif
#  if defined(weAut_01)
// Port C  C6 /CS memory card  C7 Ethernet CS
#    define nicIsSelected() (!(PORTC & (1<<7)))
#  endif
#endif

#if defined(weAut_00) && defined(weAut_01)
# error "both weAut_00 and weAut_01 defined"
#endif
/** Chip-select a SPI 2 device: the memory card (holder)
 *
 *  This (low level helper) function activates the chip select of a SPI 2
 *  device and de-selects all others at the same SPI bus. For configurations
 *  without that device nothing is done.
 *
 *  This select / de-select is done without any checks or waits.
 *
 *   \sa deSelectSPI2()
 */
void selectMemCard(void)  __attribute__((always_inline));

/** The memory card (inserted) is chip-selected
 *
 *  This evaluates to true if the memory card chip select is active.
 *  \sa selectMemCard
 */
#define memCardIsSelected() (!(PORTC & (1<<6)))
// weAut_00 :  Port C  C6 /CS memory card  Port D  D5 Ethernet CS
// weAut_01 :  Port C  C6 /CS memory card  C7 Ethernet CS


/** Divisor register for UART as SPI (UBRR) setting for 10 MHz
 *
 *  This value gives the highest possible (UART as) SPI baud rate, that is
 *  CPU clock / 2. That would be 10 MHz on a 20 MHz clocked Controller, as
 *  for example an ATmega1284 at max. speed. An ATmega 2560 would yield max.
 *  8 MHz with this setting.
 *
 *  In the same sense \ref SPI5MHz gives corresponding results of 5 vs. 4 MHz.
 *  \ref SPI2MHz, \ref SPI2MHz, \ref SPI400kHz, \ref SPI200kHz and
 *  \ref SPI100kHz on the other hand give their nominal value at both 20 MHz
 *  and 16 MHz CPU clock; that is on ref intro_secH "weAut_01",
 *  \ref intro_secH "ArduinoMega2560", \ref intro_secH  "ArduinoUno" and
 *  others.
 *  \showinitializer
 */
#define SPI10MHz 0

/** Divisor register for UART as SPI (UBRR)) setting for 5 MHz
 *
 *  This value gives the second highest possible (UART as) SPI baud rate, that
 *  is CPU clock / 4. That would be 5 MHz on a 20 MHz clocked Controller and
 *  4 MHz with a 16 MHz clock
 *  (ATmega1284 at max. speed e.g.). (An ATmega 2560 would yield max. 8 MHz
 *  with this setting.
 *  \showinitializer
 *  \sa SPI10MHz
 */
#define SPI5MHz 1

#if F_CPU == 20000000
#define SPI2MHz 4 //!< UBRR value to set SPI  clock to 2 MHz  \sa SPI10MHz
#define SPI1MHz 9 //!< control value to set SPI 2 clock to 1 MHz  \sa SPI10MHz
#define SPI400kHz 24 //!< control value to set SPI 2 clock to 400 kHz \sa SPI10MHz
#define SPI200kHz 49 //!< control value to set SPI 2 clock to 200 kHz \sa SPI10MHz
#define SPI100kHz 99 //!< control value to set SPI 2 clock to 100 kHz \sa SPI10MHz
#else
#define SPI2MHz 3 //!< UBRR value to set SPI  clock to 2 MHz  \sa SPI10MHz
#define SPI1MHz 7 //!< control value to set SPI 2 clock to 1 MHz  \sa SPI10MHz
#define SPI400kHz 19 //!< control value to set SPI 2 clock to 400 kHz \sa SPI10MHz
#define SPI200kHz 39 //!< control value to set SPI 2 clock to 200 kHz \sa SPI10MHz
#define SPI100kHz 79 //!< control value to set SPI 2 clock to 100 kHz \sa SPI10MHz
#endif



/** The SPI (2) clock frequency used for the Ethernet driver chip
 *
 *  This is the control value used to set the SPI (2) clock frequency (aka
 *  baud rate) if used with the \ref selectEtherChip "Ethernet driver".
 *
 *  Hint 1: For 20 MHz \ref FCPU_S STR "CPU frequency" the SPI (2) frequency
 *          can be set in the range 10 MHz (value 0) down to 40 kHz
 *          (value 249). <br />
 *  Hint 2: Due to different and un-synchronous clocks for the CPU (20 MHz)
 *          and the Ethernet chip (ENC28J60 25 MHz) not all baudrates go
 *          well. \ref SPI2MHz "2 MHz" proofed rock solid reliable and is
 *          the default even if 5 or 10 MHz would give more performance.
 *
 *  default / init: \ref  SPI2MHz
 *  \sa SPI5MHz
 */
extern uint8_t spi2EtherChipBaud;


/** The SPI (2) clock frequency used for the memory card (holder)
 *
 *  This is the control value used to set the SPI (2) clock frequency (aka
 *  baud rate) if used with the small \ref selectMemCard "memory card".
 *
 *  Hint 1: For 20 MHz \ref FCPU_S STR "CPU frequency" the SPI (2) frequency
 *          can be set in the range 10 MHz (value 0) down to 40 kHz
 *          (value 249). \ref SPI5MHz "5 MHz" is the default or reset
 *          initialisation value. <br />
 *  Hint 2: Most small memory cards require \ref SPI400kHz "400 kHz" or
 *          slower during initialisation (power up and idle state). This
 *          can be done by \ref smcSetIdle "smcSetIdle(1)" no matter the
 *          \ref  spi2MemCardBaud value
 *          After initialisation and depending on the card type
 *          information read a higher frequency can and should be set (here).
 *
 *  \sa SPI50MHz
 *  \sa SPI400kHz
 */
extern uint8_t spi2MemCardBaud;


//--------------------------------------------------------------------------

/** Send one byte over SPI 2
 *
 *  This function implies that the SPI device has been selected correctly
 *  before. It waits for empty transmit buffer, writes the byte
 *  \c sendB and clears the transmit complete flag.
 *
 *  If waiting for empty transmit buffer takes too long this function will set
 *  the appropriate \ref spi2Errors "error bits" and does nothing else.
 *
 *  \param sendB the byte to be sent
 *  \sa spi2Receive
 */
void spi2Transmit(uint8_t sendB) __attribute__((always_inline));

/** Receive one byte from SPI 2
 *
 *  This function implies that at least one byte has been sent over the SPI
 *  to trigger or clock the receive shift and the respective byte has not yet
 *  been consumed. The function waits for the receive complete flag and than
 *  return the byte read by SPI as result.
 *
 *  \note If said precondition is not met the wait will
 *  \ref SPI2COND_ERR_RET "time out" and this function will set the
 *  appropriate \ref spi2Errors "error bits" and return 0.
 *
 *  \sa spi2Transmit
 *  \sa spi2Tranceive
 *  \sa spi2Tranceive2
 *  \return the byte read from SPI (0 may but must not indicate an
 *          \ref spi2Errors "error"
 */
uint8_t spi2Receive(void) __attribute__((always_inline));

/** Send and receive one byte via SPI 2
 *
 *  This function implies that the SPI device has been is selected correctly
 *  before.
 *
 *  It effectively combines \ref spi2Transmit "spi2Transmit(sendB)" and
 *  \ref spi2Receive "return spi2Receive()" and guarantees the correct pairing
 *  of SPI transmit and receive.
 *
 *  On the other hand it does neither allow useful work instead of the second
 *  (TXC) wait nor utilise the SPI2 (USAT2 as SPI) double buffering.
 *
 *  \param sendB the byte to be sent
 *  \return the byte read from SPI (0 may but must not indicate an
 *          \ref spi2Errors "error")
 *  \sa spi2Transmit
 *  \sa spi2Receive
 *  \sa spi2Tranceive2
 */
uint8_t spi2Tranceive(uint8_t sendB) __attribute__((always_inline));

/** Send two bytes and receive one byte via SPI 2
 *
 *  This function acts as if calling \ref spi2Tranceive twice (returning the
 *  second call's result). Compared to that proceeding this function is 20%
 *  faster by utilising the SPI2's (USAT2 as SPI) double buffering. The device
 *  sees a continuous stream of 16 clocks instead of 2 times 8 clocks with
 *  a gap.
 *
 *  \param sendB1 the byte to be sent first
 *  \param sendB2 the second byte to be sent
 *  \return the byte read from SPI while transmitting \c sendB2 (0 may but
 *          must not indicate an \ref spi2Errors "error")
 *  \sa spi2Transmit
 *  \sa spi2Receive
 *  \sa spi2Tranceive
 *  \sa spi2TranceiveN
 */
uint8_t spi2Tranceive2(uint8_t sendB1, uint8_t sendB2);


/** Send a byte n times and receive one byte via SPI 2
 *
 *  This function sends the byte \c sendB one or more times via SPI 2
 *  discarding all receptions except for the last one.
 *  Compared to doing this by \ref spi2Tranceive "spi2Tranceive(sendB)" \c n
 *  times (in a loop) this function is at least 20 % faster cause of no gaps
 *  in the byte (MoSi) respectively clock (sClk) stream.
 *
 *  On the other hand no useful work or \ref PT_YIELD yielding can be done
 *  while the SPI (2) interface is busy with this action. Insofar \c n should
 *  be reasonably small.
 *
 *  \param sendB the byte to be sent \c n times
 *  \param n number to repeat \c sendB (0 is taken as 1)
 *  \return the byte read from SPI while transmitting \c sendB2 (0 may but
 *          must not indicate an \ref spi2Errors "error")
 *  \sa spi2Transmit
 *  \sa spi2Receive
 *  \sa spi2Tranceive
 *  \sa spi2Tranceive2
 */
uint8_t spi2TranceiveN(uint8_t sendB, uint8_t n);

/**  Receive n bytes to a buffer via SPI 2 with optional skip (before)
 *
 *  This function does \c skip + \c n receptions via SPI2 (putting all ones
 *  on MoSi). The (last) \c n bytes received are put into the buffer
 *  \c receiveB.
 *
 *  Compared to doing this by \ref spi2Tranceive "spi2Tranceive(0xFF)"
 *  \c skip + \c n times this function is at least 20 % faster. On the other
 *  hand no useful work or yielding can be done while the SPI (2) interface is
 *  busy with the action. Insofar \c skip + \c n should be reasonably small.
 *
 *  \param receiveB pointer to the buffer to receive \c n bytes to
 *                  (must not be NULL if \c n != 0)
 *  \param skip     if > 0 \c skip bytes are received and forgotten
 *                  before \c receiveB will be filled
 *  \param n        number of bytes to be received and filled into \c receiveB
 */
void spi2ReceiveSN(uint8_t * receiveB, uint16_t skip, uint16_t n);

/** Receive n bytes to a buffer via SPI 2 with optional skip (afterwards)
 *
 *  This function does \c skip + \c n receptions via SPI2 (putting all ones
 *  on MoSi). The (first) \c n bytes received are put into the buffer
 *  \c receiveB.
 *
 *  Compared to doing this by \ref spi2Tranceive "spi2Tranceive(0xFF)"
 *  \c n  + \c skip times this function is at least 20 % faster. On the other
 *  hand no useful work or yielding can be done while the SPI (2) interface is
 *  busy with the action. Insofar \c n + \c skip should be reasonably small.
 *
 *  \param receiveB pointer to the buffer to receive \c n bytes to
 *                  (must not be NULL if \c n != 0)
 *  \param n        number of bytes to be received and filled into \c receiveB
 *  \param skip     if > 0 \c skip bytes are received and forgotten
 *                  after \c filling \c n bytes into \c receiveB
 */
void spi2ReceiveNS(uint8_t * receiveB, uint16_t n, uint8_t skip);

/** Accumulated errors at SPI 2
 *
 *  Driver function of SPI 2 (UART1 as SPI) record seen errors in this byte.
 *  It may be reset by user or system software (especially after logging or
 *  displaying the error(s)).software.
 *
 *  Bit 7 (8): any error. This bit will be set on every recognised SPI 2
 *             error. It may be reset alone leaving all others set.  <br/>
 *  Bit  (4)6: any memory card error. This bit will be set on every
 *             recognised SPI 2 error if at that time the
 *             \ref memCard_sec "small memory" card holder
 *             is \ref memCardIsSelected() "selected".
 *  Bit 5 (2): any network interface / ENC28J60 error. This bit will be set
 *             on every recognised SPI 2 error if at that time the Ethernet
 *             chip is \ref nicIsSelected() "selected".
 *
 *  Bit 3 (8): network interface controller (NIC, ENC28J60) stays busy. This
 *             bit will be set if waiting for the \ref MISTAT "MISTAT.BUSY"
 *             bit to go away timed out. <br />
 *  Bit 2 (4): transmit complete failed. This bit will be set if waiting
 *             for transmission complete (TXC) timed out. <br />
 *  Bit 1 (2): transmitter buffer empty failed. This bit will be set if
 *             waiting for empty transmitter buffer (UDRE) timed out. <br />
 *  Bit 0 (1): receive complete failed. This bit will be set if waiting for
 *             receive complete (RXC) timed out. <br />
 */
extern uint8_t spi2Errors; // Bit 7 new error Bit 0 receive complete failed

/** Check a SPI2 condition with timeout
 *
 *  This helper macro replaces the endless spin wait for SPI 2 conditions.
 *  It may be used without trailing semicolon. Example: \code
   SPI2COND_ERR_RET(UCSR1A & (1<<UDRE1), 0x82, ) // wait with timeout
   //  while ( !( UCSR1A & (1<<UDRE1))); // spin wait empty transmit buffer
 *  \endcode
 *  Compared to the (endless) while wait the use of this macro does not add
 *  to runtime if the condition is met from start.
 *
 *  The maximum waiting time or timeout is 2 * (UBRR1 + 1) * 10 clocks. That
 *  is ~20% longer than minimal necessary &mdash; but far better than endless
 *  waiting leading to watchdog reset.
 *
 *  \param cond conditional expression: true if ready to proceed
 *  \param err  error bits to be set in \ref spi2Errors in case of waiting for
 *        \c cond times out (by more than  <br />
 *        The xyzIsSelected bits are set by this macro
 *  \param ret return value / expression for return statement in error case;
 *        leave blank for use within void functions
 *  \hideinitializer
 */
#define SPI2COND_ERR_RET(cond, err, ret) \
   if (!(cond)) { \
      uint16_t stpCnt = (UBRR1 + 1) << 1; \
      while ( !(cond)) { \
         if (!--stpCnt) { \
            spi2Errors |= err; /* record error: new, receive; */ \
            if(nicIsSelected()) spi2Errors |=  0x20;  \
            if(memCardIsSelected()) spi2Errors |=  0x40; \
            return ret; \
         } \
      } \
   }


//--------------------------------------------------------------------------


#ifndef   UCPHA1
# define  UCPHA1 1   //!< UCSR1C register (bit number)
#endif

#endif /* SPI_H */

/** @} */ // Doxygen addtogroup   embDriv
/** @} */ // Doxygen addtogroup   llSySer
/** @} */ // Doxygen addtogroup   weAutSys
