/*
 * Copyright (c) 2011 Albrecht Weinert, Bochum
 *
 * This file is part of \ref intro_sec "weAutSys".
 *
 * Author: Albrecht Weinert <albrecht@a-weinert.de>
 *
 * $Id: system.h 2 2017-01-25 16:49:03Z albrecht $
 */

// Hint: this file's version will be used as weAutSys version.

/** \addtogroup weAutSys
 * @{
 *   \addtogroup sysSer
 * @{
 *
 *  \file system.h
 *
 *  \ref intro_sec "weAutSys" definition of system calls and services
 *           to be used by application / user software
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 *  This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 *  &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 *  \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 *  $Revision: 2 $
 *  $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $


/** @} */ // Doxygen addtogroup sysSer
#ifndef _SYSTEM_H_
#define _SYSTEM_H_

#include <avr/pgmspace.h>
#include "pt/pt.h"



/**
 * \addtogroup appLaySup
 * @{
 * \addtogroup comTypes
 * @{
 */

/** The runtime's name
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref SYST_DAT
 *  \showinitializer
 */
#define SYST_NAM "weAutSys"


/** The runtime's revision
 *
 *  \internal To be correct: This is the SVN revision text of this file. <br />
 *  The value is a "quoted" string like "$Revision: 2 $".
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref SYST_DAT  \ref SYST_MOD
 *  \showinitializer
 */
#define SYST_REV "$Revision: 2 $"
   //             0123456789.123456

/** This file's last modification date (SVN)
 *
 *  This file's subversion (SVN) modification date and
 *  \ref SYST_REV "revision number" are taken as the
 *  \ref intro_sec "weAutSys" runtime's revision data.
 *
 *  The value is a "quoted" string.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref DEFAULT_START_TIME  \ref SYST_REV
 *     \ref SYST_MOD
 *  \showinitializer
 */
#define SYST_DAT "$Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $"

/** This file's last modifier respectively SVN author
 *
 *  The value is a "quoted" string.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref DEFAULT_START_TIME  \ref SYST_REV
 *       \ref SYST_DAT
 */
#define SYST_MOD "$Author: albrecht $"



/** @} */ // addtogroup comTypes
/** @} */ // addtogroup appLaySup



/**
 * \addtogroup utilHelp
 * @{
 * \addtogroup utilText
 * @{
 */

/** The name of the runtime system \ref intro_sec "weAutSys"
 *
 *  This is a string in flash memory.
 *  \sa systWlc
 */
extern char const systNam[];


/** A greeting line with three leading feeds and the system name
 *
 *  This is a string in flash memory.
 *  \sa systNam
 *  \sa systBye
 */
extern char const systWlc[];

/** An farewell (abort) line with three leading feeds and the system name
 *
 *  This is a string in flash memory.
 *  \sa systNam
 *  \sa systWlc
 */
extern char const systBye[];

/** The system's revision
 *
 *  It is the SVN revision of the (this) file system.h. If not treated with
 *  beautifying tools like de.frame4j.CVSkeys (.java) prior to make it will
 *  be the full SVN tag in Dollars ($tagname: token $).
 *
 *  This is a string in flash memory.
 *  \sa getFirstSVNtokenP(char*, prog_char*, uint8_t)
 */
extern char const systRev[];

/** The system's last modification date
 *
 *  It is in fact the \ref SYST_DAT "SVN date" of the (this) file system.h.
 *  If not treated with beautifying tools like de.frame4j.CVSkeys (.java)
 *  prior to the make process it will be the full SVN tag in Dollars
 *  ($tagname: token token token $).
 *
 *  Un-beautified &mdash; the usual vase &mdash; this is a horrible date
 *  time zone (text) format; the first token is the date like 2011-03-11
 *  (yyyy-mm-dd). This can be extracted by
 *  \ref getFirstSVNtokenP()
 *
 *  This is a string in flash memory.
 */
extern char const systDat[];

/** The system's last modifier
 *
 *  It is the SVN user name of the person who did the latest modification
 *  of the (this) file system.h. If not treated with beautifying tools like
 *  \code de.frame4j.CVSkeys \endcode (.java, see frame4j.de) prior to make
 *  it will be the full SVN tag in Dollars ($tagname: token $).
 *
 *  de.frame4j.CVSkeys can be configured to replace SVN or system account
 *  names with the human's name, like e.g.<ul>
 *  <li>albrecht by Albrecht Weinert</li>
 *  <li>ralf by Ralf Seidel   etc.</li></ul>
 *
 *  This is a string in flash memory.
 *  \sa getFirstSVNtokenP(char*, prog_char*, uint8_t)
 */
extern char const systMod[];

/** The author of \ref intro_sec "weAutSys"
 *
 *  It is the author's name and his personal domain.
 *
 *  This is a string in flash memory.
 *  \sa getSomeCharsP(char*, prog_char*, uint8_t)
 */
extern char const systAut[];

/** \ref intro_sec "weAutSys"'s copyright notice.
 *
 *  This is a string in flash memory.
 *  \sa getSomeCharsP(char*, prog_char*, uint8_t)
 */
extern char const systCop[];

/** The build date and time
 *
 *  It is the word "build" followed by the date and time where the
 *  C-preprocessor rolled over the (this) file system.h. This will be time
 *  where the \ref intro_sec "weAutSys" runtime was build (if you get the text from a living
 *  system.
 *
 *  This is a string in flash memory.
 *  \sa getSomeCharsP(char*, prog_char*, uint8_t)
 */
extern char const systBld[];

/** Blanks and left opening brace
 *
 *  It is "  (". <br />
 *  This is just a building block for Strings held in flash memory.
 *  \sa getSomeCharsP(char*, prog_char*, uint8_t)
 */
extern char const sepLoB[];

/** Right closing brace and blanks
 *
 *  It is ")  ". <br />
 *  This is just a building block for Strings held in flash memory.
 *  \sa getSomeCharsP(char*, prog_char*, uint8_t)
 */
extern char const sepRcB[];


//                             0123456789.123456
extern char const encEthSta[]; //!< "  ENC/Eth st. : " flash text building block
//extern char const encMACset[]; //!< "  ENC set MAC : " flash text building block
//extern char const encHWtrbl[]; //!< "  ENC HW trouble" flash text building block
// extern char const encMACgot[]; //!< "  ENC got MAC : " flash text building block
extern char const const macAdd_is[]; //!< "  MAC address : " flash text building block
extern char const ip4Add_is[]; //!< "  IP4 address : " flash text building block
//                             0123456789.123456
extern char const ipDefault[]; //!< "  IP4 default : " flash text building block
extern char const const arpEntries[];//!< "  ARP entries   " flash text building block
//                             0123456789.123456
extern char const arpEmpty[];  //!< " no valid entries in ARP table \n\n" flash text building block

extern char const ipConIpAd[]; //!< "  IP4 address : " flash text building block
extern char const ipConNMsk[]; //!< "  IP4 netmask : " flash text building block
extern char const ipConDefR[]; //!< "  def. router : " flash text building block
extern char const ipConDNSa[]; //!< "  DNS address : " flash text building block
extern char const ipConNTPa[]; //!< "  NTP address : " flash text building block
extern char const ipConDHCu[]; //!< "  use DHCP "     flash text building block
extern char const ipConDHCs[]; //!< "  DHCP set "     flash text building block
extern char const ipConDefS[]; //!< "  def. set "     flash text building block
//                             0123456789.12345


/** Copy some characters from program space to a string
 *
 *  This function copies up to \c mxLen-1 characters from \c src to \c dst
 *  and appends a trailing 0. Returned is the address of the last character
 *  in \c dst modified (the 0's address).
 *
 *  The operation will stop after \c mxLen characters transferred resp.
 *  modified or if a terminating 0 if found in \c src.
 *
 *  \param dst the destination to modify (in RAM, not null!)
 *  \param src the source (string) to copy from (in flash memory, not null!)
 *  \param mxLen the maximum number of characters to modify in \c dst
 *         including the trailing 0 appended. The maximum advance of the
 *         return value to parameter \c dst
 *  \return dst + number of characters modified; NULL only if \c dst is NULL
 */
char * getSomeCharsP(char * dst, char const * src, uint8_t mxLen);


//-----------------------------------------------------------------------

extern char const wdSetSharp[]; //!< report: watchdog set sharp (flash text)
extern char const wdSetShort[];   //!< report: watchdog set short (flash text)
extern char const wdSetLeni[];  //!< report: watchdog set long / lenient (flash text)r
extern char const wdSetNormal[];//!< report: watchdog set normal (flash text)r

extern char const commWrong[];  //!< command report: wrong (flash text)
extern char const commAmbig[];  //!< command report: ambiguous (flash text)
extern char const optAmbig[];   //!< option report: ambiguous (flash text)
extern char const optWrong[];   //!< option report: wrong (flash text)
extern char const noUserCLI[];  //!< report: no user CLI (flash text)

extern char const comHelp[];    //!< command: show help (flash text)

extern char const comWatchD[]; //!< command: watchdog  [abort | sharp | lenient | normal]
extern char const comWDog[];   //!< command: wdog     60 ms  | 120 ms | 250 ms
extern char const comBoot[];   //!< command: boot -load | -reset  (flash text)

extern char const comPLContrl[];  //!< command: [-stop | -start]  PLC cycles (flash text)
extern char const comAutomation[]; //!< command: [-stop | -start]  PLC cycles  (flash text)

extern char const comVersion[];  //!< command: show version info (flash text)
extern char const comRunInfo[];  //!< command: show run time info (flash text)
extern char const comMacAddr[];  //!< command: show / set MAC address (flash text)
extern char const comEthAddr[];  //!< command: show address (ARP) info (flash text)
extern char const comIPconfig[]; //!< command: show IP configuration info (flash text)
extern char const comIPdefAd[];  //!< command: show / set IP address (flash text)
extern char const comUART[];     //!< command: show / set UART options
extern char const comTime[];     //!< command: show / set (local) time
extern char const comDate[];     //!< command: show / set (local) date
extern char const comZone[];     //!< command: show / set time zone and DST state
extern char const comNTP[];  //!<  command:  show / set NTP state and configuration
extern char const comDHCP[]; //!<  command:  show / restart DHCP
extern char const comDNS[];  //!<  command:  resolve a name by DNS
extern char const comTelnet[]; //!<  command:  telnet [-stop] : Telnet [close]
extern char const comModbus[]; //!<  command:  modbus  [-stop -off -reset | ...

extern char const comNIC[]; //!<  command: NICont  [-off | -on | -reset | -restart]
extern char const comENC[]; //!<  command: ENC28J60  network interface (same as NICont)
extern char const comSMC[]; //!<  command: SMCard  [ -on | -start | reset | -rest...
extern char const comDirlist[]; //!<  command: dirList [-option] [dirPath] memory card directory
extern char const comTypeFile[]; //!< command: typeFile filename   display a file

// extern char const comADoff[];   //!< prepared user command: set  all (8) inputs digital (flash text)
// extern char const comADinputs[];  //!< prepared user command: set selected inputs analogue (flash text)


/*
  comHelp,  // 1
  comWatchD, comWDog,  bLF1,   // 2 3 (4 is gap)
  comBoot,  comPLContrl, comAutomation, // 5 6 7
  comVersion,    comRunInfo,    // 8 9
  comMacAddr, comEthAddr,      // 10 11
  comIPconfig,  comIPdefault, // 12 13
  comUARTcnt, // 14
  comTime, ComDate, // 15, 16
  comZone, comNTP, // 17, 18
  comDHCP,   // 19
  comDNS,  // 20
  comTelnet, // 21
  comNIC, comENC, // 22 23
  comSMC,  // 24
  comDirlistNum, //  25
*/

// Do always keep the following consistent with char* systemCommands[]
#define comHelpNum   1  //!< command: show help (flash text)

#define comWatchDNum 2  //!< command: watchdog  [abort | sharp | lenient | normal]
#define comWDogNum   3  //!< command: wdog     60 ms  | 120 ms | 250 ms
#define comBootNum   5  //!< command: boot -load | -reset   (flash text)

#define comPLContrlNum   6 //!< command: [-stop | -start]  PLC cycles (flash text)
#define comAutomationNum 7 //!< command: [-stop | -start]  PLC cycles  (flash text)

#define comVersionNum   8  //!< command: show version info (flash text)
#define comRunInfoNum   9  //!< command: show run time info (flash text)
#define comMacAddrNum  10  //!< command: show / set MAC address (flash text)
#define comEthAddrNum  11  //!< command: show address (ARP) info (flash text)
#define comIPconfigNum 12  //!< command: show IP configuration info (flash text)
#define comIPdefAdNum  13  //!< command: show / set IP address (flash text)
#define comUARTNum  14  //!< command: show / set UART flow control (flash text)
#define comTimeNum  15  //!< command: show / set (local) time
#define comDateNum  16  //!< command: show / set (local) date
#define comZoneNum  17  //!< command: show / set time zone and DST state
#define comNTPNum   18  //!<  command:  show / set NTP state and configuration
#define comDHCPNum  19  //!<  command:  show / restart DHCP
#define comDNSNum   20  //!<  command:  resolve a name by DNS
#define comTelnetNum 21 //!<  command:  telnet [-stop]
#define comModbusNum 22 //!<  command:  modbus  [-stop -off -reset | ...

#define comNICNum   23  //!<  command: NICont  [-off | -on | -reset | -restart]
#define comENCNum   24  //!<  command: ENC28J60  network interface (same as NICont)
#define comSMCNum   25  //!<  command: SMCard  [ -on | -start | reset | -rest...
#define comDirlistNum 26 //!< command: dirList [-option] [dirPath] memory card directory
#define comTypeFileNum 27 //!< command: typeFile filename   display a file

extern char const * const systemCommands[];  //!< flash array of the (flash) system command definitions

/** List and definitions of command options
 *
 *  This is the flash array of the (flash) option definitions. Options
 *  are recognised by \ref setCliLine() if prefixed by - (minus) and given
 *  as first parameter of a command, like e.g. "help -options". Like with
 *  commands, options may be abbreviated if the abbreviation is not ambiguous,
 *  as would be -res instead of -reset respectively -restart. <br />
 *  As commands options are not case sensitive.
 *
 *  Unrecognised options may be (silently) or may be not ignored by system
 *  and application commands.
 *  \sa  setCliLine()
 */
extern char const * const systemOptions[];

extern char const optionHeader[];  //!< options:  * * *   Command options

extern char const optHelp[];   //!< options:  help     command specific help
extern char const optQuest[];  //!< options: ?        like -help
extern char const optSilent[]; //!< options:  silent   no or only urgent output
extern char const optQuiet[];  //!< options:  quiet    like -silent
extern char const optInfo[];   //!< options:  inform   normal / informing output
extern char const optVerbose[];//!< options:  verbose  lengthy output
extern char const optDebug[];  //!< options:  debug    lengthy output for trouble shooting
extern char const optOn[];     //!< options:  on       turn on / start
extern char const optStart[];  //!< options:  start    like -on (in most cases)
extern char const optRestart[];//!< options:  restart  like -start (in most cases)
extern char const optOff[];    //!< options:  off      turn off / stop
extern char const optStop[];   //!< options:  stop     like -off (in most cases)
extern char const optReset[];  //!< options:  reset    like -stop or -restart (it depends)
extern char const optAbort[];  //!< options:  abort    a hard reset
extern char const optSharp[];  //!< options:  sharp    more exact / less forgiving
extern char const optLenient[];//!< options:  lenient  more forgiving

extern char const optSlow[];   //!< options:  slow     slow(er) speed  \n";
extern char const optNormal[]; //!< options:  normal   standard behaviour / speed
extern char const optFast[];   //!< options:  fast      fast(er)  speed \n";

extern char const optSystem[]; //!< options:  system related
extern char const optApplic[]; //!< options:  application related
extern char const optOption[]; //!< options:  options related
extern char const optLogHere[];//!< options:  use this device as log output
extern char const optNoLogs[]; //!< options:  do not use this device as log output
extern char const optFlowCn[]; //!< options:  ask or set flow control behaviour

extern char const optRead[]; //!< options: perform read  operation or set read mode
extern char const optWrite[];//!< options: perform write operation or set write mode
extern char const optLoad[]; //!< options: perform load  operation or set load mode

extern char const optPrimar[]; //!< options:   use or restart with primary resource
extern char const optAltern[]; //!< options:   use or restart with alternative



// Do always keep the following consistent with char* systemOptions[]
#define optHelpNum   2 //!< options:  help     command specific help
#define optQuestNum  1 //!< options: ?        like -help
#define optSilentNum 3 //!< options:  silent   no or only urgent output
#define optQuietNum  4 //!< options:  quiet    like -silent
#define optInfoNum   5 //!< options:  inform   normal / informing output
#define optVerboseNum  6 //!< options:  verbose  lengthy output
#define optDebugNum  7  //!< options:  debug    lengthy output for trouble shooting
#define optOnNum     8 //!< options:  on       turn on / start
#define optStartNum  9 //!< options:  start    like -on (in most cases)
#define optRestartNum 10 //!< options:  restart  like -start (in most cases)
#define optOffNum   11 //!< options:  off      turn off / stop
#define optStopNum  12 //!< options:  stop     like -off (in most cases)
#define optResetNum 13 //!< options:  reset    like -stop or -restart (it depends)
#define optAbortNum 14 //!< options:  abort    a hard reset
#define optSharpNum 15 //!< options:  sharp    more exact / less forgiving
#define optLenientNum 16 //!< options:  lenient  more forgiving

#define optSlowNum   17  //!< options:  slow     slow(er) speed
#define optNormalNum 18  //!< options:  normal   standard behaviour / speed
#define optFastNum   19  //!< options:  fast     fast(er)  speed

#define optSystemNum  20 //!< options:  system related
#define optApplicNum  21 //!< options:  application related
#define optOptionNum  22 //!< options:  options related
#define optLogHereNum 23 //!< options:  use this device as log output
#define optNoLogsNum  24 //!< options:  do not use this device as log output
#define optFlowCnNum  25 //!< options:  ask or set flow control behaviour

#define optReadNum  26 //!< options: perform read  operation or set read mode
#define optWriteNum 27 //!< options: perform write operation or set write mode
#define optLoadNum  28 //!< options: perform load  operation or set load mode

#define optPrimarNum 29 //!< options:   use or restart with primary resource
#define optAlternNum 30 //!< options:   use or restart with alternative


#define optNotGivenNum  255 //!< options, commands, parameters: no option given / set
#define optAmbigousNum  254 //!< options:   option given ambiguously


/** Headline for application commands overview (help list)
 *
 *  This flash text must be the first entry in
 *  \ref appCLIreg_t "userCommands[]" if user software chooses to
 *  \ref registerAppCli "register" application specific commands.
 */
extern char const helpUserCm[];

/** Headline for system commands overview (help list)
 *
 *  This flash text is the first entry in \ref systemCommands[].
 */
extern char const helpHeader[];

extern char const sysRunSectorN[];  //!< report: LF  SMC sector buffered: 0x


/** @} */ // addtogroup utilText
/** @} */ // addtogroup utilHelp

// ------   Software instrumentation  -----------------------
/**
 * \addtogroup llSySer
 * @{
 * \addtogroup swInstrm
 * @{
 */


// ----  test LEDs     ---------------------------------------


/** Set the state of the green (test) LED
 *
 *  If no green test LED is available on the respective platform but a yellow
 *  one (Arduinos e.g.) the latter is used.
 *
 *  If neither green nor yellow single status / test LED is available this
 *  procedure does nothing.
 *
 *  \param state  0 / false: turn off; != 0: turn on
 */
void setStatusLedGn(uint8_t state);


/** Change the state of the green (test) LED
 *
 *  If no such test LED is available on the respective platform this
 *  procedure does nothing.  Otherwise the LED's state is toggled.
 *
 * \sa setStatusLedGn
 */
void toggleStatusLedGn(void);

/** Set the state of the red (test) LED
 *
 *  If no red test LED is available on the respective platform, this
 *  procedure does nothing.
 *
 *  \param state  0 / false: turn off; != 0: turn on
 */
void setStatusLedRd(uint8_t state);

/** Change the state of the red (test) LED
 *
 *  If no red test LED is available on the respective platform, this
 *  procedure does nothing. Otherwise the LED's state is toggled.
 *  \sa setStatusLedRd
 */
void toggleStatusLedRd(void);

// -----------    Test Pins ---------------------------------------

/** Initialisation of the (potential) test pins
 *
 *  This function sets Port \ref setTestPin0 "C1 / testPin0" and
 *  Port \ref setTestPin1 "C0 / testPin1" as output and does
 *  \ref setTestPin0 "setTestPin0(ON)" and
 *  \ref setTestPin1 "setTestPin1(OFF)".
 *
 *  Hint: Using the testpins 0 & 1 with the module \ref intro_secH "weAut_01"
 *        inhibits the usage of Port C1 / C0 for other purposes like IC /
 *        twowire. Testpin0 JP3pin1 (would be twoWire SDA) and Testpin1 is
 *        JP3pin4 (would be SCL).
 *
 *  Hint: If the hardware is not \ref intro_secH "weAut_01" the test pins may
 *        be assigned to other ports and pins.
 *  \param TP0start initial value for Testpin 0 (On or OFF)
 *  \param TP1start initial value for Testpin 1 (On or OFF)
 */
void initTestPins(uint8_t TP0start, uint8_t TP1start);

/** Set the state of Test-Pin 0
 *
 *  This is to be used to measure the behaviour resp. timing of software
 *  by digital analysers, oscilloscopes or other instruments.
 *
 *  \param state 0 / false for off; != 0 for on
 */
void setTestPin0(uint8_t state) __attribute__((always_inline));

/** Set the state of Test-Pin 1
 *
 *  This is to be used to measure the behaviour resp. timing of software
 *  by digital analysers, oscilloscopes or other instruments.
 *
 *  \param state  0 / false for off; != 0 for on
 */
void setTestPin1(uint8_t state) __attribute__((always_inline));


#endif /* _SYSTEM_H_ */

/** @} */ // Doxygen addtogroup swInstrm
/** @} */ // Doxygen addtogroup llSySer
/** @} */ // Doxygen addtogroup weAutSys

