// Copyright (c) 2012 Albrecht Weinert, Bochum
// telnet.h 2 2017-01-25 16:49:03Z albrecht

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup appLaySup
 * @{
 * \addtogroup telnetServer
 * @{
 *
 * \file telnet.h
 *
 * \ref intro_sec "weAutSys"' system calls, services and types for the Telnet server
 *
 * This file contains the definitions for \ref intro_sec "weAutSys"'
 * system calls, services and types for the build in Telnet server. It
 * implies an \ref uip "uIP" used as TCP/IP stack.
 *
 * This is system software and must not be modified for user or
 * application programs.
 *
 * This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012  Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert

/** @} */ // Doxygen addtogroup telnetServer
/** @} */ // Doxygen addtogroup appLaySup

#ifndef TELNET_H
#define TELNET_H

#include "we-aut_sys/network.h"
#include "we-aut_sys/uart0.h"
#include "we-aut_sys/timing.h"

/**
 * \addtogroup appLaySup
 * @{
 * \addtogroup telnetServer
 * @{
 */


/** @name Defines for Telnet command, state and options
 *
 *  This is a subset of the command and option codes defined for the
 *  Telnet protocol. The use of the others not put here is discouraged.
 */
///@{

/** Telnet: interpret as command (IAC)
 *
 *  All Telnet commands consist of a sequence of at least a two
 *  bytes: <ol>
 *  <li> the "Interpret as Command" (IAC) escape character </li>
 *  <li> the code for the command. </li></ol>
 *
 *  The commands dealing with option negotiation are three byte
 *  sequences: <ol>
 *  <li> the "Interpret as Command" (IAC) escape character </li>
 *  <li> the code for the command </li>
 *  <li> the code for the option referenced in command. </li></ol>
 *
 *  Sub-negotiation are at least five bytes long and are structured as: <br />
 *  &nbsp; \ref IAC  \ref TEL_SB "SB"
 *  information byte(s)  \ref IAC  \ref TEL_SE "SE"
 *
 *  The value of \ref IAC is 255 or 0xFF. <br />
 *  The value will also be used as (Telnet command interpreter) state.
 *
 *  If \c 0xFF is part of the data it has to be escaped (by doubling it).
 *  I.e. \ref IAC \ref IAC does not act as command but as normal data
 *  byte \c 0xFF.
 *  \hideinitializer
 */
#define IAC 255

/** Telnet: option negotiation command respectively answer
 *
 *  \c DONT demands (other) no longer to perform indicated option. The
 *  positive acknowledge is a corresponding \ref WONT command. <br />
 *  The value will also be used as (Telnet command interpreter) state.
 *  \showinitializer
 */
#define DONT 254

/** Telnet: option negotiation command respectively answer
 *
 *  \c DO requests indicated option. The positive acknowledge is \ref WILL
 *  the negative answer is \ref WONT. <br />
 *  The value will also be used as (Telnet command interpreter) state.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref DONT
 *  \showinitializer
 */
#define DO 253

/** Telnet: option negotiation command respectively answer
 *
 *  \c WONT refuses (to continue) the indicated optiond. The
 *  positive acknowledge is a corresponding \ref DONT command. <br />
 *  The value will also be used as (Telnet command interpreter) state.
 *  \showinitializer
 */
#define WONT 252

/** Telnet: option negotiation command respectively answer
 *
 *  \c WILL wants or confirms the indicated option. The positive acknowledge
 *  is \ref DO the negative answer is \ref DONT. <br />
 *  The value will also be used as (Telnet command interpreter) state.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref WONT
 *  \showinitializer
 */
#define WILL 251

#define LINEMODE 34 //!< Telnet option code \showinitializer

#define SUPPRESS_GA 3 //!< Telnet option suppress go ahead \showinitializer

/** Telnet command opening brace SB
 *
 *  \ref IAC \c SB indicates the begin of sub-negotiation data. <br />
 *  The value will also be used as (Telnet command interpreter) state.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref TEL_SE "SE"
 *  \showinitializer
 */
#define TEL_SB 250

/** Telnet command closing  brace SE
 *
 *  \ref IAC \c SE indicates the end of sub-negotiation data and the
 *  sub-negotiation command as whole.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref TEL_SB "SB"
 *  \showinitializer
 */
#define TEL_SE 240

// other defines see revision ~308 .. 311  Do not use them

///@} // end member group Telnet command (state) and options (negotiation)



/** Initialise the Telnet (server)
 *
 *  This function initialises the Telnet server. It is (indirectly) called
 *  whenever the Ethernet link goes up and hence usually after reset.
 *
 *  At present the implementation does nothing, but that may change in
 *  future releases.
 */
void telnetInit(void);


/** Handle Telnet server events
 *
 *  This function handles all Telnet events. And it is a as a protothread
 *  (part of) the Telnet server state machine.
 *
 *  This Telnet server implements the basic communication, a CLI (command line
 *  interpreter) and a \ref bufLogCons "log buffer consumer".
 *  The
 *  \ref uip_tcp_appstate_t "appstate" part of the connection state hence is
 *  thread data of type \ref thr_data_t in the \ref cliThr_data_t variant.
 *
 *  The current appstate.flag usage is: <br />
 *  Bit 0 (1): command interpreter active <br />
 *  Bit 1 (2): prompt output (after last output) pending <br />
 *  Bit 3 (4): consume and send the pending \ref bufLogStreams output
 *           (before prompt). <br />
 */
ptfnct_t telnetAppcall(void);


/** @} */ // Doxygen addtogroup telnetServer
/** @} */ // Doxygen addtogroup appLaySup

#endif /* TELNET_H */
/** @} */ // Doxygen addtogroup weAutSys
