/*
 * Copyright (c) 2005, Swedish Institute of Computer Science
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file is part of the uIP TCP/IP stack
 */

/**
 * \file dhcpc.c
 *
 * Implementation of the DHCP client
 *
 * The Dynamic Host Configuration Protocol (DHCP) is used for getting IP
 * configuration data from a server. The scope of this configuration and
 * the (call back) interface have been enhanced.
 *
 *  Adam Dunkels' original copyright notice <br /> &nbsp;
 *  Copyright (c) 2005, Swedish Institute of Computer Science.
 *  All rights reserved. <br />
 *  (contained in this file) still holds. For modifications:
 *
 *  <br /> &nbsp;
 * \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author Adam Dunkels <adam@dunkels.com>
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

#include <string.h>

#include "uip/uip.h"
#include "uip/apps/dhcpc.h"
#include "pt/pt.h"

#include "we-aut_sys/common.h"
#include "we-aut_sys/timing.h"
#include "we-aut_sys/network.h"

#if  DEBUG_DHCP
# include "we-aut_sys/log_streams.h"
INFLASH(char const dhcpInitMes[]) = "\n  DHCP init'd.  \n";
//                             0 123456789.1234567 8
INFLASH(char const dhcpNoCoMes[]) = "\n  DHCP no con.  \n";
#endif


struct dhcpState_t dhcpState;

static uint8_t xid[4] = {0xEA, 0x01, 0x12, 0x23};
static const uint8_t magic_cookie[4] = {99, 130, 83, 99};

static uint8_t * add_msg_type(uint8_t * optptr, uint8_t type){
  *optptr++ = DHCP_OPTION_MSG_TYPE;
  *optptr++ = 1;
  *optptr++ = type;
  return optptr;
} //  add_msg_type(uint8_t *, uint8_t)


static uint8_t * add_server_id(uint8_t * optptr){
  *optptr++ = DHCP_OPTION_SERVER_ID;
  *optptr++ = 4;
  memcpy(optptr, dhcpState.serverid, 4);
  return optptr + 4;
} // add_server_id(uint8_t *)


static uint8_t * add_req_ipaddr(uint8_t * optptr){
  *optptr++ = DHCP_OPTION_REQ_IPADDR;
  *optptr++ = 4;
  memcpy(optptr, dhcpState.ipaddr, 4);
  return optptr + 4;
} // add_req_ipaddr(uint8_t *)


static uint8_t * add_req_options(uint8_t * optptr){
  *optptr++ = DHCP_OPTION_REQ_LIST;
  *optptr++ = 9;
  *optptr++ = DHCP_OPTION_SUBNET_MASK;  // 1
  *optptr++ = DHCP_OPTION_ROUTER;
  *optptr++ = DHCP_OPTION_DNS_SERVER;
  *optptr++ = DHCP_OPTION_NTP_SERVERS;  // 4  -
  *optptr++ = DHCP_OPTION_CLIENT_NAME;   // 5   -
  *optptr++ = DHCP_OPTION_TIME_OFFSET;   // 6    > 4..9 added by A. Weinert
  *optptr++ = DHCP_OPTION_LEASE_TIME;   // 7    /
  *optptr++ = DHCP_OPTION_RENEW_TIME;   // 8    |
  *optptr++ = DHCP_OPTION_DOMAIN_NAME;  // 15  / -
  return optptr;
} // add_req_options(uint8_t *)


static uint8_t *  add_end(uint8_t * optptr){
  *optptr++ = DHCP_OPTION_END;
  return optptr;
} // add_end(uint8_t *)


static void create_msg(register struct dhcpMsg_t * m){
  m->op    = DHCP_REQUEST;
  m->htype = DHCP_HTYPE_ETHERNET;
  m->hlen  = DHCP_HLEN_ETHERNET; // dhcpState.mac_len;
  m->hops  = 0;
  memcpy(m->xid, xid, sizeof(m->xid));
  m->secs = 0;
  m->flags = HTONS(BOOTP_BROADCAST); /*  Broadcast bit. */
  /*  uip_ipaddr_copy(m->ciaddr, uip_hostaddr);*/
  memcpy(m->ciaddr, uip_hostaddr, sizeof(m->ciaddr));
  memset(m->yiaddr, 0, sizeof(m->yiaddr));
  memset(m->siaddr, 0, sizeof(m->siaddr));
  memset(m->giaddr, 0, sizeof(m->giaddr));
  memcpy(m->chaddr, dhcpState.mac_addr, DHCP_HLEN_ETHERNET); // dhcpState.mac_len);
  memset(&m->chaddr[DHCP_HLEN_ETHERNET], 0, sizeof(m->chaddr) - DHCP_HLEN_ETHERNET);

#ifndef UIP_CONF_DHCP_LIGHT
  m->sname[0] = 0; // is sufficient for empty  (0-terminated)
  //  memset(m->sname, 0, 2); // sizeof(m->sname));  // 4 inst. of 64 as is 0 term
  m->file[0] = 0; // is sufficient for empty  (0-terminated)
  // memset(m->file, 0, 2);  // sizeof(m->file));  experi 4 inst of 128 0-termin.
#endif

  memcpy(m->options, magic_cookie, sizeof(magic_cookie));
} // create_msg(register struct dhcpMsg_t)


static void send_discover(void){
  struct dhcpMsg_t *m = (struct dhcpMsg_t *)uip_appdata;
  create_msg(m);

  uint8_t *end;
  end = add_msg_type(&m->options[4], DHCPDISCOVER);
  end = add_req_options(end);
  end = add_end(end);

  uip_send(uip_appdata, end - (uint8_t *)uip_appdata);
#if  DEBUG_DHCP > 1
  bufLogMark("DHCPsndDi", 9);
#endif
} // send_discover(void)


static void send_request(void){
  struct dhcpMsg_t *m = (struct dhcpMsg_t *)uip_appdata;
  create_msg(m);

  uint8_t *end;
  end = add_msg_type(&m->options[4], DHCPREQUEST);
  end = add_server_id(end);  // request server ID is response to DHCPOFFER
  end = add_req_ipaddr(end);
  end = add_end(end);
  
  uip_send(uip_appdata, end - (uint8_t *)uip_appdata);
#if  DEBUG_DHCP > 1
  bufLogMark("DHCPsndRe", 9);
#endif
} // void send_request()


static uint8_t parse_options(uint8_t * optptr, int len){
  uint8_t *end = optptr + len;
  uint8_t type = 0;

  while(optptr < end) {
     if (!(*optptr)) {
        ++optptr;
        continue;
     } // pad option (0)
    dhcpcGotOption((struct dhcpOption_t*) optptr);
    switch(*optptr) {
     /* Those options were put from here [AD]  to the dhcpcGotOption callback
      *  function [A.W.]:
      * DHCP_OPTION_SUBNET_MASK,  DHCP_OPTION_ROUTER,
      * DHCP_OPTION_DNS_SERVER, DHCP_OPTION_LEASE_TIME: */
    case DHCP_OPTION_MSG_TYPE:
      type = *(optptr + 2);
      // put state machine switch here for sake of call back dhcpcGotOption()
      if (type == DHCPOFFER) {
         dhcpState.state = STATE_OFFER_RECEIVED;
      } else if (type == DHCPACK) {
         dhcpState.state = STATE_CONFIG_RECEIVED;
      }
      break;
    case  DHCP_OPTION_RENEW_TIME:
       if (dhcpState.state == STATE_CONFIG_RECEIVED) {
#if  DEBUG_DHCP  > 2         // 9
          bufLogMark("DHCPstT53", 9);
          startTimer(&dhcpState.timer, 53, sec_dur_TIMER_TYPE);
#else
          const char * const vP = ((struct dhcpOption_t*) optptr)->value;
          startTimer(&dhcpState.timer,
                __builtin_bswap32(*((uint32_t *)vP)), sec_dur_TIMER_TYPE);
#endif
       }
       break;

    case DHCP_OPTION_SERVER_ID:
      memcpy(dhcpState.serverid, optptr + 2, 4);
      break;
    case DHCP_OPTION_END:
      return type;
    }
    optptr += optptr[1] + 2;
  }
  return type;
} //  parse_options(uint8_t, int)


static uint8_t parse_msg(void){
   struct dhcpMsg_t *m = (struct dhcpMsg_t *)uip_appdata;

   if(m->op == DHCP_REPLY &&
         memcmp(m->xid, xid, sizeof(xid)) == 0 &&
         memcmp(m->chaddr, dhcpState.mac_addr, DHCP_HLEN_ETHERNET) == 0) {
      memcpy(dhcpState.ipaddr, m->yiaddr, 4);
      return parse_options(&m->options[4], uip_datalen());
   }
   return 0;
} // parse_msg(void)

//------------------------------------------------------------------------

ptfnct_t dhcpc_appcall(void){
  PT_BEGIN(&dhcpState.pt);

  dhcpState.state = STATE_SENDING;
  initTimer(&dhcpState.timer, 1002, ms_TIMER_TYPE); // init timer with 1s

  for (;;) {  // in state STATE_SENDING -> STATE_OFFER_RECEIVED
    send_discover();
    startNextInterval(&dhcpState.timer); // (re-) start timer
    PT_YIELD_UNTIL(&dhcpState.pt, uip_newdata()
                                          || !timerRunning(dhcpState.timer));
    if(uip_newdata()) {
       if ( parse_msg() == DHCPOFFER) {
        //  dhcpState.state = STATE_OFFER_RECEIVED; // done in parse_msg()
#if  DEBUG_DHCP > 1
          bufLogMark("DHCPrecOf", 9);
#endif
          break;
       } // parse_msg() == DHCPOFFER
#if  DEBUG_DHCP > 1
       bufLogMark("DHCPrecF1", 9);
#endif
    } // uip_newdata()
    if(dhcpState.timer.period.v8[1] < 0xCF) {
       dhcpState.timer.period.v8[1] += 0x30; // add to timer until minute
    } // push timer up to about a minute
  } // for  in state STATE_SENDING -> STATE_OFFER_RECEIVED
  
  initTimer(&dhcpState.timer, 1002, ms_TIMER_TYPE); // init timer with 1s

  for (;;) {  // in state STATE_OFFER_RECEIVED -> STATE_CONFIG_RECEIVED
    send_request();
    startNextInterval(&dhcpState.timer); // (re-) start timer
    PT_YIELD_UNTIL(&dhcpState.pt, uip_newdata()
                                       || !timerRunning(dhcpState.timer));

    if(uip_newdata() && parse_msg() == DHCPACK) {
      // dhcpState.state = STATE_CONFIG_RECEIVED; // done in parse_msg()
#if  DEBUG_DHCP > 1
      bufLogMark("DHCPrecAq", 9);
#endif
      break;
    }
    if((dhcpState.timer.period.v8[1] += 4) > 0x2A) {  // > about 11 s
#if  DEBUG_DHCP > 1   //  9
       bufLogMark("DHCP11sRs", 9);
#endif

       // here we should signal time out instead (decide if and how)
       PT_RESTART(&dhcpState.pt); // request/response timeout, sum ~ 1 minute
    } //
  } // for  in state STATE_OFFER_RECEIVED -> STATE_CONFIG_RECEIVED
  
  dhcpc_configured(DHCPACK, dhcpState.ipaddr); // that's it, (well) done

  // A.W. 2: to stop useless periodic in this DHCP ready state we "close"
  dhcpState.conn->lport = 0;  // this is a close

  // next time this function will be called after dhcpReset() (or dhcpInit()),
  // but then it will start from PT_BEGIN above

  // or it will be called on the lapse of dhcpState.timer.
  // The latter will be the end of the renewal time.
  PT_WAIT_WHILE(&dhcpState.pt, timerRunning(dhcpState.timer));

  dhcpInit(&actMACadd); // poor man's renew = mostly start fresh (works)
  PT_END(&dhcpState.pt)
} // dhcpc_appcall(void)


void dhcpReset(void){
  dhcpState.state = STATE_INITIAL;
  freeTimer(&dhcpState.timer);
  if (dhcpState.conn) { // had connection
     dhcpState.conn->lport = 0; // release local port
     dhcpState.conn  = NULL;
  }
  // dhcpState.ticks = 2;
  PT_INIT(&dhcpState.pt);
#if  DEBUG_DHCP > 1
  bufLogMark("DHCPreset", 9);
#endif
} //  dhcpcReset()

void dhcpInit(void const * mac_addr){
  dhcpReset();
  dhcpState.mac_addr = mac_addr;
  xid[3] = ((uint8_t*)mac_addr)[2]; // unique XID by MAC ..
  xid[2] = ((uint8_t*)mac_addr)[5]; // ... hopefully
  xid[1] = msTimers.end.v8[0];  // random XID look at internal clock

  uip_ipaddr_t addr;
  uip_ipaddr(addr, 255,255,255,255);
  dhcpState.conn = uip_udp_new(&addr, HTONS(DHCPC_SERVER_PORT));
  if(dhcpState.conn != NULL) {
    uip_udp_bind(dhcpState.conn, HTONS(DHCPC_CLIENT_PORT));
#if  DEBUG_DHCP
  bufLogTxt_P(dhcpInitMes); //
#endif
   } else {
#if  DEBUG_DHCP
      bufLogTxt_P(dhcpNoCoMes);
#endif
   }
} //  dhcpcInit(const void *)

