// Copyright (c) 2012 Albrecht Weinert, Bochum
// $Id: smc2fs.c 2 2017-01-25 16:49:03Z albrecht $

/**
 * \file smc2fs.c
 *
 *  weAutSys' file system adaption to a small memory card
 *
 *  This file contains (part of) the implementation of weAutSys' adaption
 *  of ChaN's fatFS to small memory cards
 *  \ref memCard_sec "(SMCs)".
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of weAutSys &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

#include "we-aut_sys/smc2fs.h"
#include "fatFS/ff.h"
#include "we-aut_sys/streams.h"
#include "we-aut_sys/utils.h"

/*  File system object (the one for SMC)
 *
 *  This is the (one) file system state object used.
 *
 *  In  \ref intro_sec "weAutSys" / \ref intro_secH "weAut_01" and alike
 *  there can be just one or zero file-system for a small memory card (SMC
 *  card inserted.
 */
FATFS fileSystSMC;

/*  Work space for file system operations
 *
 *  This structure holds the state of the system or application software's
 *  ongoing file system operations. In a small realtime system file system
 *  operations on SMCs (interfaced via SPI) are relatively expensive in terms
 *  of both total processor time and size of indivisible (by yielding)
 *  sub-steps.
 *
 *  Hence file system operations should occur in one (well designed / planned)
 *  thread at a time only. This structure \ref fsWork of type \ref FS_WORK is
 *  to hold all file system related state and should remain singleton for
 *  said reasons.
 *
 *
 *  \sa fsWork
 */
FS_WORK fsWork;

INFLASH(char isDir[]) = "<dir ";

/*  Write file info data to output
 *
 *  This function outputs the information on the file \c filInf to the
 *  standard output stream.
 *
 *  The bits of the parameter \c inf control the information output: <br />
 *  Bit 7 (0x80) : output leading blank  <br />
 *  Bit 5 (0x20) : output last modified date / time <br />
 *  Bit 4 (0x10) : output attributes <br />
 *  Bit 3 (0x08) : output length <br />
 *  Bit 0 (0x01) : output trailing new line
 *
 *  \param filInf pointer to the \ref FILINFO "file information" structure
 *  \param inf    the informations to output
 *  \return       0: nothing was output
 */
uint8_t stdPutFilInf(FILINFO * filInf, const uint8_t inf){
   if (!filInf) return 0;     // no string no output
   if (filInf->fname[0] == 0) return 0; // no name
   if (inf & 0x80) stdPutC(' ');
#if _USE_LFN
   stdPutS(filInf->lfname ? filInf->lfname : filInf->fname);
#else
   stdPutS(filInf->fname);
#endif
   for (uint8_t i = filInf->namLen; i <= 13; ++i) {  stdPutC(' '); }

   if (inf & 0x20) { // date  time
      char s[33];
      char * beg = formFATdate(s, filInf->fdate);
      *beg++ = ' ';
      beg = formFATtime(beg, filInf->ftime);
      *beg++ = ' ';  *beg++ = ' '; *beg = 0;
      stdPutS(s);
   }

   uint8_t a = filInf->fattrib & AM_MASK;
   if (a & AM_DIR) { // directory
      stdPutS_P(isDir);
   } else { // dir / no dir
      if ((inf & 0x10) && (a & ~AM_VOL)) {  // attribute
         stdPutC(a & AM_RDO ? 'r' : '+');
         stdPutC(a & AM_SYS ? 's' : '-');
         stdPutC(a & AM_ARC ? 'a' : '-');
         stdPutC(a & AM_HID ? 'h' : '+');
         stdPutC(' ');
      }  // attribute
      if (inf & 0x08) { // size
         char s[12]; s[9] = ' '; s[10] = 0;
         eightHexs(s, filInf->fsize);
         stdPutS(s);
      } // size
   } // no directory


   if (inf & 0x01) stdPutC('\n');
   return 1;
} // putFilInf(..)

/*  Format a FAT time
 *
 *  This function formats a 16 bit time delivered as (old fashined)
 *  \ref getFATtime "FAT time" structure. <br />
 *  The format will be \code "23:59:58" \endcode. <br />
 *  The length of the text delivered (to  *s) is 9 characters including a
 *  trailing zero.
 *
 *  Note that the seconds in that (old DOS) FAT time format are 2 s
 *  increments so no odd value for seconds is possible.
 *
 *  \param s        start position within string resp. char array
 *  \param fatTime  the 16 bit FAT time structure
 *  \return pointer to the (0-) character following the formatted time
 *  \sa getFATtime()
 */
char * formFATtime(char * s, uint16_t fatTime){
   //  0123456789
   // "23:59:58";
   uint8_t v = fatTime & 0x001F;
   twoDigs(&(s[6]), v + v); // seconds
   v = (uint8_t)(fatTime >> 8);
   v = ((v << 3) & 0x38)  + ((uint8_t)(fatTime & 0x00E0) >> 5);
   twoDigs(&(s[3]), v); // min
   v = ((uint8_t)(fatTime >> 8)) >> 3;
   twoDigs(s, v); // hour
   s[2] = s[5] = ':';
   s += 8;
   *s = 0;
   return s;
 } // formFATtime(char *, uint16_t)

/*  Format a FAT date
 *
 *  This function formats a 16 bit delivered as (old fashined)
 *  \ref getFATtime "FAT date" structure (upper 16 bits 16..31). <br />
 *  The format (ISO 8601 / EN 28601) will be \code "2009-01-31" \endcode. <br />
 *  The length of the text delivered (to  *s) is 11 characters including a
 *  trailing zero.
 *
 *  \param s        start position within string resp. char array
 *  \param fatDate  the 16 bit FAT date structure
 *  \return pointer to the (0-) character following the formatted date
 *  \sa getFATtime()
 */
char * formFATdate(char * s, uint16_t fatDate){
    //  0123456789.
    // "2010-01-31";
    uint8_t v = fatDate & 0x001F;
    twoDigs(&(s[8]), v);  // day
    v = ((uint8_t)(fatDate & 0x00E0) >> 5);
    if (fatDate & 0x0100) v += 8;
    twoDigs(&(s[5]), v);
    v = (uint8_t)(fatDate >> 8) >> 1;  // year from 1980
    if (v > 19) { // 2000...
       s[0] = '2';
       v -= 20;
       if (v < 100) {
          s[1] = '0';
       } else if (v < 200) {
          s[1] = '1';
          v -= 100;
       } else {
          s[1] = '2';
          v -= 200;
       }
       twoDigs(&(s[2]), v);
    } else { // 1980 .. 1999
       s[0] = '1';
       s[1] = '9';
       if (v >= 10) {
          s[2] = '9';
          v -= 10;
       } else {
          s[2] = '0';
       }
       s[3] = '0' + v;
    }
      s[4] = s[7] = '-';
      s += 10;
      *s = 0;
      return s;
} // formFATdate(char *, uint16_t)


