// Copyright (c) 2012 Albrecht Weinert, Bochum *
// streams.c 2 2017-01-25 16:49:03Z albrecht

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup appLaySup
 * @{
 * \addtogroup stdStreams
 * @{
 *
 *
 * \file streams.c
 *
 * Stream I/O
 *
 * This file contains the implementations for the basics of weAutSys'
 * I/O via C standard respectively avr_stdio standard like streams.
 * This is system software and must not be modified for user or
 * application programs.
 *
 *
 * Please find detailed description of the variables, functions etc.
 * in the module \ref stdStreams "Streams" and in the include
 * file streams.h.
 *
 * This file is part of weAutSys &nbsp;
 *           <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */


/** @} */ // doxygen addtogroup stdStreams
/** @} */ // doxygen addtogroup appLaySup
/** @} */ // doxygen addtogroup weAutSys

#include "we-aut_sys/common.h"
#include "we-aut_sys/streams.h"
#include "we-aut_sys/log_streams.h"
#include "we-aut_sys/utils.h"
#include "we-aut_sys/uart0.h"
#include <string.h>

#if DEBUG_TEL > 2
#include "we-aut_sys/utils.h"   // Telnet intensive debug only
#endif

#include "uip/uip.h"

//------------    common stream handling -------------------------------

/* Function giving number of characters buffered for input
 *
 *  A function of this type returns the number of bytes that can be fetched
 *  from the stream's input without getting EOF.
 *
 *  For buffered devices with known state the answer will be correct.
 *
 *  In other cases the number of characters to be got in one run without EOF
 *  is unknown. Then a large number with bit 15 set (0xFFFF
 *  \ref AVAILABLE_UNKNOWN) is returned.
 */
uint16_t inBufferd(const FILE * stream) {
   if (!stream) return 0; // no stream no input
   struct streamClassDescript_t *streamData = fdev_get_udata(stream);
   if (!streamData) return AVAILABLE_UNKNOWN; // no stream data no info
   str2InBufferd inBufferdFunc = streamData->inBufferdFunc;
   if (!inBufferdFunc) return AVAILABLE_UNKNOWN; // no function no info

   // The case self handled as null could be handled with some overhead.
   // if (inBufferdFunc == &inBufferd) return AVAILABLE_UNKNOWN;

   return (*inBufferdFunc)(stream);
} // inBufferd(FILE *)


/*  Function giving buffer space available for output
 *
 *  This function returns returns the number of bytes that can
 *  currently be output to the stream in one run without waiting or overflow.
 *
 *  For buffered devices with known state the answer will be correct.
 *
 *  In other cases a large number with bit 15 set (0xFFFF
 *  \ref AVAILABLE_UNKNOWN) is returned.
 *  \sa str2CheckOutSpace()
 */
uint16_t outSpace(const FILE * stream){
   if (!stream) return 0; // no stream no output space
   struct streamClassDescript_t *streamData = fdev_get_udata(stream);
   if (!streamData) return AVAILABLE_UNKNOWN; // no stream data no info
   str2OutSpace  outSpaceFunc = streamData->outSpaceFunc;
   if (!outSpaceFunc) return AVAILABLE_UNKNOWN; // no function no info
   return (*outSpaceFunc)(stream);
} // outSpace(FILE *)


/* Function checking the space available for output
 *
 *  A function of this type checks the requested space against the buffer
 *  space \ref  outSpace "available".
 *
 *  If the stream is the UART the function checks
 *  the requested space against the UART \ref UART_OUT_BUF_CAP "buffer size"
 *  or available \ref uartOutSpace() "space". Return values are 0, 2 or 4.
 *
 *  If the stream is the LAN respectively the currently handled IP
 *  connection the function checks the requested space against the actual
 *  connection's
 *  \ref lanStreamsOutSpace() "remaining output space". Return values are 0, 3
 *  or 4.
 *
 *  For an unknown device 1 is returned and for no (NULL) device 5.
 *
 *  \param reqSpace the number of bytes requested for output in one complete
 *         step
 *  \param stream the stream
 *  \return 0: is available. I.e. go ahead with planned output.<br />
 *       1: may be available, but can't be determined for
 *          the current (standard) output device. I.e. blindly go ahead or
 *          give up;<br />
 *       2: is not available, but may become (automatically in the future,
 *          i.e. yielding / waiting is OK;<br />
 *       3: is not available, but may become by own activity (flushing,
 *          sending). I.e. just yielding / waiting would be endless.
 *       4: will never be available. I.e. give up or request less space.
 *  \sa outSpace(stdout)
 *  \sa PT_YIELD_OUT_SPACE
 */
uint8_t checkOutSpace(const uint16_t reqSpace, const FILE * stream){
   if (!stream) return 4;   // no stream no output space
   if (!reqSpace) return 0; // always space for 0 bytes
   struct streamClassDescript_t *streamData = fdev_get_udata(stream);
   if (!streamData) return 1; // no stream data no info, i.e. may be
   str2CheckOutSpace checkOutSpaceFunc = streamData->checkOutSpaceFunc;
   if (!checkOutSpaceFunc) { // no check function
      str2OutSpace  outSpaceFunc = streamData->outSpaceFunc;
      if (!outSpaceFunc) return 1; // no space function no info, i.e. may be
      uint16_t space = (*outSpaceFunc)(stream);
      return reqSpace > space ? 3 : 0;  // may get avail. by own action : is
   } // no check function

#if DEBUG_TEL > 2
   if (stream == &lanStreams) {
      //           0123456789.123456789 .
      char tS[] = " ## lanC+ ... ";
      threeDigs(&tS[10], reqSpace);
      uartPutChars(tS, 14);
   }
#endif

   return (*checkOutSpaceFunc)(reqSpace, stream);
} // strCheckOutSpace(uint16_t, FILE *)


/*  Write a RAM string to output
 *
 *  This is the common function implementing the standard behavior. It
 *  delegates all else cases (late binding) to the stream type specific
 *  function if such is set in the stream (type)
 *  \ref streamClassDescript_t "specific data".
 *
 *  For this function's behavior see the
 *  \ref str2PutS() "description" in its
 *  \ref str2PutS() "pointer / type".
 */
int putSt(char * src, const FILE * stream){
   if (!stream) return 0;   // no stream no output
   if (!src) return 0;     // no string no output
   struct streamClassDescript_t * const streamData = fdev_get_udata(stream);
   if (streamData) {
      str2PutS putSfunc = streamData->putSfunc;
      if (putSfunc) return (*putSfunc)(src, stream);
   } // extra data
   const uint16_t retVal = fputs(src, (FILE*)stream); // no specialised function
   return retVal & 0x8000 ? 0 : retVal;  // "big" / negative (EOF) as 0
} // putSt

/*  Function writing a flash memory string to output
 *
 *  This is the common function implementing the standard behavior. It
 *  delegates all else cases (late binding) to the stream type specific
 *  function if such is set in the stream (type)
 *  \ref streamClassDescript_t "specific data".
 */
int putSt_P(char const * src, const FILE * stream){
   if (!stream) return 0;   // no stream no output
   if (!src) return 0;     // no string no output
   struct streamClassDescript_t *streamData = fdev_get_udata(stream);
   if (streamData) {
      str2PutS_P putS_P_func = streamData->putS_P_func;
      if (putS_P_func) return (*putS_P_func)((char *)src, stream);
   } // extra data
   const uint16_t retVal = fputs_P(src, (FILE *)stream); // no specialised function
   return retVal & 0x8000 ? 0 : retVal;  // "big" / negative (EOF) as 0
} // putSt_P


INFLASH(char const stackP[]) = "\n * * * Stack[";

/*  Log the stack frame to stdout
 *
 *  \param maxBytes the max. number of bytes logged including the stack frame
 */
void logStackS(uint8_t maxBytes){
   uint8_t * frame = (uint8_t *)(&frame + 2);
   if (!maxBytes) return;
   putSt_P(stackP, stdout);
   char tmp[5];
   fourHexs(tmp, (uint16_t)frame); tmp[4] = 0;
   putSt(tmp, stdout);
   tmp[0] = ']'; tmp[2] = '='; tmp[1] = tmp[3] = ' ';
   putSt(tmp, stdout); tmp[2] = ',';
   while(maxBytes) {
      twoHexs(tmp, *frame);
      if (++frame == (uint8_t *)(RAMEND +1)) maxBytes = 1;
      if (! (--maxBytes)) {
         tmp[2] = tmp[3] ='\n';
      }
      putSt(tmp, stdout);
   } // while
} // logStackS(uint8_t)


//-------------------------------------  nul streams  ------------------------


/* Open the nul streams
 *
 *  This function sets the standard streams stdin/out/err to the
 *  \ref nulStreams.<br />
 *  This "switching" will be rescinded by \ref streamsClose() .
 */
void nulStreamsOpen(void){ stdout = stderr = stdin  = &nulStreams; }

/* The buffer space available for nul output
 *
 *  This function always says 4095.
 */
uint16_t nulStreamsOutSpace(FILE * streams){ return 4096; }

/* Check the space available nul streams output
 *
 * This function always returns 0 i.e. available.
 */
uint8_t nulCheckOutSpace(uint16_t reqSpace, FILE * streams){ return 0; }

/* The number of characters buffered from nul input
 *
 *  This function returns always 0 as there will be no input from nul.
 */
uint16_t nulInBufferd(FILE * streams){ return 0; }

/* Get one byte from nul stream input
 *
 *  \return EOF (always)
 */
int nulGetC(void){ return EOF; }


/* Put one byte to nul stream output
 *
 *  This function never blocks and does nothing. All other
 *  output functions to nul do have no effect, also.
 *
 *  \param c the character to be output
 */
void nulPutC(char c){  }

/* Put one byte to nul stream output
 *
 *  \param c the character to be output
 *  \param stream unused, irrelevant (signature required by stdio)
 *  \return c always (as success, even if nothing is done)
 */
int nulPutChar(char c, FILE *stream){ return c; }

/*  Get one byte from nul stream input (always empty, EOF) */
int nulGetChar(FILE *stream) { return EOF; }

/** Put a string (some characters) to nul stream output
 *
 *  This function never blocks. It returns 1 no matter the parameters and does
 *  nothing else.
 *
 *  \param src the characters to be output (not NULL !)
 *  \param stream the current Ethernet connection as stream (not used)
 *  \return the number of characters output
 */
int nulPutSt(char * const src, const FILE * stream){ return src ? 1 : 0; }

/* The stream "device" acting as nul streams */
FILE nulStreams = FDEV_SETUP_STREAM(nulPutChar, nulGetChar,
                                                          _FDEV_SETUP_RW );
struct streamClassDescript_t nulStreamsData = {
      .inBufferdFunc = (str2InBufferd)nulInBufferd,
      .outSpaceFunc =  (str2OutSpace)nulStreamsOutSpace,
      .checkOutSpaceFunc = (str2CheckOutSpace)nulCheckOutSpace,
      .putSfunc  = (str2PutS)nulPutSt,
      .putS_P_func = (str2PutS_P)nulPutSt
};


//----------------------------------------------------------------------

/* The stream "device" connected to the serial IO / UART(0)  */
FILE serStreams = FDEV_SETUP_STREAM(serPutChar, uartGetChar,
                                                          _FDEV_SETUP_RW );
struct streamClassDescript_t serStreamsData = {
      .inBufferdFunc = (str2InBufferd)serInBufferd,
      .outSpaceFunc =  (str2OutSpace)serStreamsOutSpace,
      .checkOutSpaceFunc = (str2CheckOutSpace)serCheckOutSpace,
      .putSfunc  = (str2PutS)uartPutSt,
      .putS_P_func = (str2PutS_P)uartPutSt_P
};

/* The stream "device" connected to the Ethernet (uIP)  */
FILE lanStreams = FDEV_SETUP_STREAM(lanPutChar, lanGetChar, _FDEV_SETUP_RW );

struct streamClassDescript_t lanStreamsData = {
      .inBufferdFunc =  (str2InBufferd)lanInBufferd,
      .outSpaceFunc =   (str2OutSpace)lanStreamsOutSpace,
      .checkOutSpaceFunc = (str2CheckOutSpace)lanCheckOutSpace,
      .putSfunc  = (str2PutS)lanPutSt,
      .putS_P_func = (str2PutS_P)lanPutSt_P
};


static FILE * initStreams;

struct streamClassDescript_t bufLogStreamsData = {
      .inBufferdFunc = (str2InBufferd)bufLogInBufferd,
      .outSpaceFunc =  (str2OutSpace)bufLogStreamsOutSpace,
      .checkOutSpaceFunc = (str2CheckOutSpace)bufLogCheckOutSpace,
      .putSfunc  = (str2PutS)bufLogPutSt,
      .putS_P_func = (str2PutS_P)bufLogPutSt_P
};


/*  Initialise the standard streams
 *
 *  This function is usually called once at system start up.
 *
 *  \param initStream pointer to initial stream for standard streams,
 *         usually & \ref serStreams  or & \ref lanSteams
 */
void initStdStreams(FILE * initStream){
   fdev_set_udata(&lanStreams, &lanStreamsData);
   fdev_set_udata(&serStreams, &serStreamsData);
   fdev_set_udata(&nulStreams, &nulStreamsData);
   fdev_set_udata(&bufLogStreams, &bufLogStreamsData);

   if (!initStream)  {
      initStream = initStreams ? initStreams : &nulStreams;
   }
   initStreams = stdin = stdout = stderr = initStream;
} // initStdStreams(FILE *)

/* Switch the standard streams to another stream
 *
 *  This function sets the standard streams stdin/out/err to the
 *  \c toStream if (and only if) not yet so set.<br />
 *  If already set so or if \c toStream is NULL nothing is done.
 *
 *  \param toStream pointer streams to switch the  for standard streams to
 *  \return 0: toStream is NULL; 1: stdIO is or was set to toStream
 */
uint8_t switchStdStreams(FILE * toStream){
   if (!toStream) return 0;
   stdout = stderr =  stdin  = toStream;
   return 1;
} // switchStdStreams(FILE *)

/*  Close other standard streams
 *
 *  This function just reverses the (stream switching) effect of
 *  \ref switchStdStreams().
 *
 *  It sets all standard streams (back) to their
 *  \ref initStdStreams "initial standard" stream(s).<br />
 */
void streamsClose(void) {
   stdout = stderr = stdin = initStreams;
} // streamsClose()


/* Write a character to \c stdout */
void  stdPutC(const char c){
#if USE_AVR_GCC_FILE
   ((FILE *)stdout)->put(c, stdout);
#else
   if (SER_IS_STDOUT) { // optimise for UART 0
      serPutChar(c, NULL);
   } else if (LAN_IS_STDOUT) {
      lanPutC(c);
   }  else {
      fputc(c, stdout);
   }
#endif
} // stdPutC(char)


/* Open the serial (UART) streams
 *
 *  This function sets the standard streams stdin/out/err to the
 *  \ref serStreams if (and only if) not yet so set.<br />
 *  This "switching" will be rescinded by  \ref serStreamsClose() .
 */
void serStreamsOpen(void){
   stdout = stderr = stdin = &serStreams;
} // serStreamsOpen()


//----------------------------  LAN communication as stream -----------------

/* Initialise the Ethernet / LAN streams
 *
 *  This function initialises the \ref lanStreams for usage with
 *  \ref uip_appdata newly received or to be sent.
 */
void lanStreamsInit(void){
#if defined(arduMega2560)
   // no network dummy
#elif defined(arduinoUno)
   // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   lanStreamsOpen();
   uip_conn->readIndex = uip_conn->writeIndex = 0;
#endif // has network
} // lanStreamsInit()

/* Set the Ethernet / LAN streams to echo input
 *
 *  This function should be called after
 *  \ref lanStreamsInit() "inititialising" the \ref lanStreams and before
 *  the first output. <br />
 *  The received data will then be set as beginning of the data to be sent.
 *  This will be a 1:1 echo if nothing else is output.
 */
void lanStreamsEcho(void){
#if defined(arduMega2560)
   // no network dummy
#elif defined(arduinoUno)
   // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   uip_conn->writeIndex = uip_datalen();
#endif // has network
} // lanStreamsEcho()

/* Open the Ethernet / LAN streams
 *
 *  This function does the same as \ref lanStreamsInit() and additionally
 *  sets the standard streams stdin/out/err to the \ref lanStreams. <br />
 *  This "switching" will be rescinded by  \ref lanStreamsClose() .
 */
void lanStreamsOpen(void){
   stdout = stderr = stdin = &lanStreams;
} // lanStreamsOpen()

/* Sends the LAN streams data (over the Ethernet)
 *
 *  This function \ref uipSend() "sends" the LAN streams data written so far
 *  and \ref lanStreamsClose "closes" the LAN streams
 */
void lanStreamSend(void){
#if defined(arduMega2560)
   // no network dummy
#elif defined(arduinoUno)
   // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   if ( uip_conn->writeIndex ) {
      uip_send(uip_appdata, uip_conn->writeIndex);
   }
#endif // has network
   streamsClose();
} // lanStreamsSend()

/* The buffer space available for Ethernet / LAN streams output */
uint16_t lanStreamsOutSpace(FILE * streams) {
#if defined(arduMega2560)
   return 0; // no network dummy
#elif defined(arduinoUno)
   return 0; // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   return UIP_APPDATA_SIZE - uip_conn->writeIndex;
#endif // has network
} // lanStreamsOutSpace()

/*  Check the space available Ethernet / LAN streams output */
uint8_t lanCheckOutSpace(uint16_t reqSpace, FILE * streams){
#if defined(arduMega2560)
   return 4; // no network dummy
#elif defined(arduinoUno)
   return 4; // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   if (reqSpace > UIP_APPDATA_SIZE) return 4; // to large : never
   uint16_t space = UIP_APPDATA_SIZE - uip_conn->writeIndex;

#if DEBUG_TEL > 2
   //           0123456789.123456789 .
   char tS[] = " ## lanCo ... > ...\n";
   threeDigs(&tS[10], reqSpace);
   threeDigs(&tS[16], space);
   uartPutChars(tS, 20);
#endif

   return reqSpace > space ? 3 : 0;
#endif // has network
} // checkLANOutSpace(uint16_t)   */


/* The number of characters buffered from LAN (stream) input */
uint16_t lanInBufferd(FILE * streams){
#if defined(arduMega2560)
   return 0; // no network dummy
#elif defined(arduinoUno)
   return 0; // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   if ( uip_conn->readIndex >= uip_datalen()) return 0;
   return uip_datalen() -  uip_conn->readIndex;
#endif // has network
} // lanInBufferd()

/*  Get one byte from LAN stream input
 *  \return a character 0..255 or EOF
 */
int  lanGetC(void){
#if defined(arduMega2560)
   return EOF; // no network dummy
#elif defined(arduinoUno)
   return EOF; // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   if ( uip_conn->readIndex >= uip_datalen()) return EOF;
   const uint8_t c = * ((char *)uip_appdata +  uip_conn->readIndex);
   uip_conn->readIndex++;
   return (int)c;
#endif // has network
} //  lanGetC(void)

/*  Get one byte from LAN stream input
 *  \return a character 0..255 or EOF
 */
int lanGetChar(FILE *stream){ return lanGetC(); }


/*  Put one byte to LAN stream output */
void  lanPutC(char const c){
#if defined(arduMega2560)
   // no network dummy
#elif defined(arduinoUno)
   // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   if ( uip_conn->writeIndex >= UIP_APPDATA_SIZE) return;
   * ((char *)uip_appdata +  (uip_conn->writeIndex++)) = c;
#endif // has network
} // lanPutC(const char)

/*  Put one byte to LAN stream output
 *
 *  This function does exactly the the same as the preferable
 *  \ref  lanPutC() except for the extra parameter of type FILE * that is
 *  for sake of the standard I/O usage.
 *
 *  \param c the character to be output
 *  \param stream (passed by stdio output) totally ignored by this function
 *  \return c in case of success, EOF if output buffers full.
 */
int lanPutChar(char const c, FILE *stream){
#if defined(arduMega2560)
   return EOF; // no network dummy
#elif defined(arduinoUno)
   return EOF; // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   if ( uip_conn->writeIndex >= UIP_APPDATA_SIZE) return EOF;
   * ((char *)uip_appdata +  (uip_conn->writeIndex++)) = c;
   return (int)c;
#endif // has network
} // lanPutChar(char, FILE*)


/* Put a RAM string (some characters) to LAN stream output
 *
 *  \param src the characters to be output (not NULL !)
 *  \param stream the current Ethernet connection as stream (not used)
 *  \return the number of characters output
 */
int lanPutSt(char * src, const FILE * stream){
   if (!src) return 0;
   char c = *src;
   if (!c) return 0;
#if defined(arduMega2560)
   return 0; // no network dummy
#elif defined(arduinoUno)
   return 0; // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   const uint16_t space = UIP_APPDATA_SIZE - uip_conn->writeIndex;
   if (!space) return 0;
   uint16_t retVal = 0;

   while (retVal < space && c ) {
      * ((char *)uip_appdata +  uip_conn->writeIndex) = c;
      ++ uip_conn->writeIndex;
      ++retVal;
      c = *(++src);
   } // while
   return retVal;
#endif // has network
}  //  lanPutSt(char *, FILE * )

/*  Put some characters from program space to LAN stream output
 *
 *  \param src the characters to be output in flash memory (not NULL !)
 *  \param stream the current Ethernet connection as stream (not used)
 *  \return the number of characters output
 */
int lanPutSt_P(char * src, const FILE * stream){
   if (!src) return 0;
   char c =  pgm_read_byte(src);
   if (!c) return 0;
#if defined(arduMega2560)
   return 0; // no network dummy
#elif defined(arduinoUno)
   return 0; // no network dummy
#else // defined(weAut_00) || defined(weAut_01)
   uint16_t space = UIP_APPDATA_SIZE - uip_conn->writeIndex;
   if (!space) return 0;
   uint16_t retVal = 0;

   while (retVal < space && c ) {
      * ((char *)uip_appdata +  uip_conn->writeIndex) = c;
      ++uip_conn->writeIndex;
      ++retVal;
      c = pgm_read_byte(++src);
   } // while
   return retVal;
#endif // has network
} //  lanPutSt_P(char *, FILE *)

