/*
 * Copyright (c) 2011 Albrecht Weinert, Bochum
 *
 * This file is part of weAutSys.
 * 
 * Author: Albrecht Weinert <albrecht@a-weinert.de>
 *
 * $Id: syst_threads.c 2 2017-01-25 16:49:03Z albrecht $
 */

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup sysSer
 * @{
 *
 * \addtogroup sysThr
 * @{
 *
 * \file syst_threads.c
 *
 * weAutSys implementation of System threads
 *
 * This is system software and must not be modified for user or
 * application programs.
 *
 * Please find detailed description of the variables, functions etc.
 * in the module \ref sysThr "system threads" and in the include file
 * syst_threads.h.
 *
 * This file is part of weAutSys &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

/** @} */ // doxygen addtogroup sysThr
/** @} */ // doxygen addtogroup sysSer
/** @} */ // doxygen addtogroup weAutSys

#include <util/atomic.h>

#include "we-aut_sys/common.h"
#include "we-aut_sys/utils.h"
#include "we-aut_sys/ll_system.h"
#include "we-aut_sys/proc_io.h"
#include "we-aut_sys/timing.h"
#include "we-aut_sys/system.h"
#include "we-aut_sys/syst_threads.h"
#include "we-aut_sys/log_streams.h" // for debug only
#include "we-aut_sys/streams.h"
#include "we-aut_sys/network.h"
#include "we-aut_sys/persist.h"
#include "we-aut_sys/smc.h"

#include "we-aut_sys/ntp.h"
#include "uip/apps/dhcpc.h"
#include "uip/apps/resolv.h"

uint8_t   msIntTickCopy, msIntTickDone;
uint16_t  msSystClockCount, msAbsClockCount;
uint8_t   msTo10Count, msTo100Count;
uint8_t   sys100msPeriodFlag, sys1sPeriodFlag;

struct pt ptSys100msThread, ptSys1sThread;

/* Initialise the threads  */
void initThreads(void){
   PT_INIT(&ptSys100msThread); // initialise system's 100ms thread
   PT_INIT(&ptSys1sThread);    // initialise system's 1s thread
   msIntTickCopy = msIntTickDone = msTo100Count = msTo10Count
                                 = sys100msPeriodFlag = sys1sPeriodFlag = 0;
   // de-register user threads
   resetAppThread(app1msThread);
   resetAppThread(app10msThread);
   resetAppThread(app100msThread);
   resetAppThread(app1sThread);
   resetAppThread(appSerInpThread);
   resetAppThread(appBgTskThread);
} // initThreads()


/* The milliseconds beat system function
 *
 * Does the basis of all time keeping,
 * sets the run flags for other (periodic) system cycle threads.
 */
void sys1msThread(){
   if ((msIntTickCopy = msIntTick) == msIntTickDone) return;

   // un-handled ticks have occurred
#if USE_TESTPINS == 4
   setTestPin1(ON); // TEST: measure ms tick time (to ..
                   // .. catch long running SMC via SPI functions)
#endif

   // unsigned overflow is correct modulo arithmetic for the difference
   const uint8_t msIntDif = msIntTickCopy - msIntTickDone; // the difference
   if (!abortCommand) { // run (application/user cyclic threads only on run)
      app1msThread.flag += msIntDif ; // flag for application 1ms thread
   } // run

   // flag for app10msThread
   uint8_t tm8p =  msTo10Count + msIntDif;
   if (tm8p == 4) { // h:m:s:xx4
#if HAVE_SMC
      if (smcState.flag & (0x0F & FL_IMP)) smcThreadF(); // SMC action thread
#endif
   } else if (tm8p == 2 || tm8p == 7 ) { // every 5 ms at xx2|7
#if HAVE_ETHERNET
      // Ethernet receive / poll  here on every 10 ms (11.09.2012 5)
      // History: This was originally every 100ms. Making it here (every 10ms)
      // gave significant NTP improvement.
      if (!(networkNotReady & 0xF6)) { // !(ENC not ready or link down)
         // see uIP 1.0 Module Documentation (page 42)
         uip_len = deviceDriverPoll(); // check if package was received
         if(uip_len) { // package received in uip_buf
            // Check if there ...
            if(BUF2EH->type == HTONS(UIP_ETHTYPE_IP)) {
# if DEBUG_ETH > 2
               bufLogMark("EthRecIP ", 9);
# endif
               uip_arp_ipin(); // may be used as new ARP entry
               uip_input();
               if(uip_len) { // if there is a package to send send it now
                  uip_arp_out();
                  deviceDriverSend();
               } // there is a package to send
            } else if(BUF2EH->type == HTONS(UIP_ETHTYPE_ARP)) {
# if DEBUG_ETH > 2
               bufLogMark("EthRecArp", 9);
# endif
               uip_arp_arpin();  // Process ARP request
               if(uip_len) {
                  deviceDriverSend(); // if there is a package to send send it now
               }
            } else {
# if DEBUG_ETH > 2
               bufLogMark("EthRec???", 9);
# endif
            }
         } // if package received etc.
      } // ENC  ready
#endif // have Ethernet
   } else if (tm8p >= 10) { // over ten ms
      u8div_t q_r = divByVal10(tm8p);
      if (!abortCommand) { // run (application cyclic threads only on run)
         app10msThread.flag += q_r.quot; // flag  application 10ms thread
      } // run
      tm8p = q_r.rem; //  -= 10;
   } // over ten ms
   msTo10Count = tm8p;


   // flag for sys100msThread
   tm8p = msTo100Count + msIntDif;
   if (tm8p >= 100) {  // over 100 ms
      if (tm8p >= 200) { // should not happen
         sys100msPeriodFlag += 2; // flag for 100ms system thread
         tm8p -= 200;
      } else {
         ++sys100msPeriodFlag; // flag for 100ms system thread
         tm8p -= 100;
      } // just 100..199 OK
   } // over 100 ms
   msTo100Count = tm8p;

   msTimers.end.v32 += msIntDif;  // ms ticks counted here and only here
   // setTestPin0(msTimers.end.v8[0] & 1); // show last digit of msCnt

   // handle all running timers
   for(;;) {
      struct timer_t * ct = msTimers.next;
      if (ct == NULL) break;
      if (leU32ModAr(ct->end.v32 , msClock())) { // ct->end <= msCount
         msTimers.next = ct->next;
         ct->next = NULL;
         ct->state = TIMER_LAPSED;
      } else break; // while
   } // for;;

   // flags for sys1sThread
   uint16_t tm16p = msSystClockCount +  msIntDif;
   if (tm16p  >= 1000) {
      tm16p  -= 1000;   // reset by 1000
      ++secDurTimers.end.v32; // run since time in seconds
      ++sys1sPeriodFlag; // flag 1s system thread to run (timers)
   }
   msSystClockCount = tm16p;

   tm16p = msAbsClockCount +  msIntDif;
   if (tm16p  >= 1000) {
      tm16p  -= 1000;   // wrap around 1000
      ++secDatTimers.end.v32; // local time in seconds since  1st 2008 UTC.
      sys1sPeriodFlag |= ABSsecADJ; // flag 1s system thread to run (date, timers)
   }
   msAbsClockCount = tm16p;

   procDIcyc();  // DI (AI) processing here (can do 8 AI in one ms if needed)

   msIntTickDone = msIntTickCopy;   // done with the difference

   if (isRTSflowC && !(PINB & 2) && uartOutBufWi != uartOutBufRi) {
      ATOMIC_BLOCK(ATOMIC_RESTORESTATE) { // keep or get it going
         if (SER_SENDBUF_NOT_EMPTY  // the buffer (array) is not empty and the
                 && UART_CAN_SEND) { // UARTs transmitter register is
            FROM_BUFFER2UART()
            // if one byte can be sent a second one may also be sent
            if (SER_SENDBUF_NOT_EMPTY  // the buffer (array) is not empty and
                    && UART_CAN_SEND) { // the UARTs transmitter register is
               FROM_BUFFER2UART()
            }
         }
      } // atomic
   } // something to send under RTS output flow control

#if USE_TESTPINS == 4
   setTestPin1(OFF); // TEST: measure ms thread time; was 8, 10 , 12 s
#endif
} // sys1msThread()

#if HAVE_ETHERNET
static uint8_t uIPperiodic; // we have 10 connections max, go round them in 1s
#if UIP_UDP
static uint8_t uIPudpPeriodic; // same for UDP conns if any
#endif
#endif // have Ethernet

/* The 100 ms system thread
 *
 * Sets the flag for the application 100 ms cycle.
 *
 * Does the main polling of the Ethernet driver ENC28J60 respectively
 * the Ethernet stack uIP.
 */
ptfnct_t sys100msThread(){
   PT_BEGIN(&ptSys100msThread);
   PT_WAIT_UNTIL(&ptSys100msThread, sys100msPeriodFlag); // wait if not flag
   if (!abortCommand) {
      app100msThread.flag += sys100msPeriodFlag; // flag for application 100ms
      wdt_reset();  // trigger watchdog (for a 1/4s WD two fails are deadly)
   } else if (abortCommand == PLCstop) {
      wdt_reset();  // trigger watchdog (but not for all other aborts)
   }
   sys100msPeriodFlag = 0; // reset own flag (never never forget!)

   if(bufLogConsIsUART && SER_SENDBUF_EMPTY) { // copy log to UART
      uartPutLogBuf();
   } // copy log to UART
#if HAVE_SMC
   smcThreadF();  // check SMC (every 10ms with pending action above)
   PT_YIELD(&ptSys100msThread); // make a touch break
#endif

#if HAVE_ETHERNET
   linkState(); // check link
   if (networkNotReady & 0xF6) { // network not ready
      uint8_t isOrWasDown = networkNotReady & 0x06;
      if (isOrWasDown == LNC_WDN) { // link coming up
         lanComStdInit();
      } else  if (isOrWasDown == LNC_IDN) { // link going down
         networkInit();
      }
      PT_EXIT(&ptSys100msThread); // ENC or link not (stable) : out here
   } // network not ready

   uint8_t repeat = (UIP_CONNS + 15) / 15;
   do { // touch TCP connections loop
#if UIP_CONNS < 10
      if (++uIPperiodic >=  10) {
         uIPperiodic = 0;
      } else if (uIPperiodic >= UIP_CONNS) break; // pause for m..10
#else
      if (++uIPperiodic >=  UIP_CONNS) uIPperiodic = 0;
#endif
      uip_periodic(uIPperiodic); // touch conn [uIPperiodic]
      if(uip_len) { // uip_periodic resulted in data to be sent
         uip_arp_out();
         deviceDriverSend();
      }
   } while (--repeat); //  touch TCP connections loop


#if UIP_UDP
   PT_YIELD(&ptSys100msThread); // make a touch break
   repeat = (UIP_UDP_CONNS + 15) / 15;
   do { // touch UDP connections loop
#if UIP_CONNS < 10
      if (++uIPudpPeriodic >=  10) {
         uIPudpPeriodic = 0;
      } else if (uIPperiodic >= UIP_UDP_CONNS) break; // pause for m..10
#else
      if (++uIPudpPeriodic >=  UIP_UDP_CONNS) uIPudpPeriodic = 0;
#endif
      uip_udp_periodic(uIPudpPeriodic);
      if(uip_len) { // uip_udp_periodic() resulted in data to be sent
         uip_arp_out();
         deviceDriverSend();
      }
   } while (--repeat); //  touch TCP connections loop
#endif /* UIP_UDP */
#endif // have EThernet
   PT_END(&ptSys100msThread)
} // 100 ms system thread.


static struct timer_t dstTimer; // day light saving (switch / check) timer
//                          0123456789.12345
INFLASH(char const dstDebug[]) = " # # DST debug ";
#if DEBUG_DHCP > 2
INFLASH(char const tlpDebug[]) = " # # tsL debug ";
INFLASH(char const tnxDebug[]) = " # # tsN debug ";
#endif

/* The one second system thread
 *
 * Sets the flag for the application 1s cycle
 */
ptfnct_t sys1sThread(){
   PT_BEGIN(&ptSys1sThread);
   if (!sys1sPeriodFlag) PT_EXIT(&ptSys1sThread); // non flagged: out here

   // absolute date time seconds increment actions -------------------------
   if (sys1sPeriodFlag & (ABSsecSET | ABSsecADJ)) {

      // seconds timers handling time date oriented
      for(;;) {
         struct timer_t * ct = secDatTimers.next;
         if (ct == NULL) break; // end of list
         if (ct->end.v32 <= secTime308Loc()) { // <= we don't care >= 78 years
            secDatTimers.next = ct->next;
            ct->next = NULL;
            ct->state &= TIMER_TYPE_MASK; // | TIMER_LAPSED;
         } else { // no lapse condition
            break; // if no more lapsed -> done
         } // no lapse condition
      } // for;;

      // time date handling here is check DST after jumps in time or DST-timer
      for (;(sys1sPeriodFlag & ABSsecSET) || timerLapsed(dstTimer);) { // DST
         uint8_t ruleNA = zoneOffsetZone;  //
         if (ruleNA <= 0x1C) {
            ruleNA = 0; // 0 : EU rule
#   if DEBUG_NTP  >= 3  //   0123456789.123456
                    //     0123456789.123456
               bufLogMark(" DST EU rules ", 14);
#   endif
         } else if (ruleNA < 0xB9 || ruleNA >0x8F) {
            break; // not EU and outside Northern America rule range
         } else {
#   if DEBUG_NTP  >= 3  //   0123456789.123456
                    //     0123456789.123456
               bufLogMark(" DST US+ rules ", 15);
#   endif
         }

         uint32_t sec308UTC = secTime308UTC();
         uint32_t divRem;
         uint16_t days = (uint16_t)div32by24(&divRem, sec308UTC, SECONDS_IN_DAY);

         uint16_t dinY;
         uint8_t marchYear =  getMarchYearByDays(&dinY, days);

         days -= dinY; // now March 1st current year.
         uint32_t sec308UTCMarch1 = mul16with17(days,
                                                   (uint16_t)SECONDS_IN_DAY);
         uint32_t secInYear = sec308UTC - sec308UTCMarch1;

         const dst_rule_year_t * currentRuleValuesP = getDSTrule(marchYear);
         uint32_t tON;
         uint32_t tOFF;
         if (ruleNA) { // north America
            tON = pgm_read_dword(&currentRuleValuesP->secondSunMarch);
            tOFF = pgm_read_dword(&currentRuleValuesP->firstSunNovember);
         } else { // EU
            tON = pgm_read_dword(&currentRuleValuesP->lastSunMarch0100offset);
            tOFF = pgm_read_dword(&currentRuleValuesP->lastSunOctober0100offset);
         } // EU

#if DEBUG_NTP >= 3     //   0123456789.123456
         bufLogDecHex_P(dstDebug,  marchYear, secTime308Loc());
         bufLogHexHex_P(dstDebug,  0, sec308UTC);
         bufLogHexHex_P(dstDebug,  1, secInYear);
         bufLogHexHex_P(dstDebug,  2, tON);
         bufLogHexHex_P(dstDebug,  3, tOFF);
#endif

         if (secInYear < tON) { // not yet
            setDST(0); // before
            startTimerAbs(&dstTimer, sec308UTCMarch1
                  + tON + combinedOffset, sec_dat_TIMER_TYPE);
#if DEBUG_NTP  >= 3   //   0123456789.123456
            bufLogHexHex_P(dstDebug, dstTimer.state, sec308UTCMarch1
                                                  + tON + combinedOffset);
            bufLogMark(" DST off before ", 16);
#endif
            break;
         }  // not yet / before

         if (secInYear >= tOFF) { // no more / after
            setDST(0); // after
            startTimerAbs(&dstTimer, sec308UTCMarch1 + SECONDS_IN_YEAR,
                  sec_dat_TIMER_TYPE);
#if DEBUG_NTP >= 3   //   0123456789.123456
            bufLogHexHex_P(dstDebug, dstTimer.state, sec308UTCMarch1
                  + SECONDS_IN_YEAR);
            bufLogMark(" DST off after ", 15);
#endif
            break;
         } // no more / after DST
         // to here only within DST
         setDST(1);
         startTimerAbs(&dstTimer, sec308UTCMarch1
               +  tOFF + combinedOffset, sec_dat_TIMER_TYPE);
#if DEBUG_NTP  >= 3  //   0123456789.123456
         bufLogHexHex_P(dstDebug, dstTimer.state, sec308UTCMarch1
               +  tOFF + combinedOffset);
         bufLogMark(" DST on within ", 15);
#endif

         break; // pseudo if
      } // check DST

      sys1sPeriodFlag &= 0x07; // leave "since seconds" bits only
      if (!sys1sPeriodFlag) PT_EXIT(&ptSys1sThread); // nothing left to to
   } // sys1sPeriodFlag & (ABSsecSET | ABSsecADJ)


   // seconds since start increment actions --------------------------------

   if (!abortCommand) { // run (app cyclic threads only on run)
      app1sThread.flag += sys1sPeriodFlag; // flag  app 1s thread
   } // run


   //  handle the seconds (duration) timers here
   for(;;) {
      struct timer_t * ct = secDurTimers.next;
      if (ct == NULL) break; // end of list
      if (ct->end.v32 <= secClock()) { // <= we don't care >= 2076 yet
#if DEBUG_DHCP > 2
         bufLogHexHex_P(tlpDebug, ct->period.v8[0], ct->end.v32);
#endif
         secDurTimers.next = ct->next;
         ct->next = NULL;
         ct->state = sec_dur_TIMER_TYPE | TIMER_LAPSED;
#if DEBUG_DHCP > 2
         ct = secDurTimers.next;
         bufLogHexHex_P(tnxDebug, ct->period.v8[0], ct->end.v32);
#endif

      } else { // no lapse condition
         break; // no more lapsed -> done
      } // no lapse condition
   } // for;;

   sys1sPeriodFlag = 0;

#if HAVE_ETHERNET
   // below network stuff only
   if (networkNotReady & 0xF0) PT_EXIT(&ptSys1sThread); // ENC not ready

   uipSecTick(); // call uIP tick function every second
#endif // have Ethernet

   PT_END(&ptSys1sThread)
} // One second system thread


//---   user thread handling  -----------------------------------------


/* Register an user / application thread   */
p2ptFun registerAppThread(struct mThr_data_t * thread, p2ptFun threadFun){
   p2ptFun fOld = thread->fun;
   if (fOld != threadFun) {   // change to
      if (threadFun == NULL) {// to null
         thread->fun = NULL;
      } else {
         thread->fun = threadFun;
         PT_INIT(&(thread->pt));   // initialise the protothread
          if (fOld == NULL) { // it is new
             thread->flag = 0;  // reset tick flags for newly introduced
             thread->retC = PT_ENDED;
          }
      } // to non null
   } // change to
   return fOld;
}  // registerAppThread(mThr_data_t, p2ptFun)

/* Register and (force) initialise an user / application thread
 *
 *  This function will register \c threadF with \c thread unconditionally
 *  and init reset all.
 */
void initSetAppThread(struct mThr_data_t * thread, p2ptFun threadFun){
   if (threadFun == NULL) { // to null
      thread->fun = NULL;
   } else {
      thread->fun = threadFun;
   }
   PT_INIT(&(thread->pt));   // initialise the protothread
   thread->flag = 0;  // reset tick flags for newly introduced
   thread->retC = PT_ENDED;
} // initSetAppThread(struct mThr_data_t * thread, p2ptFun threadFun)



/* Register a user / application serial input processing thread  */
p2ptFunC registerAppSerInpThread(p2ptFunC  threadF){
   const p2ptFun fOld = registerAppThread(
                   (struct mThr_data_t *)&appSerInpThread, (p2ptFun)threadF);
   // further initialisation (not necessary yet)
   return (p2ptFunC)fOld;
} // registerAppSerInpThread(p2ptFunC)

/* Register a user / application background task thread  */
p2ptFun  registerAppBgTskThread(p2ptFun  thread){
   return registerAppThread(&appBgTskThread, thread);
} // registerAppBgTskThread(p2ptFun)
