// Copyright (c) 2012 Albrecht Weinert, Bochum *
// system.c 2 2017-01-25 16:49:03Z albrecht

/**
 *  \file system.c
 *
 *  weAutSys  implementation of system calls and services
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of weAutSys &nbsp;
 *           <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */

// the next two includes are rubbish for Eclipse Indigo .h troubles only
//#include <avr/io.h>
//#undef _AVR_IOXXX_H_
//#  include <avr/iom1284p.h>
// delete three # above as soon as Indigo can handle nested .h files

#include <avr/interrupt.h>

#include "we-aut_sys/common.h"
#include "arch/config.h"
#include "we-aut_sys/utils.h"
#include "we-aut_sys/ll_system.h"
#include "we-aut_sys/system.h"
#include "we-aut_sys/streams.h" // for debug only
#include "we-aut_sys/log_streams.h" // for debug only
#include "we-aut_sys/timing.h"
#include "we-aut_sys/uart0.h"

#if HAVE_ETHERNET
# include "we-aut_sys/ntp.h"
#endif
#include "we-aut_sys/network.h"
#include "we-aut_sys/spi.h"
#if HAVE_SMC
# include "we-aut_sys/smc.h"
#endif

#include <string.h>
#include "pt/pt.h"  // for sake of Eclipse only

// ----------- common Network state -- (also for platforms w/o Ethernet) ----

struct uip_eth_addr actMACadd;

/*  Parse a MAC address
 *
 *  \param macAddr pointer to the result (eth_addr_t is a structure
 *             containing an array type); the result is changed only
 *             if a MAC address could be parsed successfully
 *  \param s   the RAM string containing the digit(s)
 *  \param si  the first digit's index in s
 *  \return    the index of the first character after the parsed IP address
 *             in case of success or 0 else
 */
uint8_t parseMACadd(eth_addr_t * macAddr, char s[], uint8_t si){
   eth_addr_t tmp;
   si = parse2hex(&tmp.addr[0] ,s, si);
   if (!si) return 0;
   for (uint8_t i = 1; i <= 5; ++i) {
      if (s[si++] != ':') return 0;
      si = parse2hex(&tmp.addr[i], s, si);
      if (!si) return 0;
   }
   memcpy(macAddr, &tmp, 6);
   return si;
} //  parseMACadd(eth_addr_t * macAddr, char s[], uint8_t si)

/* Set the stack's MAC address
 *
 *  This sets the the Ethernet stack's i.e. \ref uip "uIP"'s
 *  MAC address but not the driver's ( ENC28J60).
 *
 *  Of course, the driver's MAC setting has to be same. This is done by
 *  \ref encSetMacAdd(),  \ref encInit() or \ref networkInit().
 */
void setMACadd(eth_addr_t * mac){
   memcpy(&actMACadd, mac, 6);
} // setMACadd(eth_addr_t *)


/*  Parse an ipV4 Ethernet address
 *
 *  \param ipAddr pointer to the result (ipaddr_t is an array type)
 *  \param s   the RAM string containing the digit(s)
 *  \param si  the first digit's index in s
 *  \return    the index of the first character after the parsed IP address
 *             in case of success or 0 else
 */
uint8_t  parseIpAdd (uip_ipaddr_t ipAddr, char s[], uint8_t si){
   uint8_t parsRes;
   si = parseByteNum(&parsRes, s, si);
   if (!si) return 0;
   uint8_t lB0 = parsRes; // low (big endian!) of ipAddr[0]

   if (s[si++] != '.') {
      // bufLogPutC('#');
      return 0;
   }

   si = parseByteNum(&parsRes, s, si);
   if (!si) return 0;
   uint8_t hB0 = parsRes; // high (big endian!) of ipAddr[0]

   if (s[si++] != '.') return 0;

   si = parseByteNum(&parsRes, s, si);
   if (!si) return 0;
   uint8_t lB1 = parsRes; // low (big endian!) of ipAddr[1]

   if (s[si++] != '.') return 0;

   si = parseByteNum(&parsRes, s, si);
   if (!si) return 0;

   ipAddr[0] = hB0 << 8 | lB0;
   ipAddr[1] = parsRes << 8 | lB1;
   return si;
} // parseIpAdd(ipaddr_t ipAddr, char s[], uint8_t si)


// -----------    Test LEDs ---------------------------------------

/* Status-LEDs    myEthernet   weAut_01     arduMega2560    arduinoUno
 * LED gn         C4           (LED19) B2   (yellow) B7     B5
 * LED rd         C5           (LED18) B4
 */

void setStatusLedGn(uint8_t state){
#if defined(GN_LED_PORT)
   if (state) {
      orPort(GN_LED_PORT, (1<<GN_LED_PIN));
   } else {
      andPort(GN_LED_PORT,  ~(1<<GN_LED_PIN));
   }
#endif
} // setStatusLedGn(uint8_t state)

void toggleStatusLedGn(void){
#if defined(GN_LED_PORT)
   xorPort(GN_LED_PORT, (1<<GN_LED_PIN));
#endif
} // toggleStatusLedGn()

void setStatusLedRd(uint8_t state) {
#if defined(RD_LED_PORT)
      if (state) {
         orPort(RD_LED_PORT, (1<<RD_LED_PIN));
      } else {
         andPort(RD_LED_PORT,  ~(1<<RD_LED_PIN));
      }
#endif
} // setStatusLedRd(uint8_t state)

void toggleStatusLedRd(void){
#if defined(RD_LED_PORT)
   xorPort(RD_LED_PORT, (1<<RD_LED_PIN));
#endif
} // toggleStatusLedRd()


// -----------    Test Pins ---------------------------------------
//
// Makros: _NO_TESTPINS_ (set if used otherwise)

/* Test-PINs    myEthernet   weAut_01
 * testPin_0    C2 (Pin18)   C1 (JP3 Pin1)
 * testPin_1    C3 (Pin19)   C0 (JP3 Pin4)
 */

void setTestPin0(uint8_t state){
#if !defined(_NO_TESTPINS_) && defined(TEST_PIN0_PORT)
   if (state) {
      orPort(TEST_PIN0_PORT, (1 << TEST_PIN0_PIN));
   } else {
      andPort(TEST_PIN0_PORT, ~(1 << TEST_PIN0_PIN));
   }
#endif
} // setTestPin0(uint8_t)

void setTestPin1(uint8_t state){
#if !defined(_NO_TESTPINS_) && defined(TEST_PIN1_PORT)
   if (state) {
      orPort(TEST_PIN1_PORT, (1 << TEST_PIN1_PIN));
   } else {
      andPort(TEST_PIN1_PORT, ~(1 << TEST_PIN1_PIN));
   }
#endif
} // setTestPin1(uint8_t)


INFLASH(char const systNam[]) = SYST_NAM;
INFLASH(char const systRev[]) = SYST_REV;
INFLASH(char const systDat[]) = SYST_DAT;
INFLASH(char const systMod[]) = SYST_MOD;
INFLASH(char const systAut[]) = SYST_AUT;
INFLASH(char const systCop[]) = SYST_COP;
INFLASH(char const systBld[]) = SYST_BLD;

//                          0 1 2 3456789.123456789z123456789d123456789v12 3
INFLASH(char const systWlc[])
                         = "\n\n\n * * *   Welcome to   weAutSys   * * * \n";
INFLASH(char const systBye[])
                         = "\n\n\n * * *   Farewell     weAutSys   * * * \n";

INFLASH(char const sepLoB[]) = "  (";
INFLASH(char const sepRcB[]) = ")  ";
//                                 01 2 3 4 5
INFLASH(char const l4nefeeds[]) = " \n\n\n\n";
INFLASH(char const bLF1[]) = " \n";
INFLASH(char const bLF2[]) = " \n\n";

INFLASH(char const systPlatf[]) = "Platform: " PLATFORM_S ;
INFLASH(char const systCPUFr[]) = "  CPU frequency " FCPU_S " Hz   \n\n";


INFLASH(char const  macAdd_is[]   ) = "  MAC address : ";
INFLASH(char const ip4Add_is[]    ) = "  IP4 address : ";
INFLASH(char const const encEthSta[]    ) = "  Link state  : ";
INFLASH(char const ipDefault[]    ) = "  IP4 defaults  ";
INFLASH(char const arpEntries[]   ) = "  ARP entries   ";
INFLASH(char const arpEmpty[]     ) = "no valid entries in ARP table \n\n";


// reports used by exemplary CLI thread
INFLASH(char const wdSetShort[])  = "watchdog set short   \n";
INFLASH(char const wdSetSharp[])  = "watchdog set sharp   \n";
INFLASH(char const wdSetLeni[])   = "watchdog set lenient \n";
INFLASH(char const wdSetNormal[]) = "watchdog set normal  \n";
//                                   0123456789.123456789z1 2


INFLASH(char const commWrong[]) = "\n * * *   Wrong command      : ";
INFLASH(char const commAmbig[]) = "\n * * *   Ambiguous command  : ";
INFLASH(char const optAmbig[] ) = "\n * * *   Ambiguous option   : ";
INFLASH(char const optWrong[] ) = "\n * * *   Wrong (unimplemented) option : ";
INFLASH(char const noUserCLI[]) = "\n * * *   No user command line interpreter (CLI) registered \n";


INFLASH(char const helpHeader[]) =  " * * *   Commands to  weAutSys               \n";
//                                   0123456789.123456789z123456789d12345
INFLASH(char const helpUserCm[]) =  " * * *   Application commands  - - -         \n";

// commands recognised by CLI (below)
INFLASH(char const comHelp[]   ) =  "help     simple list of commands and options   \n"
                                    "         type help -help for more information. \n"
FOLLOW_UP "  options are -system -application -options -verbose \n"
FOLLOW_UP "  -system -application: list system resp. application commands. \n"
FOLLOW_UP "  no option: list system and application commands. \n"
FOLLOW_UP "  -options:  list all possible options.    \n"
          "  -verbose:  list all commands and options \n"
FOLLOW_UP "  The option -help or -? works with every command. This would   \n"
          "  display the command specific help.  \n"
FOLLOW_UP "  The command line interpreter (CLI) parses commands and options \n"
FOLLOW_UP "  case insensitive and recognises un-ambiguous abbreviations. \n   \n\0";
//                           0123456
INFLASH(char const comWatchD[]) =   "watchdog [-abort | -sharp | -normal | -lenient] \n"
FOLLOW_UP "  According to the option given this command sets the watchdog   \n"
FOLLOW_UP "  timeout to 60ms (-abort), 120ms (-sharp), 500ms (-lenient) or  \n"
FOLLOW_UP "  to 250ms (-normal, no option or all else as default). A system \n"
FOLLOW_UP "  well designed and running should go on  - and be tested - with \n"
FOLLOW_UP "  -sharp even under SMC and Ethernet load.   \n   \n\0";
//                           0123456 89.12345678
INFLASH(char const comWDog[] )  =   "wdog    like watchdog. Type watchdog -help \n"
                                    "        for more information.    \n";

INFLASH(char const comBoot[] )    = "boot -reset  do a hard reset by watchdog     \n"
                              "boot -load   enter the serial bootloader     \n";
//                             0123456789.12345678.123456789.123456789v1234 5
INFLASH(char const comPLContrl[]) = "PLCcontrol [-stop | -start]  PLC cycles      \n";
INFLASH(char const comAutomation[])="Automation same as  PLCcontrol               \n";
INFLASH(char const comVersion[] ) = "version  show welcome and version            \n";
INFLASH(char const comRunInfo[] ) = "runInfo  running status                      \n";
INFLASH(char const comIPconfig[]) = "ipConfig show IP configuration               \n";
INFLASH(char const comEthAddr[] ) = "ethAddr  show address relations (ARP)        \n";
INFLASH(char const comIPdefAd[] ) = "ipDefAd [ad] show / set default IP address   \n";
INFLASH(char const comMacAddr[] ) = "macAddr [ad] show / set MAC address          \n";
INFLASH(char const comUART[] )    = "UART    [-flow [in CTS | out RTS | both      \n"
                              "           | none]  | -logHere | -noLog ]    \n";
INFLASH(char const comTime[]    ) = "timeLoc [hh:mm:ss] show / set local time     \n";
INFLASH(char const comDate[]    ) = "dateLoc [yy-mm-dd] show / set (local) date   \n";
INFLASH(char const comZone[]    ) = "zoneInf show / set time zone and DST state   \n";
INFLASH(char const comNTP[]     ) = "NTPconf [-silent|-verbose|-debug|-on| ..]    \n"
                              "        show / set Net Time Protocol state   \n";
INFLASH(char const comDHCP[]    ) = "DHCPstat   [-restart] show / restart DHCP    \n";
INFLASH(char const comDNS[]     ) = "DNSresolve [option] [name]  use / test DNS   \n"
                              "        as yet this is test / experimental   \n";
INFLASH(char const comTelnet[] ) = "Telnet  [-stop -off -reset | abort]          \n"
                              "        close or abort (this) telnet session \n";
INFLASH(char const comModbus[] ) = "Modbus  [-stop -off -reset | abort | start   \n"
                              "        other options ...]                   \n";

INFLASH(char const comNIC[] )    = "NICont  [-off | -on | -reset | -[re]start]   \n"
                              "        sleep | wake| reset | initialise the \n";
INFLASH(char const comENC[] )    = "ENC28J60  network interface (same as NICont) \n";
INFLASH(char const comSMC[] )    = "SMCard  [ -on | -start | reset | -restart    \n"
                              "        | -slow | -normal | -fast ]  show /  \n"
                              "        change small memory card state       \n";
INFLASH(char const comDirlist[]) =  "dirList [-quiet] [dirname]  list a (memory   \n"
                              "        card) directory; -quiet = only names \n";
INFLASH(char const comTypeFile[])=  "typeFile filename        display a (memory   \n"
                              "        card) file \n";

INFLASH(char const * const systemCommands[]) = { helpHeader, // 0 no command just headline
      comHelp,  // 1
      comWatchD, comWDog,  bLF1,   // 2 3 (4 is gap)
      comBoot,  comPLContrl, comAutomation, // 5 6 7
      comVersion,    comRunInfo,    // 8 9
      comMacAddr, comEthAddr,     // 10 11
      comIPconfig,  comIPdefAd,  // 12 13
      comUART, // 14
      comTime, comDate, // 15, 16
      comZone, comNTP, // 17, 18
      comDHCP,   // 19
      comDNS, comTelnet, comModbus, // 20 21 22
      comNIC, comENC, // 23 24
      comSMC, comDirlist,  // 25 26
      comTypeFile, // 27
      NULL };

//                              0 1 23456789.123456789z123456789d12345 6 7 8
INFLASH(char const optionHeader[]) =
      " * * *   Command options       - - -                      \n"
      "  Commands may recognise an option if prefixed with a -   \n"
      "  and given after the command as in e.g. dirList -quiet   \n"
      "  Like commands options may be unambiguously abbreviated. \n";
INFLASH(char const optQuest[] ) = "? | ";
INFLASH(char const optHelp[]  ) =     "help command specific help    \n";
INFLASH(char const optSilent[]) = "silent   no or only urgent output \n";
INFLASH(char const optQuiet[] ) = "quiet    minimal / short output   \n";
INFLASH(char const optInfo[]  ) = "inform   normal / informing output \n";
INFLASH(char const optVerbose[])= "verbose  lengthy output  \n";
INFLASH(char const optDebug[] ) = "debug    output for trouble shooting  \n";
INFLASH(char const optOn[]    ) = "on | ";
INFLASH(char const optStart[] ) =      "start  turn on / start  \n";
INFLASH(char const optRestart[])= "restart     like -start (in most cases)  \n";
INFLASH(char const optOff[]   ) = "off | ";
INFLASH(char const optStop[]  ) =       "stop  turn off / stop   \n";
INFLASH(char const optReset[] ) = "reset    like -stop, -restart or -abort (it depends) \n";
INFLASH(char const optAbort[] ) = "abort    a hard reset  \n";
INFLASH(char const optSharp[] ) = "sharp    more exact / less forgiving  \n";
INFLASH(char const optLenient[])= "lenient  more forgiving  \n";
INFLASH(char const optSlow[])   = "slow     slow(er) speed  \n";
INFLASH(char const optNormal[]) = "normal   standard behaviour / speed \n";
INFLASH(char const optFast[])   = "fast     fast(er)  speed \n";
INFLASH(char const optSystem[]) = "system | ";
INFLASH(char const optApplic[]) = "application | ";
INFLASH(char const optOption[]) = "options  system, app  or options related  \n";
INFLASH(char const optLogHere[])= "logHere |  ";
INFLASH(char const optNoLogs[]) = "noLogs  switch on, off logging to respective device \n";
INFLASH(char const optFlowCn[]) = "flowcontrol  ask or set flow control behaviour  \n";
INFLASH(char const optRead[])   = "read | ";
INFLASH(char const optWrite[])  = "write | ";
INFLASH(char const optLoad[])   = "load  perform the operation named    \n"
                                  "                 or set this mode.    \n";
INFLASH(char const optPrimar[]) = "primary  | ";
INFLASH(char const optAltern[]) = "alternative  use or restart with the \n"
                   "              resource designated   \n";


INFLASH(char const * const systemOptions[]) = { optionHeader, // 0 no option just headline
      optQuest, optHelp,   // 1, 2
      optSilent, optQuiet, // 3, 4
      optInfo, optVerbose, optDebug, // 5, 6, 7
      optOn, optStart, optRestart, // 8, 9, 10
      optOff, optStop,  optReset, // 11, 12, 13
      optAbort,                  // 14
      optSharp,  optLenient,  // 15 , 16
      optSlow, optNormal, optFast,//  17, 18, 19
      optSystem,  optApplic,  optOption, //  20, 21, 22
      optLogHere, optNoLogs,  // 23, 24
      optFlowCn, // 25
      optRead, optWrite, optLoad, // 26, 27, 28
      optPrimar, optAltern, // 29, 30
      NULL };

//----------------------------------------------------------------------

/* Output system info (greeting) to stdout, the thread function
 *
 *  \param pt the thread state (pointer to raw structure; not NULL)
 *  \param toStream streams to switch the standard streams to (if not MULL)
 *  \return \ref PT_YIELDED if waiting for conditions that may become
 *           automatically true; \ref PT_WAITING if blocked by conditions
 *           that other's action may set true;  \ref PT_EXITED if stopped
 *           by conditions that never will become true (sorry, not all work
 *           done and never will; \ref PT_ENDED if ready with all work
 */
ptfnct_t outSysInfoThreadF(pt_t * pt, FILE * toStream){
   PT_BEGIN(pt);

   PT_YIELD_OUT_SPACE(pt, 60, toStream);
   stdPutS_P(systWlc);
   stdPutC('\n');
   PT_YIELD_OUT_SPACE(pt, 90, toStream);
   stdPutS_P(systCop);
   PT_YIELD_OUT_SPACE(pt, 90, toStream);

   char s[50] = "R.";
   char * tmp = getFirstSVNtokenP(&(s[2]), systRev, 7);
   tmp = getSomeCharsP(tmp, sepLoB, 5);
   tmp = getFirstSVNtokenP(tmp, systDat, 13);
   tmp = getSomeCharsP(tmp, sepRcB, 5);
   getSomeCharsP(tmp, systBld, 40);
   stdPutS(s);
   stdPutS_P(bLF2); // 2 line feed

   PT_YIELD_OUT_SPACE(pt, 40, toStream);
   stdPutS_P(systPlatf); stdPutS_P(systCPUFr);

   PT_END(pt)
} // outSysInfoThreadF()



/*  Initialise the output a flash string array to standard output thread
 *
 *  \param theFlashStrings2out Pointer to (flash) array
 *                             of (flash) strings to be output
 *  \param noOfFlashStrings2out  Number of (flash) strings to be output
 */
void initOutFlashTextThread(outFlashTextThr_data_t * outFlashTextThread,
     char const * const * theFlashStrings2out,  uint8_t noOfFlashStrings2out){
   PT_INIT(&outFlashTextThread->pt);
   outFlashTextThread->indToChar = 0;
   outFlashTextThread->noOfFlashStrings2out = noOfFlashStrings2out;
   outFlashTextThread->theFlashStrings2out = theFlashStrings2out;
} // initOutFlashTextThread(outFlashTextThr_data_t*, char**, uint8_t)


/* Output a flash string array to std output, the thread function
 *
 *  \param outFlashTextThread the thread state data
 *  \param toStream pointer streams to switch the standard streams to (if not MULL)
 *  \return \ref PT_YIELDED if waiting for conditions that may become
 *           automatically true; \ref PT_WAITING if blocked by conditions
 *           that other's action may set true;  \ref PT_EXITED if stopped
 *           by conditions that never will become true (sorry, not all work
 *           done and never will; \ref PT_ENDED if ready with all work
 *
 */
ptfnct_t outFlashTextThreadF(outFlashTextThr_data_t * outFlashTextThread,
                                                             FILE * toStream){
   PT_BEGIN(&outFlashTextThread->pt);
   for (;;) {

#if DEBUG_TEL > 2
         /*        0123456789.123456789v12 345 6     */
      char tS[] = " ## yieOs ... > ... ..\n";
      twoHexs(&tS[20], checkOutSpace(120,  toStream));
      threeDigs(&tS[10], 120);
      threeDigs(&tS[16], outSpace(toStream));
      bufLogTxt(tS, 23);
#endif

      PT_YIELD_OUT_SPACE(&outFlashTextThread->pt, 120, toStream);
      int8_t const noOfS = outFlashTextThread->noOfFlashStrings2out;
      if(!noOfS ) break;
      char* actTxt = (char*)pgm_read_word(outFlashTextThread->theFlashStrings2out);
      if (!actTxt) break;
      actTxt += outFlashTextThread->indToChar;
      const uint16_t noOfCout = stdPutS_P(actTxt);
      if (noOfS < 0) { // special pseudo string array case
         actTxt += noOfCout; // must point to \0\x03\x01\x02
         if(pgm_read_byte(actTxt++)) break; // no \0
         if(pgm_read_byte(actTxt++) != 3) break; // no \x03
         if(pgm_read_byte(actTxt++) != 1) break; // no \x01
         if(pgm_read_byte(actTxt)   != 2) break; // no \x02
         outFlashTextThread->indToChar += (noOfCout + 4);
         continue;
      } else {
        if (!(outFlashTextThread->noOfFlashStrings2out = noOfS -1)) break;
      }
      ++outFlashTextThread->theFlashStrings2out;
      outFlashTextThread->indToChar = 0; // start of next string
   } // while
   outFlashTextThread->noOfFlashStrings2out = 0;
   PT_END(&outFlashTextThread->pt)
} // outFlashTextThreadF(outFlashTextThr_data_t*, FILE*)



//----------------------------------------------------------------------

INFLASH(char const sysRunMeldS[]) =  " * * *    RAM for stack (&heap) 0x";
INFLASH(char const sysRunMldSP[]) =  " * * *    sampled min. stack pointer:  0x";
//                                     0123456789.123456789
INFLASH(char const sysRunMeldA[]) =  " * * * ";
INFLASH(char const sysRunMeldB[]) =  " since ";
INFLASH(char const sysRunZOffs[]) =  "  Time zone offs = ";
INFLASH(char const sysRunZOf_0[]) =  " 0 (= GMT) ";
INFLASH(char const sysRunZOf_1[]) =  " 1 (= CET) ";
INFLASH(char const sysRunZ_DST[]) =  "DST (+1h) ";

//                             0123456789.123456789
INFLASH(char const sysUART0st[]) = "\n  UART(0) :";
INFLASH(char const sysUCSR0A[])  =  " UCSR0A= ";
INFLASH(char const sysUBRR0[])   =  " UBRR0= ";
INFLASH(char const sysUCSR0B[])  =  " UCSR0B= ";
INFLASH(char const sysUCSR0C[])  =  " UCSR0C= ";
INFLASH(char sysUinErr[])  =  "  err= ";
//                                 0123456789x123456789v
//INFLASH(char const uartFlowControl[]) = " UART flow control: ";
INFLASH(char const uartFlowCntInp[]) = "  input(CTS)   \n";
INFLASH(char const uartFlowCntOut[]) = "  output(RTS)  \n";
INFLASH(char const uartFlowCntNon[]) = "  noFlowCntl   \n";
INFLASH(char const uartFlowCntBot[]) = "  flC(RTS/CTS) \n";

//                              0123456789.123456789
INFLASH(char const sysRunNTPst[]) = "\n  NTP state ";
INFLASH(char const sysRunNTP0s[]) =  "no server ";
INFLASH(char const sysRunNTPvc[]) = " VCO ";
INFLASH(char const sysRunNTPdi[]) = " dif ";
INFLASH(char const sysRunNTPtr[]) = "  NTP trip  ";
INFLASH(char const sysRunNTPt1[]) = "  \n  NTP T1or  ";
INFLASH(char const sysRunNTPt4[]) = "  \n  NTP T4rx  ";

//                              0123456789.123456789.1
INFLASH(char const sysRunSPIst[]) = "\n  SPI 2 state :";
INFLASH(char const sysRunSPIer[]) =  ", errors=  ";
INFLASH(char const sysRunSMCst[]) = "\n  SMC state: card V.";
INFLASH(char const sysRunSMC2H[]) = "2 HC";

INFLASH(char const sysRunSMCrORC[])= ", OCR= ";
INFLASH(char const sysRunSMCsecC[])= ", #sc= ";
INFLASH(char const sysRunSMCresp[])= ", resp= ";
INFLASH(char const sysRunSectorN[])= "\n  SMC sector buffered: 0x";
INFLASH(char const sysRunSMCrCSD[])= "\n  SMC CSD = ";

extern uint8_t  _end;
extern uint8_t  __stack;
uint16_t minStckP = 0xFFFF;

/*  Output some system runtime and state info to std output
 *
 *  Single Bits of the \c flag signal which informations are to be
 *  output: <br />
 *  Bit 0 : SMC (small memory card) basic driver state info <br />
 *  Bit 1 : SPI2 (UART1 as SPI) info <br />
 *  Bit 2 : UART (0) info <br />
 *  Bit 3 : NTP info <br />
 *  Bit 4 : running time, reset cause and RAM usage info <br />
 *  Bit 5 : current local date <br />
 *  Bit 6 : current local time <br />
 *  Bit 7 : zone offset and DST <br />
 *
 *  \return \ref PT_YIELDED if waiting for conditions that may become
 *           automatically true; \ref PT_WAITING if blocked by conditions
 *           that other's action may set true;  \ref PT_EXITED if stopped
 *           by conditions that never will become true (sorry, not all work
 *           done and never will; \ref PT_ENDED if ready with all work
 */
ptfnct_t outRunInfoThreadF(hierThr_data_t * threadD, FILE * toStream) {
   if (!(threadD->flag) && !threadD->flag2) {
      threadD->flag = 0;
      PT_INIT(&threadD->pt2);
      PT_LEAVE(&threadD->pt);
   } // quick out nope to do

   char s[31];
   // switchStdStreams(toStream);
   PT_BEGIN(&threadD->pt);

   PT_YIELD_OUT_SPACE(&threadD->pt, 69, toStream); // re-entry point <<<
#if HAVE_ETHERNET
   if (threadD->flag2) { // some LAN ENC Ethernet output
      const uint8_t outFlashErg = outEncLanInfoThreadF(threadD, toStream);
      if (outFlashErg <= PT_YIELDED) return outFlashErg; // re-ent ^^^
      if (outFlashErg == PT_EXITED)  PT_EXIT(&threadD->pt);
      if (!threadD->flag) PT_LEAVE(&threadD->pt);
   } // some LAN ENC Ethernet output
#endif

   if (threadD->flag & 0x10) { // Bit 4 : run time and reset cause
      PT_YIELD_OUT_SPACE(&threadD->pt, 69, toStream);
      stdPutS_P(sysRunMeldA);
      uint32_t dur =  secClock(); // (prelim.) as duration
      datdur_t  sysRun;
      sec2datdur(&sysRun, dur);
      formTimDur(s, &sysRun);
      stdPutS(s);
      stdPutS_P(sysRunMeldB);

#if USE_BOOTLOADER >= 2
      copyChars_P(s, resetCauseText, 27);
      stdPutS(s);
#else
      stdPutS_P(resetCauseText);
#endif
      stdPutC('\n');
      stdPutS_P(sysRunMeldS);
      {
         //           0123456789x123456789 v
         char tt[] = "...  <= SP <= ...  \n";

      fourHexs(&tt[0],(uint16_t)(&_end));
      fourHexs(&tt[14], (uint16_t)&__stack);
      stdPutS(tt);
      }
#if ! NO_SPMIN_SAMPLE
      if (minStckP != 0xFFFF) {
         stdPutS_P(sysRunMldSP);
         char tt[] = "....  \n";
         fourHexs(&tt[0], minStckP);
         stdPutS(tt);
         minStckP = 0xFFFF;
      }
#endif
      if (!(threadD->flag &= ~0x10)) PT_LEAVE(&threadD->pt);
   } // Bit 4 : run time and reset cause

   if (threadD->flag & 0x04) { // Bit 2 : UART info,
      PT_YIELD_OUT_SPACE(&threadD->pt, 69, toStream);
      stdPutS_P(sysUART0st);
      stdPutS_P(sysUCSR0A);
      twoHexs(s, UCSR0A); s[2] = 0;  stdPutS(s);

      stdPutS_P(sysUBRR0);
      fourDigs(s, UBRR0);  s[4] = 0;  stdPutS(s);

      stdPutS_P(sysUCSR0B);
      twoHexs(s, UCSR0B); s[2] =  0;  stdPutS(s);

      stdPutS_P(sysUCSR0C);
      twoHexs(s, UCSR0C);  stdPutS(s);

      const uint8_t errOcc = uartInErrors();
      if (errOcc) {
         stdPutS_P(sysUinErr);
         twoHexs(s, errOcc);  stdPutS(s);
      }

      if(isCTSflowC) {
         stdPutS_P( isRTSflowC ? uartFlowCntBot : uartFlowCntInp);
      } else if(isRTSflowC) {
         stdPutS_P(uartFlowCntOut);
      } else {
         stdPutS_P(uartFlowCntNon);
      }
      PT_YIELD_OUT_SPACE(&threadD->pt, 35, toStream);
      stdPutS_P(systCPUFr);

      if (!(threadD->flag &= ~0x04)) PT_LEAVE(&threadD->pt);
   } //  Bit 2 : UART info,

   if (threadD->flag & 0x02) { // Bit 1 : SPI 2 state
#if USE_SPI2
      PT_YIELD_OUT_SPACE(&threadD->pt, 49, toStream);
      stdPutS_P(sysRunSPIst);   //  16
      stdPutS_P(sysUBRR0);    // 7
      fourDigs(s, UBRR1);  s[4] = 0;  stdPutS(s);  // 4
      if (spi2Errors) {
         stdPutS_P(sysRunSPIer); // 11
         twoHexs(s, spi2Errors);
         spi2Errors = s[2] = 0;
         stdPutS(s);                // 2
      }

      stdPutS_P(bLF1);               // 3
#endif // use SPI 2
      if (!(threadD->flag &= ~0x02)) PT_LEAVE(&threadD->pt);
   } // Bit 1 : SPI 2 state

   if (threadD->flag & 0x01) { // Bit 0 : SMC info
#if HAVE_SMC
      PT_YIELD_OUT_SPACE(&threadD->pt, 69, toStream);
      stdPutS_P(sysRunSMCst);   //  19
      switch (smcState.cardType & CT_TYPE_MASK) {
         case CT_V_1 : stdPutC('1'); break;
         case CT_V_2 : stdPutC('2'); break;
         case CT_V_3 : stdPutC('3'); break;
         case CT_V_2 | CT_HC : stdPutS_P(sysRunSMC2H); break;
         default : stdPutC('?'); break;
      } // switch clear text type
      s[0] = ' ';
      twoHexs(s+1, smcState.cardType);
      s[3] = ','; s[5] = s[9] = 0;
      s[4] = PIND & 0x20 ? 'o' : 'i';
      stdPutS(s);           // 4
      stdPutS_P(sysRunSMCresp); // 8

      twoHexs(s, smcState.lastCmdResp[0]);
      twoHexs(s+3, smcState.lastCmdResp[1]);
      s[8] = s[2] = ' ';
      stdPutS(s);           // 4
      if (smcTypeD()) { // type determined
         PT_YIELD_OUT_SPACE(&threadD->pt, 69, toStream);
         stdPutS_P(sysRunSMCrORC);
         eightHexsBE(s, *((uint32_t*)smcState.ocr) ); s[8] = ' '; s[9] = 0;
         stdPutS(s);
         PT_YIELD_OUT_SPACE(&threadD->pt, 79, toStream);
         stdPutS_P(sysRunSMCrCSD);

         eightHexsBE(s, *((uint32_t*)smcState.csd)); s[8] = ' '; s[9] = 0;
         stdPutS(s);
         eightHexsBE(s, *((uint32_t*)smcState.csd + 1));
         stdPutS(s);
         eightHexsBE(s, *((uint32_t*)smcState.csd + 2));
         stdPutS(s);
         eightHexsBE(s, *((uint32_t*)smcState.csd + 3));
         stdPutS(s);

         stdPutS_P(sysRunSMCsecC);
         eightHexs(s, smcGetSectCount()); stdPutS(s);
      } // type determined
      if (smcState.bufferState & FL_RDS) { // sector buffer in sync
         PT_YIELD_OUT_SPACE(&threadD->pt, 49, toStream);
         stdPutS_P(sysRunSectorN);
         eightHexs(s, smcState.rwSector); s[8] = ' '; s[9] = 0;
         stdPutS(s);
      } // sector buffer in sync

      stdPutS_P(bLF1);
#endif // have SMC
      if (!(threadD->flag &= ~0x01)) PT_LEAVE(&threadD->pt);
   } // Bit 0 : SMC state



   if (threadD->flag & 0x08) { // Bit 3 : NTPinfo
#if HAVE_ETHERNET
      PT_YIELD_OUT_SPACE(&threadD->pt, 51, toStream);
      stdPutS_P(sysRunNTPst);
      if (!ntpState.state) {
         stdPutS_P(sysRunNTP0s);
      } else { // with NTP
         twoHexs(s, ntpState.state);
         s[3] = 'a'; s[4] = 'd';
         s[2] = s[5] = ' '; s[8] = 0;
         twoHexs(&s[6], ntpState.adjuState);
         stdPutS(s);
         uint8_t ind = 0;
         if (ntpState.adjuState & ~NTP_ADJUSTAT_PRV_MSK) {
            stdPutS_P(sysRunNTPvc);
            twoHexs(&s[0], getVCOsetting());
            s[2] = 0;
            stdPutS(s);
            stdPutS_P(sysRunNTPdi);

            s[4] = s[9] = ' ';
            if (ntpState.adjuState & NTP_ADJUSTAT_MINUS) {
               s[0] = '-';
               threeDigs(&s[1], 1000 - ntpState.msDif);
            } else {
               s[0] = '+';
               threeDigs(&s[1], ntpState.msDif);
            }

            if (ntpState.adjuState & NTP_ADJUSTAT_RNG_1S) {
               fourHexs(&s[5], ntpState.secDif);
               ind = 10;
            } else  ind = 5;

         } // with adjust
         s[ind] = '\n';
         s[ind+1] = 0;
         stdPutS(s);
         if (ntpState.logLev >= optInfoNum) {
            PT_YIELD_OUT_SPACE(&threadD->pt, 95, toStream);
            stdPutS_P(sysRunNTPtr);
            struct ntpTimestamp_t roundTrip;
            ntpTimestampDif(&roundTrip, &ntpState.lastT21, &ntpState.lastT34);
            fourHexsBE(s, roundTrip.seconds.v16[1]);
            s[4] = 's'; s[5] = ' '; s[10] = 0;
            fourHexsBE(&s[6], roundTrip.fraction.v16[0]);
            stdPutS(s);
            stdPutS_P(sysRunNTPt1);
            eightHexsBE(s, ntpState.lastSendTimestamp.seconds.v32);
            s[8] = '.'; s[17]  = 0;
            eightHexsBE(&s[9], ntpState.lastSendTimestamp.fraction.v32);
            stdPutS(s);
            stdPutS_P(sysRunNTPt4);
            eightHexsBE(s, ntpState.lastReceiveTimestamp.seconds.v32);
            s[17] = s[18] = ' ';  s[19] = '\n'; s[20] = 0;
            eightHexsBE(&s[9], ntpState.lastReceiveTimestamp.fraction.v32);
            stdPutS(s);
            // optVerboseNum  6
            if (ntpState.logLev >= optDebugNum) {
               threadD->flag |= 0x60; // output date time next
            }
         } // >= optInfoNum
      } // with NTP server
#endif // with Ethernet
      if (!(threadD->flag &= ~0x08)) PT_LEAVE(&threadD->pt);
   } // // Bit 3 : NTPinfo


   if (threadD->flag & 0x60) { // Bit 5 : local time, 6 : date

      PT_YIELD_OUT_SPACE(&threadD->pt, 69, toStream);

      char * beg = s;
      uint32_t dur =  secTime308Loc();
      datdur_t  sysTim;
      sec2datdur(&sysTim, dur);
      if (threadD->flag & 0x40) { // Bit 6 : date
         date_t sysDat;
         setDatByDays(&sysDat, sysTim.d);
         beg = formWdShort(beg, sysDat.wd);
         *beg++ = ' ';
         beg = formDateIso(beg, &sysDat);
         *beg++ = ' ';
      }
      beg = formTimOfD(beg, &sysTim);
      *beg++ = ' '; *beg++ = '\n'; *beg = 0;
      stdPutS(s);
      if (!(threadD->flag &= ~0x60)) PT_LEAVE(&threadD->pt);
   } // Bit 6 : local time


   if (threadD->flag & 0x80) { // Bit 7 : zone info

      PT_YIELD_OUT_SPACE(&threadD->pt, 69, toStream);
      stdPutS_P(sysRunZOffs);
      uint8_t ind = 0;

      if (!zoneOffsetSec) {
         stdPutS_P(sysRunZOf_0);
      } else if (zoneOffsetSec == 3600) {
         stdPutS_P(sysRunZOf_1);
      } else {
         eightHexs(s, zoneOffsetSec); // todo change to /3600 and +/- decimal
         s[8] = ' ';
         ind = 9;
      }
      if (isDST) {    //         01234567
         memcpy_P(&s[ind], sysRunZ_DST, 10);
         ind += 10;
      }
      s[ind] = 0;
      stdPutS(s);
      stdPutC('\n');
      if (!(threadD->flag &= ~0x80)) PT_LEAVE(&threadD->pt);
   } // Bit 7 : zone info

    PT_END(&threadD->pt)
} // outRunInfoThreadF()



/* Copy the first SVN token or a full string from program space  */
char *  getFirstSVNtokenP(char * dest, char const * src, uint8_t mxLen){
   if (src == NULL || dest == NULL || mxLen == 0) return dest;
   char ac = 0;
   uint8_t isSVN = --mxLen;
   if (isSVN) {
     ac = pgm_read_byte(src);
     isSVN = ac == '$';
   }
   if (isSVN) { // goto start of first SVN token
      do { // search :
         ++src;
         ac = pgm_read_byte(src);
      } while (ac && ac != ':');  // search :
      if (ac == ':') do { // skip blank
         ++src;
         ac = pgm_read_byte(src);
      } while (ac == ' ');  // skip blank
   }
   while (ac) { // copy loop
      if (isSVN && (ac == '$' || ac <= ' ')) break;
      *dest = ac;
      ++dest;
      if (!(--mxLen)) break;
      ++src;
      ac = pgm_read_byte(src);
    } // copy loop
   *dest = 0;
   return dest;
} // getFirstSVNtokenP(char*, prog_char*, uint8_t)


/* Copy some characters from program space to a string   */
char *  getSomeCharsP(char * dst, char const * src, uint8_t mxLen){
   if (!mxLen || !src || !dst) return dst;
   for(;;) {
      if(!--mxLen) {
         *dst  = 0;
         break;
      }
      char c = pgm_read_byte(src++);
      *dst = c;
      if (c) {
         ++dst;
      } else break;
   }
   return dst;
} // getSomeCharsP(char*, prog_char*, uint8_t)


/** The watchdog interrupt
 *
 *  This handler is called on first watchdog (WD) time out (TO). <br />
 *  The next WD TO will abort the system.
 */
ISR(WDT_vect) {
   setAbortCommand(WDtiOut);
} // watchdog interrupt



