/**
 *  \file utils.c
 *
 *  weAutSys  utility / library functions
 *              to be used also by application / user software
 *
 *  This file contains the implementations of \ref intro_sec "weAutSys"'
 *  (weAut_01) utility / library functions to be used also by application
 *  / user software.
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of weAutSys &nbsp;
 *           <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert

#include "we-aut_sys/common.h"
#include "we-aut_sys/utils.h"
#include "we-aut_sys/ll_system.h"
#include "we-aut_sys/ntp.h"
#include <string.h>

//---------------------------------------------------  Arithmetic   ---------

/*  Set bit masks (table in flash memory)
 *
 *  This array of length 8 in flash memory contains the masks to set
 *  bits 0 .. 7 by ORing. Hence the values are 1, 2, 4 .... 128 (0x80).
 */
  INFLASH(uint8_t const setBitMasks[8]) = {1, 2, 4, 8,  16, 32, 64, 128};

/*  Clear bit masks (table in flash memory)
 *
 *  This array of length 8 in flash memory contains the masks to clear
 *  bits 0 .. 7 by ANDing. Hence the values are ~1, ~2, ~4 .... ~128 (0x7F).
 */
  INFLASH(uint8_t const clearBitMasks[8]) = {0xFE, 0xFD, 0xFC, 0xF7,
                                             0xEF, 0xDF, 0xCF, 0x7F};


/* A 32 bit unsigned greater equal (ge) comparison for modulo arithmetic */
uint8_t geU32ModAr(uint32_t a, uint32_t b){
   return (int32_t)(a - b) >= 0;
}

/* A 32 bit unsigned less or equal (le) comparison for modulo arithmetic */
uint8_t leU32ModAr(uint32_t a, const uint32_t b){
#if 0
   a = (int32_t)(a - b - 1); // GCC makes a real 32 bit -1, then -b
#else
   asm volatile(""
         "sec          ; set carry for -1" "\n\t"
         "sbc %A0, %A1 ; a- b" "\n\t"
         "sbc %B0, %B1 ; a- b" "\n\t"
         "sbc %C0, %C1 ; a- b" "\n\t"
         "sbc %D0, %D1 ; a- b" "\n\t"
        : "+r"(a)
        : "r"(b)
       );
#endif
   return ((ucnt32_t)a).v8[3] & 0x80;
} //  leU32ModAr(uint32_t, const uint32_t)

//--------------------------------------------------------------------------

/* Divide unsigned byte by 10
 *
 * This is a multiply by inverse implementation.
 * It is faster than direct usage of avr-C's / and % and also
 * faster as using stdlib.h's div(...,10).
 */
u8div_t divByVal10(uint8_t div){
   u8div_t retV;   // (lb-A: quot, hb-B:rem)
   asm volatile(
         "ldi  %A0, 25  ; \n\t"
         "mul  %A0, %1  ;  25 * div   \n\t"
         "mov  %A0, r1  ;  to retV.quot (hi byte of product) \n\t"
         "ldi  %B0, 10  ; \n\t"
         "mul  %B0, %A0 ;  10 * quot   \n\t"
         "mov  %B0, %1  ;  rem = div   \n\t"
         "sub  %B0, r0  ;  rem = div - 10 * quot \n\t"
         "clr  r1       ;  restore zero_reg \n\t"
         "cpi  %B0, 10  ;  rem - 10  \n\t"
         "brcs __noC_%= ; if (retV.rem >= 10)  corriger la fortune \n\t"
         "subi  %B0, 10 ; rem -= 10  \n\t"
         "subi  %A0, 0xFF ; qout += 1  \n\t"
         "__noC_%=:"
         : "=&r" (retV)
         : "r" (div)
      );
   return retV;

/*   C- implementation
   uint16_t temp = div * 25;  // multiply inverse (a lil bit too low)
   u8div_t retV = {.quot = HBYTE(temp)};
   retV.rem = div - LBYTE(retV.quot * 10);
   if (retV.rem >= 10) { // corriger la fortune
         retV.rem -= 10;
         ++retV.quot;
   } // corriger la fortune
   return retV;
   */
} // divByVal10(uint8_t)

/* Divide unsigned word by 10 for a byte result
 *
 * This is a multiply by inverse implementation for small 16 bit
 * dividends (0..999).
 * It will fail totally if the \c div is 1286 or above!
 *
 * It is faster than C / and % and faster as using stdlib.h's div(...,10).
 */
u8div_t divWByVal10toByte(uint16_t div){
 //  uint16_t temp = div * 51;  // multiply inverse (a lil bit too low)
 //  u8div_t retV = {.quot = temp >> 9};
   u8div_t retV;   // (lb-A: quot, hb-B:rem)
   uint8_t t51 = 51;

   asm volatile(
      "mul  %A2, %1 ;  lo(div) * 51   \n\t"
      "mov  %A0, r1 ;  to retV.quot (hi byte of product) \n\t"
      "mul  %B2, %1 ;  hi(div) * 51   \n\t"
      "mov  %B0, r1 ;  save hibyte prod temp. in retV.rem \n\t"
      "clr  r1      ;  restore zero_reg \n\t"
      "add  %A0, r0 ;  add lo byte to retV.quot \n\t"
      "adc  %B0, r1 ;  add carry into hi byte of prod. \n\t"
      "lsr  %B0     ;  and lo bit of result into carry \n\t"
      "ror  %A0     ;  retV.quot >> 1 with carry in    \n\t"
      "ldi  %1, 10  ;   \n\t"
      "mul  %A0, %1 ;  retV.quot * 10  (only low byte of interest) \n\t"
      "mov  r1, %A2 ;  r1 = div.low \n\t"
      "sub  r1, r0  ;  r1 = div.low - quot * 10  \n\t"
      "mov  %B0, r1 ;  save as retV.rem (may be false by 1 * 10) \n\t"
      "clr  r1      ;  restore zero_reg \n\t"
      : "=&r" (retV), "+r" (t51)
      : "r" (div)
   );
   //  retV.rem = div - retV.quot * 10;
   if (retV.rem >= 10) { // corriger la fortune
         retV.rem -= 10;
         ++retV.quot;
   } // corriger la fortune
    return retV;
} // divWByVal10toByte(uint16_t)


/* Divide unsigned word by 1000.
 *
 * This is a "divide by slightly to big dividend and correct" implementation.
 *
 * It is faster than C / and % and faster as using stdlib.h's div(...,1000).
 */
u16div_t divWByVal1000(uint16_t div){
   u16div_t retV;  // 0: BA quot, DC rem
   asm volatile(
      "mov %A0, %B1 ; retV.quot.lo = div.hi \n\t"
      "lsr %A0      ; >> \n\t"
      "lsr %A0      ;   2 \n\t"
      "ldi %B0, 24  ;  \n\t"
      "mul %B0, %A0 ;  24 * quot \n\t"
      "movw %C0, %1 ; ret.rem = div \n\t"
      "andi %D0, 3  ; rem by 1024  \n\t"
      "add %C0, r0  ; + 24 * .. \n\t"
      "adc %D0, r1  ; + 24 * .. \n\t"

      "ldi %B0, 0x03 ; hi(1000) \n\t"
      "cpi %C0, 0xE8 ; rem.lo - lo(1000) \n\t"
      "cpc %D0, %B0  ; rem.hi - hi(1000) \n\t"
      "brcs __ntd_%= ; rem < 1000 : ready \n\t"
      "subi %C0, 0xE8 ; rem.lo -= lo(1000) \n\t"
      "sbc  %D0, %B0  ; rem.hi -= hi(1000) \n\t"
      "subi %A0, 0xFF ; retV.quot.lo += 1  \n\t"

      "cpi %C0, 0xE8 ; rem.lo - lo(1000) \n\t"
      "cpc %D0, %B0  ; rem.hi - hi(1000) \n\t"
      "brcs __ntd_%= ; rem < 1000 : ready \n\t"
      "subi %C0, 0xE8 ; rem.lo -= lo(1000) \n\t"
      "sbc  %D0, %B0  ; rem.hi -= hi(1000) \n\t"
      "subi %A0, 0xFF ; retV.quot.lo += 1  \n\t"

      "__ntd_%=:"
      "eor %B0, %B0 ; retV.quot.hi = 0 \n\t"
      "clr  r1      ;  restore zero_reg \n\t"
      : "=&r" (retV)
      : "r" (div)
   );
   return retV;
} //  divWByVal1000(uint16_t)


#if USE_BOOTLOADER < 2  //---------------------------------------------------

// the following are put to ll_common.h and implemented
// in weAutSys' serial bootloader:
// uint16_t div16(uint16_t * rem, uint16_t dividend, uint16_t divisor);
// uint32_t div32(uint32_t * rem, uint32_t dividend, uint32_t divisor);
// They may be used there if USE_BOOTLOADER is 2 or greater


/*  Unsigned 16 Bit divide
 *
 *  This function is an unconfined unsigned 16 bit divide. <br />
 *  This function does in effect
 *    *rem  = dividend % divisor;
 *    return = dividend / divisor;
 *
 * \param result pointer to result, quotient and remainder
 * \param dividend the number to be divided
 * \param divisor the divisor
 */
uint16_t div16(uint16_t * rem, uint16_t dividend, uint16_t divisor){
     //result->quot = dividend / divisor;
     //result->rem  = dividend % divisor;
     // Hint (internal): symbols starting with __ and not __u won't make it
     // into (cross) symbol table boot109_<target>_ext.tab

   uint16_t dividend2 = 0;
   uint16_t resAcc = 1;
   asm volatile("__loopS_%=:"
         "lsl %A0" "\n\t"
         "rol %B0"  "\n\t"
         "rol %A1"   "\n\t"
         "rol %B1"    "\n\t"

        //  "brcs __doSub_%=" "\n\t" // carry out == is bigger (can't happen at 16/16)
         "cp  %A1, %A3" "\n\t"
         "cpc  %B1, %B3" "\n\t" // dividend2 - divisor
         "brcc __doSub_%=" "\n\t" // if carry out ==  is lower: stay (no substract)
         "lsl %A2" "\n\t" // << 0 into resAccu
         "rol %B2"  "\n\t"
         "brcc __loopS_%=" "\n\t" // loop while no carry (init 1) out
         "rjmp __copyRes_%=" "\n\t" // ready

         "__doSub_%=:"
         "sub  %A1, %A3" "\n\t"
         "sbc  %B1, %B3"  "\n\t" // dividend2 -= divisor
         "sec"   "\n\t"   // set 1 << into result
         "rol %A2" "\n\t"
         "rol %B2"  "\n\t"
         "brcc __loopS_%=" "\n\t" // loop while no carry (init 1) out

         "__copyRes_%=:"

        : "+r"(dividend), "+r"(dividend2), "+r"(resAcc)
        : "r"(divisor)
        );

   if(rem) *rem = dividend2;
   return resAcc;
} //  div16(uint16_t dividend, uint16_t divisor)


/*  Unsigned 32 Bit divide
 *
 *  This function does in effect
 *
 *    *rem  = dividend % divisor;
 *    return = dividend / divisor;
 *
 * \param rem  pointer to remainder; if null the remainder is lost
 * \param dividend the number to be divided
 * \param divisor  the divisor
 * \return the quotient \c divisor / \c dividend
 */
uint32_t div32(uint32_t * rem, uint32_t dividend, uint32_t divisor){
  // result->quot.v32 = dividend / divisor;
  // result->rem.v32  = dividend % divisor;

   uint32_t dividend2 = 0;
   uint32_t resAcc = 1;
   asm volatile("__loopS_%=:"
         "lsl %A0" "\n\t"
         "rol %B0"  "\n\t"
         "rol %C0"   "\n\t"
         "rol %D0"    "\n\t"
         "rol %A1" "\n\t"
         "rol %B1"  "\n\t"
         "rol %C1"    "\n\t"
         "rol %D1"     "\n\t"

        //  "brcs __doSub_%=" "\n\t" // carry out == is bigger (can't happen at 16/16)
         "cp  %A1, %A3" "\n\t"
         "cpc  %B1, %B3" "\n\t" // dividend2 - divisor
         "cpc  %C1, %C3"  "\n\t"
         "cpc  %D1, %D3"  "\n\t"
         "brcc __doSub_%=" "\n\t" // if carry out ==  is lower: stay (no substract)
         "lsl %A2" "\n\t" // << 0 into resAccu
         "rol %B2"  "\n\t"
         "rol %C2"   "\n\t"
         "rol %D2"    "\n\t"
         "brcc __loopS_%=" "\n\t" // loop while no carry (init 1) out
         "rjmp __copyRes_%=" "\n\t" // ready

         "__doSub_%=:"
         "sub  %A1, %A3" "\n\t"
         "sbc  %B1, %B3"  "\n\t" // dividend2 -= divisor
         "sbc  %C1, %C3"   "\n\t"
         "sbc  %D1, %D3"   "\n\t"
         "sec"   "\n\t"   // set 1 << into result
         "rol %A2" "\n\t"
         "rol %B2"  "\n\t"
         "rol %C2"   "\n\t"
         "rol %D2"    "\n\t"
         "brcc __loopS_%=" "\n\t" // loop while no carry (init 1) out

         "__copyRes_%=:"

        : "+r"(dividend), "+r"(dividend2), "+r"(resAcc)
        : "r"(divisor)
        );

   if(rem) *rem = dividend2;
   return resAcc;
} // div32

/*  Universal unsigned 24 bit divide
 *
 *  This functions's effect is \code
 *    result (return)  = dividend / divisor;
 *    *rem  = dividend % divisor;
 *  \endcode
 *  This function is quite similar to \ref div32. It also shares the 32 bit
 *  parameter and return types (as avr-gcc C has no 24 bit type).
 *  Nevertheless it just operates on and returns 24 bit unsigned numbers
 *  &mdash; just ignoring respectively setting 0 the upper byte.
 *
 *  So it is about twice as fast as the optimised \ref div32.
 *
 * \param rem  pointer to the 32 bit variable for the 24 bit remainder;
 *             if null the remainder is lost; upper byte will be 0.
 * \param dividend the 24 bit number to be divided (upper byte ignored)
 * \param divisor  the 24 bit divisor (upper byte ignored)
 * \return the 24 bit quotient \c divisor / \c dividend (upper byte 0)
 * \sa div16
 */
uint32_t div24(uint32_t * rem, uint32_t dividend, uint32_t divisor){
   // result->quot.v32 = dividend / divisor;
   // result->rem.v32  = dividend % divisor;

    uint32_t dividend2 = 0;
    uint32_t resAcc = 1;
    asm volatile("__loopS_%=:"
          "lsl %A0  ;  \n\t"
          "rol %B0  ;   \n\t"
          "rol %C0  ; shift dividend one left into dividend2   \n\t"
          "rol %A1" "\n\t"
          "rol %B1"  "\n\t"
          "rol %C1"    "\n\t"

         //  "brcs __doSub_%=" "\n\t" // carry out == is bigger (can't happen at 24/24)
          "cp  %A1, %A3" "\n\t"
          "cpc  %B1, %B3" "\n\t" // dividend2 - divisor
          "cpc  %C1, %C3"  "\n\t"
          "brcc __doSub_%=" "\n\t" // if carry out ==  is lower: stay (no substract)
          "lsl %A2" "\n\t" // << 0 into resAccu
          "rol %B2"  "\n\t"
          "rol %C2"   "\n\t"
          "brcc __loopS_%=" "\n\t" // loop while no carry (init 1) out
          "rjmp __copyRes_%=" "\n\t" // ready

          "__doSub_%=:"
          "sub  %A1, %A3" "\n\t"
          "sbc  %B1, %B3"  "\n\t" // dividend2 -= divisor
          "sbc  %C1, %C3"   "\n\t"
          "sec"   "\n\t"   // set 1 << into result
          "rol %A2" "\n\t"
          "rol %B2"  "\n\t"
          "rol %C2"   "\n\t"
          "brcc __loopS_%=" "\n\t" // loop while no carry (init 1) out

          "__copyRes_%=:"
          "eor %D2, %D2  ; upper byte of quotient  0   \n\t"
          "eor %D1, %D1  ; upper byte of remainder 0   \n\t"
         : "+r"(dividend), "+r"(dividend2), "+r"(resAcc)
         : "r"(divisor)
         );

    if(rem) *rem = dividend2;
    return resAcc;
 } // div24

/*  Unsigned 32 bit divide by 24 bit
 *
 *  This functions's effect is \code
 *    result (return)  = dividend / divisor;
 *    *rem  = dividend % divisor;
 *  \endcode
 *  This function is quite similar to \ref div32 respectively div24 or a mix
 *  of both. It shares the 32 bit parameter and return types (as avr-gcc C has
 *  no 24 bit type).
 *
 *  The divisor and remainder are 24 bit numbers (upper 32 bit byte ignored or
 *  set 0). The dividend and the quotient are 32 bit.
 *
 * \param rem  pointer to the 32 bit variable for the 24 bit remainder;
 *             if null the remainder is lost; upper byte will be 0.
 * \param dividend the 32 bit number to be divided
 * \param divisor  the 24 bit divisor (upper 32 bit byte ignored)
 * \return the 32 bit quotient \c divisor / \c dividend (upper byte 0)
 * \sa div16
 */
uint32_t div32by24(uint32_t * rem, uint32_t dividend, uint32_t divisor){
   // result->quot.v32 = dividend / divisor;
   // result->rem.v32  = dividend % divisor;

    uint32_t dividend2 = 0;
    uint32_t resAcc = 1;
    asm volatile("__loopS_%=:"
          "lsl %A0" "\n\t"
          "rol %B0"  "\n\t"
          "rol %C0"   "\n\t"
          "rol %D0  ; shift dividend one left into dividend2   \n\t"
          "rol %A1" "\n\t"
          "rol %B1"  "\n\t"
          "rol %C1"    "\n\t"
          "brcs __doSub_%=" "\n\t" // carry out == is bigger (can happen at 32/24)

          "cp  %A1, %A3" "\n\t"
          "cpc  %B1, %B3" "\n\t" // dividend2 - divisor
          "cpc  %C1, %C3"  "\n\t"
          "brcc __doSub_%=" "\n\t" // if carry out ==  is lower: stay (no substract)
          "lsl %A2" "\n\t" // << 0 into resAccu
          "rol %B2"  "\n\t"
          "rol %C2"   "\n\t"
          "rol %D2"    "\n\t"
          "brcc __loopS_%=" "\n\t" // loop while no carry (init 1) out
          "rjmp __copyRes_%=" "\n\t" // ready

          "__doSub_%=:"
          "sub  %A1, %A3" "\n\t"
          "sbc  %B1, %B3"  "\n\t" // dividend2 -= divisor
          "sbc  %C1, %C3"   "\n\t"
          "sec"   "\n\t"   // set 1 << into result
          "rol %A2" "\n\t"
          "rol %B2"  "\n\t"
          "rol %C2"   "\n\t"
          "rol %D2"    "\n\t"
          "brcc __loopS_%=" "\n\t" // loop while no carry (init 1) out

          "__copyRes_%=:"
          "eor %D1, %D1  ; upper byte of remainder 0   \n\t"

         : "+r"(dividend), "+r"(dividend2), "+r"(resAcc)
         : "r"(divisor)
         );

    if(rem) *rem = dividend2;
    return resAcc;
 } // div32by24

/*  Unsigned 32 bit divide by 16 bit
 *
 *  This functions's effect is \code
 *    result (return)  = dividend / divisor;
 *    *rem  = dividend % divisor;
 *  \endcode
 *  This function is quite similar to \ref div32 respectively \ref div16 or a
 *  mix of both.
 *
 *  The divisor and remainder are 16 bit numbers. The dividend and the
 *  quotient are 32 bit.
 *
 *  It is about 35% faster than \ref div32 (which is quite a lot on a RICS
 *  without division).
 *
 * \param rem  pointer to the 16 bit remainder;
 *             if null the remainder is lost
 * \param dividend the 32 bit number to be divided
 * \param divisor  the 16 bit divisor
 * \return the 32 bit quotient \c divisor / \c dividend
 * \sa div24
 */
uint32_t div32by16(uint16_t * rem, uint32_t dividend, uint16_t divisor){
   // result->quot.v32 = dividend / divisor;
   // result->rem.v32  = dividend % divisor;

    uint32_t dividend2 = 0;
    uint32_t resAcc = 1;
    asm volatile("__loopS_%=:"
          "lsl %A0" "\n\t"
          "rol %B0"  "\n\t"
          "rol %C0"   "\n\t"
          "rol %D0  ; shift dividend one left into dividend2   \n\t"
          "rol %A1" "\n\t"
          "rol %B1"  "\n\t"
          "brcs __doSub_%=" "\n\t" // carry out == is bigger (can happen at 32/24)

          "cp  %A1, %A3" "\n\t"
          "cpc  %B1, %B3" "\n\t" // dividend2 - divisor
          "brcc __doSub_%=" "\n\t" // if carry out ==  is lower: stay (no substract)
          "lsl %A2" "\n\t" // << 0 into resAccu
          "rol %B2"  "\n\t"
          "rol %C2"   "\n\t"
          "rol %D2"    "\n\t"
          "brcc __loopS_%=" "\n\t" // loop while no carry (init 1) out
          "rjmp __copyRes_%=" "\n\t" // ready

          "__doSub_%=:"
          "sub  %A1, %A3" "\n\t"
          "sbc  %B1, %B3"  "\n\t" // dividend2 -= divisor
          "sec"   "\n\t"   // set 1 << into result
          "rol %A2" "\n\t"
          "rol %B2"  "\n\t"
          "rol %C2"   "\n\t"
          "rol %D2"    "\n\t"
          "brcc __loopS_%=" "\n\t" // loop while no carry (init 1) out

          "__copyRes_%=:"
         : "+r"(dividend), "+r"(dividend2), "+r"(resAcc)
         : "r"(divisor)
         );

    if(rem) *rem = dividend2;
    return resAcc;
 } // div32by16


#endif  // --- use above from bootloader ----- (copy out of bootLib.c) -----



/*  Multiply unsigned 16 bit with 8 bit
 *
 *  This function calculates the the product of an 8 and a 16 bit number
 *  and returns the (24 bit) result as 32 bit.
 *
 *  It is far more efficient and less error prone as what could be formulated
 *  in C and then made by (AVR-) GCC.
 *
 *  \param f16 the 16 bit (unsigned) factor
 *  \param f8  the 8 bit (unsigned) factor
 *  \return the product f16 * f8 (It is a 24 bit value, bits 24..31 are 0)
 */
uint32_t mul16with8(const uint16_t f16, const uint8_t f8) {
   uint32_t result;
   asm volatile(
      "mul  %A1, %2 ;  lo(f16) * f8   \n\t"
      "movw %A0, r0 ;  to lower 16 bit \n\t"
      "clr  %C0     ;  upper 16 bit 0   \n\t"
      "clr  %D0     ;  upper 16 bit 0  \n\t"
      "mul  %B1, %2 ;  hi(f16) * f8   \n\t"
      "add  %B0, r0 ;  add to middle \n\t"
      "adc  %C0, r1 ;  add to midhi \n\t"
      "clr  r1 ;  restore zero_reg \n\t"
      : "=&r" (result)
      : "r" (f16), "r" (f8)
   );
   return result;
} // mul16with8(uint16_t, uint8_t)

/*  Multiply unsigned 16 bit with 16 bit
 *
 *  This function calculates the the product of two 16 bit numbers
 *  and returns the 32 bit result.
 *
 *  The (asm) implementation is far more efficient and the usage is less
 *  error prone as what could be expressed in C and then made by (AVR-) GCC.
 *
 *  \param ab the first 16 bit (unsigned) factor
 *  \param cd  the second 16 bit (unsigned) factor
 *  \return the product ab * cd
 */
uint32_t mul16(const uint16_t ab, const uint16_t cd){
   uint32_t result;
   asm volatile(
      "mul  %A1, %A2;  lo(ab) * lo(cd)  \n\t"
      "movw %A0, r0 ;  to lower 16 bit  \n\t"
      "mul  %B1, %B2;  hi(ab) * hi(cd)  \n\t"
      "movw %C0, r0 ;  to upper 16 bit  \n\t"

      "mul  %A1, %B2;  lo(ab) * hi(cd)  \n\t"
      "add  %B0, r0 ;  add to middle   \n\t"
      "adc  %C0, r1 ;  add to midhi   \n\t"
      "clr  r1 ;  restore zero_reg   \n\t"
      "adc  %D0, r1 ;  carry to upper hi  \n\t"

      "mul  %B1, %A2;  hi(ab) * lo(cd)  \n\t"
      "add  %B0, r0 ;  add to middle   \n\t"
      "adc  %C0, r1 ;  add to midhi   \n\t"
      "clr  r1 ;  restore zero_reg   \n\t"
      "adc  %D0, r1 ;  carry to upper hi  \n\t"
      : "=&r" (result)
      : "r" (ab), "r" (cd)
   );
   return result;
} // mul16(2 * uint16_t)

/*  Multiply unsigned 16 bit with 17 bit
 *
 *  This function calculates the the product of a 16 bit number with a
 *  17 bit number and returns the 32 bit result. The second factor \c cd is
 *  supplied as 16 bit unsigned number with bit 16 implied as 1.
 *
 *  The (asm) implementation is far more efficient and the usage is less
 *  error prone as what could be expressed in C and then made by (AVR-) GCC.
 *
 *  \param ab the first 16 bit (unsigned) factor
 *  \param cd  the second 17 bit (unsigned) factor
 *  \return the product ab * (cd + 0x10000)
 *  see  mul16with8  mul16with17
 */
uint32_t mul16with17(const uint16_t ab, const uint16_t cd){
   uint32_t result;
   asm volatile(
      "mul  %A1, %A2;  lo(ab) * lo(cd)   \n\t"
      "movw %A0, r0 ;  to lower 16 bit  \n\t"
      "mul  %B1, %B2;  hi(ab) * hi(cd)   \n\t"
      "movw %C0, r0 ;  to upper 16 bit  \n\t"

      "add  %C0, %A1;  add  * 0x1000 to upper \n\t"
      "adc  %D0, %B1;  add dto. (hi)  \n\t"

      "mul  %A1, %B2;  lo(ab) * hi(cd)  \n\t"
      "add  %B0, r0 ;  add to middle   \n\t"
      "adc  %C0, r1 ;  add to midhi   \n\t"
      "clr  r1 ;  restore zero_reg   \n\t"
      "adc  %D0, r1 ;  carry to upper hi \n\t"

      "mul  %B1, %A2;  hi(ab) * lo(cd)  \n\t"
      "add  %B0, r0 ;  add to middle   \n\t"
      "adc  %C0, r1 ;  add to midhi   \n\t"
      "clr  r1 ;  restore zero_reg   \n\t"
      "adc  %D0, r1 ;  carry to upper hi  \n\t"
      : "=&r" (result)
      : "r" (ab), "r" (cd)
   );
   return result;
} // mul16with17(2 * uint16_t)

/*  Multiply unsigned 32 bit with 512
 *
 *  This function calculates the the product of a 32 bit number with the
 *  constant 512 and returns the 32 bit result. This multiplication is needed
 *  for certain memory card operations (sector number to byte address e.g.).
 *
 *  This implementation is far more efficient as what could be expressed
 *  in C and then made by (AVR-) GCC.
 *
 *  \param fac the 32 bit (unsigned) factor <br />
 *             fac's upper 9 bits are ignored and should be 0 to get a
 *             arithmetically correct result
 *  \return the product fac * 512
 */
uint32_t mul32withVal512(const uint32_t fac){
   uint32_t result;
   asm volatile(
         "mov  %D0, %C1;  fourth byte of result  before * 2  \n\t"
         "mov  %C0, %B1;  third  byte of result  before * 2  \n\t"
         "mov  %B0, %A1;  second byte of result  before * 2  \n\t"
         "mov  %A0, r1;   low byte of result 0 (zero reg) \n\t"

         "lsl  %B0;     second byte of result  * 2  \n\t"
         "rol  %C0;     third  byte of result  * 2  \n\t"
         "rol  %D0;     fourth byte of result  * 2  \n\t"
         : "=&r" (result)
           : "r" (fac)
   );
   return result;
} //  mul32withVal512(uint32_t)

/*  Multiply unsigned 16 bit with 512
 *
 *  This function calculates the the product of a 16 bit number with the
 *  constant 512 and returns the 16 bit result. This multiplication is needed
 *  for certain memory card operations (sector number to byte address e.g.).
 *
 *  This implementation is far more efficient as what could be expressed
 *  in C and then made by (AVR-) GCC.
 *
 *  \param fac the 16 bit (unsigned) factor <br />
 *             fac's upper 9 bits are ignored and should be 0 to get a
 *             arithmetically correct result
 *  \return the product fac * 512
 */
uint16_t mul16withVal512(const uint16_t fac){
   uint16_t result;
   asm volatile(
         "mov  %B0, %A1;  second byte of result  before * 2  \n\t"
         "mov  %A0, r1;   low byte of result 0 (zero reg) \n\t"

         "lsl  %B0;     second byte of result  * 2  \n\t"
         : "=&r" (result)
           : "r" (fac)
   );
   return result;
} //  mul16withVal512(uint32_t)



/*  Divide unsigned 32 bit by 512
 *
 *  This function calculates the the quotient of a 32 bit number resulting
 *  from division by 512 and returns the 32 bit result. This division is
 *  needed for certain memory card operations (with 512 byte sectors e.g.).
 *
 *  This implementation is far more efficient as what could be expressed
 *  in C and then made by (AVR-) GCC.
 *
 *  \param div the 32 bit (unsigned) dividend
 *  \return the quotient div / 512
 */
uint32_t div32byVal512(const uint32_t div){
   uint32_t result;
   asm volatile(   // todo wrong its mul copy
         "mov  %A0, %B1;  low byte of result before / 2  \n\t"
         "mov  %B0, %C1;  second byte of result bef / 2  \n\t"
         "mov  %C0, %D1;  third byte of result  b.  / 2  \n\t"
         "mov  %D0, r1;   forth byte  0 (zero reg)  \n\t"

         "lsr  %C0;     second byte of result  / 2  \n\t"
         "ror  %B0;     third  byte of result  / 2  \n\t"
         "ror  %A0;     fourth byte of result  / 2  \n\t"
         : "=&r" (result)
           : "r" (div)
   );
   return result;
} // div32byVal512(const uint32_t)


/*  Divide unsigned 16 bit by 512
 *
 *  This function calculates the the quotient of a 16 bit number resulting
 *  from division by 512 and returns the 16 bit result. This division is
 *  needed for certain memory card operations (with 512 byte sectors e.g.).
 *
 *  This implementation is far more efficient as what could be expressed
 *  in C and then made by (AVR-) GCC.
 *
 *  \param div the 16 bit (unsigned) dividend
 *  \return the quotient div / 512
 */
uint16_t div16byVal512(const uint16_t div){
   uint16_t result;
   asm volatile(   // todo wrong its mul copy
         "mov  %A0, %B1;  low byte of result  before / 2  \n\t"
         "mov  %B0, r1;   second byte of result 0 (zero reg) \n\t"

         "lsr  %A0;       low byte of result  / 2  \n\t"
         : "=&r" (result)
           : "r" (div)
   );
   return result;
} // div16byVal512(const uint16_t)


/*  Divide unsigned 32 bit by 1024
 *
 *  This function calculates the the quotient of a 32 bit number resulting
 *  from division by 1024 and returns the 32 bit result. This division is
 *  needed for certain memory card operations.
 *
 *  This implementation is far more efficient as what could be expressed
 *  in C and then made by (AVR-) GCC.
 *
 *  \param div the 32 bit (unsigned) dividend
 *  \return the quotient div / 1024
 */
uint32_t div32byVal1024(const uint32_t div){
   uint32_t result;
   asm volatile(
         "mov  %A0, %B1;  low byte of result before / 2  \n\t"
         "mov  %B0, %C1;  second byte of result bef / 2  \n\t"
         "mov  %C0, %D1;  third byte of result  b.  / 2  \n\t"
         "mov  %D0, r1;   forth byte  0 (zero reg)  \n\t"

         "lsr  %C0;     second byte of result  / 2  \n\t"
         "ror  %B0;     third  byte of result  / 2  \n\t"
         "ror  %A0;     fourth byte of result  / 2  \n\t"

         "lsr  %C0;     second byte of result  / 2  \n\t"
         "ror  %B0;     third  byte of result  / 2  \n\t"
         "ror  %A0;     fourth byte of result  / 2  \n\t"
         : "=&r" (result)
           : "r" (div)
   );
   return result;
} // div32byVal1024(const uint32_t)

/*  Divide unsigned 32 bit by 2048
 *
 *  This function calculates the the quotient of a 32 bit number resulting
 *  from division by 2048 and returns the 32 bit result. This division is
 *  needed for certain memory card operations.
 *
 *  This implementation is far more efficient as what could be expressed
 *  in C and then made by (AVR-) GCC.
 *
 *  \param div the 32 bit (unsigned) dividend
 *  \return the quotient div / 2048
 *  \sa  mul32withVal512
 */
uint32_t div32byVal2048(const uint32_t div){
   uint32_t result;
   asm volatile(
         "mov  %A0, %B1;  low byte of result before / 2  \n\t"
         "mov  %B0, %C1;  second byte of result bef / 2  \n\t"
         "mov  %C0, %D1;  third byte of result  b.  / 2  \n\t"
         "mov  %D0, r1;   forth byte  0 (zero reg)  \n\t"

         "lsr  %C0;     second byte of result  / 2  \n\t"
         "ror  %B0;     third  byte of result  / 2  \n\t"
         "ror  %A0;     fourth byte of result  / 2  \n\t"

         "lsr  %C0;     second byte of result  / 2  \n\t"
         "ror  %B0;     third  byte of result  / 2  \n\t"
         "ror  %A0;     fourth byte of result  / 2  \n\t"

         "lsr  %C0;     second byte of result  / 2  \n\t"
         "ror  %B0;     third  byte of result  / 2  \n\t"
         "ror  %A0;     fourth byte of result  / 2  \n\t"
         : "=&r" (result)
           : "r" (div)
   );
   return result;
} // div32byVal1024(const uint32_t)

/*  Divide unsigned 32 bit by 256
 *
 *  This function calculates the the quotient of a 32 bit number resulting
 *  from division by 2048 and returns the 32 bit result. This division is
 *  needed for certain memory card operations.
 *
 *  This implementation is far more efficient as what could be expressed
 *  in C and then made by (AVR-) GCC.
 *
 *  \param div the 32 bit (unsigned) dividend
 *  \return the quotient div / 256
 *  \sa  mul32withVal512
 */
uint32_t div32byVal256(const uint32_t div){
   uint32_t result;
   asm volatile(
         "mov  %A0, %B1;  low byte of result  \n\t"
         "mov  %B0, %C1;  second byte of result  \n\t"
         "mov  %C0, %D1;  third byte of result   \n\t"
         "mov  %D0, r1;   forth byte  0 (zero reg)  \n\t"

           : "=&r" (result)
           : "r" (div)
   );
   return result;
} // div32byVal256(const uint32_t)

/*  Divide unsigned 32 bit by 128
 *
 *  This function calculates the the quotient of a 32 bit number resulting
 *  from division by 2048 and returns the 32 bit result. This division is
 *  needed for certain memory card operations.
 *
 *  This implementation is far more efficient as what could be expressed
 *  in C and then made by (AVR-) GCC.
 *
 *  \param div the 32 bit (unsigned) dividend
 *  \return the quotient div / 128
 *  \sa  mul32withVal512
 */
uint32_t div32byVal128(const uint32_t div){
   uint32_t result;
   asm volatile(
         "mov  %A0, %B1;  low byte of result before * 2  \n\t"
         "mov  %B0, %C1;  second byte of result bef * 2  \n\t"
         "mov  %C0, %D1;  third byte of result  b.  * 2  \n\t"

         "mov  %D0, %A1;  forth byte holds  low byte of div  \n\t"
         "lsr  %D0;       and C = high bit  \n\t"

         "mov  %D0, r1;   forth byte  0 (zero reg)  \n\t"

         "ror  %A0;  low byte of result * 2  \n\t"
         "ror  %B0;  second byte of result * 2  \n\t"
         "ror  %C0;  third byte of result * 2  \n\t"
         "ror  %D0;  forth byte of result * 2  \n\t"

           : "=&r" (result)
           : "r" (div)
   );
   return result;
} // div32byVal128(const uint32_t)


//------------------------------------------------------------------------

/*  Modulo 7 of an unsigned 16 bit value
 *
 *  This is an efficient result = val % 7 without using any divide functions.
 *  \param val the unsigned number
 *  \return   the remainder by 7 for val (0..6)
 */
uint8_t mod16byVal7(uint16_t val){
   uint16_t a = val;  // %1 = %0
   asm volatile(
         "or  %B1, %B1  ; Hi(a) \n\t"
         "breq hibyteZero_%=     \n\t"
         "hibyteNonZero_%=: ; (2)   \n\t"
         "andi %A1, 0xf8    ; sim. /8  \n\t"

         "lsr %B1 ; hi.a>>1       \n\t"
         "ror %A1 ; lo.a>>1        \n\t"
         "sub %A0, %A1 ; val -= a lo  \n\t"
         "sbc %B0, %B1 ; val -= a hi   \n\t"
         "lsr %B1 ; hi.a>>1       \n\t"
         "ror %A1 ; lo.a>>1        \n\t"
         "sub %A0, %A1 ; val -= a lo  \n\t"
         "sbc %B0, %B1 ; val -= a hi   \n\t"
         "lsr %B1 ; hi.a>>1       \n\t"
         "ror %A1 ; lo.a>>1        \n\t"
         "sub %A0, %A1 ; val -= a lo  \n\t"
         "sbc %B0, %B1 ; val -= a hi   \n\t"
         "movw %1, %0  ; a = val  \n\t"
         "or  %B1, %B1 ; Hi(a)   \n\t"
         "brne hibyteNonZero_%= ; loop1  (14*4)  \n\t"

         "hibyteZero_%=:          \n\t"
         "andi %A1, 0xf8    ;  sim. /8  \n\t"
         "breq over7Zero_%= ; (3)  \n\t"
         "over7NonZero_%=:          \n\t"

         "lsr %A1      ; a>>1       \n\t"
         "sub %A0, %A1 ; val -= a   \n\t"
         "lsr %A1      ; a>>1       \n\t"
         "sub %A0, %A1 ; val -= a   \n\t"
         "lsr %A1      ; a>>1       \n\t"
         "sub %A0, %A1 ; val -= a   \n\t"
         "mov %A1, %A0 ; a = val  \n\t"
         "andi %A1, 0xf8    ;  sim. /8  \n\t"
         "brne over7NonZero_%= ; loop2 (9*3)   \n\t"

         "over7Zero_%=:   \n\t"
         "cpi %A0, 7 ; val == 7\n\t"
         "brne mod7LabEnd_%=  \n\t"
         "eor %A0,%A0 ;     val = 0    (4) \n\t"
         "mod7LabEnd_%=:           \n\t"
    :  "+r"(val),  "+r"(a)
   );
   return (uint8_t)val;
}  // mod16byVal7(uint16_t)

/** Modulo 7 of an unsigned 8 bit value
 *
 *  This is an efficient result = val % 7 without using any divide functions.
 *  \param val the unsigned number
 *  \return   the remainder by 7 for val (0..6)
 */
uint8_t mod8byVal7(uint8_t val){
   uint8_t a = val;  // %1 = %0
   asm volatile(
         "andi %A1, 0xf8    ;  sim. /8  \n\t"
         "breq over7Zero_%= ; (3)    \n\t"
         "over7NonZero_%=:          \n\t"

         "lsr %A1      ; a>>1       \n\t"
         "sub %A0, %A1 ; val -= a   \n\t"
         "lsr %A1      ; a>>1       \n\t"
         "sub %A0, %A1 ; val -= a   \n\t"
         "lsr %A1      ; a>>1       \n\t"
         "sub %A0, %A1 ; val -= a   \n\t"
         "mov %A1, %A0 ; a = val  \n\t"
         "andi %A1, 0xf8    ;  sim. /8  \n\t"
         "brne over7NonZero_%= ; loop2 (9*3)   \n\t"

         "over7Zero_%=:   \n\t"
         "cpi %A0, 7 ; val == 7\n\t"
         "brne mod7LabEnd_%=  \n\t"
         "eor %A0,%A0 ;     val = 0    (4) \n\t"
         "mod7LabEnd_%=:           \n\t"
    :  "+r"(val),  "+r"(a)
   );
   return val;
} //  mod8byVal7(uint8_t)

//  -------- computer architecture --------------------------------------


/* Toggle the endianess of a 32 bit value */
void toggle32endian(ucnt32_t * value){
   uint8_t t = (*value).v8[3];
   (*value).v8[3] = (*value).v8[0];
   (*value).v8[0] = t;
   t = (*value).v8[2];
   (*value).v8[2] = (*value).v8[1];
   (*value).v8[1] = t;
} // toggle32endian(ucnt32_t*)


/* Toggle the endianess of a 16 bit value */
void toggle16endian(ucnt16_t * value){
   uint8_t t = (*value).v8[1];
   (*value).v8[1] = (*value).v8[0];
   (*value).v8[0] = t;
} // toggle16endian(ucnt16_t *)

/* Convert the endianess of a 16 bit value
 *
 * This, quite efficiently, is just swap (r24,r25)  by three mov rx,ry
 * consuming one helper register (r18).
 */
uint16_t convert16endian(uint16_t value){
   ucnt16_t ret;
   ret.v8[0] = ((ucnt16_t)value).v8[1];
   ret.v8[1] = ((ucnt16_t)value).v8[0];
   return ret.v16;
} // convert16endian(uint16_t)

/*  Convert the endianess of a 32 bit value
 *
 *  \param value the value to convert
 *  \return value with inverted endianess
 */
uint32_t convert32endian(uint32_t value){
   ucnt32_t ret;
   ret.v8[0] = ((ucnt32_t)value).v8[3];
   ret.v8[1] = ((ucnt32_t)value).v8[2];
   ret.v8[2] = ((ucnt32_t)value).v8[1];
   ret.v8[3] = ((ucnt32_t)value).v8[0];
   return ret.v32;
} // convert32endian(uint32_t)


/* Add a normal (little endian) 16 bit value to 32 bit big endian */
void add16littleTo32bigEndian(ucnt32_t * opRes, uint16_t op2) {
/*
   ucnt32_t tmp;
   tmp.v8[0] = (*opRes).v8[3];
   tmp.v8[1] = (*opRes).v8[2];
   tmp.v8[2] = (*opRes).v8[1];
   tmp.v8[3] = (*opRes).v8[0];
   tmp.v32 += op2;
   (*opRes).v8[0] = tmp.v8[3];
   (*opRes).v8[1] = tmp.v8[2];
   (*opRes).v8[2] = tmp.v8[1];
   (*opRes).v8[3] = tmp.v8[0];

    reduced code size by inline asm below:
    it just saves 2 instructions and 5 register usages */

   asm volatile(
         "movw r30, %0" "\n\t"
         "ldd  __tmp_reg__, Z + 3" "\n\t"
         "add  __tmp_reg__, %A1"     "\n\t"
         "std  Z + 3, __tmp_reg__"   "\n\t"
         "ldd  __tmp_reg__, Z + 2" "\n\t"
         "adc  __tmp_reg__, %B1"     "\n\t"
         "std  Z + 2, __tmp_reg__"   "\n\t"
         "ldd  __tmp_reg__, Z + 1" "\n\t"
         "adc  __tmp_reg__, __zero_reg__" "\n\t"
         "std  Z + 1, __tmp_reg__"       "\n\t"
         "ld   __tmp_reg__, Z"         "\n\t"
         "adc  __tmp_reg__, __zero_reg__" "\n\t"
         "st   Z, __tmp_reg__"   "\n\t"
         :
         : "r"(opRes), "r"(op2)
         : "memory", "r30", "r31"
  );
} // add16littleTo32bigEndian(ucnt32_t *, uint16_t)


// internal: moved here from ntp.c due to inline asm bug
// Bug 48435

/* Sum of two NTP time stamps
 *
 *  This function calculates *result = * a + *b; <br />
 *  No pointer must be null, but two or three of them might point to the
 *  same \ref ntpTimestamp_t structure.
 *
 *  \param a points to first operand (not NULL)
 *  \param b   points to second operand (not NULL)
 *  \param result points to result structure (not NULL)
 */
void ntpTimestampSum(struct ntpTimestamp_t * result,
      struct ntpTimestamp_t * a,
         struct ntpTimestamp_t * b){
   asm volatile(""
         "adiw r26, 7 ; x[7]    (2C) \n\t"

#if ! defined NTP_FRACT_DROP || NTP_FRACT_DROP == 0
         "ldd r24, %a1+7  ;  a*  (2C) \n\t"
         "ldd r25, %a2+7  ;  b*  (2C) \n\t"
         "add r24, r25    ;  a* + b*  \n\t"
         "st  X, r24      ;  result* = (2C)  \n\t"
#else
         "st  X,  __zero_reg__  ; NTP_FRACT_DROP  1..3  \n\t"
         "clc                   ; clear carry \n\t"
#endif

#if ! defined NTP_FRACT_DROP || NTP_FRACT_DROP < 2
         "ldd r24, %a1+6  ;  a* next byte  (2C)  \n\t"
         "ldd r25, %a2+6  ;  b* next byte   (2C) \n\t"
         "adc r24, r25    ;  a + b + carry   \n\t"
         "st  -X, r24     ; result next byte (2C) \n\t"
#else
         "st  -X,  __zero_reg__  ; NTP_FRACT_DROP  2..3  \n\t"
#endif

#if ! defined NTP_FRACT_DROP || NTP_FRACT_DROP < 3
         "ldd r24, %a1+5   \n\t"
         "ldd r25, %a2+5  \n\t"
         "adc r24, r25   \n\t"
         "st  -X, r24   \n\t"
#else
         "st  -X,  __zero_reg__  ; NTP_FRACT_DROP  3  \n\t"
#endif

         "ldd r24, %a1+4   \n\t"
         "ldd r25, %a2+4  \n\t"
         "adc r24, r25   \n\t"
         "st  -X, r24   \n\t"

         "ldd r24, %a1+3   \n\t"
         "ldd r25, %a2+3  \n\t"
         "adc r24, r25   \n\t"
         "st  -X, r24   \n\t"


         "ldd r24, %a1+2   \n\t"
         "ldd r25, %a2+2  \n\t"
         "adc r24, r25   \n\t"
         "st  -X, r24   \n\t"

         "ldd r24, %a1+1   \n\t"
         "ldd r25, %a2+1  \n\t"
         "adc r24, r25   \n\t"
         "st  -X, r24   \n\t"

         "ld  r24, %a1   \n\t"
         "ld  r25, %a2  \n\t"
         "adc r24, r25   \n\t"
         "st  -X, r24   \n\t"
        :
        : "x"(result), "y"(a), "z"(b)
        : "r24", "r25"
       );
  } // ntpTimestampSum(...)



//  -------- formatting  --------------------------------------


/*  Set a three decimal digit number with leading zeroes into a string
 *
 *  This function converts a number 0..999 into three characters "000" ..
 *  "999". Values of 1000 and above will get "***".
 */
void threeDigs(char * begin, uint16_t value){
   if (value == 0) {
      begin[0] = begin[1] = begin[2] = '0';
      return;
   }
   if (value >= 1000) {
      begin[0] = begin[1] = begin[2] = '*';
      return;
   }
   u8div_t q_r = divWByVal10toByte(value); // q:ht r:o
   begin[2] = '0' + q_r.rem;
   q_r = divByVal10(q_r.quot); // q:h r:t
   begin[1] = '0' + q_r.rem;
   begin[0] = '0' + q_r.quot;
} //  threeDigs(char*, uint16_t)

/*  Set a three decimal digit number with leading zeroes into a string
 *
 *  This function converts an eight bit number 0..255 into three characters
 *  "000" .. "255".
 */
void threeDigsB(char * begin, uint8_t const value){
   u8div_t q_r = divByVal10(value); // q:ht r:o
   uint8_t const dig3 = '0' + q_r.rem;
   q_r = divByVal10(q_r.quot); // q:h r:t
   begin[0] = '0' + q_r.quot;
   begin[1] = '0' + q_r.rem;
   begin[2] = dig3;
} //  threeDigsB(char*, uint16_t)

/** Set a four decimal digit number into a string
 *
 *  This function converts a number 0..9999 into four characters "   0" ..
 *  "9999". Values of 10000 and above will get " **0".
 */
void fourDigs(char * begin, uint16_t value){
   begin[0] = begin[1] = begin[2] = ' ';
   begin[3] = '0';
   if (value == 0) return;

   if (value >= 1000) {
      if (value >= 10000) {
         begin[1] = begin[2] =  '*';
         return;
      }
      u16div_t q_R = divWByVal1000(value);
      begin[0] = '0' + q_R.quot;
      begin[1] = begin[2] = '0';
      if (!(value = q_R.rem)) return;
   } // > 1000
   u8div_t q_r = divWByVal10toByte(value); // q:ht r:o
   begin[3] = '0' + q_r.rem;
   q_r = divByVal10(q_r.quot); // q:h r:t
   if (q_r.quot) {
      begin[2] = '0' + q_r.rem;
      begin[1] = '0' + q_r.quot;
   } else if (q_r.rem) {
      begin[2] = '0' + q_r.rem;
   }
} //  fourDigs(char*, uint16_t)


/*  Set two decimal digit number with leading zeroes into a string
 *
 *  This function converts a number 0..99 into two characters "00" ..
 *  "99".
 */
void twoDigs(char * begin, uint8_t value) {
   if (value >= 100) {
      begin[0] = begin[1]  = '*';
      return;
   }
   u8div_t q_r = divByVal10(value); // q:t r:o
   begin[1] = '0' + q_r.rem;
   begin[0] = '0' + q_r.quot;
} // twoDigs(char*, uint8_t)


/*  Set a two hexadecimal digit number with leading zeroes into a string
 *
 *  This function converts a number 0..255 into two characters "00" ..
 *  "FF".
 */
void twoHexs(char * s, uint8_t value){
   uint8_t c = (value >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   *s++ = c;
   c = (value & 0x0F) + '0';
   if (c > '9') c += 'A' - ':';
   *s = c;
} // twoHexs(char *,  uint8_t)


/*  Set a four hexadecimal digit number with leading zeroes into a string
 *
 *  This function converts a 16 bit number into four characters "0000" ..
 *  "FFFF".
 */
void fourHexs(char * begin, uint16_t value){
   uint8_t c = (HBYTE(value) >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   *begin = c;
   c = (HBYTE(value) & 0x0F) + '0';
   if (c > '9') c += 'A' - ':';
   begin[1] = c;
   c = (LBYTE(value) >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   begin[2] = c;
   c = (LBYTE(value) & 0x0F) + '0';
   if (c > '9') c += 'A' - ':';
   begin[3] = c;
   SAMPLE_MIN_SP(0);
} // fourHexs(char *,  uint16_t)

/*  Set a four hexadecimal digit number with leading zeroes into a string
 *
 *  This function converts a 16 bit number into four characters "0000" ..
 *  "FFFF".
 */
void fourHexsBE(char * begin, uint16_t const value){
   uint8_t c = (LBYTE(value) >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   *begin = c;
   c = (LBYTE(value) & 0x0F) + '0';
   if (c > '9') c += 'A' - ':';
   begin[1] = c;
   c = (HBYTE(value) >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   begin[2] = c;
   c = (HBYTE(value) & 0x0F) + '0';
   if (c > '9') c += 'A' - ':';
   begin[3] = c;
} // fourHexsBE(char *,  uint16_t)



/*  Set an eight digit hexadecimal number with leading zeroes into a string
 *
 *  This function converts a 32 bit number into eight characters "00000000"
 *  .. "FFFFFFFF".
 */
void eightHexs(char * begin, uint32_t value){
   uint8_t c = (((ucnt32_t)(value)).v8[3] >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   begin[0] = c;
   c = (((ucnt32_t)(value)).v8[3] & 0x0F) + '0';
   if (c > '9') c += 'A' - ':';
   begin[1] = c;
   c = (((ucnt32_t)(value)).v8[2] >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   begin[2] = c;
   c = (((ucnt32_t)(value)).v8[2] & 0x0F)  + '0';
   if (c > '9') c += 'A' - ':';
   begin[3] = c;

   c = (((ucnt32_t)(value)).v8[1] >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   begin[4] = c;
   c = (((ucnt32_t)(value)).v8[1] & 0x0F) + '0';
   if (c > '9') c += 'A' - ':';
   begin[5] = c;
   c = (((ucnt32_t)(value)).v8[0] >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   begin[6] = c;
   c = (((ucnt32_t)(value)).v8[0] & 0x0F)  + '0';
   if (c > '9') c += 'A' - ':';
   begin[7] = c;
} // eightHexs(char *,  uint32_t)


/* Set an eight digit big endian hexadecimal number with into a string
 *
 *  This function does the same as \ref eightHexs except for the wrong
 *  andianess of the \c value parameter.
 *
 *  \param begin start position within the target string resp. char array
 *  \param value the big endian value to convert
 */
void eightHexsBE(char * begin, uint32_t value){
   uint8_t c = (((ucnt32_t)(value)).v8[0] >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   begin[0] = c;
   c = (((ucnt32_t)(value)).v8[0] & 0x0F) + '0';
   if (c > '9') c += 'A' - ':';
   begin[1] = c;
   c = (((ucnt32_t)(value)).v8[1] >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   begin[2] = c;
   c = (((ucnt32_t)(value)).v8[1] & 0x0F)  + '0';
   if (c > '9') c += 'A' - ':';
   begin[3] = c;

   c = (((ucnt32_t)(value)).v8[2] >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   begin[4] = c;
   c = (((ucnt32_t)(value)).v8[2] & 0x0F) + '0';
   if (c > '9') c += 'A' - ':';
   begin[5] = c;
   c = (((ucnt32_t)(value)).v8[3] >> 4) + '0';
   if (c > '9') c += 'A' - ':';
   begin[6] = c;
   c = (((ucnt32_t)(value)).v8[3] & 0x0F)  + '0';
   if (c > '9') c += 'A' - ':';
   begin[7] = c;
} // eightHexsBE(char *, uint32_t)

/*  Digits buffer for parsing
 *
 *  Hint length must be # of digits (4) + one trailer and header for
 *  parseWordnum. Used in  form16BytSeq, formN20BytSeq too.
 */
// uint8_t digs[69];

/*  Format a sequence of 16 bytes in four hex groups and as characters
 *
 *  This function formats the first 16 bytes pointed to by \c seq in four
 *  groups of 8 hexadecimal digits separated by space followed by said
 *  16 bytes as 16 character text. There control characters (0..0x20 and
 *  0x80..0xA0) are put as blank space.
 *
 *  The result's length is 55 characters including a trailing blank and
 *  terminating 0.<br />
 *  The result looks like <code>
 *  20746578 74206F6E 20612032 4720534D  text on a 2G SM </code>
 *
 *  \param s     start position within string respectively char array
 *  \param seq pointer to the 16 byte sequence to be formatted
 *  \return points to the (0-) character behind the formated result
 *  \sa formN20BytSeq
 */
char * form16BytSeq(char * s, uint8_t * seq){
   uint32_t * begin = (uint32_t *)seq;
   eightHexsBE(s, *begin++); s+=8; *s++ = ' ';
   eightHexsBE(s, *begin++); s+=8; *s++ = ' ';
   eightHexsBE(s, *begin++); s+=8; *s++ = ' ';
   eightHexsBE(s, *begin);   s+=8; *s++ = ' '; *s++ = ' ';
   char const * const stop = s + 20;
   for (; s < stop;){
      const uint8_t c = *seq++;
      *s++ = (c & 0x7F) < 0x20 ? ' ' : c;
   }
   *s++ = ' '; *s = 0;
   return s;
}  // form16BytSeq(uint8_t *)

/*  Format a sequence of 0..20 bytes in five hex groups and as characters
 *
 *  This function formats the \c c (max. 20) bytes pointed to by \c seq in
 *  five groups of 8 hexadecimal digits separated by dot (.) followed by 20
 *  bytes as 20 character text. In this text field control characters
 *  (0..0x20 and 0x80..0xA0) as well as the bytes \c [>= \c n] are put as
 *  blank space.
 *
 *  The maximum length of the result (for \c n=20) is 68 characters
 *  including a trailing blank and terminating 0. <br />
 *  The result (for \c n=17)  looks like <code>
 *  20746578.74206F6E.20612032.4720534D.43....... text on a 2G SMC  </code>
 *
 *  \param s     start position within string resp. char array
 *  \param seq pointer to the up to 20 byte sequence to be formatted
 *  \param n   0..20 number of bytes
 *  \return points to the (0-) character behind the formated result
 *  \sa form16BytSeq
 */
char * formN20BytSeq(char * s, uint8_t * seq, uint8_t n){
   memset(s, '.', 45);
   memset(&s[45], ' ', 22);
   if (!n || !seq) goto done; // nothing to do

   char * toHx = s;
   char * toCh = &s[46];
   if (n > 20) { n = 20; }
   uint8_t * endS = seq + n;
   n = 4;
   for( ; seq != endS;  ++toCh) {
      const uint8_t value = *seq++;
      if ((value & 0x7F) > 0x20) { *toCh = value; }

      uint8_t c = (value >> 4) + '0';
       if (c > '9') c += 'A' - ':';
       *toHx++ = c;
       c = (value & 0x0F) + '0';
       if (c > '9') c += 'A' - ':';
       *toHx++ = c;
       if (! (--n)) {
          ++toHx;
          n = 4;
       }
   } // for over seq

   done:
   s += 67; *s = 0;
   return s;
}  // formN20BytSeq(uint8_t * seq, uint8_t n)


INFLASH(char const  modbTelPat[]) =
      " #   0 +.... u0  fx.. .. . .. . .. . .. . .. . .. . .. . ";
//     0123456789.123456789v123456789t123456789q123456789c1234567

/*  Format a Modbus telegram
 *
 *  This function formats a Modbus TCP/IP telegram provided. This consists
 *  of seven bytes so called MBAB header + 11..13 bytes so called PDU.
 *  From the MBAB the sequence (#), bytes to follow (+) and the unit (u)
 *  are formatted. The PDU part starts with the function code (fx) followed by
 *  0 .. 6 words (16 bit) The number and interpretation of the latter is
 *  dependent of the function code and direction (request / response).
 *
 *  The format returned is <br />
 *  " #.... +.... u.. fx.. .... .... .... .... .... .... .... " <br />
 *  All values are formated hexadecimal; 16 bit numbers are interpreted as
 *  (wrong) big endian.
 *
 *  \param s     start position within string resp. char array
 *  \param modTeleg pointer to the Modbus TCP/IP telegram
 *  \return points to the (0-) character behind the formated result
 */
char * formModTelegr(char * s, struct modTelegr_t * modTeleg){
   memcpy_P(s, modbTelPat, 54);
   ucnt16_t t16;
   t16.v16 = modTeleg->transId.v16;
   if (t16.v16) fourHexsBE(&s[2], t16.v16);
   uint8_t t8 = modTeleg->unitNumb;
   if(t8) twoHexs(&s[14], t8);
   t8 = modTeleg->fCode;
   twoHexs(&s[19], t8);

   t16.v16 = modTeleg->byToFol.v16;
   fourHexsBE(&s[8], t16.v16);

   t8 = (t16.v8[0] || t16.v8[1] > 14) ? 14 : t16.v8[1];
   if (t8 < 3) {
      s += 22;
      goto done;
   }
   t8 -= 2;
   uint8_t i = 0;
   uint8_t j = 22;
   for (;; t8 -= 2, ++i) {
      if (t8 == 1) {
         twoHexs((char *)&s[j], modTeleg->pduCont[i].v8[0]);
         s += j + 3;
         break;
      }
      fourHexsBE((char *)&s[j], modTeleg->pduCont[i].v16);
      j+=5;
      if (t8 < 3) {
         s += j;
         break;
     }
   }
   done:
   *s = 0;
   return s;
}  // formModTelegr(struct modTelegr_t *)

//-----------------------------------------------------   parsing   --------

/*  Search the first token in a short (RAM) string
 *
 *  This function returns the start index (>=si) of the token found in
 *  string s.
 *  The begin of a token may be any non control code except space, comma,
 *  semicolon and equals ( ,;=). <br />
 *  255 is returned if no token was found before [len-1].
 *
 *  \param s   the RAM string to search in
 *  \param si  the index to start search (usually 0 for first token)
 *  \param len the maximum length of s to search in
 */
uint8_t searchFirstToken(char s[], uint8_t si, uint8_t len){
   while (si < len) {
      char c = s[si];
      if((c & 0x7F) < ' ' || c == ' ' || c == ',' || c == ';' || c == '=') {
         ++si;
         continue;
      }
      return si;
   } // while
   return 255;
} // searchFirstToken(char[] ...

/*  Search the next token in a short (RAM) string
 *
 *  This function returns the start index (>=si) of the token found in
 *  string s.
 *  The begin of a token may be any non control code except space, comma,
 *  semicolon and equals ( ,;=). <br />
 *  255 is returned if no token was found before [len-1] or a 0-code.
 *  This 0-code condition is the only difference to \ref searchFirstToken().

 *
 *  \param s   the RAM string to search in
 *  \param si  the index to start search
 *  \param len the maximum length of s to search in
 */
uint8_t searchTokenStart(char s[], uint8_t si, uint8_t len){
   while (si < len) {
      char c = s[si];
      if (!c) break;
      if((c & 0x7F) < ' ' || c == ' ' || c == ',' || c == ';' || c == '=') {
         ++si;
         continue;
      }
      return si;
   } // while
   return 255;
} // searchTokenStart(char[] ...

/*  Search the the end of a token in a short (RAM) string
 *
 *  This function returns the index of the first character following
 *  the the token starting at [sI] in string s or the length of s if no
 *  end was found.
 *
 *  If the return value is == si, no token was found.
 *
 *  Any control character, space, comma, semicolon and equals ( ,;=) is
 *  considered as end of a token.
 *
 *  \param s   the RAM string to search in
 *  \param si  the (found) token's begin index
 *  \param len the maximum length of s to search in
 */
uint8_t searchTokenEnd(char s[], uint8_t si, uint8_t len){
   if (si >= len) return si;
   while (++si < len) {
      char c = s[si];
      if (c <= ' ' || c == ',' || c == ';' || c == '=') break;
   }
   return si;
} // searchTokenEnd(char[] ...

/*  Match a (short) token in a (RAM) string to a list of flash strings
 *
 *  The comparison is case insensitive.
 *
 *  \param s   the RAM string containing the token to search for
 *  \param si  the token's start index
 *  \param se  the token's end index + 1
 *  \param list the flash array of flash strings to match to
 *              (ending with NULL)
 *  \param list2 the second flash array of flash strings to match to
 *              (ending with NULL) 120 is added to the return match index
 *              to distinguish the result from a match in \c list
 *
 *  \return    the index of a match in list; or the index + 120 of a
 *             match in list 2; or 255 if no match; or 254 if the token
 *             in s was too short to match unambiguously
 */
uint8_t searchTokenIn(char s[], uint8_t si, uint8_t se,
                     char const * const * list, char const * const * list2){
   if (si >= se || !s) return 255; // no token
   uint8_t inL2 = 0;
   if (!list) { // no list 1
      if (!list2) return 255; // no lists
      inL2 = INDEX_OFFSET_LIST2; // start in list2
      list = list2;
   } // no list 1

   const char first = s[si] | 0x20; // first char in token (upper case)
   uint8_t listRet = 255; // preset as not (yet) found
   uint8_t listInd = 255; // start index in list as -1
   while (1) {
      char* actTxt =  (char*)pgm_read_word(&list[++listInd]);
      if (!actTxt || listInd >= INDEX_OFFSET_LIST2) { // NULL is list end
         if (inL2 || !list2 ) break; // in list 2 already or no list2
         inL2 = INDEX_OFFSET_LIST2; // start in list2
         list = list2;
         listInd = 255; // start index in list2 as -1
         continue; // with list2 now
      } // list end
      char c = pgm_read_byte(actTxt);
      if (!c) continue; // empty list element
      if ((c | 0x20) != first) continue; // first character mismatch
      int i = si;
      while (++i < se) {
         c = pgm_read_byte(++actTxt);
         if (c <= ' ') break; // list element shorter than token inner while
         char act = s[i] | 0x20;
         if ((c | 0x20) != act) break; // character mismatch
      } // over the token
      if (i == se) { // match found
         if (listRet != 255) return 254; // 2nd match: ambiguous
         c = pgm_read_byte(actTxt); // next char in list element
         if (c <= ' ')  return listInd; // at end of list element
         listRet = listInd + inL2; // remember found
      } // match found
   } // over the list
   return listRet;
} // searchTokenIn(char[] ....


/*  Copy the first SVN token or a full string from flash memory to RAM
 *
 *  This function copies the first SVN-token &mdash; i.e. Subversion
 *  keyword's (first) value &mdash;  from \c src to \c dst plus a trailing 0.
 *  Returned is the address of the last character in \c dst modified (i.e.
 *  the 0). In the example<br />
 *  "$Dte: 2012-12-19 14:13:57 +0100 (Mi, 19 Dez 2012) $"<br />
 *  the first SVN token would be "2012-12-19".
 *
 *  The operation will stop after \c mxLen characters transferred resp.
 *  modified. The \c mxLen count includes the 0 appended.<br />
 *  If \c src does not begin with a $ \c src will be
 *  transferred as is.
 *
 *  \param dst the destination to modify (in RAM)
 *  \param src the source (the SVN tag) to copy from (in flash memory)
 *  \param mxLen the maximum number of characters to modify in \c dst
 *         including the trailing 0 appended. The maximum advance of the
 *         return value to parameter \c dst
 *  \return dst + number of characters modified; NULL only if dst is NULL
 */
char * copyFirstSVNtokenP(char * dst, ADDR_T src, uint8_t mxLen){
   if (!src || dst == NULL || mxLen == 0) return dst;

   char ac = 0;
   uint8_t isSVN = --mxLen;
   if (isSVN) {
     ac = pgm_read_byte(src);
     isSVN = ac == '$';
   }

   if (isSVN) { // goto start of first SVN token
      do { // search :
         ++src;
         ac = pgm_read_byte(src);
      } while (ac && ac != ':');  // search :
      if (ac == ':') do { // skip blank
         ++src;
         ac = pgm_read_byte(src);
      } while (ac == ' ');  // skip blank
   }

   while (ac) { // copy loop
      if (isSVN && (ac == '$' || ac <= ' ')) break;
      *dst = ac;
      ++dst;
      if (!(--mxLen)) break;
      ++src;
      ac = pgm_read_byte(src);
    } // copy loop

   *dst = 0;
   return dst;
} // copySVNtokenP(int32_t, char *, uint8_t)




/*  Recognise one digit
 *
 *  This is just a helper function returning the digit value 0..9 for
 *  characters '0'..'9' and 10..15 for characters 'a'..'f' or 'A'..'F'
 *  If parameter c is in neither range 16 is returned.
 */
uint8_t asDigit(uint8_t c){
   if (c < '0') return 16;
   if (c <= '9') return c - '0';
   c |= 0x20; // poor man's to lower case
   if (c < 'a' || c > 'f') return 16;
   return c - ('a' -10);
} // asDigit


/*  Parse a two digit hexadecimal number
 *
 *  \param res pointer to the result changed if a a hexadecimal number was
 *             parsed successfully
 *  \param s   the RAM string containing the 2 digit hex (00..FF/ff)
 *  \param si  the first digit's index in s
 *  \return    the index of the first character after the parsed number
 *             in case of success or 0 else
 */
uint8_t parse2hex(uint8_t * res, char s[], uint8_t si) {
   uint8_t c = s[si++];
   while (c == ' ') c = s[si++]; // skip blank
   const uint8_t dig1 = asDigit(c);
   if (dig1 & 16) return 0;
   c = s[si];
   const uint8_t dig2 = asDigit(c);
   if (dig2 & 16) {
      *res = dig1;
      return si;
   }
   *res = dig1 << 4 | dig2;
   return si + 1;
} // parse2hex(uint8_t *, char[], uint8_t)

/* Parse a one byte number
 *
 *  This function tries to interpret a part of the string \c s starting at
 *  \c si as an eight bit (one byte) number. Returned is the index of the
 *  first character after that (parsed) number in case of success and 0
 *  otherwise.
 *
 *  Leading blanks will be skipped. One sign character '+' or '-' is
 *  accepted. A minus will complement the result, notwithstanding it being
 *  declared unsigned. Leading zeroes don't, of course change, the result
 *  and they do not lead to octal interpretation.
 *
 *  A leading 0x or 0X will have the following one or two characters be
 *  interpreted as hexadecimal digits. The parsing then will stop at the
 *  latest after the second digit even if the character following (where the
 *  return value points to) is a legal hex digit character.
 *
 *  \param res pointer to the result; changed only if a number was
 *             parsed successfully
 *  \param s   the RAM string containing the digit(s)
 *  \param si  the first digit's index in s
 *  \return    the index of the first character after the parsed number
 *             in case of success or 0 else
 */
uint8_t parseByteNum(uint8_t * res, char s[], uint8_t si){
   uint8_t c = s[si++];
   while (c == ' ') c = s[si++]; // skip blanks
   const uint8_t sign = c == '-';
   if (sign || c == '+')  c = s[si++]; // skip one sign
   if (c < '0' || c > '9') return 0;  // must start with digit

   uint8_t res8 = 0; // used as boolean isHex first
   if (c == '0') { // start with 0
       c = s[si++]; // next
       if ((c | 0x20) == 'x') { // hex
          res8 = c = s[si++]; // next after 0x
       }
       while (c == '0') c = s[si++]; // skip leading zeroes
   } // start with 0

   if (res8) { // hex
      uint8_t dig1 = asDigit(c);
      if (dig1 == 16) { // first digit out of range
         if (res8 == '0') { // but we have 0x0...
            *res = 0;
            return si-1;
         }  // but we have 0x0...
         return 0; // 0xilegal
      } // first digit out of range
      res8 = asDigit( s[si++]); // res8 is digit 0
     if (res8 == 16) { // second digit illegal
        --si;
        res8 = dig1;
        goto    end; // yeah ugly! But it's C, isn't it
     }
     res8 |= (uint8_t)(dig1 * 16); // look if C makes shift
     goto    end; // yeah again
   } // hex

   // decimal
   if (c < '0' || c > '9') { // optimise zero
      *res = 0;
      return si-1;
   } // optimise for zero we might have so far

   // optimise for two digit
   res8 = c - '0';
   c = s[si];
   if (c < '0' || c > '9') goto end;

   res8 = 10 * res8 + c - '0'; // 8 bit arithmetic sufficient
   // now we have two digit 0..99
   c = s[++si];
   if (c < '0' || c > '9') goto end;

   // from here on we have 3 digits and need 16 bit arithmetic
   uint16_t ret =  res8 * 10 +  c - '0';
   if (ret & 0xFF00) return 0; // overflow
   res8 = (uint8_t) ret; ++si;

   end:
   if (sign) res8 = -res8;
   *res = res8;
   return si;
} // parseByteNum

/*  Parse a 16 bit number
 *
 *  This function tries to interpret a part of the string s starting at
 *  si a decimal number. Returned is the index of the first character after
 *  that (parsed) number in case of success and 0 otherwise.
 *
 *  Leading blanks will be skipped.
 *
 *  \param res pointer to the result; changed only if a decimal number was
 *             parsed successfully
 *  \param s   the RAM string containing the digit(s)
 *  \param si  the first digit's index in s
 *  \return    the index of the first character after the parsed number
 *             in case of success or 0 else
 */
uint8_t parseWordNum(uint16_t * res, char s[], uint8_t si){
   uint8_t c = s[si++];
   while (c == ' ') c = s[si++]; // skip blank
   if (!c) return 0;
   uint8_t sign = 0;
   if ((sign = (c == '-')) || c == '+')  c = s[si++]; // skip one sign
   if (c < '0' || c > '9') return 0; // must start with digit

   uint8_t res8 = 0; // used as boolean isHex first
   if (c == '0') { // start with 0
       c = s[si++]; // next
       if ((c | 0x20) == 'x') { // hex
          res8 = c = s[si++]; // next after 0x
       }
       while (c == '0') c = s[si++]; // skip leading zeroes
   } // start with 0

   if (res8) { // hex
      uint8_t digs[6];
      digs[0] = 0;
      digs[1] = asDigit(c);
      if (digs[1] == 16) { // first digit out of range
         if (res8 == '0') { // but we have 0x0...
            *res = 0;
            return si-1;
         }  // but we have 0x0...
         return 0; // 0xilegal
      } // first digit out of range
      res8 = 1;
      for(;;)  {
         if ((digs[++res8] = asDigit(s[si++])) == 16) {
            --si;
            break;
         }
         if (res8 == 5) break;
      }
      c = res8 > 3 ? digs[res8-3] |  digs[res8-4] << 4 : 0;
      const uint16_t r =  c << 8 | ((uint8_t)(digs[res8-1] | digs[res8-2] << 4));
      *res = sign ? -r : r;
      return si;
   } // hex

   // optimise for zero we might have so far
   if (c < '0' || c > '9') {
      *res = 0;
      return si-1;
   }
   // optimise for two digit
   res8 = c - '0';
   c = s[si];
   if (c < '0' || c > '9') {
      if (sign) res8 = -res8;
      *res = (uint16_t)res8;
      return si;
   }
   res8 = 10 * res8 + c - '0'; // 8 bit arithmetic sufficient
   // now we have two digit 0..99
   c = s[++si];
   if (c < '0' || c > '9') {
      if (sign) res8 = -res8;
      *res = (uint16_t)res8;
      return si;
   }
   // from here on we have 3 digits and need 16 bit arithmetic

   uint16_t ret =  res8 * 10 +  c - '0';
   while ((c = s[++si]) >= '0' && c <= '9') {
      ret = ret * 10 + (c - '0');
   }
   if (sign) ret = -ret;
   *res =  ret;
   return si;
} // parseWordNum

/** Parse a four byte number
 *
 *  This function tries to interpret a part of the string \c s starting at
 *  \c si as a 32 bit (type uint32_t) number. Returned is the index of the
 *  first character after that (parsed) number in case of success and 0
 *  otherwise.
 *
 *  Leading blanks will be skipped. One sign character '+' or '-' is
 *  accepted. A minus will complement the result, notwithstanding it being
 *  declared unsigned or eventually parsed as hex. Leading zeroes don't, of
 *  course change, the result and they do not lead to octal interpretation.
 *
 *  A leading 0x or 0X will have the following one to eight characters be
 *  interpreted as hexadecimal digits not counting leading zeros after the
 *  'x'. The parsing then will stop at the latest after the eighth digit even
 *  if the character following (where the returned value points to) is a
 *  legal hex digit character.
 *
 *  \param res pointer to the result; changed only if a number was
 *             parsed successfully
 *  \param s   the RAM string containing the digit(s)
 *  \param si  the first digit's index in s
 *  \return    the index of the first character after the parsed number
 *             in case of success or 0 else
 *  \sa   parseByteNum
 */
uint8_t parseDWordNum(uint32_t * res, char s[], uint8_t si){
   uint8_t c = s[si++];
   while (c == ' ') c = s[si++]; // skip blank
   if (!c) return 0;
   uint8_t sign = 0;
   if ((sign = (c == '-')) || c == '+')  c = s[si++]; // skip one sign
   if (c < '0' || c > '9') return 0; // must start with digit

   uint8_t res8 = 0; // used as boolean isHex first
   if (c == '0') { // start with 0
       c = s[si++]; // next
       if ((c | 0x20) == 'x') { // hex
          res8 = c = s[si++]; // next after 0x
       }
       while (c == '0') c = s[si++]; // skip leading zeroes
   } // start with 0

   if (res8) { // hex
      uint8_t digs[10];
      digs[0] = 0;
      digs[1] = asDigit(c);
      if (digs[1] == 16) { // first digit out of range
         if (res8 == '0') { // but we have 0x0...
            *res = 0;
            return si-1;
         }  // but we have 0x0...
         return 0; // 0xilegal
      } // first digit out of range
      res8 = 1;
      for(;;)  {
         if ((digs[++res8] = asDigit(s[si++])) == 16) {
            --si;
            break;
         }
         if (res8 == 9) break;
      }
      ucnt32_t r;
      for (c = 0; c < 4; ++c) {
         r.v8[c] = res8 > 1 ?(digs[res8-1] | digs[res8-2] << 4) : 0;
         res8 -=2;
      }
      *res = sign ? -r.v32 : r.v32;
      return si;
   } // hex

   // optimise for zero we might have so far
   if (c < '0' || c > '9') {
      *res = 0;
      return si-1;
   }
   // optimise for two digit
   res8 = c - '0';
   c = s[si];
   if (c < '0' || c > '9') {
      if (sign) res8 = -res8;
      *res = (uint16_t)res8;
      return si;
   }
   res8 = 10 * res8 + c - '0'; // 8 bit arithmetic sufficient
   // now we have two digit 0..99
   c = s[++si];
   if (c < '0' || c > '9') {
      if (sign) res8 = -res8;
      *res = (uint16_t)res8;
      return si;
   }
   // from here on we have 3 digits and need 16 bit arithmetic

   uint32_t ret =  res8 * 10 +  c - '0';
   while ((c = s[++si]) >= '0' && c <= '9') {
      ret = ret * 10 + (c - '0');
   }
   if (sign) ret = -ret;
   *res =  ret;
   return si;
} // parseWordNum

/*  Check if a character is contained in a RAM string
 *
 *  \param s   the (0-terminated) RAM string to be searched for \c c
 *  \param c   the (non 0) character looked for
 *  \return    0 : \c c was not found in \c s, \c c : \c c was found in \c s
 */
uint8_t isContainedIn(char s[], const uint8_t c){
   if (!s) return 0;
   if (c) for(;;) {
      uint8_t ca = *s++;
      if (!ca) return 0;
      if (ca == c) break;
   }
   return c;
} // isContainedIn

/*  Check if a character is contained in a flash string
 *
 *  \param s   the (0-terminated) string in flash / program memory to be searched for \c c
 *  \param c   the (non 0) character looked for
 *  \return    0 : \c c was not found in \c s, \c c : \c c was found in \c s
 */
uint8_t isContainedIn_P(char s[], const uint8_t c){
   if (!s) return 0;
   if (c) for(;;++s) {
      uint8_t ca = pgm_read_byte(s);
      if (!ca) return 0;
      if (ca == c) break;
   }
   return c;
} // isContainedIn

